# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.9 $

package Cddb;

use Trace;
use IO::File qw();
use IO::Socket qw();
use strict;

my $cddb = {};

sub configure {
    my %parms = @_;

    # URL - CDDB HTTP server access URL

    if (defined($parms{'URL'}) && ($parms{'URL'} =~ m/^http\:\/\/([\w\.]+):?(\d*)\/(.*)$/)) {
	$cddb->{'Host'} = $1;
	$cddb->{'Port'} = $2 || 'http(80)';
	$cddb->{'Script'} = $3;
	if ($1 =~ /freedb/) {
	    $cddb->{'Submit Host'} = 'freedb.freedb.org';
	    $cddb->{'Submit Script'} = '~cddb/submit.cgi';
	}
	else {
	    $cddb->{'Submit Host'} = $1;
	    $cddb->{'Submit Script'} = '~cddb/submit.cgi';
	}
    }
    else {
	Trace::error("Missing or invalid mandatory 'URL' parameter");
	return undef;
    }

    # User - Name of user, used in messages to CDDB server

    if (!defined($parms{'User'})) {
	Trace::error("Missing mandatory 'User' parameter");
	return undef;
    }
    $cddb->{'User'} = $parms{'User'};

    # Host - Domain name, used in messages to CDDB server to indicate
    # domain name part of users email address.

    if (!defined($parms{'Host'})) {
	Trace::error("Missing mandatory 'Host' parameter");
	return undef;
    }
    $cddb->{'Domain'} = $parms{'Host'};

    # Cache - Command cache, place to store the result of CDDB commands

    if (defined($parms{'Cache'})) {
	$cddb->{'Cache'} = $parms{'Cache'};
	if (! -e $cddb->{'Cache'}) {
	    if (! mkdir($cddb->{'Cache'}, 0755)) {
		Trace::error("Cache dir did not exist and could not be created");
		return undef;
	    }
	}
    }

    # Program - Program name, used in messages to CDDB server

    if (!defined($parms{'Program'})) {
	Trace::error("Missing mandatory 'Program' parameter");
	return undef;
    }
    $cddb->{'Program'} = $parms{'Program'};

    # Version - Program version, used in messages to CDDB server

    if (!defined($parms{'Version'})) {
	Trace::error("Missing mandatory 'Version' parameter");
	return undef;
    }
    $cddb->{'Version'} = $parms{'Version'};
    return 1;
}

# configured 

sub configured {
    if (defined($cddb->{'Host'})) {
	return 1;
    }
    return 0;
}

# command() - Issue a CDDB command.

sub command {
    my %parms = @_;

    my $command = $parms{'Command'};
    if (!defined($command)) {
	Trace::error("Missing mandatory parameter 'Command'");
	return undef;
    }

    # Try and retrieve it from the local cache

    if (defined($cddb->{'Cache'})) {
	my $cacheFile = $cddb->{'Cache'}.'/'.quotemeta($command);
	if (-e $cacheFile && -r $cacheFile) {
	    my $fh = IO::File->new;
	    if (!$fh->open('<'.$cacheFile)) {
		Trace::error("Cannot open $cacheFile: $!");
	    }
	    else {
		$fh->read(my $result, (-s $cacheFile));
		$fh->close;
		Trace::debug("Using cached version of '$command'") if Trace::check(Trace::DEBUG3);
		return ($result);
	    }
	}
    }

    $command =~ s/\s/\+/g;
    $command = "cmd=".$command;
    $command .= "&hello=".$cddb->{'User'}."+".$cddb->{'Domain'}."+".$cddb->{'Program'}."+".$cddb->{'Version'};
    $command .= "&proto=3";

    Trace::debug("Sending '$command'") if Trace::check(Trace::DEBUG3);

    my $sh = IO::Socket::INET->new(PeerAddr => $cddb->{'Host'}, PeerPort => $cddb->{'Port'}, Proto => 'tcp', Timeout => 30);
    if (!$sh) {
	Trace::error("Could not get response from %s: $!", $cddb->{'Host'});
	return undef;
    }

    $sh->autoflush;
    $sh->printf("GET /%s?$command\n", $cddb->{'Script'});
    my $result = '';
    while (<$sh>) {
	$result .= $_;
    }
    $sh->close;

    # Make sure the result is valid and worth saving, valid means a
    # CDDB return code < 400

    if ($result =~ /^(\d+)\s/) {
	if ($1 >= 400) {
	    Trace::error('Error from CDDB: '.$result);
	    return undef;
	}
    }
    else {
	# Might be an invalid URL
	Trace::error('Does not look like a CDDB response: '.$result);
	return undef;
    }

    # Save to the local cache

    if (defined($cddb->{'Cache'})) {
	my $cacheFile = $cddb->{'Cache'}.'/'.quotemeta($command);
	my $fh = IO::File->new;
	if (!$fh->open('>'.$cacheFile)) {
	    Trace::error("Could not open $cacheFile: $!");
	}
	else {
	    $fh->write($result, length($result));
	    $fh->close;
	}
    }

    return ($result);
}

# submit() - Submit a CDDB database entry.

sub submit {
    my %parms = @_;

    my $data = $parms{'Data'};
    if (!defined($data)) {
	Trace::error("Missing mandatory parameter 'Data'");
	return;
    }
    my $category = $parms{'Category'};
    if (!defined($category)) {
	Trace::error("Missing mandatory parameter 'Category'");
	return;
    }
    my $cddbId = $parms{'CddbId'};
    if (!defined($cddbId)) {
	Trace::error("Missing mandatory parameter 'CddbId'");
	return;
    }

    Trace::debug('Submitting database entry') if Trace::check(Trace::DEBUG3);

    my $sh = IO::Socket::INET->new(PeerAddr => $cddb->{'Submit Host'}, PeerPort => $cddb->{'Port'}, Proto => 'tcp', Timeout => 30);
    if (!$sh) {
	Trace::error("Could not get response from %s: $!", $cddb->{'Submit Host'});
	return;
    }

    $sh->printf("POST /%s HTTP/1.0\n", $cddb->{'Submit Script'});
    $sh->printf("Category: %s\n", $category);
    $sh->printf("Discid: %s\n", $cddbId);
    $sh->printf("User-Email: %s@%s\n", $cddb->{'User'}, $cddb->{'Domain'});
    $sh->printf("Submit-Mode: submit\n");
    $sh->printf("X-Cddbd-Note: Problems? visit http://emptytreeseedy.sourceforge.net/\n");
    $sh->printf("Content-Length: %d\n\n", length $data);
    print $sh $data;
    $sh->flush;

    my $result = '';
    while (<$sh>) {
	$result .= $_;
    }
    $sh->close;

    if ($result =~ /^(\d+)\s/m) {
	if ($1 != 200) {
	    Trace::error('Error from CDDB: '.$result);
	    return;
	}
	return 1;
    }
    else {
	# Might be an invalid URL
	Trace::error('Did not get CDDB response: '.$result);
	return;
    }

    return ($result);
}

# sites - Load the latest list of sites from the configured CDDBP server.

sub sites {
    my @sites = ();
    my $result = Cddb::command(Command => 'sites');

    return if !defined($result);

    my @lines = split(/\n/, $result);
    foreach my $line (@lines) {
	# Just http
	if ($line =~ /^(\S+)\s+http\s+(\d+)\s+(\S+)\s+\S+\s+\S+\s+(.*)\r?$/) {
	    my $host = $1;
	    my $port = $2;
	    my $cgi = $3;
	    my $desc = $4;
	    if ($port == 80) {
		push @sites, { URL => "http://$host$cgi", Desc => $desc };
	    }
	    else {
		push @sites, { URL => "http://$host:$port$cgi", Desc => $desc };
	    }
	}
    }
    return @sites;
}

1;
