# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.21 $

package Cdrom;

use Thing;
@ISA = qw( Thing );
use CdromH;
use Info::Cd;
use Command;
use Format::Wav;
use IO::File;
use Trace;
use strict;

# Status message `parms' are a hash ref containing `state', `cause'
# and `attributes' fields (denoted by `field') generated as below. A
# common attribute `configureError' if present denotes some problem
# with configuration and that the cdrom is in an unuseable state as
# described by this short message.

# State     Cause        To
# -----     -----        --
#
# Stopped   None         From  Cdrom is stopped with `cd' in the drive.
#
# Stopped   Message      All   Cdrom is stopped with `cd' in the drive due to
#                              action upon message receipt.
#
# Stopped   End          All   Cdrom is stopped with `cd' in the drive due to
#                              end of play or rip.
#
# Stopped   Ignored      From  Attempt to Stop or Pause the cdrom was ignored
#                              because it is already in the Stopped state.
#
# Stopped   Invalid      From  Attempt to Play or Rip had invalid parameters.
#                              e.g. lastTrackIndex past end of cd etc.
#
# Stopped   RipStuck     All   Ripping was aborted after `time' for rip 
#                              message with `parms' because it was taking too
#                              long.
#
# Stopped   RipError     All   The ripping command returned an error described
#                              by `diagnostic' for the rip message
#                              with `parms'.
#
# Stopped   Error        All   An unexpected Cdrom error occurred as described
#                              by `diagnostic'.
# 
# NoCd      None         From  Cdrom's `tray' is TrayOpen or TrayClosed with
#                              no cd known to be in it. `tray' is a common 
#                              attribute for all Status messages with a NoCd
#                              state.
#
# NoCd      Message      All   As above due to action upon message receipt.
#
# NoCd      Ignored      From  An attempt to stop, rip, play, pause, open an
#                              already open tray was ignored.
#
# NoCd      Error        All   An unexpected Cdrom error occurred as described
#                              by `diagnostic'.
#
# Playing   None         All   Cdrom is playing the numbered `trackIndex' from 
#                              `cd' and has seconds of play `left'. Sent at 
#                              the start of each track.
#
# Playing   Message      All   Cdrom is playing the numbered `trackIndex' from 
#                              `cd' and has seconds of play `left'. Sent
#                              following action upon message receipt.
#
# Playing   Invalid      From  Attempt to play or rip had invalid parameters.
#
# Playing   Error        All   An unexpected Cdrom error occurred as described
#                              by `diagnostic'.
#
# Paused    None         All   Playing of `trackIndex' from `cd' is paused
#                              with seconds of play `left'.
#
# Paused    Ignored      From  Play is already paused.
#
# Ripping   None         All   Cdrom is ripping the numbered `trackIndex' from 
#                              `cd' and is `percentageComplete'. Sent 
#                              periodically.
#
# Ripping   Message      All   Cdrom is ripping the numbered `trackIndex' from 
#                              `cd' and is `percentageComplete'. Sent
#                              following action upon message receipt.
#
# Ripping   Invalid      From  Attempt to play or rip had invalid parameters. 
#
# Ripping   Error        All   An unexpected Cdrom error occurred as described
#                              by `diagnostic'.
#

# Linux CDROM ioctls

BEGIN {
  use vars qw( $cdrom_ioctls );
  my $sys = `uname -smp`;
  chomp $sys;

  if ($sys =~ /Linux/) {
    $cdrom_ioctls = 1;
  }
  else {
    $cdrom_ioctls = 0;
  }
}

sub CDROM_LBA ()             {0x01;}
sub CDROM_MSF ()             {0x02;}
sub CDROM_DATA_TRACK ()      {0x04;}
sub CDROMPAUSE ()            {0x5301;}
sub CDROMRESUME ()           {0x5302;}
sub CDROMPLAYMSF ()          {0x5303;}
sub CDROMPLAYTRKIND ()       {0x5304;}
sub CDROMSTOP ()             {0x5307;}
sub CDROMSTART ()            {0x5308;}
sub CDROMREADTOCHDR ()       {0x5305;}
sub CDROMREADTOCENTRY ()     {0x5306;}
sub CDROMCLOSETRAY ()        {0x5319;}
sub CDROMEJECT ()            {0x5309;}
sub CDROMMULTISESSION ()     {0x5310;}
sub CDROM_GET_MCN ()         {0x5311;}
sub CDROM_GET_CAPABILITY ()  {0x5331;}
sub CDROM_MEDIA_CHANGED	()   {0x5325;}
sub CDROM_DRIVE_STATUS ()    {0x5326;}
sub CDROM_LOCKDOOR ()	     {0x5329;}
sub CDROM_DEBUG ()	     {0x5330;}
my $cdromTocHdrStruct = 'cc';
my $cdromTocEntryStruct = 'ccccccc';
my $cdromMultiSessionStruct = 'cccccc';
my $cdromPlayMsf = 'cccccc';
my $cdromPlayTrkind = 'cccc';

sub new {
  my $proto = shift;
  my $type = ref($proto) || $proto;
  my $cdrom = $type->SUPER::new(@_);
  if ($cdrom->error) {
    return $cdrom;
  }

  $cdrom->{'handle'} = new IO::File;
  $cdrom->{'ripMonitorInterval'} = 15;
  $cdrom->{'state'} = NoCd;
  $cdrom->{'configureError'} = 'Not configured';
  $cdrom->{'cause'} = None;
  $cdrom->{'attributes'} = {};
  $cdrom->{'lock'} = 0;
  $cdrom->{'autoDetect'} = 1;
  $cdrom->at(name => '_autoDetect', to => $cdrom, time => time + 2);
  $cdrom->{'monitors'} = {};
  # Limit rip time based on the file size, 60 seconds should have
  # ripped in 10 minutes.
  $cdrom->{'timeLimitForSixty'} = 600;
  $cdrom->{'tag'} = 1;

  $cdrom->register(name => State, handler => 'state');
  $cdrom->register(name => Play, handler => 'play');
  $cdrom->register(name => Pause, handler => 'pause');
  $cdrom->register(name => Seek, handler => 'seek');
  $cdrom->register(name => Rip, handler => 'rip');
  $cdrom->register(name => Stop, handler => 'stop');
  $cdrom->register(name => Eject, handler => 'eject');
  $cdrom->register(name => Close, handler => 'close');
  $cdrom->register(name => '_playMonitor', handler => '_playMonitor');
  $cdrom->register(name => '_ripMonitor', handler => '_ripMonitor');
  $cdrom->register(name => '_autoDetect', handler => '_autoDetect');
  $cdrom->register(name => Thing::ConfigureRequest, handler => '_configure');
  $cdrom->register(name => Thing::ShutdownRequest, handler => '_shutdown');

  return $cdrom;
}

# _configure()

sub _configure {
  my $cdrom = shift;
  my %tParms = @_;
  my $parms = $tParms{'parms'};

  $cdrom->{'configureError'} = '';

  if (!$cdrom_ioctls) {
    Trace::error('The cdrom is only supported under Linux');
    $cdrom->{'configureError'} = $parms->{'error'} = Trace::lastError;
    $cdrom->_status;
    return;
  }

  if (!defined($parms->{'db'})) {
    Trace::error('Need to specify database parameters');
    $cdrom->{'configureError'} = $parms->{'error'} = Trace::lastError;
    $cdrom->_status;
    return;
  }

  if (!Info::Persist::configure(db => $parms->{'db'}->{'local'})) {
    $cdrom->{'configureError'} = $parms->{'error'} = Trace::lastError;
    $cdrom->_status;
    return;
  } 

  if (!Info::Cd::configure(remotes => $parms->{'db'}->{'remotes'})) {
    $cdrom->{'configureError'} = $parms->{'error'} = Trace::lastError;
    $cdrom->_status;
    return;
  }

  if (!defined($parms->{'repo'})) {
    Trace::error('Need repo configured');
    $cdrom->{'configureError'} = $parms->{'error'} = Trace::lastError;
    $cdrom->_status;
    return;
  }
  $cdrom->{'repo'} = $cdrom->monitor($parms->{'repo'});

  if (!defined($parms->{'device'})) {
    Trace::error('Need to specify a device');
    $cdrom->{'configureError'} = $parms->{'error'} = Trace::lastError;
    $cdrom->_status;
    return;
  }

  if (defined($parms->{'autoDetect'})) {
    if (!$cdrom->{'autoDetect'} && $parms->{'autoDetect'}) {
      $cdrom->at(name => '_autoDetect', to => $cdrom, time => time + 2);
    }
    $cdrom->{'autoDetect'} = $parms->{'autoDetect'};
  }

  if (defined($parms->{'lock'})) {
    $cdrom->{'lock'} = $parms->{'lock'};
  }

  if ($cdrom->{'lock'} && $cdrom->{'autoDetect'}) {
    $parms->{'error'} = 'Auto detection disables drive locking'; # Warning
    $cdrom->{'lock'} = 0;
  }

  if (!defined($cdrom->{'device'}) || ($cdrom->{'device'} ne $parms->{'device'})) {

    # New device

    $cdrom->_stop;		# Stop whatever the current one is doing

    $cdrom->{'device'} = $parms->{'device'};
    $cdrom->{'state'} = NoCd;

    if ($parms->{'debug'}) {
      if (!$cdrom->{'handle'}->open($cdrom->{'device'}, O_NONBLOCK|O_RDONLY)) {
	Trace::error('Cannot open cdrom: '.$!);
	$parms->{'error'} = Trace::lastError;
	$cdrom->{'configureError'} = 'Problems with the device: '.Trace::lastError;
	$cdrom->_status;
	return;
      }
      if (!defined(ioctl($cdrom->{'handle'}, CDROM_DEBUG, 1))) {
	Trace::error("Cannot debug: $!");
	$cdrom->{'handle'}->close;
	$parms->{'error'} = Trace::lastError;
	$cdrom->{'configureError'} = 'Problems with the device: '.Trace::lastError;
	$cdrom->_status;
	return;
      }
      $cdrom->{'handle'}->close;
    }

    if (!$cdrom->{'handle'}->open($cdrom->{'device'}, O_NONBLOCK|O_RDONLY)) {
      Trace::error('Cannot open cdrom: '.$!);
      $parms->{'error'} = Trace::lastError;
      $cdrom->{'configureError'} = 'Problems with the device: '.Trace::lastError;
      $cdrom->_status;
      return;
    }
    $cdrom->{'handle'}->close;
    # Start with the door closed, sends status, only practical
    # when the Repo is alive to take ownership of the Cd object.
    if ($cdrom->{'repo'}->currentState eq Thing::Handle::Alive) {
      $cdrom->_close;
    }
    else {
      $cdrom->{'configureError'} = 'Not configured';
    }
  }
  $cdrom->_lock();
}

# monitorChange

sub monitorChange {
  my $cdrom = shift;
  my $handle = shift;

  Trace::debug('Monitor change for %s to %s', $handle->key, $handle->currentState) if Trace::check(Trace::DEBUG3);

  if (($handle == $cdrom->{'repo'}) && ($handle->currentState eq Thing::Handle::Alive) && ($cdrom->{'configureError'} eq 'Not configured')) {
    $cdrom->{'configureError'} = '';
    $cdrom->_close;
  }
}

sub _shutdown {
  my $cdrom = shift;
  $cdrom->_stop;
}

# state() - State enquiry.

sub state {
  my $cdrom = shift;
  my %tParms = @_;
  $cdrom->_addToMonitors($tParms{'from'});
  $cdrom->_status(to => $tParms{'from'});
}

sub _status {
  my $cdrom = shift;
  my %parms = @_;

  my $attr = $cdrom->{'attributes'};
  $attr->{'configureError'} = $cdrom->{'configureError'} if ($cdrom->{'configureError'} ne '');

  if (($cdrom->{'state'} eq Stopped) || ($cdrom->{'state'} eq Playing) || ($cdrom->{'state'} eq Paused) || ($cdrom->{'state'} eq Ripping)) {

    $attr->{'cd'} = $cdrom->{'cd'};

    if (($cdrom->{'state'} eq Playing) || ($cdrom->{'state'} eq Paused) || ($cdrom->{'state'} eq Ripping)) {
      $attr->{'trackIndex'} = $cdrom->{'trackIndex'};
    }
    else {
      if (defined($cdrom->{'seek'})) {
	$attr->{'trackIndex'} = $cdrom->{'seek'}->{'trackIndex'};
      }
      else {
	$attr->{'trackIndex'} = $cdrom->{'cd'}->firstAudioTrackIndex;
      }
    }

    if ($cdrom->{'state'} eq Ripping) {
      $attr->{'percentageComplete'} = $cdrom->{'ripping'}->{'percentageComplete'};
    }
    elsif ($cdrom->{'state'} eq Playing) {
      $attr->{'left'} = $cdrom->{'playing'}->{'end'} - time;
    }
    elsif ($cdrom->{'state'} eq Paused) {
      $attr->{'left'} = $cdrom->{'playing'}->{'left'};
    }
    else {
      if (defined($cdrom->{'seek'}) && defined($cdrom->{'seek'}->{'offset'})) {
	$attr->{'left'} = $attr->{'cd'}->track($attr->{'trackIndex'})->duration - $cdrom->{'seek'}->{'offset'};
      }
      else {
	$attr->{'left'} = $attr->{'cd'}->track($attr->{'trackIndex'})->duration;
      }
    }
  }
  else {
    $attr->{'tray'} = $cdrom->{'tray'};
  }
  $attr->{'time'} = time;

  # Send on down
  my @dest = ();
  if (defined($parms{'to'})) {
    push @dest, $parms{'to'};
  }
  else {
    @dest = values(%{ $cdrom->{'monitors'} });
  }
  foreach my $dest (@dest) {
    if (!$cdrom->message(name => Status, 
			 to => $dest,
			 parms =>
			 {
			  state => $cdrom->{'state'},
			  cause => $cdrom->{'cause'},
			  attributes => $cdrom->{'attributes'},
			 }
			)) {
      Trace::debug('Removed unresponsive state monitor: %s', $dest->unique) if (Trace::check(Trace::DEBUG3));
      delete $cdrom->{'monitors'}->{ $dest->unique };
    }
  }
  $cdrom->{'cause'} = None;
  $cdrom->{'attributes'} = {};
}

# play() - Play from `firstTrackIndex' to `lastTrackIndex' request. If
# either parameter is unspecified the cd's values are used or a prior
# seek request. Will not play if state is NoCd.

sub play {
  my $cdrom = shift;
  my %tParms = @_;
  my $parms = $tParms{'parms'};
  $cdrom->_addToMonitors($tParms{'from'});

  if ($cdrom->{'state'} eq NoCd) {
    $cdrom->{'cause'} = Ignored;
    $cdrom->_status(to => $tParms{'from'});
    return;
  }
  elsif ($cdrom->{'state'} eq Ripping) {
    if (!$cdrom->_stop()) {
      $cdrom->_status();
      return;
    }
  }

  return $cdrom->_play(from => $tParms{'from'}, %{ $parms });
}

sub _play {
  my $cdrom = shift;
  my %parms = @_;

  my $open = 0;
  my $errorSub = sub { 
    Trace::error(@_) if ($#_ != -1);
    $cdrom->{'cause'} = Error;
    $cdrom->{'attributes'}->{'diagnostic'} = Trace::lastError;
    $cdrom->{'handle'}->close if ($open);
    $cdrom->_status(to => $parms{'from'});
    return 0;
  };
  my $invalidSub = sub { 
    $cdrom->{'cause'} = Invalid;
    $cdrom->{'attributes'}->{'diagnostic'} = shift;
    $cdrom->{'handle'}->close if ($open);
    $cdrom->_status(to => $parms{'from'});
    delete $cdrom->{'seek'};
    return 0;
  };

  my $cd = $cdrom->{'cd'};

  if (!$cdrom->{'handle'}->open($cdrom->{'device'}, O_NONBLOCK|O_RDONLY)) {
    return &$errorSub('Cannot open cdrom: %s', $!);
  }
  $open = 1;

  my $time;
  if ($cdrom->{'state'} eq Paused) {
    if (!defined(ioctl($cdrom->{'handle'}, CDROMRESUME, 0))) {
      return &$errorSub('Cannot resume play: %s', $!);
    }
    $time = time + $cdrom->{'playing'}->{'left'};
    $cdrom->{'playing'}->{'end'} = $time;
  }
  else {
    my $fti;
    my $lti;
    my $offset = 0.0;

    if (defined($cdrom->{'seek'})) {
      # Unexecuted seek request, take play values from here
      if (defined($cdrom->{'seek'}->{'trackIndex'})) {
	$fti = $cdrom->{'seek'}->{'trackIndex'};
      }
      if (defined($cdrom->{'seek'}->{'offset'})) {
	$offset = $cdrom->{'seek'}->{'offset'};
      }
      delete $cdrom->{'seek'};
    }

    if (!defined($fti)) {
      $fti = defined($parms{'firstTrackIndex'}) ? $parms{'firstTrackIndex'} : $cd->firstAudioTrackIndex;
    }

    $lti = defined($parms{'lastTrackIndex'}) ? $parms{'lastTrackIndex'} : $cd->lastAudioTrackIndex;

    if ($cd->track($fti)->data || $cd->track($lti)->data) {
      return &$invalidSub('Cannot play data track');
    }

    if (($fti > $lti) || ($fti < $cd->firstAudioTrackIndex) || ($lti > $cd->lastAudioTrackIndex)) {
      return &$invalidSub('Out of bound play request '.$fti.' to '.$lti);
    }

    if ($offset >= $cd->track($fti)->duration) {
      return &$invalidSub('Seek offset out of bounds');
    }

    if (!defined(ioctl($cdrom->{'handle'}, CDROMSTART, 0))) {
      return &$errorSub('Cannot start cdrom: %s', $!);
    }

    # Having troubles with CDROMPLAYTRKIND, so I'll use
    # CDROMPLAYMSF instead, from comments in other code
    # (kdemultimedia/kscd/cdrom.c) I'm not alone.

    my $leadoutTi = $lti + 1;
    if ($leadoutTi > $cd->lastTrackIndex) {
      $leadoutTi = Info::Track::LEADOUTTN;
    }
    my @startMsf = $cd->track($fti)->msf;
    if ($offset) {
      $startMsf[0] += POSIX::floor($offset / 60);
      $startMsf[1] += POSIX::floor($offset) % 60;
      $startMsf[2] += POSIX::floor(($offset - POSIX::floor($offset)) * 75);
      if ($startMsf[2] >= 75) {
	$startMsf[2] -= 75;
	$startMsf[1]++;
      }
      if ($startMsf[1] >= 60) {
	$startMsf[1] -= 60;
	$startMsf[0]++;
      }
    }
    my $playMsf = pack($cdromPlayMsf, @startMsf, $cd->track($leadoutTi)->msf); 
    if (!defined(ioctl($cdrom->{'handle'}, CDROMPLAYMSF, $playMsf))) {
      return &$errorSub('Cannot play tracks (%d to %d): %s', $fti, $lti, $!);
    }

    # Save `playing' information to inform clients when the next
    # track starts and when playing has stopped.

    $time = time + $cd->track($fti)->duration - $offset;
    $cdrom->{'playing'} = { lastTrackIndex => $lti, end => $time };
    $cdrom->{'trackIndex'} = $fti;
  }
  $cdrom->{'handle'}->close;
  $cdrom->at(name => '_playMonitor', to => $cdrom, time => $time, parms => { tag => ++$cdrom->{'tag'} });

  $cdrom->{'state'} = Playing;
  $cdrom->{'cause'} = Message;
  $cdrom->_status();
  return;
}

# _playMonitor - Finished a track, move on to the next or signal its the last.

sub _playMonitor {
  my $cdrom = shift;
  my %tParms = @_;
  my $parms = $tParms{'parms'};

  # If the Cd was stopped and started or replaced and started then
  # this message is no longer for this Cd.

  my $cd = $cdrom->{'cd'};
  return if (!defined($cd));
  return if (($cdrom->{'state'} ne Playing) || ($parms->{'tag'} ne $cdrom->{'tag'}));

  my $finished = $cdrom->{'trackIndex'};
  my $track = $cd->track($finished);
  if ($track) {
    $track = $track->monitor;
    $track->updateStatsAfterPlay;
  }

  if ($cdrom->{'trackIndex'} == $cdrom->{'playing'}->{'lastTrackIndex'}) {
    delete $cdrom->{'playing'};
    $cdrom->{'state'} = Stopped;
    $cdrom->{'cause'} = End;
  }
  else {
    # Skip data tracks
    do {
      $cdrom->{'trackIndex'}++;
    } while (($cdrom->{'trackIndex'} <= $cdrom->{'playing'}->{'lastTrackIndex'}) && $cd->track($cdrom->{'trackIndex'})->data);
    if ($cdrom->{'trackIndex'} > $cdrom->{'playing'}->{'lastTrackIndex'}) {
      # Last tracks were data
      delete $cdrom->{'playing'};
      $cdrom->{'state'} = Stopped;
      $cdrom->{'cause'} = End;
    }
    else {
      my $time = time + $cd->track($cdrom->{'trackIndex'})->duration;
      $cdrom->{'playing'}->{'end'} = $time;
      $cdrom->at(name => '_playMonitor', to => $cdrom, time => $time, parms => { tag => ++$cdrom->{'tag'} });
    }
  }

  $cdrom->_status();
}

# pause() - Pause play or Stop ripping.

sub pause {
  my $cdrom = shift;
  my %tParms = @_;
  my $parms = $tParms{'parms'};
  $cdrom->_addToMonitors($tParms{'from'});

  if (($cdrom->{'state'} eq NoCd) ||
      ($cdrom->{'state'} eq Stopped) ||
      ($cdrom->{'state'} eq Paused)) {
    $cdrom->{'cause'} = Ignored;
    $cdrom->_status(to => $tParms{'from'});
    return;
  }
  elsif ($cdrom->{'state'} eq Ripping) {
    if (!$cdrom->_stop()) {
      $cdrom->_status();
      return;
    }
    $cdrom->{'cause'} = Message;
    $cdrom->_status();
  }
  else {
    if (!$cdrom->{'handle'}->open($cdrom->{'device'}, O_NONBLOCK|O_RDONLY)) {
      Trace::error('Cannot open cdrom: '.$!);
      $cdrom->{'attributes'}->{'diagnostic'} = Trace::lastError;
      $cdrom->{'cause'} = Error;
      $cdrom->_status(to => $tParms{'from'});
      return;
    }

    if (!defined(ioctl($cdrom->{'handle'}, CDROMPAUSE, 0))) {
      Trace::error('Cannot pause: '.$!);
      $cdrom->{'attributes'}->{'diagnostic'} = Trace::lastError;
      $cdrom->{'cause'} = Error;
      $cdrom->_status(to => $tParms{'from'});
      $cdrom->{'handle'}->close;
      return;
    }
    $cdrom->{'playing'}->{'left'} = $cdrom->{'playing'}->{'end'} - time;
    $cdrom->{'state'} = Paused;
    $cdrom->{'cause'} = Message;
    $cdrom->_status();
  }
}

# seek() - Seek to 'trackIndex' and 'offset' in the cd. Will not seek
# when NoCD or Ripping.

sub seek {
  my $cdrom = shift;
  my %tParms = @_;
  my $parms = $tParms{'parms'};
  $cdrom->_addToMonitors($tParms{'from'});

  if (($cdrom->{'state'} eq NoCd) || ($cdrom->{'state'} eq Ripping)) {
    $cdrom->{'cause'} = Ignored;
    $cdrom->_status(to => $tParms{'from'});
    return;
  }
  elsif ($cdrom->{'state'} eq Paused) {
    if (!$cdrom->_stop()) {
      $cdrom->_status();
      return;
    }
  }

  my $invalidSub = sub { 
    $cdrom->{'cause'} = Invalid;
    $cdrom->{'attributes'}->{'diagnostic'} = shift;
    $cdrom->_status(to => $tParms{'from'});
    return 0;
  };

  if (!defined($parms->{'trackIndex'})) {
    return &$invalidSub('Seek requires a track index');
  }

  my $cd = $cdrom->{'cd'};
  my $ti = $parms->{'trackIndex'};
  if (($ti < $cd->firstAudioTrackIndex) || ($ti > $cd->lastAudioTrackIndex)) {
    return &$invalidSub('Seek track index is out of bounds');
  }
  if ($cd->track($ti)->data) {
    return &$invalidSub('Seek track is a data track');
  }

  if (defined($parms->{'offset'}) && ($parms->{'offset'} >= $cd->track($ti)->duration)) {
    return &$invalidSub('Seek offset is outside the track');
  }

  $cdrom->{'seek'} = $parms;
  if ($cdrom->{'state'} eq Playing) {
    return $cdrom->_play(from => $tParms{'from'});
  }
  $cdrom->_status();
  return;
}

# rip() - Rip `trackIndex' to `file'. Will not rip if state is NoCd.

sub rip {
  my $cdrom = shift;
  my %tParms = @_;
  my $parms = $tParms{'parms'};
  $cdrom->_addToMonitors($tParms{'from'});

  if ($cdrom->{'state'} eq NoCd) {
    $cdrom->{'cause'} = Ignored;
    $cdrom->_status(to => $tParms{'from'});
    return;
  }
  elsif (($cdrom->{'state'} eq Ripping) || ($cdrom->{'state'} eq Paused) || ($cdrom->{'state'} eq Playing)) {
    if (!$cdrom->_stop()) {
      $cdrom->_status();
      return;
    }
  }

  my $cd = $cdrom->{'cd'};

  my $trackIndex = $parms->{'trackIndex'};
  if (!defined($trackIndex) || ($trackIndex < $cd->firstTrackIndex) || ($trackIndex > $cd->lastTrackIndex)) {
    Trace::error('Invalid or missing trackIndex');
    $cdrom->{'cause'} = Invalid;
    $cdrom->_status(to => $tParms{'from'});
    return;
  }

  my $track = $cd->track($trackIndex);
  if ($track->data) {
    Trace::error('Data track, will not rip');
    $cdrom->{'cause'} = Invalid;
    $cdrom->_status(to => $tParms{'from'});
    return;
  }

  my $file = $parms->{'file'};
  if (!defined($file)) {
    Trace::mandatory('file');
    $cdrom->{'cause'} = Invalid;
    $cdrom->_status(to => $tParms{'from'});
    return;
  }

  Trace::debug('Ripping track: %d', $trackIndex) if (Trace::check(Trace::DEBUG3));

  my $cmd = Command->new(Shell => sprintf('cdparanoia -d %s %d %s', $cdrom->{'device'}, $trackIndex, Command::quotemeta($file)));
  if (!$cmd || !$cmd->run) {
    Trace::error('Cannot execute rip command');
    $cdrom->{'cause'} = RipError;
    $cdrom->{'attributes'}->{'diagnostic'} = 'Command returned: '.$cmd->return."\n".$cmd->output;
    $cdrom->{'attributes'}->{'parms'} = $parms;
    Trace::debug($cmd->return) if (Trace::check(Trace::DEBUG3));
    Trace::debug($cmd->output) if (Trace::check(Trace::DEBUG3));
    $cdrom->_status(to => $tParms{'from'});
    return;
  }

  $cdrom->{'trackIndex'} = $trackIndex;
  $cdrom->{'ripping'} = {};
  $cdrom->{'ripping'}->{'command'} = $cmd;
  $cdrom->{'ripping'}->{'file'} = $file;
  $cdrom->{'ripping'}->{'percentageComplete'} = 0;
  $cdrom->{'ripping'}->{'format'} = Format::Wav->new;
  $cdrom->{'ripping'}->{'format'}->duration($track->duration);
  $cdrom->{'ripping'}->{'parms'} = $parms;
  $cdrom->{'ripping'}->{'start'} = time;
  $cdrom->{'ripping'}->{'timeLimit'} = ($track->duration * $cdrom->{'timeLimitForSixty'}) / 60;

  Trace::debug('Giving %d seconds', $cdrom->{'ripping'}->{'timeLimit'}) if Trace::check(Trace::DEBUG6);

  $cdrom->at( name => '_ripMonitor', to => $cdrom, time => time + $cdrom->{'ripMonitorInterval'} );

  $cdrom->{'state'} = Ripping;
  $cdrom->{'cause'} = Message;
  $cdrom->_status();
}

sub _ripMonitor {
  my $cdrom = shift;
  my %tParms = @_;
  my $parms = $tParms{'parms'};

  my $cd = $cdrom->{'cd'};
  return if (!defined($cd));
  return if ($cdrom->{'state'} ne Ripping);

  my $cmd = $cdrom->{'ripping'}->{'command'};
  if (($cmd->state eq Command::RUNNING) && (time > ($cdrom->{'ripping'}->{'start'} + $cdrom->{'ripping'}->{'timeLimit'}))) {
    Trace::debug('Exceeded time limit') if Trace::check(Trace::DEBUG4);
    $cdrom->{'cause'} = RipStuck;
    $cdrom->{'attributes'}->{'time'} = int($cdrom->{'ripping'}->{'timeLimit'});
    $cdrom->_stop();
  }
  elsif ($cmd->state eq Command::RUNNING) {
    # Check up again later
    $cdrom->at( name => '_ripMonitor', to => $cdrom, time => time + $cdrom->{'ripMonitorInterval'});
    # Work out percentage complete
    my $pc = 100 * ((-s $cdrom->{'ripping'}->{'file'}) / $cdrom->{'ripping'}->{'format'}->expFileSize);
    $cdrom->{'ripping'}->{'percentageComplete'} = int($pc);
    Trace::debug('%d%% complete', $cdrom->{'ripping'}->{'percentageComplete'}) if (Trace::check(Trace::DEBUG3));
  }
  else {
    # Command is finished
    $cdrom->{'state'} = Stopped;
    if (($cmd->state eq Command::FINISHED) && 
	defined($cmd->return) &&
	($cmd->return == 0)) {
      # Possible good return, check size
      my $fileSize = -s $cdrom->{'ripping'}->{'file'};
      my $expSize = $cdrom->{'ripping'}->{'format'}->expFileSize;
      if ((($expSize - $fileSize) / $expSize) > 0.005) {
	# Size is more than -0.5% off
	Trace::error('Rip is %d bytes should be closer to %d bytes', $fileSize, $expSize);
	$cdrom->{'cause'} = RipError;
	$cdrom->{'attributes'}->{'diagnostic'} = Trace::lastError;
	$cdrom->{'attributes'}->{'parms'} = $parms;
      }
      else {
	$cdrom->{'cause'} = End;
      }
    }
    else {
      # Bad return
      $cdrom->{'cause'} = RipError;
      $cdrom->{'attributes'}->{'diagnostic'} = 'Command returned: '.$cmd->return."\n".$cmd->output;
      $cdrom->{'attributes'}->{'parms'} = $parms;
      Trace::dump($cdrom->{'attributes'}->{'diagnostic'});
    }
    delete $cdrom->{'ripping'};
  }
  $cdrom->_status();
}

# stop() - Stop a play or rip. 

sub stop {
  my $cdrom = shift;
  my %tParms = @_;
  $cdrom->_addToMonitors($tParms{'from'});

  if (($cdrom->{'state'} eq Stopped) || ($cdrom->{'state'} eq NoCd)) {
    $cdrom->{'cause'} = Ignored;
    $cdrom->_status(to => $tParms{'from'});
    return;
  }
  if ($cdrom->_stop()) {
    $cdrom->{'cause'} = Message;
  }
  $cdrom->_status();
}

sub _stop {
  my $cdrom = shift;

  if ($cdrom->{'state'} eq Ripping) {
    $cdrom->{'ripping'}->{'command'}->kill('TERM');
    delete $cdrom->{'ripping'};
  }
  elsif (($cdrom->{'state'} eq Playing) || ($cdrom->{'state'} eq Paused)) {
    if (!$cdrom->{'handle'}->open($cdrom->{'device'}, O_NONBLOCK|O_RDONLY)) {
      Trace::error('Cannot open cdrom: '.$!);
      $cdrom->{'attributes'}->{'diagnostic'} = Trace::lastError;
      $cdrom->{'cause'} = Error;
      return 0;
    }
    if (!defined(ioctl($cdrom->{'handle'}, CDROMSTOP, 1))) {
      Trace::error("Cannot stop Cd: $!");
      $cdrom->{'attributes'}->{'diagnostic'} = Trace::lastError;
      $cdrom->{'cause'} = Error;
      $cdrom->{'handle'}->close;
      return 0;
    }
    $cdrom->{'handle'}->close;
  }
  $cdrom->{'state'} = Stopped;
  return 1;
}

# eject()

sub eject {
  my $cdrom = shift;
  my %tParms = @_;
  my $parms = $tParms{'parms'};
  $cdrom->_addToMonitors($tParms{'from'});

  if (($cdrom->{'state'} eq NoCd) && defined($cdrom->{'tray'}) && ($cdrom->{'tray'} eq TrayOpen)) {
    $cdrom->{'cause'} = Ignored;
    Trace::debug('Eject ignored, already open') if Trace::check(Trace::DEBUG4);
    $cdrom->_status(to => $tParms{'from'});
    return;
  }

  if (($cdrom->{'state'} eq Playing) || ($cdrom->{'state'} eq Paused) || ($cdrom->{'state'} eq Ripping)) {
    Trace::debug('Stop prior to eject') if Trace::check(Trace::DEBUG4);
    if (!$cdrom->_stop()) {
      $cdrom->_status();
      return;
    }
  }

  if (!$cdrom->{'handle'}->open($cdrom->{'device'}, O_NONBLOCK|O_RDONLY)) {
    Trace::error('Cannot open cdrom: '.$!);
    $cdrom->{'attributes'}->{'diagnostic'} = Trace::lastError;
    $cdrom->{'cause'} = Error;
    $cdrom->_status(to => $tParms{'from'});
    return;
  }

  if (($cdrom->{'lock'} && !defined(ioctl($cdrom->{'handle'}, CDROM_LOCKDOOR, 0))) ||
      !defined(ioctl($cdrom->{'handle'}, CDROMEJECT, 1)) ||
      ($cdrom->{'lock'} && !defined(ioctl($cdrom->{'handle'}, CDROM_LOCKDOOR, 1)))) {
    Trace::error("Cdrom error: $!");
    $cdrom->{'attributes'}->{'diagnostic'} = Trace::lastError;
    $cdrom->{'cause'} = Error;	
    $cdrom->_status(to => $tParms{'from'});	
    $cdrom->{'handle'}->close;
    return;
  }
  $cdrom->{'handle'}->close;
  $cdrom->{'state'} = NoCd;
  $cdrom->{'cause'} = Message;
  $cdrom->{'tray'} = TrayOpen;
  if ($cdrom->{'cd'}) {
    $cdrom->{'cd'}->demonitor;
    $cdrom->{'artist'}->demonitor;
    map { $_->demonitor } @{ $cdrom->{'tracks'} };
    delete $cdrom->{'cd'};
    delete $cdrom->{'seek'};
  }

  $cdrom->_status();
}

# close()

sub close {
  my $cdrom = shift;
  my %tParms = @_;

  $cdrom->_addToMonitors($tParms{'from'});
  return $cdrom->_close($tParms{'from'});
}

sub _close {
  my $cdrom = shift;
  my $from = shift;
  my $tracks = {};

  if (($cdrom->{'state'} eq Ripping) || ($cdrom->{'state'} eq Playing) || ($cdrom->{'state'} eq Paused) || ($cdrom->{'state'} eq Stopped)) {
    Trace::debug('Already closed') if Trace::check(Trace::DEBUG4);
    $cdrom->{'cause'} = Ignored;
    $cdrom->_status(to => $from) if $from;
    return;
  }

  # Allow close when `tray' indicates it is already closed in case a
  # manual close has occured. Its no harm anyway.

  # (($cdrom->{'state'} eq NoCd) && defined($cdrom->{'tray'}) && 
  # ($cdrom->{'tray'} eq TrayClosed))) 

  if (!$cdrom->{'handle'}->open($cdrom->{'device'}, O_NONBLOCK|O_RDONLY)) {
    Trace::error('Cannot open cdrom: '.$!);
    $cdrom->{'attributes'}->{'diagnostic'} = Trace::lastError;
    $cdrom->{'cause'} = Error;
    $cdrom->_status(to => $from) if $from;
    return;
  }
  if ($cdrom->{'lock'} && !defined(ioctl($cdrom->{'handle'}, CDROM_LOCKDOOR, 0))) {
    Trace::error("Cannot unlock: $!");
    $cdrom->{'attributes'}->{'diagnostic'} = Trace::lastError;
    $cdrom->{'cause'} = Error;
    $cdrom->_status(to => $from) if $from;
    $cdrom->{'handle'}->close;
    return;
  }
  $cdrom->{'handle'}->close;

  # A blocking call will close the door.

  $cdrom->{'tray'} = TrayClosed; 
  if (!$cdrom->{'handle'}->open($cdrom->{'device'}, O_RDONLY)) {
    Trace::debug('Blocking open failed: %s', $!) if Trace::check(Trace::DEBUG4);
    # 2.4.x Linux kernels changed things, try a non blocking and an specific close
    if (!$cdrom->{'handle'}->open($cdrom->{'device'}, O_NONBLOCK | O_RDONLY) || !defined(ioctl($cdrom->{'handle'}, CDROMCLOSETRAY, 1))) {
      # Failure here indicates that there is perhaps no CD in the drive
      Trace::debug('No CD in drive?: %s', $!) if Trace::check(Trace::DEBUG4);
      $cdrom->{'cause'} = Message;
      $cdrom->_status();
      return;
    }
  }
  $cdrom->{'handle'}->close;

  if (!$cdrom->{'handle'}->open($cdrom->{'device'}, O_NONBLOCK|O_RDONLY)) {
    Trace::error('Cannot open cdrom: '.$!);
    $cdrom->{'attributes'}->{'diagnostic'} = Trace::lastError;
    $cdrom->{'cause'} = Error;
    $cdrom->_status();
    return;
  }
  if ($cdrom->{'lock'} && !defined(ioctl($cdrom->{'handle'}, CDROM_LOCKDOOR, 1))) {
    Trace::error("Cannot lock: $!");
    $cdrom->{'attributes'}->{'diagnostic'} = Trace::lastError;
    $cdrom->{'cause'} = Error;
    $cdrom->_status();
    $cdrom->{'handle'}->close;
    return;
  }

  # Get the first and last track numbers

  my $tocHdr = pack($cdromTocHdrStruct, 1, 1);
  if (!defined(ioctl($cdrom->{'handle'}, CDROMREADTOCHDR, $tocHdr))) {
    # Failure here indicates that there is perhaps no CD in the drive
    Trace::debug('Header read failed: %s', $!) if Trace::check(Trace::DEBUG4);
    $cdrom->{'handle'}->close;
    $cdrom->{'cause'} = Message;
    $cdrom->_status();
    return;
  }
  (my $firstIndex, my $lastIndex) = unpack($cdromTocHdrStruct, $tocHdr);

  # Get the MSF value for the leadout track.

  my $tocEntry = pack($cdromTocEntryStruct, Info::Track::LEADOUTTN, 0, CDROM_MSF, 0, 0, 0, 0, 0);
  if (!defined(ioctl($cdrom->{'handle'}, CDROMREADTOCENTRY, $tocEntry))) {
    Trace::error("Cannot read table of contents entry: $!");
    $cdrom->{'attributes'}->{'diagnostic'} = Trace::lastError;
    $cdrom->{'cause'} = Error;
    $cdrom->_status();
    $cdrom->{'handle'}->close;
    return;
  }
  my @tocEntryArray = unpack($cdromTocEntryStruct, $tocEntry);

  $tracks->{ Info::Track::LEADOUTTN } = 
    { data => ($tocEntryArray[1] & (CDROM_DATA_TRACK << 4)),
      lba => (($tocEntryArray[4] * 60) + $tocEntryArray[5]) * 75 + $tocEntryArray[6],
      index => Info::Track::LEADOUTTN };

  # Get CD-Plus/Extra info.

  my $ms = pack($cdromMultiSessionStruct, 0, 0, 0, 0, 0, CDROM_MSF);
  if (!defined(ioctl($cdrom->{'handle'}, CDROMMULTISESSION, $ms))) {
    Trace::error("Cannot get multisession: $!");
    $cdrom->{'attributes'}->{'diagnostic'} = Trace::lastError;
    $cdrom->{'cause'} = Error;
    $cdrom->_status();
    $cdrom->{'handle'}->close;
    return;
  }

  my @msArray = unpack($cdromMultiSessionStruct, $ms);
  $tracks->{ Info::Track::MSTN } = 
    { data => $msArray[4],
      lba => (($msArray[0] * 60) + $msArray[1]) * 75 + $msArray[2],
      index => Info::Track::MSTN };

  # Get the LBA value for each track on the CD.

  for (my $index = $firstIndex; $index <= $lastIndex; $index++) {

    my $tocEntry = pack($cdromTocEntryStruct, $index, 0, CDROM_MSF, 0, 0, 0, 0, 0);
    if (!defined(ioctl($cdrom->{'handle'}, CDROMREADTOCENTRY, $tocEntry))) {
      Trace::error("Cannot read table of contents entry: $!");
      $cdrom->{'attributes'}->{'diagnostic'} = Trace::lastError;
      $cdrom->{'cause'} = Error;
      $cdrom->_status();
      $cdrom->{'handle'}->close;
      return;
    }
    my @tocEntryArray = unpack($cdromTocEntryStruct, $tocEntry);

    $tracks->{ $index } = 
      { data => ($tocEntryArray[1] & (CDROM_DATA_TRACK << 4)),
	lba => (($tocEntryArray[4] * 60) + $tocEntryArray[5]) * 75 + $tocEntryArray[6],
	index => $index };
  }

  $cdrom->{'handle'}->close;

  if (!keys(%{ $tracks })) {
    # No tracks, Unexpected
    Trace::error('No tracks on CD');
    $cdrom->{'attributes'}->{'diagnostic'} = 'No tracks on CD';
    $cdrom->{'cause'} = Error;
    $cdrom->_status();
    return;
  }

  Trace::debug('Found Cd with %d tracks', scalar keys(%{ $tracks })) if Trace::check(Trace::DEBUG4);

  my $cd = Info::Cd->find(owner => $cdrom->{'repo'}, trackinfo => $tracks, create => 1);
  if (!defined($cd)) {
    $cdrom->{'attributes'}->{'diagnostic'} = 'Error creating Cd:'."\n";
    $cdrom->{'attributes'}->{'diagnostic'} .= Trace::lastError;
    $cdrom->{'cause'} = Error;
    $cdrom->_status();
    return;
  }

  $cdrom->{'state'} = Stopped;
  $cdrom->{'cause'} = Message;
  if ($cdrom->{'cd'}) {
    $cdrom->{'cd'}->demonitor;
    $cdrom->{'artist'}->demonitor;
    map { $_->demonitor } @{ $cdrom->{'tracks'} };
  }

  # Hold a monitored version of a cd to ensure that status messages
  # have the most up to date value. Hold a monitored version of artist
  # and tracks because cd->demonitor will not 'demonitor' them.

  $cdrom->{'artist'} = $cd->artist->monitor;
  my @ctracks = $cd->tracks;
  @ctracks = map { $_->monitor } @ctracks;
  $cdrom->{'tracks'} = \@ctracks;
  $cdrom->{'cd'} = $cd->monitor;

  $cdrom->_status();
  if (!$cd->known && Cddb::configured) {
    $cd->remoteLookup(forking => 1, chooseExact => 1);
  }
}

# _autoDetect - Auto detection of Cds following manual Cdrom open and close.

sub _autoDetect {
  my $cdrom = shift;

  return if (!$cdrom->{'autoDetect'});

  if ($cdrom->{'device'}) {
    if (!$cdrom->{'handle'}->open($cdrom->{'device'}, O_NONBLOCK|O_RDONLY)) {
      Trace::error('Cannot open cdrom: '.$!);
      return;
    }

    my $tocHdr = pack($cdromTocHdrStruct, 1, 1);
    my $tocEntry = pack($cdromTocEntryStruct, Info::Track::LEADOUTTN, 0, CDROM_MSF, 0, 0, 0, 0, 0);
    if (defined(ioctl($cdrom->{'handle'}, CDROMREADTOCHDR, $tocHdr)) &&
	defined(ioctl($cdrom->{'handle'}, CDROMREADTOCENTRY, $tocEntry))) {
      $cdrom->{'handle'}->close;

      # Header is read to ensure drive was not ejected, leadout
      # entry to compare against current Cd.

      my @tocEntryArray = unpack($cdromTocEntryStruct, $tocEntry);
      my $lba = (($tocEntryArray[4] * 60) + $tocEntryArray[5]) * 75 + $tocEntryArray[6];
      if (!defined($cdrom->{'cd'}) || ($cdrom->{'cd'}->track(Info::Track::LEADOUTTN)->lba != $lba)) {
	# A cd has been manually inserted
	Trace::debug('Cd manually inserted') if Trace::check(Trace::DEBUG3);
	$cdrom->_close;
      }
    }
    else {
      $cdrom->{'handle'}->close;

      if ($cdrom->{'state'} ne NoCd) {
	# Cdrom has been manually ejected
	Trace::debug('Cd manually ejected') if Trace::check(Trace::DEBUG3);
	$cdrom->_stop if ($cdrom->{'state'} eq Ripping);
	$cdrom->{'tray'} = TrayOpen;
	$cdrom->{'state'} = NoCd;
	if ($cdrom->{'cd'}) {
	  $cdrom->{'cd'}->demonitor;
	  $cdrom->{'artist'}->demonitor;
	  map { $_->demonitor } @{ $cdrom->{'tracks'} };
	  delete $cdrom->{'cd'};
	  delete $cdrom->{'seek'};
	}
	$cdrom->_status();
      }
    }
  }
  # Check again in 2 seconds
  $cdrom->at(name => '_autoDetect', to => $cdrom, time =>  time + 2);
}

# _lock - Change the lock state of the drive

sub _lock {
  my $cdrom = shift;

  if ($cdrom->{'device'}) {
    if (!$cdrom->{'handle'}->open($cdrom->{'device'}, O_NONBLOCK|O_RDONLY)) {
      Trace::error('Cannot open cdrom: '.$!);
      return;
    }
    if (!defined(ioctl($cdrom->{'handle'}, CDROM_LOCKDOOR, $cdrom->{'lock'} ? 1 : 0))) {
      Trace::error("Cannot %s: $!", $cdrom->{'lock'} ? 'lock' : 'unlock');
    }
    $cdrom->{'handle'}->close;
  }
  return;
}

# _addToMonitors - Anything that sends us a command should be informed
# of state or spool changes.

sub _addToMonitors {
  my $cdrom = shift;
  my $monitor = shift;

  my $monitors = $cdrom->{'monitors'};
  if (!defined($monitors->{ $monitor->unique })) {
    Trace::debug('New state monitor: %s', $monitor->unique) if (Trace::check(Trace::DEBUG3));
    $monitors->{ $monitor->unique } = $monitor;
  }
  return 1;
}

1;
