# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.8 $

package Command;

use strict;
use IO::File qw();
use IO::Select qw ();
use POSIX qw(:fcntl_h);
use POSIX qw( _exit );
use Trace;

# Default remote exec command
sub RSH { 'rsh'; }

# Command not started yet
sub IDLE {1;}
# Command in progress
sub RUNNING {2;}
# Command finished, return codes are valid
sub FINISHED {3;}
# Command failed to complete, return codes were not read
sub ERROR {4;}
# Command was killed, return codes were not read
sub KILLED {5;}
# Command timed out, return codes were not read
sub TIMEDOUT {6;}

# new() - create a command to run `Shell' or `Function'. Can be
# executed on a different machine with `Remote'. A `Timeout' value may
# be specified.

sub new {
    my $type = shift;
    my %params = @_;
    my $command = {};

    if (defined($params{'Shell'})) {
	$command->{'Shell'} = $params{'Shell'};

	if (defined($params{'Remote'})) {
	    # Capture the exit status (super escape needed, levels of interp)
	    $command->{'Shell'} .= ' 2>&1; echo __REMOTEEXIT__:\\\\\\\\\\\\\\$?:';
	    # Remote command gets wrapped with a sh -c \\\\\\"...\\\\\\"
	    $command->{'Shell'} = 'sh -c \\\\\\"'.$command->{'Shell'}.'\\\\\\"';
	    # And a ssh/rsh \"...\"
	    my $rsh = $params{'Rsh'};
	    if (!defined($rsh)) {
		$rsh = Command::RSH;
	    }
	    $command->{'Shell'} = $rsh.' '.$params{'Remote'}.' \"'.$command->{'Shell'}.'\"';
	}

	if (defined($params{'STDOUT'}) && !$params{'STDOUT'}) {
	    $command->{'Shell'} .= ' > /dev/null';
	}

	# Capture the exit status
	$command->{'Shell'} .= ' 2>&1; echo __EXIT__:\$?:';

	# Commands get wrapped with a sh -c "..."
	$command->{'Shell'} = 'sh -c "'.$command->{'Shell'}.'"';

	Trace::debug('Full command: %s', $command->{'Shell'}) if Trace::check(Trace::DEBUG3);
    }
    elsif (defined($params{'Function'})) {	
	$command->{'Function'} = $params{'Function'};
    }
    else {
	Trace::debug('Function or shell command required') if Trace::check(Trace::DEBUG3);
	return 0;
    }

    if (defined($params{'Timeout'})) {

	# Synchronous
	$command->{'Timeout'} = $params{'Timeout'};
    }
    else {
	
	# Asynchronous
	$command->{'Timeout'} = 0;
    }

    $command->{'Output'} = '';
    $command->{'Pid'} = 0;
    $command->{'State'} = IDLE;

    bless $command, $type;
    return ($command);
}

# run() - execute the command. Synchronous commands will block until
# they exit or the Timeout is reached. Returns 1 on success, 0
# otherwise.

sub run {
    my $command = shift;

    if ($command->{'State'} != IDLE) {
	Trace::error('Already running');
	return 0;
    }

    my $openString = '';
    if (defined($command->{'Shell'})) {
	$openString = $command->{'Shell'}.'|';
    }
    else {
	$openString = '-|';
    }

    Trace::debug('Started') if Trace::check(Trace::DEBUG3);
    $command->{'Start Time'} = time;

    my $fh = IO::File->new;
    my $pid = $fh->open($openString);
    if (!defined($pid)) {
	Trace::error('Could not fork');
	return 0;
    }

    if ($pid == 0) {
	# Child, so this must be a function command
	select(STDOUT); $| = 1;
	my $rc = &{ $command->{'Function'} };
	printf(STDOUT "__EXIT__:$rc:\n");
	_exit(0);
    }

    $command->{'State'} = RUNNING;

    $fh->fcntl(F_SETFL, O_NONBLOCK);
    $command->{'Fh'} = $fh;

    if ($command->{'Timeout'}) {

	# Synchronous, wait first for some output, an EXIT at least is
	# expected.
	my $sel = IO::Select->new($fh);
	my $runTime;
        TIMEOUT: while (($runTime = (time - $command->{'Start Time'})) < $command->{'Timeout'}) {
	    Trace::debug('Trying to read for %d seconds', $command->{'Timeout'} - $runTime) if Trace::check(Trace::DEBUG6);
	    if ($sel->can_read($command->{'Timeout'} - $runTime)) {
		$command->output;
	    }
	    last TIMEOUT if ($command->{'State'} == FINISHED);
	}
	
	# If it is still running kill it.
	if ($command->{'State'} == RUNNING) {
	    $fh->close;
	    $command->kill;
	    $command->{'State'} = TIMEDOUT;
	    Trace::debug('Killed due to timeout') if Trace::check(Trace::DEBUG3);
	}
    }
    else {

	# Asynchronous, return immediately
	$command->{'Pid'} = $pid;
    }
    return (1);
}

# output - Try and read the command output. If the command filehandle
# is blocking (default) this will block until the command exits.

sub output {
    my $command = shift;

    if ($command->{'State'} != RUNNING) {
	return ($command->{'Output'});
    }

    my $fh = $command->{'Fh'};
    my $output;
    my $bytesRead = 0;
    my $thisRead;
    my $timeout = 0;
    do {
	$thisRead = sysread $fh, $output, 1024;
	if (!defined($thisRead)) {
	    Trace::debug('Undefined sysread') if Trace::check(Trace::DEBUG6);
	}
	elsif ($thisRead < 0) {
	    Trace::debug('Negative sysread: %s', $!) if Trace::check(Trace::DEBUG6);
	}
	elsif ($thisRead) {
	    Trace::debug('Read %d bytes', $thisRead) if Trace::check(Trace::DEBUG6);
	    if ($output =~ m/(.*)__EXIT__:(\d+):(.*)/s) {
		$command->{'Return'} = $2;
		$output = $1.$3;
	    }
	    if ($output =~ m/(.*)__REMOTEEXIT__:(\d+):(.*)/s) {
		$command->{'RemoteReturn'} = $2;
		$output = $1.$3;
	    }
	    $command->{'Output'} .= $output;
	}
	else {
	    Trace::debug('Read %d bytes', $thisRead) if Trace::check(Trace::DEBUG6);
	}

	# Break out for timeouts
	if ($command->{'Timeout'} && ((time - $command->{'Start Time'}) >= $command->{'Timeout'})) {
	    $timeout = 1;
	}
    } while (!$timeout && (defined($thisRead) && ($thisRead > 0)));

    if (defined($command->{'Return'})) {
	$command->{'State'} = FINISHED;
	Trace::debug('Finished: %d', $command->{'Return'}) if Trace::check(Trace::DEBUG3);
	$fh->close;
    }

    return ($command->{'Output'});
}

sub return {
    my $command = shift;

    return ($command->{'Return'});
}

sub remoteReturn {
    my $command = shift;

    return ($command->{'RemoteReturn'});
}

sub state {
    my $command = shift;
    
    # Scan the ouput looking for a new state
    $command->output;
    return ($command->{'State'});
}

sub fh { $_[0]->{'Fh'}; }

sub wait {
    my $command = shift;

    # Block on the output
    $command->{'Fh'}->fcntl(F_SETFL, 0);
    $command->output;
    return 1;
}

sub kill {
    my $command = shift;
    my $signal = shift;

    # Need to kill the PID of the shell spawned process

    my $pid = $command->pid;
    if (!defined($pid)) {
	$command->{'State'} = KILLED;
	Trace::debug('Already dead?') if Trace::check(Trace::DEBUG3);
	$command->{'Fh'}->close;
	return;
    }

    if (defined($signal)) {
	kill $signal, $pid;
    }
    else {
	kill $pid;
    }
    $command->{'State'} = KILLED;
    Trace::debug('Killed by user') if Trace::check(Trace::DEBUG3);
    $command->{'Fh'}->close;
    return;
}

sub duration {
    my $command = shift;
    return (time - $command->{'Start Time'});
}

sub pid { 
    my $command = shift;

    # Need to find the PID of the shell spawned process

    # PPID   PID  PGID   SID TTY

    #    1  1168  1168   379 pts/1  0:00 Encoder
    # 1168  1175  1168   379 pts/1  0:00  \_ sh -c 8hz-mp3 -b 192 \/stor
    # 1175  1176  1168   379 pts/1  6:19      \_ /home/gavin/bin/8hz-mp3
  
    if (!defined($command->{'RealPid'})) {
	my @children = Command::children($command->{'Pid'});
	if (!@children || ($#children == -1)) {
	    Trace::debug('Dead?') if Trace::check(Trace::DEBUG3);
	    return undef;
	}
	$command->{'RealPid'} = $children[0];
    }
    return $command->{'RealPid'};
}

# CLASS function

# quotemeta() - need to \\\ the $'s and the "

sub quotemeta {
    my $str = shift;
    $str =~ s/\$/\\\$/g; 
    $str =~ s/\"/\\\"/g; 
    return CORE::quotemeta($str);    
}

# children

sub children {
    if ($#_ != 0) {
	return undef;
    }
    my $pid;
    if (ref($_[0]) eq 'Command') {
	$pid = $_[0]->pid;
    }
    else {
	$pid = $_[0];
    }
    return undef if (!defined($pid));
    
    my $cmd = Command->new(Shell => 'ps xj', Timeout => 60);
    if (!$cmd || !$cmd->run || ($cmd->state != Command::FINISHED) || (defined($cmd->return) && ($cmd->return != 0))) {
	Trace::error('Cannot execute PS to gather children');
	Trace::error($cmd->output);
	return undef;
    }

    my @children = ();
    my @procs = split (/\n/, $cmd->output);
    foreach my $proc (@procs) {
	if ($proc =~ /^\s*(\d*)\s+(\d*)\s+/) {
	    if (int($1) == $pid) {
		Trace::debug('Command %d spawned %d', $pid, $2) if Trace::check(Trace::DEBUG3);
		push @children, $2;
	    }
	}
    }
    return @children;
}

1;
