# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.16 $

package Controller;

# Controller - Manages ripping and encoding, auto rip and auto
# play. Particularly useful as a single point to deal with the dynamic
# nature of host availability.

# See docs/controllerRippingSM.dot
# See docs/controllerEncodingSM.dot 

use Thing;
@ISA = qw( Thing );
use ControllerH;
use RepoH;
use CdromH;
use EncoderH;
use Format::Wav;
use Format::Mp3;
use Info::Cd;
use Info::Track;
use Utility;
use strict;

# Extra Internal States, for encoders and rippers.

sub Playing    { '__Playing'; }
sub Checking   { '__Checking'; }
sub OutOfSpace { '__OutOfSpace'; }
sub Hold       { '__Hold'; }

sub new {
    my $proto = shift;
    my $type = ref($proto) || $proto;
    my $controller = $type->SUPER::new(@_);
    if ($controller->error) {
	return $controller;
    }
    
    $controller->register(name => Cdrom::Status, handler => 'cdromStatus');
    $controller->register(name => 'info', handler => 'repoInfo');
    $controller->register(name => 'stored', handler => 'repoStored');
    $controller->register(name => 'failure', handler => 'repoFailure');
    $controller->register(name => Encoder::Status, handler => 'encoderStatus');

    $controller->register(name => EncodeState, handler => 'encodeState');
    $controller->register(name => EncodeHalt, handler => 'encodeHalt');
    $controller->register(name => EncodeResume, handler => 'encodeResume');
    $controller->register(name => RipState, handler => 'ripState');
    $controller->register(name => RipperRip, handler => 'ripperRip');
    $controller->register(name => RipperAdd, handler => 'ripperAdd');
    $controller->register(name => RipperRemove, handler => 'ripperRemove');
    $controller->register(name => Thing::Info::Update, handler => 'objectUpdate', last => 1);

    $controller->register(name => Thing::ConfigureRequest, handler => '_configure');
    $controller->register(name => Thing::ShutdownRequest, handler => '_shutdown');

    $controller->register(name => '_poll', handler => '_poll');
    $controller->register(name => '_encoderAwake', handler => '_encoderAwake');

    my $rip = $controller->{'rip'} = {};
    # Format of ripped files
    $rip->{'format'} = Format::Wav->new(samplesPerSecond => 44100, bitsPerSample => 16, channels => 2);
    $rip->{'rippers'} = {};
    $rip->{'monitors'} = {};
    $rip->{'storePending'} = [];

    my $encode = $controller->{'encode'} = {};
    # Two lists are maintained of the encode jobs, there is also a
    # virtual list, the current work underway.
    $encode->{'tracks'} = [];
    $encode->{'storePending'} = [];
    # Overall encoding state is initially active
    $encode->{'state'} = Running;
    $encode->{'monitors'} = {};
    $encode->{'encoders'} = {};
    $encode->{'rates'} = [];
    $encode->{'repoQueried'} = 0;

    $controller->{'configured'} = 0;
    $controller->{'pollScheduled'} = 0;

    # A global flag to indicate out of space condition. Allows us to
    # back off ripping and favour encoding.
    $controller->{'outOfSpace'} = 0;
    # Once out of space is hit ripping is cut back to one at a time
    # until more space is made available through configuration or
    # manually and the user requests ripperRip. 
    $controller->{'outOfSpaceHobbled'} = 0;

    return $controller;
}

# _configure()

sub _configure {
    my $controller = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};

    my $errorS = sub { if (defined($parms->{'error'})) { $parms->{'error'} .= "\n".$_[0]; } else { $parms->{'error'} = $_[0]; } };

    Trace::debug('Configuration request') if Trace::check(Trace::DEBUG2);

    $controller->{'configured'} = 0;

    if (!defined($parms->{'repo'})) {
	Trace::mandatory('repo');
	&$errorS(Trace::lastError);
    }    
    if (defined($controller->{'repo'}) && !Thing::Handle::cmpNameHost($controller->{'repo'}, $parms->{'repo'})) {
	$controller->demonitor($controller->{'repo'});
    }
    $controller->{'repo'} = $controller->monitor($parms->{'repo'});

    if (!defined($parms->{'tmpdirs'}) || ($#{ $parms->{'tmpdirs'} } == -1)) {
	Trace::error('No temporary directories configured. Without at least one configured no ripping or encoding will be done.');
	# The repo now builds the temp dirs from its music dirs so we don't want to double report the error
	# $parms->{'error'} = '...';
    }
    else {
	if (!$controller->_initTmp(tmpdirs => $parms->{'tmpdirs'})) {
	    &$errorS(Trace::lastError);
	}
    }

    if (defined($parms->{'db'}) && !Info::Persist::configure(db => $parms->{'db'}->{'local'})) {
	&$errorS(Trace::lastError);
    }

    my $rip = $controller->{'rip'};
    my $encode = $controller->{'encode'};
    my @pendingDeletes = ();

    # Stop all ripping

    foreach my $ripper (values(%{ $rip->{'rippers'} })) {
	Trace::debug('Deconfiguring ripper %s', _describeRipper($ripper)) if Trace::check(Trace::DEBUG3);
	$controller->demonitor($ripper);
	if ($ripper->{'state'} eq Ripping) {
	    $controller->message(name => Cdrom::Stop, to => $ripper);
	    if (defined($ripper->{'file'})) {
		push @pendingDeletes, $ripper->{'file'};
	    }
	}
    }    

    # Stop all encoding

    foreach my $encoder (values(%{ $encode->{'encoders'} })) {
	Trace::debug('Deconfiguring encoder %s', _describeEncoder($encoder)) if Trace::check(Trace::DEBUG3);
	$controller->demonitor($encoder);
	if ($encoder->{'state'} eq Encoding) {
	    $controller->message(name => Encoder::Stop, to => $encoder);
	    if (defined($encoder->{'file'})) {
		push @pendingDeletes, $encoder->{'file'};
	    }
	}
    }    
    $encode->{'tracks'} = [];	# Forget about the previous encode job list.

    # Ignore status messages created by stops

    $rip->{'rippers'} = {};
    $encode->{'encoders'} = {};
    $controller->wait(event => sub { 0; }, timeout => 5);
    map { unlink $_ if (-e $_) } @pendingDeletes; 

    # Configure rippers

    my $rippers = {};
    if (defined($parms->{'cdroms'})) {
	foreach my $cdrom (@{ $parms->{'cdroms'} }) {	    
	    my $ripper;
	    my $name = $cdrom->{'handle'}->key;
	    Trace::debug('Config for (%s)', $name) if Trace::check(Trace::DEBUG3);
	    $ripper = $rippers->{ $name } = $cdrom->{'handle'};
	    $ripper->{'repoQueried'} = 0;
	    $controller->monitor($ripper);

	    if (!defined($cdrom->{'autoPlay'})) {
		$ripper->{'autoPlay'} = 0;
	    }
	    else {
		$ripper->{'autoPlay'} = $cdrom->{'autoPlay'};
	    }
	    if (!defined($cdrom->{'autoRip'})) {
		$ripper->{'autoRip'} = 0;
	    }
	    else {
		$ripper->{'autoRip'} = $cdrom->{'autoRip'};
	    }
	    
	    # Initial state - due to auto rip.
	    if ($ripper->{'autoRip'}) {
		$ripper->{'state'} = Idle;
	    }
	    else {
		$ripper->{'state'} = Halted;
	    }

	    # Get state from Cdrom
	    if (!$controller->message(name => Cdrom::State, to => $ripper)) {
		$ripper->{'state'} = Dead;
	    }
	}
    }
    $rip->{'rippers'} = $rippers;

    # Configure encoders

    my $encoders = {};	
    if (defined($parms->{'encoders'})) {
	foreach my $encoder (@{ $parms->{'encoders'} }) {
	    my $name = $encoder->{'handle'}->key;
	    Trace::debug('Config for (%s)', $name) if Trace::check(Trace::DEBUG3);
	    $encoder = $encoders->{ $name } = $encoder->{'handle'};
	    $controller->monitor($encoder);

	    # Initial state - due to overall state
	    if ($encode->{'state'} eq Running) {
		$encoder->{'state'} = Idle;
	    }
	    else {
		$encoder->{'state'} = Halted;
	    }

	    # Get state from Encoder
	    if (!$controller->message(name => Encoder::State, to => $encoder)) {
		$encoder->{'state'} = Dead;
	    }
	}
    }
    $encode->{'encoders'} = $encoders;
    if (scalar keys %{ $encode->{'encoders'} } == 0) {
	&$errorS('No encoders configured. Without at least one encoder configured no encoding will be performed.');
    }

    # Query encoding work

    if (($controller->{'repo'}->currentState eq Thing::Handle::Alive) && 
	$controller->message(name => 'query', to => $controller->{'repo'},
			     parms => { type => 'all', 'format' => $rip->{'format'}, _tag => 'init', }, )) {
	$encode->{'repoQueried'} = 1;
    }
    else {
	$encode->{'repoQueried'} = 0;
    }

    # Rates.

    if (defined($parms->{'rates'}) && ($#{ $parms->{'rates'} } != -1)) {
        $encode->{'rates'} = $parms->{'rates'};
    }
    else {
	&$errorS('No encode rates configured. Without at least one rate configured no encoding will be performed.');
	$encode->{'rates'} = [];
    }

    # Clear out of space conditions

    $controller->_clearOutOfSpace();
    $controller->_clearOutOfSpaceHobbled();

    $controller->{'configured'} = 1;
    return 1;
}

sub _shutdown {
    my $controller = shift;
    my $encode = $controller->{'encode'};
    my $rip = $controller->{'rip'};
    
    foreach my $encoder (values(%{ $encode->{'encoders'} })) {
	if (defined($encoder->{'file'})) {
	    unlink $encoder->{'file'} if (-e $encoder->{'file'});
	}
    }

    foreach my $ripper (values(%{ $rip->{'ripers'} })) {
	if (defined($ripper->{'file'})) {
	    unlink $ripper->{'file'} if (-e $ripper->{'file'});
	}
    }
    return 1;
}

# monitorChange

sub monitorChange {
    my $controller = shift;
    my $handle = shift;
    my $encode = $controller->{'encode'};
    
    Trace::debug('Monitor change for %s to %s', $handle->key, $handle->currentState) if Trace::check(Trace::DEBUG3);

    return if (!$controller->{'configured'});

    if (Thing::Handle::cmpNameHost($handle, $controller->{'repo'}) && ($handle->currentState eq Thing::Handle::Alive) && !$encode->{'repoQueried'}) {

	if (!$controller->message(name => 'query', to => $controller->{'repo'},
				  parms => { type => 'all', 'format' => $controller->{'rip'}->{'format'}, _tag => 'init', }, )) {
	    Trace::error('Need to contact repo for proper operation');
	    $encode->{'repoQueried'} = 0;
	}
	else {
	    $encode->{'repoQueried'} = 1;
	}
        return;
    }

    foreach my $encoder (values(%{ $encode->{'encoders'} })) {
	next if (!Thing::Handle::cmpNameHost($handle, $encoder));
	if ($encoder->currentState eq Thing::Handle::Dead) {
	    if ($encoder->{'state'} eq Encoding) {
		# This track may not have finished encoding, so put
		# the job back on the encode list and remove the
		# encode attempt file.
		$controller->_returnToEncodeSchedule($encoder->{'track'}, $encoder->{'format'}->rate);
		unlink $encoder->{'file'} if (-e $encoder->{'file'});
		$controller->_encodeMachine();
	    }
	    if ($encoder->{'state'} ne Dead) {
		Trace::debug('Encoder %s is dead', _describeEncoder($encoder)) if Trace::check(Trace::DEBUG4);
		$encoder->{'state'} = Dead;
		$controller->_encodeStatus();
	    }
	}
	elsif (($encoder->previousState ne Thing::Handle::Alive) && ($encoder->currentState eq Thing::Handle::Alive)) {
	    Trace::debug('Encoder %s is alive', _describeEncoder($encoder)) if Trace::check(Trace::DEBUG4);
	    $controller->message(name => Encoder::State, to => $encoder);
	}
	return;
    }

    my $rip = $controller->{'rip'};
    foreach my $ripper (values(%{ $rip->{'rippers'} })) {
	next if (!Thing::Handle::cmpNameHost($handle, $ripper));
	if ($ripper->currentState eq Thing::Handle::Dead) {
	    if ($ripper->{'state'} eq Ripping) {
		unlink $ripper->{'file'} if (-e $ripper->{'file'});
	    }
	    if ($ripper->{'state'} ne Dead) {
		Trace::debug('Ripper %s is dead', _describeRipper($ripper)) if Trace::check(Trace::DEBUG4);
		delete $ripper->{'cd'};
		delete $ripper->{'tracks'};
		$ripper->{'state'} = Dead;
	    }
	}
	elsif (($ripper->previousState ne Thing::Handle::Alive) && ($ripper->currentState eq Thing::Handle::Alive)) {
	    Trace::debug('Ripper %s is alive', _describeRipper($ripper)) if Trace::check(Trace::DEBUG4);
	    $controller->message(name => Cdrom::State, to => $ripper);
	}
	return;
    }
    return 1;
}

# objectUpdate - Info object has changed.

sub objectUpdate {
    my $controller = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};
    my $ioid = $parms->{'ioid'};
    my $object = $controller->infoFind(ioid => $ioid);

    if (!$object) {
	Trace::debug('Not monitoring: %s', $ioid) if (Trace::check(Trace::DEBUG3));
	return;
    }

    my $rip = $controller->{'rip'};

    if ($object->isa('Info::Album')) {
	foreach my $ripper (values(%{ $rip->{'rippers'} })) {
	    next if (!$ripper->{'cd'});
	    next if ($ripper->{'cd'}->ioid ne $object->ioid); # For another cd/album
	    if (!$ripper->{'repoQueried'} && $ripper->{'cd'}->known) {
		# Automatically query the need to rip this Cd but only if it is known.
		$controller->_repoQuery($ripper);
	    }
	}
    }
    return 1;
}

# ENCODING

# encodeState() - Request to monitor or stop monitoring the state of the encoding machine.

sub encodeState {
    my $controller = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};

    my $encode = $controller->{'encode'};
    my $monitors = $encode->{'monitors'};

    if (defined($parms->{'monitor'})) {
	if (!defined($monitors->{ $tParms{'from'}->unique })) {
	    Trace::debug('New encode monitor: %s', $tParms{'from'}->unique) if Trace::check(Trace::DEBUG4);
	    $monitors->{ $tParms{'from'}->unique } = $tParms{'from'};
	}
	$controller->_encodeStatus(to => $tParms{'from'});    
    }
    elsif (defined($parms->{'demonitor'})) {
	if (defined($monitors->{ $tParms{'from'}->unique })) {
	    delete $monitors->{ $tParms{'from'}->unique };
	}
    }
    return 1;
}

# _encodeStatus() - Inform a requestor or all monitors the state of
# the encoding machine. 

sub _encodeStatus {
    my $controller = shift;
    my %parms = @_;

    my $encode = $controller->{'encode'};
    my $monitors = $encode->{'monitors'};

    my $status = {};

    # Overall state
    $status->{'state'} = $encode->{'state'};
    
    # Copy the tracks
    my @activeEncoders = grep { ($_->{'state'} eq Encoding) && defined($_->{'format'}) } values(%{ $encode->{'encoders'} });
    my @underway = map 
    { 
	{
	    track => $_->{'track'}, 
	    rate => $_->{'format'}->rate,
	    handle => $_->handle, 
	    percentageComplete => $_->{'percentageComplete'},
	    start => $_->{'start'},
	}
    } @activeEncoders;
    my @storePending = map
    {
	{
	    track => $_->{'track'}, 
	    rate => $_->{'rate'}, 
	    storePending => 1,
	    time => $_->{'time'}, 
	} 
    } @{ $encode->{'storePending'} };
    $status->{'tracks'} = [ @underway, @{ $encode->{'tracks'} }, @storePending ];
    
    # Copy the encoders
    foreach my $encoder (values(%{ $encode->{'encoders'} })) {
	my $state;
	if (($encoder->{'state'} eq Encoding) || ($encoder->{'state'} eq Dead) || ($encoder->{'state'} eq Halted)) {
	    $state = $encoder->{'state'};
	    push @{ $status->{'encoders'} }, { handle => $encoder->handle, state => $state, };
	}
	else {
	    $state = Idle;
	    my $reason;
	    if ($encoder->{'state'} eq Idle) {
		if (!$controller->{'configured'}) {
		    $reason = 'Not configured';
		}
		elsif ($#{ $encode->{'tracks'} } == -1) {
		    $reason = 'No tracks to encode';
		}
		else {
		    $reason = 'Unknown';
		}
	    }
	    elsif ($encoder->{'state'} eq OutOfSpace) {
		$reason = 'There is not enough disk space to encode';
	    }
	    elsif ($encoder->{'state'} eq Hold) {
		$reason = $encoder->{'reason'};
	    }	    	    
	    push @{ $status->{'encoders'} }, { handle => $encoder->handle, state => $state, reason => $reason, };
	}
    }
    
    # Send on down
    my @dest = ();
    if (defined($parms{'to'})) {
	push @dest, $parms{'to'};
    }
    else {
	@dest = values(%{ $monitors });
    }
    foreach my $dest (@dest) {
	if (!$controller->message(name => EncodeStatus, to => $dest, parms => $status)) {
	    Trace::debug('Removed unresponsive encode monitor: %s', $dest->unique) if Trace::check(Trace::DEBUG4);
	    delete $monitors->{ $dest->unique };
	}
    }
    return 1;
}

# _encodeDiagnostic() 

sub _encodeDiagnostic {
    my $controller = shift;
    my %parms = @_;

    if (defined($parms{'diagnostic'}) && (length $parms{'diagnostic'} > 512)) {
	$parms{'diagnostic'} = substr $parms{'diagnostic'}, 0, 512;
    }

    my $encode = $controller->{'encode'};
    my $monitors = $encode->{'monitors'};

    foreach my $dest (values(%{ $monitors })) {
	if (!$controller->message(name => EncodeDiagnostic, to => $dest, parms => \%parms)) {
	    Trace::debug('Removed unresponsive encode monitor: %s', $dest->unique) if Trace::check(Trace::DEBUG4);
	    delete $monitors->{ $dest->unique };
	}
    }
    return 1;
}

# encodeHalt - Request to halt the encoding machine.

sub encodeHalt {
    my $controller = shift;

    my $encode = $controller->{'encode'};

    # Stop all encoding

    foreach my $encoder (values(%{ $encode->{'encoders'} })) {
	if ($encoder->{'state'} eq Encoding) {
	    # Abort the current encode job - Stop the encoder, put the
	    # the job back on the encode list and delete the file when
	    # the encoder will have finished with it.
	    $controller->message(name => Encoder::Stop, to => $encoder);
	    $controller->_returnToEncodeSchedule($encoder->{'track'}, $encoder->{'format'}->rate);
	}
	if ($encoder->{'state'} ne Dead) {
	    $encoder->{'state'} = Halted;
	}
    }
    $encode->{'state'} = Halted;

    $controller->_encodeStatus();
    return 1;
}

# encodeResume - Request to resume the encoding machine.

sub encodeResume {
    my $controller = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};
    my $encode = $controller->{'encode'};
    my $rip = $controller->{'rip'};

    $controller->_clearOutOfSpace();

    foreach my $encoder (values(%{ $encode->{'encoders'} })) {
	if ($encoder->{'state'} ne Dead) {
	    $encoder->{'state'} = Idle;
	}
    }
    $encode->{'state'} = Running;

    $controller->_encodeMachine();
    return 1;
}

# encoderStatus() - Status from Encoder.

sub encoderStatus {
    my $controller = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};
    
    my $encode = $controller->{'encode'};
    my $encoder = $encode->{'encoders'}->{ $tParms{'from'}->key };
    if (!defined($encoder)) {
	Trace::debug('Status message for unknown encoder (%s), reports %s', $tParms{'from'}->key,
		     ($parms->{'state'} eq Encoder::Idle) ? 'Idle' : 'Encoding') if Trace::check(Trace::DEBUG2);
	return;
    }
    
    Trace::debug('Encoder %s reports %s, cause %s', _describeEncoder($encoder),
		 ($parms->{'state'} eq Encoder::Idle) ? 'Idle' : 'Encoding', 
		 defined($parms->{'cause'}) ? $parms->{'cause'} : 'undef') if Trace::check(Trace::DEBUG3);

    if ($parms->{'state'} eq Encoder::Idle) {

	if ($encoder->{'state'} eq Idle) {
	    if (defined($parms->{'attributes'}->{'outfile'}) && (-e $parms->{'attributes'}->{'outfile'})) {
		# Possible synch problem.
		unlink $parms->{'attributes'}->{'outfile'};
	    }
	}

	elsif ($encoder->{'state'} eq Encoding) {

	    $encoder->{'state'} = Idle;	    

	    if ($parms->{'cause'} eq Encoder::End) {
		Trace::debug('Successful encode for %s', _describeEncoder($encoder)) if Trace::check(Trace::DEBUG4);
		$controller->message(name => 'store', to => $controller->{'repo'},
				     parms => { track => $encoder->{'track'}, file => $encoder->{'file'}, 'format' => $encoder->{'format'},
						remove => 1, _tag => 'encodedStore', }, );
		push @{ $encode->{'storePending'} }, { track => $encoder->{'track'}, rate => $encoder->{'format'}->rate, time => time,
						       file => $encoder->{'file'}, 'format' => $encoder->{'format'}, };
		$controller->_schedulePoll();
	    }

	    elsif ($parms->{'cause'} eq Encoder::NotConfigured) {
		# Encoder is not configured - Put the track back on
		# the job list and retry the encoder in five seconds.
		$controller->_returnToEncodeSchedule($encoder->{'track'}, $encoder->{'format'}->rate);
	        $encoder->{'state'} = Hold;
		$controller->at(name => '_encoderAwake', parms => { encoder => $encoder }, time => time + 5);
		$encoder->{'reason'} = 'Encoder not configured';
	    }

	    elsif ($parms->{'cause'} eq Encoder::Error) {
		# An error from the encoder - Put the track back on
		# the job list and remove the encode attempt.
		$controller->_returnToEncodeSchedule($encoder->{'track'}, $encoder->{'format'}->rate);
		unlink $parms->{'attributes'}->{'outfile'} if (-e $parms->{'attributes'}->{'outfile'});
		# Typically a space problem check availability in temp space.
		my $avail = Utility::available($encoder->{'file'});
		if (defined($avail) && ($avail == 0)) {
		    Trace::debug('%s: Not enough space while encoding', _describeEncoder($encoder)) if Trace::check(Trace::DEBUG4);
		    $controller->_encodeDiagnostic(handle => $encoder->handle, 
						   summary => 'Ran out of disk space when encoding track '.$encoder->{'track'}->name,
						   diagnostic => $parms->{'attributes'}->{'output'});
		}
		else {
		    # Not a space problem, bad configuration, bad rip
		    # file, synchronization issue. Hold the encoder
		    # for sixty seconds to avoid trashing. TODO a bad
		    # rip file will just be bounced between encoders
		    # indefinately.
		    $encoder->{'state'} = Hold;
		    $controller->at(name => '_encoderAwake', parms => { encoder => $encoder }, time => time + 60);
		    $encoder->{'reason'} = 'Error from encoder';
		    $controller->_encodeDiagnostic(handle => $encoder->handle,
						   summary => 'Error from encoder',
						   diagnostic => $parms->{'attributes'}->{'error'});
		}
	    }

	    elsif ($parms->{'cause'} eq Encoder::Stuck) {
		# Stuck encode job.
		$controller->_returnToEncodeSchedule($encoder->{'track'}, $encoder->{'format'}->rate);
		unlink $parms->{'attributes'}->{'outfile'} if (-e $parms->{'attributes'}->{'outfile'});
		$controller->_encodeDiagnostic(handle => $encoder->handle,
					       summary => 'Encoding was aborted after '.$parms->{'attributes'}->{'time'}.' seconds');
	    }

	    elsif (($parms->{'cause'} eq Encoder::Ignored) || ($parms->{'cause'} eq Encoder::None) || ($parms->{'cause'} eq Encoder::Stopped)) {
		# Synchronization problem - We think it is encoding,
		# it thinks otherwise but not for a valid error. Can
		# happen with network outages.  This track was never
		# encoded put it back on the job list.
		$controller->_returnToEncodeSchedule($encoder->{'track'}, $encoder->{'format'}->rate);
		if (defined($parms->{'attributes'}->{'outfile'}) && (-e $parms->{'attributes'}->{'outfile'})) {
		    unlink $parms->{'attributes'}->{'outfile'};
		}
	    }

	    else {
		Trace::error('Invalid cause (%s) for Idle state for encoder %s', $parms->{'cause'}, _describeEncoder($encoder));
	    }
	}

	elsif ($encoder->{'state'} eq Halted) {
	    if (defined($parms->{'attributes'}->{'outfile'}) && (-e $parms->{'attributes'}->{'outfile'})) {
		# Possible synch problem.
		unlink $parms->{'attributes'}->{'outfile'};
	    }
	}

	elsif ($encoder->{'state'} eq Dead) {
	    $encoder->{'state'} = Idle;
	    if (defined($parms->{'attributes'}->{'outfile'}) && (-e $parms->{'attributes'}->{'outfile'})) {
		# Possible synch problem.
		unlink $parms->{'attributes'}->{'outfile'};
	    }
	}

	elsif ($encoder->{'state'} eq Hold) {
	    if (defined($parms->{'attributes'}->{'outfile'}) && (-e $parms->{'attributes'}->{'outfile'})) {
		# Possible synch problem.
		unlink $parms->{'attributes'}->{'outfile'};
	    }
	}

	elsif ($encoder->{'state'} eq OutOfSpace) {
	    if (defined($parms->{'attributes'}->{'outfile'}) && (-e $parms->{'attributes'}->{'outfile'})) {
		# Possible synch problem.
		unlink $parms->{'attributes'}->{'outfile'};
	    }
	}

	else {
	    Trace::error('Unhandled state for encoder %s', _describeEncoder($encoder));
	}

	$controller->_encodeAttempt(encoder => $encoder);	    
    }
    elsif ($parms->{'state'} eq Encoder::Encoding) {

	if ($encoder->{'state'} eq Idle) {
	    $controller->_encodeEncodingResync($encoder);
	}

	elsif ($encoder->{'state'} eq Encoding) {

	    if ($parms->{'cause'} eq Encoder::None) {
		if (defined($parms->{'attributes'}->{'percentageComplete'})) {
		    Trace::debug('Encoder %s %d%% complete', _describeEncoder($encoder),
				 $parms->{'attributes'}->{'percentageComplete'}) if Trace::check(Trace::DEBUG6);
		    $encoder->{'percentageComplete'} = $parms->{'attributes'}->{'percentageComplete'};
		}
	    }

	    elsif ($parms->{'cause'} eq Encoder::Ignored) {
		# The encoder ignored our encode request because it is
		# in the middle of another - resync with the encoder,
		# the job will be put back on the list when the
		# encoder tells us its idle.
		$controller->_encodeEncodingResync($encoder);
	    }

	    else {
		Trace::error('Invalid cause (%s) for Encoding state for encoder %s', $parms->{'cause'}, _describeEncoder($encoder));
	    }
	}

	elsif ($encoder->{'state'} eq Halted) {
	    $controller->_encodeEncodingResync($encoder);
	}

	elsif ($encoder->{'state'} eq Dead) {
	    $controller->_encodeEncodingResync($encoder);
	}

	elsif ($encoder->{'state'} eq Hold) {
	    $controller->_encodeEncodingResync($encoder);
	}

	elsif ($encoder->{'state'} eq OutOfSpace) {
	    $controller->_encodeEncodingResync($encoder);
	}

	else {
	    Trace::error('Bad state for encoder %s', _describeEncoder($encoder));
	}
    }
    else {
	Trace::error('Unexpected state (%s) from encoder %s', $parms->{'state'}, _describeEncoder($encoder));
    }

    $controller->_encodeStatus();
    return 1;
}

# _encoderAwake

sub _encoderAwake {
    my $controller = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};
    
    my $encoder = $parms->{'encoder'};
    $encoder->{'state'} = Idle;
    $controller->_encodeAttempt(encoder => $encoder);	    
    return 1;
}

# _encodeEncodingResync - Resync with an encoder where the encoder thinks it
# is still doing an encode job.

sub _encodeEncodingResync {
    my $controller = shift;
    my $encoder = shift;

    $controller->message(name => Encoder::Stop, to => $encoder);
    $controller->_encodeDiagnostic(handle => $encoder->handle,
				   summary => 'Reconnecting with encoder',
				   diagnostic => 'This might happen if the controller restarted or the encoder was off the network for some time.');
    return 1;
}

# _encodeMachine()

sub _encodeMachine {
    my $controller = shift;
    my $encode = $controller->{'encode'};

    return if (!$controller->{'configured'});
    return if ($encode->{'state'} ne Running);

    ENC: foreach my $encoder (values(%{ $encode->{'encoders'} })) {
	next ENC if ($encoder->{'state'} ne Idle);
	$controller->_encodeAttempt(encoder => $encoder);
    }

    $controller->_encodeStatus();    
    return 1;
}

# _encodeAttempt() - Try and do an encode, change state accordingly.

sub _encodeAttempt {
    my $controller = shift;
    my %parms = @_;
    my $encode = $controller->{'encode'};
    my $encoder = $parms{'encoder'};

    return 0 if (!$controller->{'configured'});
    return 0 if ($encode->{'state'} ne Running);
    return 0 if ($#{ $encode->{'tracks'} } == -1);

    my $pos = 0;
    my $neededSize = 0;
    foreach my $ti (@{ $encode->{'tracks'} }) {
	my $track = $ti->{'track'};
	my $rate = $ti->{'rate'};
	my $duration = $track->{'ripformat'}->duration;
	my $size;
	my $format = Format::Mp3->new(rate => $rate);
	$format->duration($duration);
	if ($format->variableBitrate) {
	    # Don't know the size ahead of time, so over compensate
	    my $oc = Format::Mp3->new(rate => '256kbps');
	    $oc->duration($duration);
	    $size = $oc->expFileSize;
	}
	else {
	    $size = $format->expFileSize;
	}		
	my $file = $controller->_tmpfile(size => $size);
	if ($file) {
	    $controller->message(name => Encoder::Encode,
				 to => $encoder,
				 parms => 
				 { 
				     infile => $track->{'ripformat'}->{'file'},
				     outfile => $file,
				     outFormat => $format,
				 },
				 );
	    Trace::debug('%s encoding %s at %s', _describeEncoder($encoder), $track->name, $format->rate) if Trace::check(Trace::DEBUG4);
	    $encoder->{'track'} = $track;
	    $encoder->{'format'} = $format;
	    $encoder->{'file'} = $file;
	    $encoder->{'state'} = Encoding;
	    $encoder->{'percentageComplete'} = 0;
	    $encoder->{'start'} = time;
	    splice @{ $encode->{'tracks'} }, $pos, 1;
	    return 1;
	}
	else {
	    if (!$neededSize || ($size < $neededSize)) {
		$neededSize = $size;
	    }
	}
	$pos++;
    }

    Trace::debug('Not enough space') if Trace::check(Trace::DEBUG4);
    $encoder->{'state'} = OutOfSpace;
    $controller->{'outOfSpace'} = 1;
    $controller->_encodeDiagnostic(handle => $encoder->handle, summary => 'Not enough disk space to do encoding',
	   diagnostic => "At least $neededSize bytes are needed. Completion of current encoding jobs might free up the required space.");
    $controller->_setOutOfSpaceHobbled();

    return 1;
}

# _addToEncodeSchedule - Follow a few sanity checks before adding a
# track to the encode list.

sub _addToEncodeSchedule {
    my $controller = shift;
    my $encode = $controller->{'encode'};
    my $track = shift;
    my $rate = shift;

    my @activeEncoders = grep { $_->{'state'} eq Encoding } values(%{ $encode->{'encoders'} });
    my @underway = map { { track => $_->{'track'}, rate => $_->{'format'}->rate } } @activeEncoders;

    if (grep { ($_->{'track'}->ioid eq $track->ioid) && ($_->{'rate'} eq $rate) } @underway) {
	Trace::debug('Skipped encode of %s at '.$rate.' because it is underway', $track->name) if Trace::check(Trace::DEBUG4);
    }
    elsif (grep { ($_->{'track'}->ioid eq $track->ioid) && ($_->{'rate'} eq $rate) } @{ $encode->{'storePending'} }) {
	Trace::debug('Skipped encode of %s at '.$rate.' because it is awaiting repo store', $track->name) if Trace::check(Trace::DEBUG4);
    }
    elsif (grep { ($_->{'track'}->ioid eq $track->ioid) && ($_->{'rate'} eq $rate) } @{ $encode->{'tracks'} }) {
	Trace::debug('Skipped encode of %s at '.$rate.' because it is already on the encode list', $track->name) if Trace::check(Trace::DEBUG4);
    }
    else {
	Trace::debug('Added %s to encode schedule at '.$rate, $track->name) if Trace::check(Trace::DEBUG4);
	push @{ $encode->{'tracks'} }, { track => $track, rate => $rate, };
    }

    return 1;
}

# _returnToEncodeSchedule - Return a non encoded track to the encode
# list.

sub _returnToEncodeSchedule {
    my $controller = shift;
    my $encode = $controller->{'encode'};
    my $track = shift;
    my $rate = shift;

    push @{ $encode->{'tracks'} }, { track => $track, rate => $rate, };
    Trace::debug('Put %s back on encode list at %s', $track->name, $rate) if Trace::check(Trace::DEBUG4);

    return 1;
}

# RIPPING

# ripState() - Request for the state of the rip machine.

sub ripState {
    my $controller = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};
    
    my $rip = $controller->{'rip'};
    my $monitors = $rip->{'monitors'};
    if (!defined($monitors->{ $tParms{'from'}->unique })) {
	Trace::debug('New rip monitor: %s', $tParms{'from'}->unique) if Trace::check(Trace::DEBUG4);
	$monitors->{ $tParms{'from'}->unique } = $tParms{'from'};
    }
    $controller->_ripStatus(to => $tParms{'from'});
}

# _ripStatus() - Inform a requestor or all monitors the state of the
# rip machine.

sub _ripStatus {
    my $controller = shift;
    my %parms = @_;

    my $rip = $controller->{'rip'};
    my $monitors = $rip->{'monitors'};

    my @dest = ();
    if (defined($parms{'to'})) {
	push @dest, $parms{'to'};
    }
    else {
	@dest = values(%{ $monitors });
    }

    my $status = [];
    
    # Copy the rippers and their view of the tracks
    foreach my $ripper (values(%{ $rip->{'rippers'} })) {
	my $this = { handle => $ripper->handle };
	my $tracks = $this->{'tracks'} = {};
	foreach my $track (values(%{ $ripper->{'tracks'} })) {
	    $tracks->{ $track->index } =
	    {
		index => $track->index,
		want => defined($track->{'want'}) ? $track->{'want'} : 0,
		ripped => defined($track->{'ripped'}) ? $track->{'ripped'} : 0,
		encoded => defined($track->{'encoded'}) ? $track->{'encoded'} : 0,
	    };
	}
	if (($ripper->{'state'} eq Ripping) || ($ripper->{'state'} eq Halted) || ($ripper->{'state'} eq Dead)) {
	    $this->{'state'} = $ripper->{'state'};
	}
	else {
	    $this->{'state'} = Idle;
	    $this->{'reason'} = undef;
	    if (grep { $_->{'want'} && (!$_->{'ripped'} || !$_->{'encoded'}) } values(%{ $tracks })) {
		if ($controller->{'outOfSpace'} || $ripper->{'state'} eq OutOfSpace) {
		    $this->{'reason'} = 'More disk space needed';
		}
		elsif ($ripper->{'state'} eq Idle) {
		    if (!$controller->{'configured'}) {
			$this->{'reason'} = 'Not configured enough for ripping';
		    }
		    else {
			if ($controller->{'outOfSpaceHobbled'}) {
			    $this->{'reason'} = 'Ripping is limited';
			}
		    }
		}
		elsif ($ripper->{'state'} eq Checking) {
		    $this->{'reason'} = 'Checking need to rip';
		}
		elsif ($ripper->{'state'} eq Playing) {
		    if ($ripper->{'autoRip'}) {
			$this->{'reason'} = 'Auto rip on hold';
		    }
		}
	    }
	}
	push @{ $status }, $this; 
    }
    
    # Send on down
    foreach my $dest (@dest) {
	if (!$controller->message(name => RipStatus, to => $dest, parms => $status)) {
	    Trace::debug('Removed unresponsive rip monitor: %s', $dest->unique) if Trace::check(Trace::DEBUG4);
	    delete $monitors->{ $dest->unique };
	}
    }
}

# _ripDiagnostic() 

sub _ripDiagnostic {
    my $controller = shift;
    my %parms = @_;

    if (defined($parms{'diagnostic'}) && (length $parms{'diagnostic'} > 512)) {
	$parms{'diagnostic'} = substr $parms{'diagnostic'}, 0, 512;
    }

    my $rip = $controller->{'rip'};
    my $monitors = $rip->{'monitors'};

    foreach my $dest (values(%{ $monitors })) {
	if (!$controller->message(name => RipDiagnostic, to => $dest, parms => \%parms)) {
	    Trace::debug('Removed unresponsive rip monitor: %s', $dest->unique) if Trace::check(Trace::DEBUG4);
	    delete $monitors->{ $dest->unique };
	}
    }
}

# ripperRip() - Request to rip tracks from a Cd.

sub ripperRip {
    my $controller = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};

    $controller->_clearOutOfSpace();
    $controller->_clearOutOfSpaceHobbled();

    my $rip = $controller->{'rip'};
    
    if (!defined($parms) || !defined($parms->{'handle'})) {
	Trace::error('Need the cdrom to rip from');
	return;
    }
    if (!defined($rip->{'rippers'}->{ $parms->{'handle'}->key })) {
	Trace::error('Rip message for unknown cdrom: %s', $parms->{'handle'}->key);
	return;
    }
    my $ripper = $rip->{'rippers'}->{ $parms->{'handle'}->key };

    $ripper->{'userHalted'} = 0;    

    if ($ripper->{'state'} eq Halted) {
	$ripper->{'state'} = Idle;
    }
    elsif (($ripper->{'state'} eq Playing) || ($ripper->{'state'} eq Ripping)) {
	return;
    }

    if (defined($ripper->{'cd'}) && !$ripper->{'repoQueried'}) {

	# Query the need to rip this Cd

	$controller->_repoQuery($ripper);
    }

    $controller->_ripMachine();
    $controller->_ripStatus();
}

# ripperAdd() - Add tracks to those we want to rip

sub ripperAdd {
    my $controller = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};

    my $rip = $controller->{'rip'};
    if (!defined($parms) || !defined($parms->{'handle'})) {
	Trace::error('Need the cdrom to rip from');
	return;
    }
    if (!defined($rip->{'rippers'}->{ $parms->{'handle'}->key })) {
	Trace::error('Rip message for unknown cdrom: %s', $parms->{'handle'}->key);
	return;
    }
    my $ripper = $rip->{'rippers'}->{ $parms->{'handle'}->key };

    Trace::debug('Ripper Add: %s', _describeRipper($ripper)) if Trace::check(Trace::DEBUG4);
    
    if (!defined($parms->{'trackIndices'})) {
	$controller->_markAllTracksForRip($ripper);
    }
    else {
	my $cd = $ripper->{'cd'};
	return if (!defined($cd));

	foreach my $index (@{ $parms->{'trackIndices'} }) {
	    if (($index >= $cd->firstTrackIndex) && ($index <= $cd->lastTrackIndex)) {
		Trace::debug('%s: Marking track %d for rip', _describeRipper($ripper), $index) if Trace::check(Trace::DEBUG6);
		$ripper->{'tracks'}->{ $index }->{'want'} = (!$ripper->{'tracks'}->{ $index }->data);
	    }
	}
    }
    $controller->_ripMachine();
    $controller->_ripStatus();
}

# ripperRemove() - Remove tracks from those we want to rip

sub ripperRemove {
    my $controller = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};

    my $rip = $controller->{'rip'};
    if (!defined($parms) || !defined($parms->{'handle'})) {
	Trace::error('Need the cdrom to rip from');
	return;
    }
    my $ripper;
    if (!defined($ripper = $rip->{'rippers'}->{ $parms->{'handle'}->key })){
	Trace::error('Rip message for unknown cdrom: %s', $parms->{'handle'}->key);
	return;
    }
    
    Trace::debug('Ripper Remove: %s', _describeRipper($ripper)) if Trace::check(Trace::DEBUG4);
    
    if (!defined($parms->{'trackIndices'})) {
	$controller->_markAllTracksForNotRip($ripper);
    }    
    else {
	my $cd = $ripper->{'cd'};
	return if (!defined($cd));

	foreach my $index (@{ $parms->{'trackIndices'} }) {
	    if (($index >= $cd->firstTrackIndex) && ($index <= $cd->lastTrackIndex)) {		
		Trace::debug('%s: Clearing track %d for rip', _describeRipper($ripper), $index) if Trace::check(Trace::DEBUG6);
		$ripper->{'tracks'}->{ $index }->{'want'} = 0;
	    }
	}
    }
    $controller->_ripMachine();
    $controller->_ripStatus();
}

# cdromStatus() - Status from cdrom.

sub cdromStatus {
    my $controller = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};

    my $rip = $controller->{'rip'};
    my $ripper = $rip->{'rippers'}->{ $tParms{'from'}->key };
    if (!defined($ripper)) {
	Trace::debug('Status message %s for unknown ripper: %s', $parms->{'state'}, $tParms{'from'}->key) if Trace::check(Trace::DEBUG4);
	return;
    }

    Trace::debug('Ripper %s reports %s, cause %s', _describeRipper($ripper), $parms->{'state'}, $parms->{'cause'}) if Trace::check(Trace::DEBUG3);

    my $cd = $parms->{'attributes'}->{'cd'};
    my $prevCd;
    my $aNewCd = 0;
    if (defined($cd)) {
	if (!defined($ripper->{'cd'})) {
	    $aNewCd = 1;
	}
	else {
	    $aNewCd = ($cd->ioid ne $ripper->{'cd'}->ioid);
	}
    }

    if ((!defined($cd) || $aNewCd) && defined($ripper->{'cd'})) {
	delete $ripper->{'tracks'};
	delete $ripper->{'cd'};
    }

    if ($aNewCd) {
	$ripper->{'cd'} = $cd->monitor;
	$controller->_markAllTracksForNotRip($ripper);
	$ripper->{'repoQueried'} = 0;
	$ripper->{'userHalted'} = 0;
    }

    if ($ripper->{'state'} eq Dead) {
	if ($ripper->{'autoRip'} && !$ripper->{'userHalted'}) {
	    $ripper->{'state'} = Idle;
	}
	else {
	    $ripper->{'state'} = Halted;
	}
    }

    # Auto play a new cd
    if (($parms->{'state'} eq Cdrom::Stopped) && $ripper->{'autoPlay'} && $aNewCd) {
	$controller->message(name => Cdrom::Play, to => $ripper);
	$ripper->{'state'} = Playing;
    }
    # Auto rip a new cd
    elsif (($parms->{'state'} eq Cdrom::Stopped) && $ripper->{'autoRip'} && $aNewCd) {
	Trace::debug('Auto ripping %s', _describeRipper($ripper)) if Trace::check(Trace::DEBUG4);
	$controller->_markAllTracksForRip($ripper);
	# State will change when the repo replies to the rip query
    }
    # Auto rip when auto play has finished or was stopped
    elsif (($parms->{'state'} eq Cdrom::Stopped) && 
	   ($ripper->{'state'} eq Playing) && (($parms->{'cause'} eq Cdrom::End) || ($parms->{'cause'} eq Cdrom::Message)) &&
	   $ripper->{'autoRip'}) {
	Trace::debug('Auto ripping %s now that play has finished', _describeRipper($ripper)) if Trace::check(Trace::DEBUG4);
	$controller->_markAllTracksForRip($ripper);
	$ripper->{'state'} = Idle;
    }
    # Rip has finished
    elsif (($parms->{'state'} eq Cdrom::Stopped) && ($ripper->{'state'} eq Ripping) && ($parms->{'cause'} eq Cdrom::End)) {
	# Store the rip
	my $index = $ripper->{'index'};
	my $track = $ripper->{'tracks'}->{ $index };	    
	$track->{'ripped'} = 1;
	$controller->message(name => 'store',
			     to => $controller->{'repo'},
			     parms => 				 
			     { 
				 track => $track,				     
				 file => $ripper->{'file'},				     
				 'format' => $rip->{'format'},				     
				 remove => 1,				     
				 _tag => 'rippedStore',
			     },
			     );
	$ripper->{'state'} = Idle;
	push @{ $rip->{'storePending'} }, { track => $track, file => $ripper->{'file'}, 'format' => $rip->{'format'}, time => time };
	$controller->_schedulePoll();
    }
    # Stopped ripping
    elsif (($parms->{'state'} eq Cdrom::Stopped) && ($ripper->{'state'} eq Ripping)) {
	if (($parms->{'cause'} eq Cdrom::None) || ($parms->{'cause'} eq Cdrom::Ignored)) {
	    # Unexpected, should only occur if two controllers are trying
	    # to work with the same Cdrom.
	    $controller->_ripDiagnostic(handle => $ripper->handle,
					summary => 'Strange response from Cdrom',
					diagnostic => 'You might want to check your configuration.');
	    $ripper->{'state'} = Halted;
	}
	elsif ($parms->{'cause'} eq Cdrom::Message) {
	    # Stopped by user intervention
	    $ripper->{'userHalted'} = 1;
	    $ripper->{'state'} = Halted;
	}
	elsif ($parms->{'cause'} eq Cdrom::RipStuck) {
	    Trace::debug('%s: Rip got stuck', _describeRipper($ripper)) if Trace::check(Trace::DEBUG4);
	    $ripper->{'state'} = Halted;
	    $controller->_ripDiagnostic(handle => $ripper->handle, 
					summary => 'Spent too long ('.$parms->{'attributes'}->{'time'}.' seconds) ripping '.
					$ripper->{'tracks'}->{ $ripper->{'index'} }->name);
	}
	else {
	    # Cdrom error.

	    # Typically a space problem check availability in temp space.
	    my $avail = Utility::available($ripper->{'file'});
	    if (defined($avail) && ($avail == 0)) {
		Trace::debug('%s: Not enough space while ripping', _describeRipper($ripper)) if Trace::check(Trace::DEBUG4);
		$controller->_ripDiagnostic(handle => $ripper->handle, 
					    summary => 'Ran out of disk space when ripping '.
					    $ripper->{'tracks'}->{ $ripper->{'index'} }->name);
		# Maybe enough space in other temp areas.
		$ripper->{'state'} = Idle;
	    }
	    else {
		$controller->_ripDiagnostic(handle => $ripper->handle, 
					    summary => 'Error from cdrom when ripping '.
					    $ripper->{'tracks'}->{ $ripper->{'index'} }->name);
		$ripper->{'state'} = Halted;
	    }
	}
	unlink $ripper->{'file'} if (-e $ripper->{'file'});
    }
    # Stopped playing
    elsif (($parms->{'state'} eq Cdrom::Stopped) && ($ripper->{'state'} eq Playing)) {
	if ($parms->{'cause'} eq Cdrom::Message) {
	    # Stopped by user intervention
	    $ripper->{'userHalted'} = 1;
	}
	else {
	    # Cdrom error, assume this error is propagated to the
	    # user by monitoring the Cdrom.
	}
	$ripper->{'state'} = Halted;
    }
    elsif (
	   # Playing while we think we are ripping
	   ((($parms->{'state'} eq Cdrom::Playing) || ($parms->{'state'} eq Cdrom::Paused)) && ($ripper->{'state'} eq Ripping)) ||
	   # Ripping when we don't think we are
	   (($parms->{'state'} eq Cdrom::Ripping) && ($ripper->{'state'} ne Ripping))) {
	$controller->_ripDiagnostic(handle => $ripper->handle,
				    summary => 'Unexpected response from Cdrom',
				    diagnostic => 'This might happen if the controller restarted.');
	if ($ripper->{'state'} eq Ripping) {
	    unlink $ripper->{'file'} if (-e $ripper->{'file'});
	}
	$ripper->{'state'} = Halted;
    }
    elsif ($parms->{'state'} eq Cdrom::NoCd) {
	if ($ripper->{'state'} eq Ripping) {
	    unlink $ripper->{'file'} if (-e $ripper->{'file'});
	}
	$ripper->{'state'} = Halted;
    }

    if (defined($ripper->{'cd'}) && !$ripper->{'repoQueried'} && $ripper->{'cd'}->known) {

	# Automatically query the need to rip this Cd but only if it is known.

	$controller->_repoQuery($ripper);
    }

    $controller->_ripMachine();
    $controller->_ripStatus();
}

# _repoQuery

sub _repoQuery {
    my $controller = shift;
    my $ripper = shift;

    my $cd = $ripper->{'cd'};
    my @tracks = values(%{ $ripper->{'tracks'} });
    Trace::debug('Querying need to rip %d tracks from: %s', $#tracks + 1, $cd->name) if Trace::check(Trace::DEBUG4);
    $controller->message(name => 'query',
			 to => $controller->{'repo'},
			 parms => {
			     type => 'track',
			     tracks => \@tracks,
			     _tag => 'ripQuery',
			     _name => $ripper->key,
			     _cd => $cd,
			 },
			 );
    if (($ripper->{'state'} eq Idle) || ($ripper->{'state'} eq Halted)) {
	$ripper->{'state'} = Checking;
    }
    $ripper->{'repoQueried'} = 1;
}

# _markAllTracksForRip

sub _markAllTracksForRip 
{
    my $controller = shift;
    my $ripper = shift;
    my $cd = $ripper->{'cd'};

    Trace::debug('Marking all tracks for rip: %s', _describeRipper($ripper)) if Trace::check(Trace::DEBUG4);

    if (scalar keys %{ $ripper->{'tracks'} }) {
	foreach my $track (values(%{ $ripper->{'tracks'} })) {
	    Trace::debug('%s: Marking track %d for rip', _describeRipper($ripper), $track->index) if Trace::check(Trace::DEBUG6);
	    $track->{'want'} = (!$track->data);
	}
    }
    else {
	$ripper->{'tracks'} = {};
	for (my $ti = $cd->firstTrackIndex; $ti <= $cd->lastTrackIndex; $ti++) {
	    my $track = $cd->track($ti)->monitor;
	    Trace::debug('%s: Marking track %d for rip', _describeRipper($ripper), $track->index) if Trace::check(Trace::DEBUG6);
	    $track->{'want'} = (!$track->data);
	    $track->{'ripped'} = undef; # Don't know
	    $track->{'encoded'} = undef; # Don't know
	    $ripper->{'tracks'}->{ $ti } = $track;
	}
    }
}

# _markAllTracksForNotRip

sub _markAllTracksForNotRip 
{
    my $controller = shift;
    my $ripper = shift;
    my $cd = $ripper->{'cd'};
    return if (!defined($cd));

    Trace::debug('Clearing all tracks for rip: %s', _describeRipper($ripper)) if Trace::check(Trace::DEBUG4);

    if (scalar keys %{ $ripper->{'tracks'} }) {
	foreach my $track (values(%{ $ripper->{'tracks'} })) {
	    Trace::debug('%s: Clearing track %d for rip', _describeRipper($ripper), $track->index) if Trace::check(Trace::DEBUG6);
	    $track->{'want'} = 0;
	}
    }
    else {
	$ripper->{'tracks'} = {};
	for (my $ti = $cd->firstTrackIndex; $ti <= $cd->lastTrackIndex; $ti++) {
	    my $track = $cd->track($ti)->monitor;
	    Trace::debug('%s: Clearing track %d for rip', _describeRipper($ripper), $track->index) if Trace::check(Trace::DEBUG6);
	    $track->{'want'} = 0;
	    $track->{'ripped'} = undef; # Don't know
	    $track->{'encoded'} = undef; # Don't know
	    $ripper->{'tracks'}->{ $ti } = $track;
	}
    }
}

# _ripMachine

sub _ripMachine {
    my $controller = shift;
    my $rip = $controller->{'rip'};
    my $encode = $controller->{'encode'};

    return if (!$controller->{'configured'});

    if ($controller->{'outOfSpaceHobbled'}) {
	Trace::debug('Ripping is hobbled') if Trace::check(Trace::DEBUG6);
	return if ($#{ $encode->{'tracks'} } != -1);
	foreach my $ripper (values(%{ $rip->{'rippers'} })) {
	    return if ($ripper->{'state'} eq Ripping);
	}
	foreach my $encoder (values(%{ $encode->{'encoders'} })) {
	    return if ($encoder->{'state'} eq Encoding);
	}
	my @encPending = grep { !defined($_->{'response'}) || ($_->{'response'} >= $_->{'time'}) } @{ $encode->{'storePending'} };
	my @ripPending = grep { !defined($_->{'response'}) || ($_->{'response'} >= $_->{'time'}) } @{ $rip->{'storePending'} };
	return if (@ripPending || @encPending);
    }

    RIPPER: foreach my $ripper (values(%{ $rip->{'rippers'} })) {

	next RIPPER if ($ripper->{'state'} ne Idle);
	next RIPPER if (!defined($ripper->{'cd'}));
	next RIPPER if (!$ripper->{'repoQueried'});

	if ($controller->{'outOfSpace'}) {
	    $ripper->{'state'} = OutOfSpace;
	    next RIPPER;
	}

	my @tracks = sort { $a->index <=> $b->index } values(%{ $ripper->{'tracks'} });
	my $cd = $ripper->{'cd'};
	
	TRACK: foreach my $track (@tracks) {
	    next TRACK if (!$track->{'want'});
	    next TRACK if (!defined($track->{'ripped'}) || $track->{'ripped'});
	    next TRACK if (!defined($track->{'encoded'}) || $track->{'encoded'});

	    my $index = $ripper->{'index'} = $track->index;
	    my $duration = $track->duration;
	    $rip->{'format'}->duration($duration);
	    my $size = $rip->{'format'}->expFileSize;
	    $ripper->{'file'} = $controller->_tmpfile(size => $size);
	    if ($ripper->{'file'}) {
		if (!$controller->message(name => Cdrom::Rip, 
					  to => $ripper,
					  parms => 
					  { 
					      trackIndex => $index,
					      file => $ripper->{'file'},
					  },
					  )) {
		    Trace::debug('Ripper %s is dead', _describeRipper($ripper)) if Trace::check(Trace::DEBUG4);
		    $ripper->{'state'} = Dead;
		}
		else {
		    Trace::debug('Ripping track %d for %s', $index, _describeRipper($ripper)) if Trace::check(Trace::DEBUG4);
		    $ripper->{'state'} = Ripping;
		}
	    }
	    else {
		# Back off in the hopes that space will come free...
		Trace::debug('Not enough space for %s, backing off ripping', _describeRipper($ripper)) if Trace::check(Trace::DEBUG4);
		$ripper->{'state'} = OutOfSpace;
		$controller->{'outOfSpace'} = 1;
		$controller->_ripDiagnostic(handle => $ripper->handle, summary => 'Not enough disk space to rip track '.$index,
					    diagnostic => "$size bytes are needed. Completion of current encoding jobs might free up the required space.");
		$controller->_setOutOfSpaceHobbled();
	    }
	    last TRACK;
	}
    }
}

# repoInfo() - Info response from repo query.

sub repoInfo {
    my $controller = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};
    my $encode = $controller->{'encode'};
    my $rip = $controller->{'rip'};

    if ($parms->{'error'}) {
	Trace::debug('Error from Repo: %s', $parms->{'error'}) if Trace::check(Trace::DEBUG4);
	return;
    }

    if (($parms->{'type'} eq 'all') && ($parms->{'_tag'} eq 'init')) {

	# Need to send a second query to find what of these ripped
	# tracks have been encoded.

	if (!defined($parms->{'tracks'}) || ($#{ $parms->{'tracks'} } == -1)) {
	    Trace::debug('No existing ripped tracks to encode') if Trace::check(Trace::DEBUG4);
	}
	else {
	    Trace::debug('Querying need to encode the %d already ripped tracks', $#{ $parms->{'tracks'} } + 1) if Trace::check(Trace::DEBUG4);
	    $controller->message(name => 'query', 
				 to => $controller->{'repo'},
				 parms => 
				 {
				     type => 'track',
				     tracks => $parms->{'tracks'},
				     _tag => 'encodeQuery',
				 },
				 );
	}
    }
    elsif (($parms->{'type'} eq 'track') && ($parms->{'_tag'} eq 'encodeQuery')) {

	# Assumed to be in order

	TRACK: foreach my $track (@{ $parms->{'tracks'} }) {
	    my @formats = values(%{ $track->{'formatsH'} });
	    $track = $track->monitor; # Recovers my ref.
	    if ($#formats == -1) {
		Trace::debug('Nothing known about track: %s', $track->name) if Trace::check(Trace::DEBUG3);
	    }
	    elsif ($#formats == 0) {
		# Only one of format for this track in response, this
		# should be the ripped WAV from earlier, encode at all
		# rates that are not underway or pending store.
		if ($formats[0]->type eq 'Wav') {
		    $track->{'ripformat'} = $formats[0];
		    foreach my $rate (@{ $encode->{'rates'} }) {
			$controller->_addToEncodeSchedule($track, $rate);
		    }
		}
		else {
		    Trace::error('Track exists but not as WAV: %s', $track->name);
		}
	    }
	    else {
		# More than one format in the response, it has been
		# encoded at some or all of the rates.

		# Find the ripped wav & mark the encoded rates
		my $ripformat;
		my %rates;
		foreach my $format (@formats) {
		    if ($format->type eq 'Wav') {
			$ripformat = $format;
		    }
		    else {
			$rates{ $format->rate } = 1;
		    }
		}		
		if (!defined($ripformat)) {
		    Trace::error('Unexpected! what happened to the rip for track: %s', $track->name);
		}
		else {
		    $track->{'ripformat'} = $ripformat;
		    # Schedule the non encoded tracks
		    my $encwork = 0;
		    foreach my $rate (@{ $encode->{'rates'} }) {
			if ($rates{ $rate }) {
			    Trace::debug('Skipped encode of %s at '.$rate.' because it is done', $track->name) if Trace::check(Trace::DEBUG4);
			}
			else {
			    $controller->_addToEncodeSchedule($track, $rate);
			    $encwork = 1;
			}
		    }
		    if (!$encwork) {
			# Unnecessary rip, can happen following a configuration change.
			Trace::debug('This rip should not be here: %s', $track->{'ripformat'}->file) if (Trace::check(Trace::DEBUG3));
			$controller->message(name => Repo::Delete,
					     to => $controller->{'repo'},
					     parms => 
					     { 
						 track => $track,
						 'format' => $rip->{'format'},
					     },
					     );
		    }
		}
	    }
	}
	$controller->_encodeMachine();
    }

    elsif (($parms->{'type'} eq 'track') && ($parms->{'_tag'} eq 'ripQuery')) {

	# Response for a check to see if we need to rip, which we do
	# if the rip does not exist and the track has not been fully
	# encoded.

    	if (!defined($rip->{'rippers'}->{ $parms->{'_name'} })) {
	    Trace::error('Repo reply for unknown ripper: %s', $parms->{'_name'});
	    return;
	}
	my $ripper = $rip->{'rippers'}->{ $parms->{'_name'} };

	# Sanity check that it is for this Cd 

	if (!defined($ripper->{'cd'}) || ($ripper->{'cd'}->ioid ne $parms->{'_cd'}->ioid)) {
	    Trace::debug('Unexpected repo info for different Cd') if Trace::check(Trace::DEBUG4);
	    return;
	}

	if ($ripper->{'state'} eq Checking) {
	    if (!$ripper->{'userHalted'} && $ripper->{'autoRip'}) {
		$ripper->{'state'} = Idle;
	    }
	    else {
		$ripper->{'state'} = Halted;
	    }
	}

	Trace::debug('Rip query response for tracks from %s', $ripper->{'cd'}->name) if Trace::check(Trace::DEBUG4);

	map { $_->{'encodeRates'} = [] } values(%{ $ripper->{'tracks'} }); # Clear existing rates knowledge

	# Get the formats found by the Repo

	foreach my $track (@{ $parms->{'tracks'} }) {
	    $ripper->{'tracks'}->{ $track->index }->{'formatsH'} = $track->{'formatsH'};
	}

	# is track ripped? 

	foreach my $track (values(%{ $ripper->{'tracks'} })) {
	    if (defined($track->{'formatsH'}->{ Format::Wav::descriptor })) {
		Trace::debug('Track %d rip exists', $track->index) if Trace::check(Trace::DEBUG6);
		$track->{'ripped'} = 1;
		$track->{'ripformat'} = $track->{'formatsH'}->{ Format::Wav::descriptor };
	    }
	}

	# is track encoded at all rates?

	foreach my $track (values(%{ $ripper->{'tracks'} })) {
	    # Encoded
	    my @formats = values(%{ $track->{'formatsH'} });
	    foreach my $rate (@{ $encode->{'rates'} }) {
		if (grep { ($_->type eq 'Mp3') && ($_->rate eq $rate) } @formats) {
		    Trace::debug('Track %d is encoded at %s', $track->index, $rate) if Trace::check(Trace::DEBUG6);
		    push @{ $track->{'encodeRates'} }, $rate;
		}
	    }
	    # Encoded and pending store?
	    my @storePending = grep { $track->ioid eq $_->{'track'}->ioid } @{ $encode->{'storePending'} };
	    map {
		Trace::debug('Track %d is encoded at %s on store pending list', $_->{'track'}->index, $->{'rate'}) if Trace::check(Trace::DEBUG6);
		push @{ $track->{'encodeRates'} }, $_->{'rate'};
	    } @storePending;
	}

	foreach my $track (values(%{ $ripper->{'tracks'} })) {
	    if (Utility::Array::same($track->{'encodeRates'}, $encode->{'rates'})) {
		Trace::debug('Track %d is fully encoded at all rates', $track->index) if Trace::check(Trace::DEBUG6);
		$track->{'encoded'} = 1;
	    }
	    else {
		$track->{'encoded'} = 0;
	    }
	    
	    # Ripped and pending store?
	    if (grep { $track->ioid eq $_->{'track'}->ioid } @{ $rip->{'storePending'} }) {
		Trace::debug('Track %d rip is pending store', $track->index) if Trace::check(Trace::DEBUG6);
		$track->{'ripped'} = 1;
	    }
	}
		
	TRACK: foreach my $track (values(%{ $ripper->{'tracks'} })) {
	    if (!defined($track->{'ripped'}) || !$track->{'ripped'}) {
		Trace::debug('Track %d is not ripped', $track->index) if Trace::check(Trace::DEBUG6);
		$track->{'ripped'} = 0;
	    }
	    if (!defined($track->{'encoded'}) || !$track->{'encoded'}) {
		Trace::debug('Track %d is not fully encoded', $track->index) if Trace::check(Trace::DEBUG6);
		$track->{'encoded'} = 0;
	    }
	}
	$controller->_ripMachine();
	$controller->_ripStatus();
    }

    return 1;
}

# repoStored()

sub repoStored {
    my $controller = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};
    my $encode = $controller->{'encode'};
    my $rip = $controller->{'rip'};

    if ($parms->{'_tag'} eq 'rippedStore') {

	my $track = $parms->{'track'}->monitor;
	$track->{'ripformat'} = $parms->{'format'};
	Trace::debug('Successful store of ripped %s for %s', $parms->{'file'}, $track->name) if Trace::check(Trace::DEBUG4);

	# Remove this from the store pending list

	my @storePending = grep { $_->{'file'} ne $parms->{'file'} } @{ $rip->{'storePending'} };
	$rip->{'storePending'} = \@storePending;

	# Could be in the process of configuration, if so this rip
	# will be picked up when the repo is queried. Rates may change
	# during configuration.

	return 1 if (!$controller->{'configured'});

	# The WAV is stored, it can be encoded now if need be.

	RATE: foreach my $rate (@{ $encode->{'rates'} }) {
	    foreach my $erate (@{ $track->{'encodeRates'} }) {
		next RATE if ($erate eq $rate);
	    }

	    # Track was not encoded at this rate when the Cd was
	    # queried.

	    $controller->_addToEncodeSchedule($track, $rate);
	}
    }
    elsif ($parms->{'_tag'} eq 'encodedStore') {

	my $track = $parms->{'track'}->monitor;
	my $rate = $parms->{'format'}->rate;
	my $file = $parms->{'file'};

	Trace::debug('Successful store of encoded %s for %s at %s', $file, $track->name, $rate) if Trace::check(Trace::DEBUG4);

	# Remove this encode job from the store pending list

	my @storePending = grep { $_->{'file'} ne $file } @{ $encode->{'storePending'} };
	$encode->{'storePending'} = \@storePending;

	# Could be in the process of configuration, if so this rates
	# may change and the WAV may be still needed.

	return 1 if (!$controller->{'configured'});

	# See if this ripped WAV is still needed, if not it can be removed
	
	my @activeEncoders = grep { $_->{'state'} eq Encoding } values(%{ $encode->{'encoders'} });
	my @underway = map { { track => $_->{'track'}, rate => $_->{'format'}->rate } } @activeEncoders;

	my @need = grep { $_->{'track'}->ioid eq $track->ioid } (@underway, @{ $encode->{'storePending'} }, @{ $encode->{'tracks'} });

	if ($#need == -1) {

	    Trace::debug('Removing no longer needed rip for %s', $track->name) if Trace::check(Trace::DEBUG4);

	    $controller->message(name => Repo::Delete,
				 to => $controller->{'repo'},
				 parms => 
				 { 
				     track => $track,
				     'format' => $rip->{'format'},
				 },
				 );

	    # Mark encoded on the rip list if it is there

	    $track->{'encoded'} = 1;
	    $track->{'ripped'} = 0;
	    $controller->_ripStatus;
	}
    }

    $controller->_clearOutOfSpace();
    $controller->_encodeMachine();
    $controller->_ripMachine();
    $controller->_ripStatus();

    return 1;
}

# repoFailure() - Failed to store. Will try and store again in both
# cases when the store pending time limit is passed. Failure to store
# in assumed to mean the Repo cannot find the necessary space.

sub repoFailure {
    my $controller = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};
    my $encode = $controller->{'encode'};
    my $rip = $controller->{'rip'};

    $controller->{'outOfSpace'} = 1;

    if ($parms->{'_tag'} eq 'rippedStore') {

	my $track = $parms->{'track'}->monitor;

	# Sanity check that it is on the pending list

	my @failed = grep { $_->{'track'}->ioid eq $track->ioid } @{ $rip->{'storePending'} };
	if ($#failed != 0) {
	    Trace::error('Cannot find this track on the rip store pending list: %s', $track->name);
	}
	else {
	    $failed[0]->{'response'} = time;
	}
	$controller->_ripDiagnostic(handle => $controller->{'repo'}->handle,
				    summary => 'Error storing ripped file.',
				    diagnostic => $parms->{'error'});
    }
    elsif ($parms->{'_tag'} eq 'encodedStore') {
	
	my $track = $parms->{'track'}->monitor;
	my $rate = $parms->{'format'}->rate;

	# Sanity check that it is on the pending list

	my @failed = grep { ($_->{'track'}->ioid eq $track->ioid) && ($_->{'rate'} eq $rate) } @{ $encode->{'storePending'} };
	if ($#failed != 0) {
	    Trace::error('Cannot find this encode job on the store pending list: %s at %s', $track->name, $rate);
	}
	else {
	    $failed[0]->{'response'} = time;
	}

	$controller->_encodeDiagnostic(handle => $controller->{'repo'}->handle,
				       summary => 'Error storing encoded file',
				       diagnostic => $parms->{'error'});
    }
    $controller->_setOutOfSpaceHobbled();
    $controller->_schedulePoll();
    return 1;
}

# _schedulePoll() - Decide if polled work is needed.

sub _schedulePoll {
    my $controller = shift;
    my $encode = $controller->{'encode'};
    my $rip = $controller->{'rip'};

    return if ($controller->{'pollScheduled'});
    $controller->{'pollScheduled'} = 0;

    # Retry pending stores every sixty seconds (note: the repo has
    # replied that the previous store attempt failed)

    my @encPending = grep { defined($_->{'response'}) && ($_->{'response'} >= $_->{'time'}) } @{ $encode->{'storePending'} };
    my @ripPending = grep { defined($_->{'response'}) && ($_->{'response'} >= $_->{'time'}) } @{ $rip->{'storePending'} };

    return if (!@encPending && !@ripPending);

    $controller->{'pollScheduled'} = 1;
    $controller->at(name => '_poll', to => $controller, time =>  time + 60);
    return 1;
}

# _poll() - Encodes and rips can be pending store.

sub _poll {
    my $controller = shift;
    my $encode = $controller->{'encode'};
    my $rip = $controller->{'rip'};

    my @encPending = grep { defined($_->{'response'}) && ($_->{'response'} >= $_->{'time'}) } @{ $encode->{'storePending'} };
    my @ripPending = grep { defined($_->{'response'}) && ($_->{'response'} >= $_->{'time'}) } @{ $rip->{'storePending'} };

    $controller->_resendStorePending(\@encPending, 'encodedStore') if (@encPending);
    $controller->_resendStorePending(\@ripPending, 'rippedStore') if (@ripPending);
    return 1;
}

sub _clearOutOfSpace {
    my $controller = shift;
    my $encode = $controller->{'encode'};
    my $rip = $controller->{'rip'};

    return 0 if (!$controller->{'outOfSpace'});

    Trace::debug('Clearing out of space condition') if Trace::check(Trace::DEBUG4);

    $controller->{'outOfSpace'} = 0;

    my @encPending = grep { defined($_->{'response'}) && ($_->{'response'} >= $_->{'time'}) } @{ $encode->{'storePending'} };
    my @ripPending = grep { defined($_->{'response'}) && ($_->{'response'} >= $_->{'time'}) } @{ $rip->{'storePending'} };

    $controller->_resendStorePending(\@encPending, 'encodedStore');
    $controller->_resendStorePending(\@ripPending, 'rippedStore');
    
    foreach my $encoder (values(%{ $encode->{'encoders'} })) {
	if ($encoder->{'state'} eq OutOfSpace) {
	    $encoder->{'state'} = Idle;
	    Trace::debug('Encoder %s is now back', _describeEncoder($encoder)) if Trace::check(Trace::DEBUG4);
	}
    }

    foreach my $ripper (values(%{ $rip->{'rippers'} })) {
	if ($ripper->{'state'} eq OutOfSpace) {
	    $ripper->{'state'} = Idle;
	    Trace::debug('Ripper %s is now back', _describeRipper($ripper)) if Trace::check(Trace::DEBUG4);
	}
    }
    return 1;
}

sub _setOutOfSpaceHobbled {
    my $controller = shift;

    if (!$controller->{'outOfSpaceHobbled'}) {
	$controller->_ripDiagnostic(handle => $controller->handle,
				    summary => 'Ripping is now limited.',
				    diagnostic => 'Ripping will only occur when there is no encoding work to be done. You can clear this condition by restarting the rip process or configuring more space.');
	$controller->{'outOfSpaceHobbled'} = 1;
    }
}

sub _clearOutOfSpaceHobbled {
    my $controller = shift;

    if ($controller->{'outOfSpaceHobbled'}) {
	$controller->_ripDiagnostic(handle => $controller->handle, summary => 'Ripping is no longer limited.');
	$controller->{'outOfSpaceHobbled'} = 0;
    }
}

sub _resendStorePending {
    my $controller = shift;
    my $pendings = shift;
    my $tag = shift;

    foreach my $pending (@{ $pendings }) {
	$controller->message(name => 'store',
			     to => $controller->{'repo'},
			     parms => 
			     { 
				 track => $pending->{'track'},
				 file => $pending->{'file'},
				 'format' => $pending->{'format'},
				 remove => 1,
				 _tag => $tag,
			     },
			     );
	Trace::debug('Retrying %s job: %s', $tag, $pending->{'track'}->name) if Trace::check(Trace::DEBUG4);
	$pending->{'time'} = time;
    }
}

# _initTmp() - Initialize the temporary file storage areas, if it is a
# new dir scrub any old temp files.

sub _initTmp {
    my $controller = shift;
    my %parms = @_;

    my $tmpdirs = $parms{'tmpdirs'};
    DIR: foreach my $dir (@{ $tmpdirs }) {
	next DIR if (grep { $_ eq $dir } @{ $controller->{'tmpdirs'} }); 
	if (!(-r $dir && -w $dir && -x $dir && -d $dir)) {
	    Trace::error('Cannot access temporary directory (%s)', $dir);
	    return undef;
	}
	if (!opendir(DIR, $dir)) {
	    Trace::error('Cannot open temporary directory: %s: %s', $dir, $!);
	    return undef;
	}
        ENTRY: foreach my $entry (readdir(DIR)) {
	    next ENTRY if ($entry !~ m/^__thing__tmp__.*$/);
	    unlink $dir.'/'.$entry;
	}
	closedir(DIR);
    }
    $controller->{'tmpdirs'} = $tmpdirs;
    return 1;
}

# _tmpfile(size) - return a temp file of `size' bytes or 0 if there is
# not enough space.

sub _tmpfile {
    my $controller = shift;
    my %parms = @_;

    my $size = $parms{'size'} || 0;
    my $tmpdirs = $controller->{'tmpdirs'};
    DIR: foreach my $dir (@{ $tmpdirs }) {
	my $avail = Utility::available($dir);
	if (!$size || (defined($avail) && ($avail > $size))) {
	    my $tmpfile;
	    do {
		$tmpfile = $dir.'/__thing__tmp__'.int(rand(99999)).'.tmp';
	    } while (-e $tmpfile);
	    Trace::debug('Allocated temp file %s (%d bytes)', $tmpfile, $size) if Trace::check(Trace::DEBUG4);
	    return $tmpfile;
	}
    }
    Trace::debug('Cannot allocate enough space for temp file (%d bytes)', $size) if Trace::check(Trace::DEBUG4);
    return 0;
}

sub _describeEncoder {
    my $encoder = shift;

    my $state;
    if ($encoder->{'state'} eq Idle) {
	$state = 'Idle';
    }
    elsif ($encoder->{'state'} eq Encoding) {
	$state = 'Encoding';
    }
    elsif ($encoder->{'state'} eq Halted) {
	$state = 'Halted';
    }
    elsif ($encoder->{'state'} eq Dead) {
	$state = 'Dead';
    }
    elsif ($encoder->{'state'} eq Hold) {
	$state = 'Hold';
    }
    elsif ($encoder->{'state'} eq OutOfSpace) {
	$state = 'OutOfSpace';
    }

    return '('.$encoder->key.' ['.$state.'])';
}

sub _describeRipper {
    my $ripper = shift;

    my $state;
    if ($ripper->{'state'} eq Idle) {
	$state = 'Idle';
    }
    elsif ($ripper->{'state'} eq Ripping) {
	$state = 'Ripping';
    }
    elsif ($ripper->{'state'} eq Playing) {
	$state = 'Playing';
    }
    elsif ($ripper->{'state'} eq Checking) {
	$state = 'Checking';
    }
    elsif ($ripper->{'state'} eq Halted) {
	$state = 'Halted';
    }
    elsif ($ripper->{'state'} eq Dead) {
	$state = 'Dead';
    }
    elsif ($ripper->{'state'} eq OutOfSpace) {
	$state = 'OutOfSpace';
    }

    return '('.$ripper->key.' ['.$state.'])';
}

1;
