# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.6 $

package Encoder;

use Thing;
@ISA = qw( Thing );
use EncoderH;
use Command;
use Utility;
use Format::Wav;
use Format::Mp3;
use Trace;
use strict;

# Status message `parms' are a hash ref containing `state', `cause'
# and `attributes' fields generated as follows:

# State     Cause           To
# -----     -----           --
#
# Idle      None            From  Encoder is in idle state.
#
# Idle      End             All   Encoder has finished encoding `infile' to 
#                                 `outfile' and is now in the 
#                                 idle state.
#
# Idle      Ignored         From  Attempt to Stop the encoder was ignored because
#                                 it is already in the Idle state.
#
# Idle      NotConfigured   From  Attempt to Encode before the encoder is configured.
#
# Idle      Error           All   An encoder specific error occurred described by
#                                 `return' and `output' when encoding `infile' to 
#                                 `outfile'.
#
# Idle      Stuck           All   Encoding for `infile' to `outfile' was aborted 
#                                 after `time' because it was taking too long.
#
# Idle      Stopped         All   Encoding was stopped by a Stop message while 
#                                 encoding `infile' to `outfile'.
#
# Encoding  None            All   Encoding is under way on `infile' to `outfile'
#                                 and is `percentComplete'. Sent periodically.
# 
# Encoding  Ignored         From  Attempt to Encode was ignored for `infile' to 
#                                 `outfile' because one is already underway on
#                                 `currentInfile' to `currentOutfile'.

sub new {
    my $proto = shift;
    my $type = ref($proto) || $proto;
    my $encoder = $type->SUPER::new(@_);
    if ($encoder->error) {
	return $encoder;
    }

    $encoder->{'state'} = Idle;
    $encoder->{'cause'} = None;
    $encoder->{'attributes'} = {};
    $encoder->{'monitors'} = {};
    $encoder->{'monitorInterval'} = 15;
    # Limit encode time based on the file size, 60 seconds should have
    # encoded in 60 minutes.
    $encoder->{'timeLimitForSixty'} = 3600;
    $encoder->register(name => Encode, handler => 'encode');
    $encoder->register(name => Stop, handler => 'stop');
    $encoder->register(name => State, handler => 'state');
    $encoder->register(name => Status, handler => Thing::Ignore);
    $encoder->register(name => '_encodeMonitor', handler => '_encodeMonitor');

    $encoder->register(name => Thing::ConfigureRequest, handler => '_configure');
    $encoder->register(name => Thing::ShutdownRequest, handler => '_shutdown');

    return $encoder;
}

# _configure

sub _configure {
    my $encoder = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};

    if (!defined($parms->{'encoder'}) || ($parms->{'encoder'} eq '')) {
	$encoder->{'encoder'} = 'lame';
    }
    else {
	if (($parms->{'encoder'} ne 'lame') && ($parms->{'encoder'} ne '8hz-mp3')) {
	    $parms->{'error'} = 'Unknown encoder: '.$parms->{'encoder'};
	}
	else {
	    $encoder->{'encoder'} = $parms->{'encoder'} ;
	}
    }
    if (!defined($parms->{'error'})) {
	if (!defined($parms->{'exe'}) || ($parms->{'exe'} eq '')) {
	    $encoder->{'exe'} = $encoder->{'encoder'};
	}
	else {
	    $encoder->{'exe'} = $parms->{'exe'};
	}
	my $error = Utility::executes($encoder->{'exe'});
	if ($error) {
	    $parms->{'error'} = 'Cannot execute: '.$encoder->{'exe'}.': '.$error;
	}
    }
}

sub _shutdown {
    my $encoder = shift;

    if (($encoder->{'state'} eq Encoding) && defined($encoder->{'command'})) {
	$encoder->{'command'}->kill('TERM');
    }
}

# state() - State enquiry. 

sub state {
    my $encoder = shift;
    my %tParms = @_;

    my $monitors = $encoder->{'monitors'};
    if (!defined($monitors->{ $tParms{'from'}->unique })) {
	Trace::debug('New state monitor: %s', $tParms{'from'}->unique) if Trace::check(Trace::DEBUG3);
	$monitors->{ $tParms{'from'}->unique } = $tParms{'from'};
    }
    $encoder->_status(to => $tParms{'from'});
}

sub _status {
    my $encoder = shift;
    my %parms = @_;

    if ($encoder->{'state'} eq Encoding) {
	$encoder->{'attributes'}->{'infile'} = $encoder->{'infile'};
	$encoder->{'attributes'}->{'outfile'} = $encoder->{'outfile'};
	$encoder->{'attributes'}->{'percentageComplete'} = $encoder->{'percentageComplete'};
    }

    # Send on down
    my @dest = ();
    if (defined($parms{'to'})) {
	push @dest, $parms{'to'};
    }
    else {
	@dest = values(%{ $encoder->{'monitors'} });
    }
    foreach my $dest (@dest) {
	if (!$encoder->message(name => Status, 
			       to => $dest,
			       parms =>
			       {
				   state => $encoder->{'state'},
				   cause => $encoder->{'cause'},
				   attributes => $encoder->{'attributes'},
			       }
			       )) {
	    Trace::debug('Removed unresponsive state monitor: %s', $dest->unique) if Trace::check(Trace::DEBUG3);
	    delete $encoder->{'monitors'}->{ $dest->unique };
	}
    }
    $encoder->{'cause'} = None;
    $encoder->{'attributes'} = {};
}

# stop() - Stop message.

sub stop {
    my $encoder = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};

    Trace::debug('Stop request from: %s', $tParms{'from'}->key) if Trace::check(Trace::DEBUG4);

    if ($encoder->{'state'} eq Idle) {
	Trace::error('Already stopped');
	$encoder->{'cause'} = Ignored;
	$encoder->_status(to => $tParms{'from'});
	return;	
    }
    $encoder->{'command'}->kill('TERM');
    $encoder->{'attributes'} = {
	infile => $encoder->{'infile'},
	outfile => $encoder->{'outfile'},
    };
    $encoder->{'cause'} = Stopped;
    $encoder->{'state'} = Idle;
    delete $encoder->{'command'};
    $encoder->_status();
}

# encode() - Encode `infile' to `outfile' by `outFormat'.

sub encode {
    my $encoder = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};

    if (!defined($encoder->{'encoder'})) {
	$encoder->{'cause'} = NotConfigured;
	$encoder->_status(to => $tParms{'from'});
	return;
    }
    $encoder->{'attributes'} = {
	infile => undef,
	outfile => undef,
    };
    my $infile = $parms->{'infile'};
    if (!defined($infile)) {
	Trace::mandatory('infile');
	$encoder->{'cause'} = Error;
	$encoder->{'attributes'}->{'return'} = 0;
	$encoder->{'attributes'}->{'output'} = Trace::lastError;
	$encoder->_status(to => $tParms{'from'});
	return;
    }
    my $outfile = $parms->{'outfile'};
    if (!defined($outfile)) {
	Trace::mandatory('outfile');
	$encoder->{'cause'} = Error;
	$encoder->{'attributes'}->{'return'} = 0;
	$encoder->{'attributes'}->{'output'} = Trace::lastError;
	$encoder->_status(to => $tParms{'from'});
	return;
    }
    $encoder->{'attributes'} = {
	infile => $infile,
	outfile => $outfile,
    };
    my $outFormat = $parms->{'outFormat'};
    if (!defined($outFormat)) {
	Trace::mandatory('outFormat');
	$encoder->{'cause'} = Error;
	$encoder->{'attributes'}->{'return'} = 0;
	$encoder->{'attributes'}->{'output'} = Trace::lastError;
	$encoder->_status(to => $tParms{'from'});
	return;
    }

    if ($encoder->{'state'} eq Encoding) {
	Trace::error('Already encoding: %s', $encoder->{'infile'});
	$encoder->{'cause'} = Ignored;
	$encoder->{'attributes'}->{'currentInfile'} = $encoder->{'infile'};
	$encoder->{'attributes'}->{'currentOutfile'} = $encoder->{'outfile'};
	$encoder->_status(to => $tParms{'from'});
	return;	
    }
    my $inFormat = Format->new(file => $infile);
    if (!defined($inFormat)) {
	Trace::error('Could not determin infile format');
	$encoder->{'cause'} = Error;
	$encoder->{'attributes'}->{'return'} = 0;
	$encoder->{'attributes'}->{'output'} = Trace::lastError;
	$encoder->_status(to => $tParms{'from'});
	return;
    }
    $outFormat->duration($inFormat->duration);

    $encoder->{'state'} = Encoding;
    $encoder->{'cause'} = None;

    Trace::debug('Encode request from: %s', $tParms{'from'}->key) if Trace::check(Trace::DEBUG4);

    $encoder->{'infile'} = $infile;
    $encoder->{'outfile'} = $outfile;
    $encoder->{'inFormat'} = $inFormat;
    $encoder->{'outFormat'} = $outFormat;
    $encoder->{'percentageComplete'} = 0;
    $encoder->{'timeLimit'} = ($inFormat->duration * $encoder->{'timeLimitForSixty'}) / 60;

    Trace::debug('Encoding %s to %s at %s', $infile, $outfile, $outFormat->rate) if Trace::check(Trace::DEBUG4);
    Trace::debug('Giving %d seconds', $encoder->{'timeLimit'}) if Trace::check(Trace::DEBUG6);

    my $str;
    if ($encoder->{'encoder'} eq 'lame') {
	if ($outFormat->variableBitrate) {
	    $str = sprintf('%s -v -V 1 -b 32 -m j -h %s %s',
			   $encoder->{'exe'},
			   Command::quotemeta($infile),
			   Command::quotemeta($outfile));
	}
	else {
	    $str = sprintf('%s -b %d %s %s',
			   $encoder->{'exe'},
			   $outFormat->kbps,
			   Command::quotemeta($infile),
			   Command::quotemeta($outfile));
	}
    }
    else { # 8hz-mp3
	if ($outFormat->variableBitrate) {
	    Trace::error('8hz-mp3 cannot do variable bitrate encoding');
	    $encoder->{'state'} = Idle;
	    $encoder->{'cause'} = Error;
	    $encoder->{'attributes'}->{'return'} = 0;
	    $encoder->{'attributes'}->{'output'} = Trace::lastError;
	    $encoder->_status(to => $tParms{'from'});
	    return;
	}
	else {
	    $str = sprintf('%s -h -b %d %s %s', $encoder->{'exe'}, $outFormat->kbps,
			   Command::quotemeta($infile), Command::quotemeta($outfile));
	}
    }
    my $cmd = Command->new(Shell => $str);
    if (!$cmd || !$cmd->run) {
	Trace::error('Cannot execute encode');
	$encoder->{'state'} = Idle;
	$encoder->{'cause'} = Error;
	$encoder->{'attributes'}->{'return'} = $cmd->return;
	$encoder->{'attributes'}->{'output'} = $cmd->output;
	$encoder->_status(to => $tParms{'from'});
	return;
    }

    $encoder->{'command'} = $cmd;
    $encoder->{'tag'} = time;
    $encoder->{'start'} = time;
    $encoder->at(name => '_encodeMonitor',
		 to => $encoder,
		 time =>  time + $encoder->{'monitorInterval'},
		 parms => {
		     tag => $encoder->{'tag'},
		 });
    $encoder->_status();
}

sub _encodeMonitor {
    my $encoder = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};

    $encoder->{'attributes'} = {
	infile => $encoder->{'infile'},
	outfile => $encoder->{'outfile'},
    };
    my $cmd = $encoder->{'command'};
    return if (!defined($cmd));
    return if ($encoder->{'state'} eq Idle);
    return if ($parms->{'tag'} ne $encoder->{'tag'});

    if (($cmd->state eq Command::RUNNING) && (time > ($encoder->{'start'} + $encoder->{'timeLimit'}))) {
	Trace::debug('Exceeded time limit') if Trace::check(Trace::DEBUG4);
	$encoder->{'command'}->kill('TERM');
	delete $encoder->{'command'};
	$encoder->{'state'} = Idle;
	$encoder->{'cause'} = Stuck;
	$encoder->{'attributes'}->{'time'} = int($encoder->{'timeLimit'});
    }
    elsif ($cmd->state eq Command::RUNNING) {
	$encoder->{'cause'} = None;
	# Check up again later.
	$encoder->at( name => '_encodeMonitor',
		      to => $encoder,
		      time =>  time + $encoder->{'monitorInterval'},
		      parms => {
			  tag => $encoder->{'tag'},
		      } );
	# Work out percentage complete
	my $pc;
	if ($encoder->{'outFormat'}->variableBitrate) {
	    # Variable bitrate
	    if (defined($encoder->{'histRate'})) {
		my $timeSpent = time - $encoder->{'start'};
		my $secondsEncoded = $timeSpent * $encoder->{'histRate'};
		$pc = 100*($secondsEncoded/$encoder->{'outFormat'}->duration);
		if ($pc > 100) {
		    $pc = 100;
		}
	    }
	    else {
		# Don't have a history of encoding rate yet
		$pc = 0;
	    }
	}
	else {
	    # Constant bitrate
	    $pc = (100 * (-s $encoder->{'outfile'})) / $encoder->{'outFormat'}->expFileSize;
	}
	$encoder->{'percentageComplete'} = $pc;
	Trace::debug('%d%% complete', int($encoder->{'percentageComplete'})) if Trace::check(Trace::DEBUG4);
    }
    else {
	# Command is finished
	$encoder->{'state'} = Idle;
	$encoder->{'cause'} = End;
	if (($cmd->state eq Command::FINISHED) && defined($cmd->return) && ($cmd->return == 0)) {
	    Trace::debug('Possibly good') if Trace::check(Trace::DEBUG4);
	    if ($encoder->{'outFormat'}->variableBitrate) {
		# Update rate history for variable bitrate
		my $rate = $encoder->{'outFormat'}->duration / (time - $encoder->{'start'});
		if (defined($encoder->{'histRate'})) {
		    $encoder->{'histRate'} = ($rate+$encoder->{'histRate'})/2;
		}
		else {
		    $encoder->{'histRate'} = $rate;
		}
		Trace::debug('Historical encode rate: %f', $encoder->{'histRate'}) if Trace::check(Trace::DEBUG4);
		# Use the output file's duration as a sanity check
		my $expDuration = $encoder->{'outFormat'}->duration;
		my $createFormat = Format::Mp3->new(file => $encoder->{'outfile'});
		if (!defined($createFormat)) {
		    $encoder->{'cause'} = Error;
		    $encoder->{'attributes'}->{'output'} = Trace::lastError();
		}
		elsif (abs($createFormat->duration - $expDuration) > 2) {
		    Trace::debug('More than two seconds out') if Trace::check(Trace::DEBUG4);
		    $encoder->{'cause'} = Error;
		    $encoder->{'attributes'}->{'output'} = 'More than two seconds out in encoded file';
		}
	    }
	    else {
		# Sanity check size for constant bitrate
		my $fileSize = -s $encoder->{'outfile'};
		my $expSize = $encoder->{'outFormat'}->expFileSize;
		if ((($expSize - $fileSize) / $expSize) > 0.005) {
		    # Size is more than -0.5% off
		    Trace::debug('Size is +/- 0.5% off') if Trace::check(Trace::DEBUG4);
		    $encoder->{'cause'} = Error;
		    $encoder->{'attributes'}->{'return'} = $cmd->return;
		    $encoder->{'attributes'}->{'output'} = 'Size of encoded file is more than -0.5% off';
		}
	    }
	}
	else {
	    # Bad return
	    $encoder->{'cause'} = Error;
	    $encoder->{'attributes'}->{'return'} = $cmd->return;
	    $encoder->{'attributes'}->{'output'} = $cmd->output;	    
	}
    }
    $encoder->_status();
}

1;
