# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.5 $

package Format::Mp3;

use Format;
@ISA = qw( Format );
use MP3::Info;
use Trace;
use strict;

sub new {
    my $proto = shift;
    my $type = ref($proto) || $proto;
    my $format = $type->SUPER::new(@_);
    if (!defined($format)) {
	return undef;
    }

    my %parms = @_;

    if (defined($parms{'file'})) {
	my $info = MP3::Info::get_mp3info($parms{'file'});
	if (!defined($info)) {
	    Trace::error('For %s', $parms{'file'});
	    return undef;
	}

	# Determin VBR based on the filename suffix first and
	# MP3::Info second because a file may have been encoded using
	# VBR but each frame may have the same rate.

	if ($info->{'VBR'} || ($parms{'file'} =~ /_vbr.mp3$/)) {
	    $format->{'rate'} = 'Variable Bitrate';
	    $format->{'variableBitrate'} = 1;
	}
	else {
	    $format->{'kbps'} = $info->{'BITRATE'};
	    $format->{'variableBitrate'} = 0;
	    $format->{'rate'} = $info->{'BITRATE'}.'kbps';
	}
	$format->{'duration'} = ($info->{'MM'} * 60) + $info->{'SS'};
	$format->{'file'} = $parms{'file'};
    }
    else {
	my $rate = $format->{'rate'} = $parms{'rate'};
	if ($rate =~ /^(\d+)kbps$/) {
	    $format->{'kbps'} = $1;
	    $format->{'variableBitrate'} = 0;
	}
	elsif ($rate =~ /^Variable Bitrate$/) {
	    $format->{'variableBitrate'} = 1;
	}
	else {
	    Trace::error('Cannot use rate: '.$rate);
	    return undef;
	}
    }

    return $format;
}

sub suffix {
    my $format = shift;
    my %parms = @_;
    my $suffix = '';

    if ($parms{'unique'}) {
	if ($format->{'variableBitrate'}) {
	    $suffix = '_vbr';
	}
	else {
	    $suffix = '_'.$format->{'kbps'};
	}
    }

    $suffix .= '.mp3';
}

sub rate { $_[0]->{'rate'}; }
sub kbps { $_[0]->{'kbps'}; }
sub variableBitrate { $_[0]->{'variableBitrate'}; }

sub rates { ('Variable Bitrate', '256kbps', '224kbps', '192kbps', '176kbps', '160kbps', '144kbps', '128kbps', 
	     '112kbps', '96kbps', '80kbps', '64kbps', '56kbps', '48kbps', '32kbps') };

# expFileSize() - Returns the expected size of a file of this format

sub expFileSize {
    if ($_[0]->{'variableBitrate'}) {
	Trace::error('Cannot determin the expected file size of a variable bitrate format');
	return undef;
    }
    else {
	return ($_[0]->{'duration'} * $_[0]->{'kbps'} * 1000) / 8;
    }
}

sub descriptor {
    my $format = shift;
    my @rates = rates;
    my $desc = 2; # Mp3 gets 2.., descriptor is based on pos in rates array
    foreach my $rate (@rates) {
	if ($rate eq $format->rate) {
	    return $desc;
	}
	$desc++;
    }
    Trace::error('No descriptor for format of rate %s', $format->rate);
    return undef;
}

sub bestQuality {
    my @formats = @_;

    my $best;
    foreach my $format (@formats) {
	next if ($format->type ne 'Mp3');
	if (!defined($best) || ($best->descriptor > $format->descriptor)) {
	    $best = $format;
	}
    }
    return $best;
}

1;
