# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.4 $

package Format::Wav;

use Format;
@ISA = qw( Format );
use IO::File qw();
use Trace;
use strict;

sub new {
    my $proto = shift;
    my $type = ref($proto) || $proto;
    my $format = $type->SUPER::new(@_);
    if (!defined($format)) {
	return undef;
    }

    my %parms = @_;

    if (defined($parms{'file'})) {

	# Try and use the suffix or get the format info by reading the
	# file header

	if ($parms{'file'} =~ m/_(\d+)hz_(\w)_(\d+)bit\.wav/) {
	    $format->{'samplesPerSecond'} = $1;
	    $format->{'bitsPerSample'} = $3;
	    if ($2 eq 'mono') {
		$format->{'channels'} = 1;
	    }
	    elsif ($2 eq 'stereo') {
		$format->{'channels'} = 2;
	    }
	    elsif ($2 =~ /(\d+)channel/) {
		$format->{'channels'} = $1;
	    }
	}
	else {
	    my $fh = IO::File->new;
	    if (!$fh->open($parms{'file'})) {
		Trace::error('Cannot open file: %s: %s', $parms{'file'}, $!);
		return undef;
	    }
	    my $header;
	    if ($fh->read($header, 44) != 44) {
		Trace::error('Cannot read Wav file header: %s', $!);
		$fh->close;
		return undef;
	    }
	    $fh->close;

	    (my $fourcc, my $rcl, my $wave, my $fmt, my $fcl, my $ftag, 
	     my $channels, my $samplesPerSecond, my $avgBytesPerSec, 
	     my $blockAlign, my $bitsPerSample) = 
		 unpack('A4lA4A4lssllss', $header);

	    $format->{'samplesPerSecond'} = $samplesPerSecond;
	    $format->{'bitsPerSample'} = $bitsPerSample;
	    $format->{'channels'} = $channels;
	}
	$format->{'file'} = $parms{'file'};
	$format->{'duration'} = ((-s $parms{'file'}) / (($format->{'samplesPerSecond'} * $format->{'bitsPerSample'} * $format->{'channels'}) / 8));
    }
    else {
	$format->{'samplesPerSecond'} = $parms{'samplesPerSecond'} || 44100;
	$format->{'bitsPerSample'} = $parms{'bitsPerSample'} || 16;
	$format->{'channels'} = $parms{'channels'} || 2;
    }

    return $format;
}

sub suffix {
    my $format = shift;
    my %parms = @_;
    my $suffix = '';

    if ($parms{'unique'}) {
	$suffix = '_'.$format->{'samplesPerSecond'}.'hz';
	if ($format->{'channels'} == 1) {
	    $suffix .= '_mono';
	}
	elsif ($format->{'channels'} == 2) {
	    $suffix .= '_stereo';
	}
	else {
	    $suffix .= '_'.$format->{'channels'}.'channel';
	}
	$suffix .= '_'.$format->{'bitsPerSample'}.'bit';
    }

    $suffix .= '.wav';
}

sub samplesPerSecond { $_[0]->{'samplesPerSecond'}; }
sub bitsPerSample { $_[0]->{'bitsPerSample'}; }
sub channels { $_[0]->{'channels'}; }

# expFileSize() - Returns the expected size of a file of this format

sub expFileSize {
    ($_[0]->{'duration'} * 
     $_[0]->{'samplesPerSecond'} * 
     $_[0]->{'bitsPerSample'} *
     $_[0]->{'channels'}) / 8;
}

sub descriptor {
    return 1; # Reserved for wav 
}

1;

