# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.9 $

package GtkEx::CList;

use Gtk;
@ISA = qw( Gtk::CList );
use Trace;
use strict;

sub new { 
    my $proto = shift;
    my $type = ref($proto) || $proto;

    my $list;
    my %parms;
    if ($#_ == 0) {
	$list = $type->SUPER::new($_[0]);
    }
    else {
	%parms = @_;
	if (defined($parms{'columns'}) && !defined($parms{'titles'})) {
	    $list = $type->SUPER::new($parms{'columns'});
	}
	elsif (defined($parms{'titles'})) {
	    $list = $type->SUPER::new_with_titles(map { "$_" } @{ $parms{'titles'} });
	}
	else {
	    return undef;
	}
	$list->{'_titles'} = $parms{'titles'};
	$list->{'_columns'} = $parms{'columns'} || ($#{ $parms{'titles'} } + 1);
    }
    return undef if (!defined($list));

    if ($parms{'style'}) {
	$list->set_style($parms{'style'});
    }
    if ($parms{'menu'}) {
	$list->{'_menu'} = $parms{'menu'};
    }
    if (defined($parms{'tooltips'}) && defined($parms{'tips'}) && defined($parms{'titles'})) {
	$list->{'_tooltips'} = $parms{'tooltips'};
	$list->{'_tips'} = $parms{'tips'};
	my $col = 0;
	foreach my $tip (@{ $parms{'tips'} }) {
	    my $label = Gtk::Label->new($parms{'titles'}->[ $col ]);
	    $label->set_justify('left');
	    $label->show();
	    my $event = Gtk::EventBox->new;
	    $event->show;
	    $parms{'tooltips'}->set_tip($event, $tip, '');
	    my $box = Gtk::HBox->new(0, 0);
	    $box->border_width(0);
	    $box->show();
	    $box->pack_start($label, 0, 0, 0);
	    $event->add($box);
	    $list->set_column_widget($col, $event);
	    $col++;
	}
    }
    if (ref($list) ne $proto) {
	bless $list, $type;
    }

    $list->signal_connect('event', 'GtkEx::CList::_event');
    $list->signal_connect('select-row', 'GtkEx::CList::_selectRow');
    return $list;
}

sub new_with_titles { 
    my $proto = shift;
    my $type = ref($proto) || $proto;
    return $type->new('titles' => \@_);
}

sub show {
    my $list = shift;
    $list->SUPER::show(@_);
    _clear(0, $list);
}

sub set_row_height {
    my $list = shift;
    $list->SUPER::set_row_height(@_);
    $list->{'_row_height'} = $_[0];
}

sub get_row_height { defined($_[0]->{'_row_height'}) ? $_[0]->{'_row_height'} : 14; }

sub setEdit {
    my $list = shift;
    my %parms = @_;

    if (defined($parms{'edit'}) && $parms{'edit'} && !defined($list->{'_edit'})) {

	my $edit = $list->{'_edit'} = {};

	# Vbox with 3 rows:
	#   the list
	#   add/edit area
	#   buttons

	my $vbox = Gtk::VBox->new(0, 5);
	$vbox->border_width(0);

	# List

	my $scroll = Gtk::ScrolledWindow->new(undef, undef);
	$scroll->set_policy('automatic', 'automatic');
	$scroll->show();
	$scroll->add_with_viewport($list);
	$list->column_titles_passive();
	$list->parent->set_shadow_type('none');
	$vbox->pack_start($scroll, 1, 1, 0);	

	# Add/Edit area

	$edit->{'table'} = my $table = GtkEx::TableEdit->new(colSpacing => 5, rowSpacing => 5, tooltips => $list->{'_tooltips'});
	$table->border_width(0);
	$table->show();

	$edit->{'fields'} = [];
	for (my $pos = 0; $pos < $list->{'_columns'}; $pos++) {

	    my $fieldName = $list->{'_titles'}->[$pos] if (defined($list->{'_titles'})) || '';

	    # Default to text entry
		
	    my $input = Gtk::Entry->new(1024);
	    $input->show;
	    $edit->{'fields'}->[$pos] = { edit => 0, entry => $input, container => $input, title => $fieldName };

	    $table->appendRow(title => $fieldName.': ', widget => $edit->{'fields'}->[$pos]->{'container'}, 
			      tip => defined($list->{'_tips'}) ? $list->{'_tips'}->[$pos] : undef);
	    $table->deactivateRow(row => $pos);

	    $list->set_column_auto_resize($pos, 1);
	}
	$vbox->pack_start($table, 0, 0, 0);

	# Buttons

	my $buttonBox = Gtk::HButtonBox->new();
	$buttonBox->show;
	$buttonBox->set_layout('edge');
	$buttonBox->set_spacing(5);
	$buttonBox->set_child_ipadding(15, 0);
	$buttonBox->set_child_size(0, 0);

	my $button = Gtk::Button->new('Add');
	$button->signal_connect('clicked', 'GtkEx::CList::_add', $list);
	$button->show();
	$buttonBox->add($button);
	$button = Gtk::Button->new('Delete');
	$button->signal_connect('clicked', 'GtkEx::CList::_delete', $list);
	$button->show();
	$buttonBox->add($button);
	$button = Gtk::Button->new('Modify');
	$button->signal_connect('clicked', 'GtkEx::CList::_modify', $list);
	$button->show();
	$buttonBox->add($button);
	$button = Gtk::Button->new('Clear');
	$button->signal_connect('clicked', 'GtkEx::CList::_clear', $list);
	$button->show();
	$buttonBox->add($button);
	$vbox->pack_start($buttonBox, 0, 0, 0);

	$list->{'_container'} = $vbox;
	#$list->column_titles_hide;
	#$vbox->set_usize(50, 150);
	$list->parent->set_usize(1, 75); # Minimum height
    }
    elsif ((!defined($parms{'edit'}) || !$parms{'edit'}) && defined($list->{'_edit'})) {
	delete $list->{'_container'};
	delete $list->{'_edit'};
    }
}

sub getEdit {
    my $list = shift;
    if (!defined($list->{'_edit'})) {
	return { edit => 0 };
    }
    else {
	return { edit => 1, container => $list->{'_container'}};
    }
}

sub setEditColumn {
    my $list = shift;
    my %parms = @_;

    return if (!defined($parms{'column'}));

    if ($parms{'edit'}) {
	my $edit = $list->{'_edit'};
	if (!defined($edit)) {
	    $list->setEdit(edit => 1);
	    $edit = $list->{'_edit'};
	}

	if (defined($parms{'container'})) {
	    # New container widget
	    $edit->{'table'}->placeRow(row => $parms{'column'},
				       title => $edit->{'fields'}->[ $parms{'column'} ]->{'title'}.': ',
				       widget => $parms{'container'});
	    $edit->{'fields'}->[$parms{'column'}]->{'container'} = $parms{'container'};
	    $edit->{'fields'}->[$parms{'column'}]->{'entry'} = $parms{'entry'};
	}
	$edit->{'table'}->activateRow(row => $parms{'column'});
	$edit->{'fields'}->[ $parms{'column'} ]->{'edit'} = 1;
    }
    else {
	my $edit = $list->{'_edit'};
	$edit->{'table'}->deactivateRow(row => $parms{'column'});
	$edit->{'fields'}->[$parms{'column'}]->{'edit'} = 0;
    }
}

sub getEditColumn {
    my $list = shift;
    my %parms = @_;

    return if (!defined($parms{'column'}));

    if (!defined($list->{'_edit'})) {
	return { edit => 0 };
    }
    else {
	return $list->{'_edit'}->{'fields'}->[$parms{'column'}];
    }
}

sub setEditRowAdd {
    my $list = shift;
    my %parms = @_;
    
    return if (!defined($parms{'handler'}));
    return if (!defined($list->{'_edit'}));

    my $h = $parms{'handler'}; delete $parms{'handler'};
    $list->{'_edit'}->{'addHandler'} = { handler => $h, parms => \%parms };
}

sub setEditRowDelete {
    my $list = shift;
    my %parms = @_;
    
    return if (!defined($parms{'handler'}));
    return if (!defined($list->{'_edit'}));

    my $h = $parms{'handler'}; delete $parms{'handler'};
    $list->{'_edit'}->{'deleteHandler'} = { handler => $h, parms => \%parms };
}

sub setEditRowModify {
    my $list = shift;
    my %parms = @_;
    
    return if (!defined($parms{'handler'}));
    return if (!defined($list->{'_edit'}));

    my $h = $parms{'handler'}; delete $parms{'handler'};
    $list->{'_edit'}->{'modifyHandler'} = { handler => $h, parms => \%parms };
}

sub setEditColumnGet {
    my $list = shift;
    my %parms = @_;

    return if (!defined($parms{'column'}));
    return if (!defined($parms{'handler'}));
    return if (!defined($list->{'_edit'}));

    my $h = $parms{'handler'}; delete $parms{'handler'};
    $list->{'_edit'}->{'fields'}->[ $parms{'column'} ]->{'getHandler'} = 
    { handler => $h, parms => \%parms };
}

sub setEditColumnSet {
    my $list = shift;
    my %parms = @_;

    return if (!defined($parms{'column'}));
    return if (!defined($parms{'handler'}));
    return if (!defined($list->{'_edit'}));

    my $h = $parms{'handler'}; delete $parms{'handler'};
    $list->{'_edit'}->{'fields'}->[ $parms{'column'} ]->{'setHandler'} = 
    { handler => $h, parms => \%parms };
}

sub setEditColumnDefault {
    my $list = shift;
    my %parms = @_;

    return if (!defined($parms{'column'}));
    return if (!defined($parms{'handler'}));
    return if (!defined($list->{'_edit'}));

    my $h = $parms{'handler'}; delete $parms{'handler'};
    $list->{'_edit'}->{'fields'}->[ $parms{'column'} ]->{'defaultHandler'} = 
    { handler => $h, parms => \%parms };
}

sub getRow {
    my $list = shift;
    my $row = shift;
    my @values;

    my $pos = 0;
    my $value;
    while ($value = $list->get_text($row, $pos)) {
	push @values, $value;
	$pos++;
    }
    return @values;
}

sub _event {
    my $list = shift;
    my $event = shift;

    return 0 if (!defined($event));

    if (1) {
	return 0 if (($event->{'type'} eq 'expose') ||
		     ($event->{'type'} eq 'map') ||
		     ($event->{'type'} eq 'unmap') ||
		     ($event->{'type'} eq 'motion_notify') ||
		     ($event->{'type'} eq 'leave_notify') ||
		     ($event->{'type'} eq 'enter_notify'));
    
	Trace::debug(ref($list).'::'.$event->{'type'}) if (Trace::check(Trace::DEBUG3));
	if (0) {
	    require 'dumpvar.pl';
	    main::dumpValue($event);
	    $| = 1;
	}
    }

    # Button two behaves as button one, button three behaves as button
    # one unless it is pressed over a selected area - in any case it
    # brings up a menu.

    if ($event->{'type'} eq 'button_press') {

	(my $row, my $col) = $list->get_selection_info($event->{'x'}, $event->{'y'});

	Trace::debug('Button press at %d, %d (%d, %d) for %s', $row, $col, $event->{'x'}, $event->{'y'},
		     $list->selection_mode) if Trace::check(Trace::DEBUG6);

	# Press on column resize gives co-ordinates in this range, so we
	# sacrifice this space and return 0 to allow resize.
	return 0 if (($event->{'x'} <= 6) && ($event->{'y'} <= 20));

	if ($list->selection_mode eq 'multiple') {

	    if (($event->{'button'} != 3) || !$list->_selected($row)) {

		if (($event->{'state'} & 0x0f) == 0) {
		    my @sel = $list->selection;
		    if (($#sel != 0) || !$list->_selected($row)) {
			$list->unselect_all();
		    }
		    $list->signal_emit_by_name('select_row', $row, $col, $event);
		    $list->{'_lastRow'} = $row;
		}
		elsif (($event->{'state'} & 0x0f) == 4) {
		    # Ctrl click
		    if ($list->_selected($row)) {
			$list->signal_emit_by_name('unselect_row', $row, $col, $event);
		    }
		    else {
			$list->signal_emit_by_name('select_row', $row, $col, $event);
		    }
		    $list->{'_lastRow'} = $row;
		}
		elsif (($event->{'state'} & 0x0f) == 1) {
		    # Shift click, select from last to point
		    $list->unselect_all;
		    if ($row >= $list->{'_lastRow'}) {
			for(; $row >= $list->{'_lastRow'}; $row--) {
			    $list->signal_emit_by_name('select_row', $row, $col, $event);
			}
		    }
		    else {
			for(; $row <= $list->{'_lastRow'}; $row++) {
			    $list->signal_emit_by_name('select_row', $row, $col, $event);
			}
		    }
		}
	    }
	}
	elsif (($list->selection_mode eq 'single') || ($list->selection_mode eq 'browse')) {

	    if (($event->{'button'} != 3) || !$list->_selected($row)) {
		$list->signal_emit_by_name('select_row', $row, $col, $event);
	    }
	}

	if (($event->{'button'} == 3) && ($list->{'_menu'})) {
	    $list->{'_menu'}->popup(0, 0, $event->{'button'}, $event->{'time'});
	}
	$list->grab_focus;
	return 1;
    }

    return 0;
}

# _selectRow

sub _selectRow {
    my $list = shift;
    my $row = shift;
    my $col = shift;
    my $event = shift;

    if ($list->row_is_visible($row) ne 'full') {
	$list->moveto($row, $col, 0.5, 0.5);
    }
    if ($list->{'_edit'}) {
	$col = 0;
	foreach my $field (@{ $list->{'_edit'}->{'fields'} }) {
	    my $value = $list->get_text($row, $col);
	    if (defined($field->{'setHandler'})) {
		&{ $field->{'setHandler'}->{'handler'} }($list, value => $value,
							 entry => $field->{'entry'},
							 container => $field->{'container'},
							 %{ $field->{'setHandler'}->{'parms'} });
	    }
	    else {
		# Assume text entry default
		$field->{'entry'}->set_text($value);
	    }
	    $col++;
	}
    }
    return 0;
}

sub _selected {
    my $list = shift;
    my $row = shift;

    my @sel = $list->selection;
    foreach my $srow (@sel) {
	if ($srow == $row) {
	    return 1;
	}
    }
    return 0;
}

sub _add {
    my $widget = shift;
    my $list = shift;

    my @values = ();
    foreach my $field (@{ $list->{'_edit'}->{'fields'} }) {
	if (defined($field->{'getHandler'})) {
	    push @values, &{ $field->{'getHandler'}->{'handler'} }($list, 
							 entry => $field->{'entry'},
							 container => $field->{'container'},
						   %{ $field->{'getHandler'}->{'parms'} });
	}
	else {
	    # Assume text entry default
	    push @values, $field->{'entry'}->get_text;
	}
    }

    if (!defined($list->{'_edit'}->{'addHandler'}) ||
	&{ $list->{'_edit'}->{'addHandler'}->{'handler'} }($list, values => \@values,
					   %{ $list->{'_edit'}->{'addHandler'}->{'parms'} })) {
	$list->append(@values);
    }

    return 1;
}

sub _modify {
    my $widget = shift;
    my $list = shift;

    my @values = ();
    foreach my $field (@{ $list->{'_edit'}->{'fields'} }) {
	if (defined($field->{'getHandler'})) {
	    push @values, &{ $field->{'getHandler'}->{'handler'} }($list, 
							 entry => $field->{'entry'},
							 container => $field->{'container'},
						   %{ $field->{'getHandler'}->{'parms'} });
	}
	else {
	    # Assume text entry default
	    push @values, $field->{'entry'}->get_text;
	}
    }

    my @sel = $list->selection;
    return 1 if (!@sel || ($#sel == -1));

    my @oldValues = $list->getRow($sel[0]);
    if (!defined($list->{'_edit'}->{'modifyHandler'}) ||
	&{ $list->{'_edit'}->{'modifyHandler'}->{'handler'} }($list, values => \@values,
							      oldValues => \@oldValues,
				      %{ $list->{'_edit'}->{'modifyHandler'}->{'parms'} })) {
	$list->remove($sel[0]);
	$list->insert($sel[0], @values);
    }

    return 1;
}

sub _delete {
    my $widget = shift;
    my $list = shift;
    my @values;

    my @sel = $list->selection;
    return if (!@sel || ($#sel == -1));

    foreach my $row (@sel) {
	@values = $list->getRow($row);

	if (!defined($list->{'_edit'}->{'deleteHandler'}) ||
	    &{ $list->{'_edit'}->{'deleteHandler'}->{'handler'} }($list, values => \@values,
								  %{ $list->{'_edit'}->{'deleteHandler'}->{'parms'} })) {
	    $list->remove($row);
	}
    }

    return 1;
}

sub _clear {
    my $widget = shift;
    my $list = shift;

    my @values = ();
    foreach my $field (@{ $list->{'_edit'}->{'fields'} }) {
	if (defined($field->{'defaultHandler'})) {
	    &{ $field->{'defaultHandler'}->{'handler'} }($list, 
							 entry => $field->{'entry'},
							 container => $field->{'container'},
						 %{ $field->{'defaultHandler'}->{'parms'} });
	}
	else {
	    # Assume text entry default
	    $field->{'entry'}->set_text('');
	}
    }

    return 1;
}

sub _testRow {
    my $list = shift;
    my @values = @_;

    return if (!defined($list->{'_testRow'}));

    foreach my $test (@{ $list->{'_testRow'} }) {

	if (UNIVERSAL::isa($test, 'CODE')) {
	    if (! &{ $test }($list, @values)) {
		return 0;
	    }
	}
	elsif ($test eq 'nonNull') {
	    foreach my $value (@values) {
		if (!defined($value) || ($value eq '')) {
		    $list->_testError('Please enter a value in each field');
		    return 0;
		}
	    }
	}
	elsif ($test eq 'unique') {
	    my @cValues;
	    my $row = 0;
	    while (@cValues = $list->getRow($row)) {
		if (join('', @values) eq join('', @cValues)) {
		    $list->_testError('This is already in the list');
		    return 0;
		}
		$row++;
	    }
	}
    }
    return 1;
}

sub _testError {
    my $list = shift;

    return if (!defined($list->{'_testError'}));

    &{ $list->{'_testError'} }(@_);
}

1;
