# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.25.2.1 $

package Gui;

use Thing;
@ISA = qw( Thing );

# User friendly attempt to load Gtk, Gtk::Gdk

BEGIN {
    foreach my $testlib ('Gtk.pm', 'Gtk/Gdk.pm') {
	eval {
	    require $testlib;
	};
	if ($@) {
	    printf(STDERR "Cannot load GUI: Gtk Perl must be installed.\n");
	    exit 0;
	}
    }
}

use Gtk 0.7005;
use Gtk::Gdk;
use Gui::Window;
use Gui::Page::Browser;
use Gui::Page::Encoding;
use Gui::Page::Config;
use Gui::Page::Diagnostic;
use Gui::Base;
use Gui::Node;
use Thing::ServiceH;
use Trace;
use strict;

sub new {
    my $proto = shift;
    my $type = ref($proto) || $proto;

    Trace::debug('Starting gui on: %s', $ENV{'DISPLAY'}) if Trace::check(Trace::DEBUG3);

    Gtk->init();

    my $rcfile = $ENV{'HOME'}."/.emptytree-seedy/rc";
    Gtk::Rc->parse($rcfile) if (-e $rcfile);

    # Password checking, Gui is not started by ../bin/thing in order
    # to prompt for the password somewhere other than the command line.

    my $pwd = Thing::readPwd;
    if (!$pwd) {

	my $canceled = 0;
	my $entered = 0;

	my $table = GtkEx::TableEdit->new(colSpacing => 5, rowSpacing => 5);
	$table->show();
	my $entry = Gtk::Entry->new(1024);
	$entry->set_usize(100, 20);
	$entry->set_text('');
	$entry->set_visibility(0);
	$entry->show;
	$table->appendRow(title => 'Password:', widget => $entry);
	my $choice = GtkEx::Utility::choicePopup(widget => $table, title => 'Emptytree Seedy Password', 
						 msg => 'Enter your Emptytree Seedy password.', 
						 okFunc => sub { $entered = 1; 0; }, cancelFunc => sub { $canceled = 1; 1; });

	while (!$canceled) {
	    Gtk->main_iteration;
	    if ($entered) {
		$pwd = $entry->get_text;
		if (!defined($pwd) || ($pwd eq '')) {
		    GtkEx::Utility::ackPopup(msg => 'Please enter a password');
		    $entered = 0;
		    next;
		}
		else {
		    $ENV{'THING_PWD'} = $pwd;
		    $choice->destroy;
		    last;
		}
	    }
	}
	if ($canceled) {
	    Gtk->exit(0);
	}
    }

    # Service port, again Gui is not started by ../bin/thing so this
    # may need to be read from a config file.

    if (!defined($ENV{THING_SERVICEPORT})) {
	Thing::Service::readConfig;
    }

    my $gui = $type->SUPER::new(@_);
    if ($gui->error) {
	my $ok = 0;
	GtkEx::Utility::ackPopup(msg => $gui->error, func => sub { $ok = 1; 1; });
	while (!$ok) {
	    Gtk->main_iteration;
	}
	return $gui;
    }

    $gui->{'pages'} = { array => [] };
    $gui->{'windows'} = [];
    $gui->{'windowId'} = 1;
    $gui->{'things'} = {};
    $gui->{'dragData'} = {};

    $gui->{'border'} = 5;
    $gui->{'spacing'} = 5;
    $gui->{'rowHeight'} = 18;

    # Default Style

    $gui->{'style'} = Gtk::Style->new();

    # Tooltips

    $gui->{'tooltips'} = Gtk::Tooltips->new;
    $gui->{'tooltips'}->enable;

    # Pixmaps

    my @dirs = split('/', $ENV{'THING_LIBPATH'});
    pop @dirs;
    push @dirs, 'pixmaps';
    $gui->{'pixmapDir'} = join('/', @dirs);

    if (!(-e $gui->{'pixmapDir'}) || !(-d _) || !(-r _)) {
	Trace::error('Problem with pixmap directory: %s: %s', $gui->{'pixmapDir'}, $!);
	$gui->error(Trace::lastError);
	return $gui;
    }

    # Music Base

    $gui->{'base'} = Gui::Base->new(gui => $gui);
    $gui->{'baseId'} = 1;

    # Music browser

    my $page = Gui::Page::Browser->new(gui => $gui, baseId => $gui->{'baseId'}++);
    if (!$page) {
	$gui->error(Trace::lastError);
	sleep 2;
	return $gui;
    }
    push @{ $gui->{'pages'}->{'array'} }, $page;

    # Initial window

    my $initial = Gui::Window->new(page => $page, level => 'toplevel', id => $gui->{'windowId'}++, gui => $gui);
    push @{ $gui->{'windows'} }, $initial;

    # Config will be loaded by the Config page, signaling a possible need to redraw

    $gui->register(name => Thing::Service::ConfigLoadResponse, handler => '_configResponse', last => 1);
    $gui->register(name => Thing::ConfigureResponse, handler => '_configResponse', last => 1);

    # Config

    $page = Gui::Page::Config->new(gui => $gui);
    if (!$page) {
	$gui->error(Trace::lastError);
	sleep 2;
	return $gui;
    }
    push @{ $gui->{'pages'}->{'array'} }, $page;
    $gui->{'pages'}->{'config'} = $page;

    # Encoding

    $page = Gui::Page::Encoding->new(gui => $gui);
    if (!$page) {
	$gui->error(Trace::lastError);
	sleep 2;
	return $gui;
    }
    push @{ $gui->{'pages'}->{'array'} }, $page;
    $gui->{'pages'}->{'encoding'} = $page;

    # Diagnostics

    $page = Gui::Page::Diagnostic->new(gui => $gui);
    if (!$page) {
	$gui->error(Trace::lastError);
	sleep 2;
	return $gui;
    }
    push @{ $gui->{'pages'}->{'array'} }, $page;
    $gui->{'pages'}->{'diagnostics'} = $page;

    # Record start for Gtk:main

    $gui->{'start'} = time;

    $DB::single = 1;

    # Pop the config window when nothing has been configured

    my $initialStart = sub {
	if (defined($gui->{'pages'}->{'config'}->initialStart)) {
	    if ($gui->{'pages'}->{'config'}->initialStart) {
		$gui->createWindow($gui->{'pages'}->{'config'});
	    }
	    return 0;
	}
	else {
	    return 1;
	}
    };
    Gtk->timeout_add(200, $initialStart);

    return $gui;
}

sub _configResponse {
    my $gui = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};

    # Config is loaded or changed by the Config page, the Gui traps
    # the responses in order to update other pages. It may not be
    # complete at this point.

    my $config = $gui->{'pages'}->{'config'}->config;
    my @allThings = map { values(%{ $_->{'things'} }) } values(%{ $config });

    my @list1 = grep { $_->{'config'}->{'active'} && ($_->{'config'}->{'type'} eq 'Controller') } @allThings;
    my $controller = 0;
    if ($#list1 != -1) {
	$controller = $list1[0];
    }

    my @list2 = grep { $_->{'config'}->{'active'} && ($_->{'config'}->{'type'} eq 'Repo') } @allThings;
    my $repo = 0;
    if ($#list2 != -1) {
	$repo = $list2[0];
    }

    # Steal the Repo db config for this context

    if ($repo && defined($repo->{'config'}->{'db'}) && (!Info::Persist::configure(db => $repo->{'config'}->{'db'}->{'local'}) ||
							!Info::Cd::configure(remotes => $repo->{'config'}->{'db'}->{'remotes'}))) {
	$gui->diagnostic->add(type => 'Database', name => 'gui', summary => 'Configuration error', 
			      diagnostic => 'Error configuring database:'."\n".Trace::lastError);
    }

    # Music base

    my @players = grep { $_->{'config'}->{'active'} && ($_->{'config'}->{'type'} eq 'Player') } @allThings;
    my @cdroms = grep { $_->{'config'}->{'active'} && ($_->{'config'}->{'type'} eq 'Cdrom') } @allThings;

    if ($controller && $repo && 
	!$gui->{'base'}->configure(repo => $repo->{'handle'}, controller => $controller->{'handle'}, 
				   players => [ map { $_->{'handle'} } @players ], cdroms => [ map { $_->{'handle'} } @cdroms ])) {
	$gui->diagnostic->add(type => 'Browser', name => 'gui', summary => 'Configuration error', diagnostic => Trace::lastError);
    }

    # Encoding status

    if ($controller && !$gui->{'pages'}->{'encoding'}->configure(controller => $controller->{'handle'})) {
	$gui->diagnostic->add(type => 'Encoding', name => 'gui', summary => 'Configuration error', diagnostic => Trace::lastError);
    }

    return 1;
}

# Attributes

sub border        { $_[0]->{'border'}; }
sub spacing       { $_[0]->{'spacing'}; }
sub tooltips      { $_[0]->{'tooltips'}; }
sub busy          { $_[0]->{'busy'}; }
sub accelGroup    { $_[0]->{'accelGroup'}; }
sub style         { $_[0]->{'style'}; }
sub pixmapDir     { $_[0]->{'pixmapDir'}; }
sub diagnostic    { $_[0]->{'pages'}->{'diagnostics'}; }
sub config        { $_[0]->{'pages'}->{'config'}; }
sub pages         { (@{ $_[0]->{'pages'}->{'array'} }); }

sub rowHeight     { $_[0]->{'rowHeight'}; }
sub eventDebug    { defined($ENV{GUI_EVENT_DEBUG}) ? $ENV{GUI_EVENT_DEBUG} : 0; }

sub base          { $_[0]->{'base'}; }

# monitorChange - Pass to pages

sub monitorChange {
    my $gui = shift;
    my $handle = shift;
    $gui->{'base'}->monitorChange($handle);
    $gui->{'pages'}->{'config'}->monitorChange($handle) if (defined($gui->{'pages'}->{'config'}));
    $gui->{'pages'}->{'encoding'}->monitorChange($handle) if (defined($gui->{'pages'}->{'encoding'}));
}

# windowDestroyed - A Gui::Window has been closed, exit if its the
# last window, hide or destroy the page.

sub windowDestroyed {
    my $gui = shift;
    my %parms = @_;

    my $window = $parms{'window'} || die;

    if ($#{ $gui->{'windows'} } == 0) {
	Gtk->exit(0);
    }

    my $pos = 0;
    foreach my $existing (@{ $gui->{'windows'} }) {
	if ($window == $existing) {
	    splice @{ $gui->{'windows'} }, $pos, 1;
	    last;
	}
	$pos++;
    }

    if ($window->page->isa('Gui::Page::Browser')) {
	$window->page->destroy;
    }
    return 1;
}

# topbarCreate - Create a generic topbar menu & busybox for a window.

sub topbarCreate {
    my $gui = shift;
    my %parms = @_;

    my $window = $parms{'window'} || die;

    my $hbox = Gtk::HBox->new(0, 0);
    $hbox->border_width(0);
    $hbox->show();

    # Menu

    $window->{'menu'} = my $menu = Gtk::MenuBar->new();
    $menu->set_shadow_type('none');
    $menu->show;
    $hbox->pack_start($menu, 1, 1, 0);

    # File
    my $menuItem = Gtk::MenuItem->new('File');
    $menuItem->show();
    $menu->append($menuItem);
    my $drop = Gtk::Menu->new();
    $drop->show;
    $menuItem->set_submenu($drop);

    # File -> New Window
    my $newWindow = Gtk::MenuItem->new('New Window');
    $newWindow->signal_connect('activate', sub { $gui->createBrowserWindow; });
    $drop->append($newWindow);
    $newWindow->show();

    # File -> Close
    my $close = Gtk::MenuItem->new('Close');
    $close->signal_connect('activate', sub { $window->destroy; });
    $drop->append($close);
    $close->show();

    # File -> Quit
    my $quit = Gtk::MenuItem->new('Quit');
    if (defined($DB::profile) && $DB::profile) {
	# SmallProf.pm with END changed to sub forceEND
	$quit->signal_connect('activate', sub { $DB::profile = 0; DB::forceEND(); Gtk->exit(0); });
    }
    else {
	$quit->signal_connect('activate', sub { Gtk->exit(0); });
    }
    $drop->append($quit);
    $quit->show();

    # View
    $window->{'viewMenu'} = $menuItem = Gtk::MenuItem->new('View');
    $menuItem->signal_connect('activate', sub { $gui->drawViewSubmenu(window => $window); return 0; });
    $menuItem->show();
    $menu->append($menuItem);
    $window->{'viewSubmenu'} = $drop = Gtk::Menu->new();
    $drop->show;
    $menuItem->set_submenu($drop);

    # Lists
    $window->{'listsMenu'} = $menuItem = Gtk::MenuItem->new('Lists');
    $menuItem->signal_connect('activate', sub { $gui->drawListsSubmenu(window => $window); return 0; });
    $menuItem->show();
    $menu->append($menuItem);
    $window->{'listsSubmenu'} = $drop = Gtk::Menu->new();
    $drop->show;
    $menuItem->set_submenu($drop);

    if (!$window->page->isa('Gui::Page::Browser')) {
	$menuItem->set_sensitive(0);
    }

    # Go
    $window->{'goMenu'} = $menuItem = Gtk::MenuItem->new('Go');
    $menuItem->signal_connect('activate', sub { $gui->drawGoSubmenu(window => $window); return 0; });
    $menuItem->show();
    $menu->append($menuItem);
    $window->{'goSubmenu'} = $drop = Gtk::Menu->new();
    $drop->show;
    $menuItem->set_submenu($drop);

    if (!$window->page->isa('Gui::Page::Browser')) {
	$menuItem->set_sensitive(0);
    }

    # Spacer
    $menuItem = Gtk::MenuItem->new();
    $menuItem->show();
    $menu->append($menuItem);
    my $sbox = Gtk::HBox->new(0, 0);
    $sbox->show;
    $menuItem->add($sbox);
    my $slabel = Gtk::Label->new;
    $slabel->show;
    $sbox->pack_start($slabel, 1, 1, 0);

    my $vsep = Gtk::VSeparator->new;
    $vsep->show;
    $hbox->pack_start($vsep, 0, 0, 0);

    # Busy Indicator

    my $pdir = $gui->pixmapDir;
    my @pixmaps = map { sprintf('%s/22x22/%04d.xpm', $pdir, $_); } (1..50);
    $hbox->{'busy'} = GtkEx::BusyIcon->new(window => $window->window, style => $gui->style, pixmaps => \@pixmaps);
    $hbox->{'busy'}->show;
    $hbox->pack_start($hbox->{'busy'}, 0, 0, 0);

    return $hbox;
}

# drawViewSubmenu

sub drawViewSubmenu {
    my $gui = shift;
    my %parms = @_;

    my $window = $parms{'window'} || die;
    my $drop = $window->{'viewSubmenu'};

    my @children = $drop->children;
    map { $drop->remove($_) } @children;

    # Show/Hide Tree

    my $mi = Gtk::MenuItem->new('Show Tree');
    $mi->show;
    $drop->append($mi);

    if (!$window->page->isa('Gui::Page::Browser')) {
	$mi->set_sensitive(0);
    }
    else {
	if (!$window->page->treeHidden) {
	    $mi->child->set_text('Hide Tree');
	    $mi->signal_connect('activate', sub { $window->page->treeHide; });
	}
	else {
	    $mi->signal_connect('activate', sub { $window->page->treeShow; });
	}
    }

    # Show/Hide Configuration

    $mi = Gtk::MenuItem->new('Show Configuration');
    $mi->show;
    $drop->append($mi);

    if (!$gui->{'pages'}->{'config'}->hidden) {
	$mi->child->set_text('Hide Configuration');
	$mi->signal_connect('activate', sub { $gui->{'pages'}->{'config'}->window->destroy; });
    }
    else {
    	$mi->signal_connect('activate', sub { $gui->createWindow($gui->{'pages'}->{'config'}); });
    }

    # Show/Hide Encoding

    $mi = Gtk::MenuItem->new('Show Encoding Status');
    $mi->show;
    $drop->append($mi);

    if (!$gui->{'pages'}->{'encoding'}->hidden) {
	$mi->child->set_text('Hide Encoding Status');
	$mi->signal_connect('activate', sub { $gui->{'pages'}->{'encoding'}->window->destroy; });
    }
    else {
    	$mi->signal_connect('activate', sub { $gui->createWindow($gui->{'pages'}->{'encoding'}); });
    }

    # Show/Hide Diagnostics

    $mi = Gtk::MenuItem->new('Show Diagnostics');
    $mi->show;
    $drop->append($mi);

    if (!$gui->{'pages'}->{'diagnostics'}->hidden) {
	$mi->child->set_text('Hide Diagnostics');
	$mi->signal_connect('activate', sub { $gui->{'pages'}->{'diagnostics'}->window->destroy; });
    }    
    else {
    	$mi->signal_connect('activate', sub { $gui->createWindow($gui->{'pages'}->{'diagnostics'}); });
    }

    return 1;
}

# drawGoSubmenu

sub drawGoSubmenu {
    my $gui = shift;
    my %parms = @_;

    my $window = $parms{'window'} || die;
    my $drop = $window->{'goSubmenu'};

    my @children = $drop->children;
    map { $drop->remove($_) } @children;

    $window->page->menuCreateGoSubmenu($drop);

    return 1;
}

# drawListsSubmenu

sub drawListsSubmenu {
    my $gui = shift;
    my %parms = @_;

    my $window = $parms{'window'} || die;
    my $drop = $window->{'listsSubmenu'};

    my @children = $drop->children;
    map { $drop->remove($_) } @children;

    $window->page->menuCreateListsSubmenu($drop);

    return 1;
}

# createBrowserWindow - Popup a new browser.

sub createBrowserWindow {
    my $gui = shift;

    my $page = Gui::Page::Browser->new(gui => $gui, baseId => $gui->{'baseId'}++);
    push @{ $gui->{'pages'}->{'array'} }, $page;

    my $window = Gui::Window->new(page => $page, level => 'toplevel', id => $gui->{'windowId'}++, gui => $gui);
    push @{ $gui->{'windows'} }, $window;

    return 1;
}

# createWindow

sub createWindow {
    my $gui = shift;
    my $page = shift;

    my $window = Gui::Window->new(page => $page, level => 'toplevel', id => $gui->{'windowId'}++, gui => $gui);
    push @{ $gui->{'windows'} }, $window;

    return 1;
}

# dragData - Save and return drag data for different sources.

sub dragData {
    my $gui = shift;
    my %parms = @_;

    if (defined($parms{'data'})) {
	$gui->{'dragData'}->{ $parms{'id'} } = $parms{'data'};
    }
    return $gui->{'dragData'}->{ $parms{'id'} };
}

# work() - Gtk polls us after the initial startup

sub work {
    my $gui = shift;

    if (defined($gui->{'start'}) && !defined($gui->{'idleAdded'})) {
	$gui->{'idleAdded'} = 1;
	Gtk->timeout_add(50, sub { $gui->SUPER::work(); return 1; });
	Gtk->main;
    }
    else {
	$gui->SUPER::work(@_);
    }
}

sub loop { $_[0]->work; }

1;
