# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.8 $

package Gui::Base;

# Manages a base view of user interface objects. Objects are
# encapsulated as nodes and arranged in a tree hiearchy suitable for
# displaying with a Gtk::CTree. Handlers can be registered for node
# addition, deletion and update.

# Node types

sub Root         { 'Root'; }
sub Repo         { 'Repo'; }
sub Artist       { 'Artist'; }
sub Album        { 'Album'; }
sub Track        { 'Track'; }
sub List         { 'List'; }
sub Lists        { 'Lists'; }
sub Player       { 'Player'; }
sub Cdrom        { 'Cdrom'; }
sub Scratchpad   { 'Scratchpad'; }

use Thing;
use Thing::ServiceH;
use Gui::Node;
use Gui::Node::Player;
use Gui::Node::Cdrom;
use Info::List;
use Info::Track;
use CdromH;
use ControllerH;
use PlayerH;
use RepoH;
use Gtk;
use GtkEx::Utility;
use Trace;
use strict;

sub new {
    my $proto = shift;
    my $type = ref($proto) || $proto;
    my %parms = @_;

    my $base = {};
    bless $base, $type;

    Thing::context->register(name => 'info', handler => sub { shift; $base->_repoInfo(@_); });
    Thing::context->register(name => Player::Spooled, handler => sub { shift; $base->_playerSpooled(@_); });
    Thing::context->register(name => Player::Status, handler => sub { shift; $base->_playerStatus(@_); });
    Thing::context->register(name => Cdrom::Status, handler => sub { shift; $base->_cdromStatus(@_); });
    Thing::context->register(name => Controller::RipStatus, handler => sub { shift; $base->_cdromRipStatus(@_); });
    Thing::context->register(name => Thing::Info::Update, handler => sub { shift; $base->_objectUpdate(@_); }, last => 1);
    Thing::context->register(name => '_lists', handler => sub { shift; $base->_lists(@_); });
    Thing::context->register(name => '_listTracks', handler => sub { shift; $base->_listTracks(@_); });

    $base->{'registry'} = {};

    $base->{'populateQ'} = {};
    $base->{'populateStack'} = [];
    $base->{'populateQid'} = 0;

    $base->{'root'} = Gui::Node->gm_new(leaf => 0, refresh => 0, homogeneous => 0, type => Root);
    $base->{'scratchpad'} = Gui::Node->gm_new(parent => $base->rootNode, leaf => 0, refresh => 0, type => Scratchpad, 
					      homogeneous => 1, childType => Track);
    $base->{'scratchpad'}->gm_populated(1);
    $base->rootNode->gm_children([ $base->{'scratchpad'} ]);

    $base->{'gui'} = $parms{'gui'};

    return $base;
}

sub gui { $_[0]->{'gui'}; }
sub controller { $_[0]->{'controller'}; }

sub configure {
    my $base = shift;
    my %parms = @_;

    if (!defined($base->{'controller'}) || !Thing::Handle::cmpNameHost($base->{'controller'}, $parms{'controller'})) {
	$base->{'controller'} = $parms{'controller'};
	if ($base->{'controller'}->currentState eq Thing::Handle::Alive) {
	    Thing::context->message(name => Controller::RipState, to => $base->{'controller'});
	}
    }

    my $playerAt;
    if (defined($parms{'repo'}) && (!defined($base->{'repo'}) || !Thing::Handle::cmpNameHost($base->{'repo'}->what, $parms{'repo'}))) {
	# New or different repo
	if (defined($base->{'repo'})) {
	    $base->_nodeDelete($base->{'repo'});
	    $base->rootNode->gm_removeChild(child => $base->{'repo'});
	}
	$base->{'repo'} = Gui::Node->gm_new(parent => $base->rootNode, what => Thing::context->monitor($parms{'repo'}), 
					    type => Repo, leaf => 0, refresh => 1, homogeneous => 1, childType => Artist);
	$base->rootNode->gm_addChild(child => $base->{'repo'}, at => 0);
	$base->_nodeAdd($base->{'repo'});
	$playerAt = 2;
    }
    elsif (!defined($parms{'repo'}) && defined($base->{'repo'})) {
	# No repo
	$base->_nodeDelete($base->{'repo'});
	$base->rootNode->gm_removeChild(child => $base->{'repo'});
	delete $base->{'repo'};
	$playerAt = 1;
    }

    if (defined($parms{'repo'}) && Info::Persist::configured) {
	# Repo and configured database
	if (!defined($base->{'lists'})) {
	    $base->{'lists'} = Gui::Node->gm_new(parent => $base->rootNode, leaf => 0, refresh => 1, type => Lists, 
						 homogeneous => 1, childType => List);
	    $base->rootNode->gm_addChild(child => $base->{'lists'}, at => 1);
	    $base->_nodeAdd($base->{'lists'});
	}
	$playerAt = 3;
    }
    elsif (defined($base->{'lists'})) {
	$base->_nodeDelete($base->{'lists'});
	$base->rootNode->gm_removeChild(child => $base->{'lists'});
	delete $base->{'lists'};
    }

    my @things = ();
    if ($parms{'players'} && ($#{ $parms{'players'} } >= 0)) {
	push @things, map { { handle => Thing::context->monitor($_), type => Player, at => $playerAt } } 
	              @{ $parms{'players'} };
    }
    if ($parms{'cdroms'} && ($#{ $parms{'cdroms'} } >= 0)) {
	push @things, map { { handle => Thing::context->monitor($_), type => Cdrom, at => undef } }
	              @{ $parms{'cdroms'} };
    }

    foreach my $thing (@things) {

	# Addition

	my $exists = 0;
	foreach my $existing ($base->rootNode->children) {
	    next if (!defined($existing->what));
	    next if (!UNIVERSAL::isa($existing->what, 'Thing::Handle'));
	    next if (!Thing::Handle::cmpNameHost($thing->{'handle'}, $existing->what));
	    $exists = 1;
	    last;
	}
	if (!$exists) {
	    my $proto;
	    if ($thing->{'type'} eq Player) {
		$proto = 'Gui::Node::Player';
	    }
	    else {
		$proto = 'Gui::Node::Cdrom';
	    }
	    my $node = $proto->gm_new(parent => $base->rootNode, what => $thing->{'handle'}, leaf => 0 , refresh => 1, 
				      homogeneous => 1, type => $thing->{'type'}, childType => Track);
	    $base->rootNode->gm_addChild(child => $node, at => $thing->{'at'});
	    $base->_nodeAdd($node);
	}
    }

    # Removal

    foreach my $existing ($base->rootNode->children) {
	next if (!defined($existing->what));
	next if (!UNIVERSAL::isa($existing->what, 'Thing::Handle'));
	next if (($existing->type ne Player) && ($existing->type ne Cdrom));
	my $removed = $existing;
	foreach my $thing (@things) {
	    next if (!Thing::Handle::cmpNameHost($thing->{'handle'}, $existing->what));
	    $removed = 0;
	    last;
	}
	if ($removed) {
	    $base->_nodeDelete($existing);
	    $base->rootNode->gm_removeChild(child => $existing);
	}
    }
    return 1;
}

sub monitorChange {
    my $base = shift;
    my $handle = shift;

    if (Thing::Handle::cmpNameHost($handle, $base->{'controller'})) {
	if ($base->{'controller'}->currentState eq Thing::Handle::Alive) {
	    Thing::context->message(name => Controller::RipState, to => $base->{'controller'});
	}
	return 1;
    }

    my @nodes = $base->findNode(what => $handle);
    if (!@nodes) {
	Trace::debug('Monitor change not one of our nodes: %s', $handle->key) if Trace::check(Trace::DEBUG3);
	return;
    }

    # If it is dead we remove any children and mark it as empty.
    if ($handle->currentState eq Thing::Handle::Dead) {
	$base->depopulateNode(group => \@nodes);
	map { $_->gm_populated(1) } @nodes;
    }
    # If it switches to alive it needs population.
    elsif (($handle->previousState eq Thing::Handle::Dead) && ($handle->currentState eq Thing::Handle::Alive)) {
	map { $_->gm_populated(0) } @nodes;	
    }
    map { $base->_nodeUpdate($_) } @nodes;
    return 1;
}

sub _objectUpdate {
    my $base = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};
    my $ioid = $parms->{'ioid'};
    my $object = Thing::context->infoFind(ioid => $ioid);

    if (!$object) {
	Trace::debug('Not monitoring: %s', $ioid) if (Trace::check(Trace::DEBUG3));
	return;
    }

    # Special handling for Cdroms
    if ($object->isa('Info::Cd')) {
	$base->cdromObjectUpdate($object);
	return;
    }
    elsif ($object->isa('Info::Artist')) {
	# Artist name may have changed
	foreach my $cdrom ($base->cdroms) {
	    if (defined($cdrom->cd) && ($cdrom->cd->artist->ioid eq $object->ioid)) {
		$base->_nodeUpdate($cdrom);
	    }
	}
    }
    elsif ($object->isa('Info::Album')) {
	# Album name may have changed
	foreach my $cdrom ($base->cdroms) {
	    if (defined($cdrom->cd) && ($cdrom->cd->ioid eq $object->ioid)) {
		$base->_nodeUpdate($cdrom);
	    }
	}
    }
    elsif ($object->isa('Info::List')) {
	# Track listing may have changed
	if (defined($parms->{'msg'})) {
	    my $list = $base->findNode(what => $object);
	    if (!defined($list)) {
		Trace::debug('List updated not one of our nodes: %s', $object->name) if (Trace::check(Trace::DEBUG3));
		return;
	    }
	    if ($parms->{'msg'} eq Info::List::AddTracks) {
		$base->_listAdded($list, $parms->{'trackIoids'}, $parms->{'at'});
	    }
	    elsif ($parms->{'msg'} eq Info::List::DeleteTracks) {
		$base->_listRemoved($list, $parms->{'trackPositions'});
	    }
	    elsif ($parms->{'msg'} eq Info::List::MoveTrack) {
		$list->gm_moveChild( %{ $parms });
		$base->_nodeMove($list, %{ $parms });
	    }
	}
    }

    my @nodes = ();
    push @nodes, $base->findNode(what => $object);
    if ($object->isa('Info::Artist')) {
	# Dependants - Albums and Tracks
	my @dependants = $base->findNode(what => sub { (($_[0]->isa('Info::Album') && ($_[0]->artist->ioid eq $ioid)) ||
							($_[0]->isa('Info::Track') && ($_[0]->album->artist->ioid eq $ioid))) });
	push @nodes, @dependants;
    }
    elsif ($object->isa('Info::Album')) {
	# Dependants - Tracks
	my @dependants = $base->findNode(what => sub { ($_[0]->isa('Info::Track') && ($_[0]->album->ioid eq $ioid)) });
	push @nodes, @dependants;
    }

    if (!@nodes) {
	Trace::debug('Object update not one of our nodes: %s', $object->name) if (Trace::check(Trace::DEBUG3));
	return;
    }

    map {
	my $node = $_;
	# Children may have been affected so force a refresh
	if ($node->populated && !$object->isa('Info::List')) {
	    # Pretty costly if children have not changed TODO, will be fixed with sync messages ala Lists
	    $base->populateNode(node => $node, force => 1, recursive => 0, action => sub { $base->_nodeUpdate($node); });
	}
	else {
	    $base->_nodeUpdate($node);
	}
    } @nodes;

    return 1;
}

# register - Register handlers for update, add and delete of nodes under an id.

sub register {
    my $base = shift;
    my %parms = @_;

    $base->{'registry'}->{ $parms{'id'} } = \%parms;
}

# deregister - Deregister handlers

sub deregister {
    my $base = shift;
    my %parms = @_;

    delete $base->{'registry'}->{ $parms{'id'} };
}

# NODE - Node operations

# rootNode

sub rootNode { $_[0]->{'root'}; }

# findNode - find the nodes in the hierarchy by `what' they represent or
# verify that a `node' exists in the hierarchy.

sub findNode {
    my $base = shift;
    my %parms = @_;

    if (!defined($parms{'what'}) && !defined($parms{'node'})) {
	Trace::mandatory('what|node');
	return;
    }

    #Trace::debug("Tree: \n%s\n", $base->_dumpTree) if (Trace::check(Trace::DEBUG3));

    my @flat = ($base->rootNode);
    my @children = @flat;
    while (@children = map { if ($_->populated) { $_->children; } else { (); } } @children) {
	push @flat, @children;
    }

    my @nodes = ();
    my $find;
    if (defined($parms{'what'})) {
	$find = $parms{'what'};
	foreach my $node (@flat) {
	    next if (!defined($node->what));
	    if ((UNIVERSAL::isa($find, 'CODE') && &$find($node->what)) || (!UNIVERSAL::isa($find, 'CODE') && ($node->what == $find))) {
		push @nodes, $node;
	    }
	}
    }
    else {
	$find = $parms{'node'};
	foreach my $node (@flat) {
	    push @nodes, $node if ($node == $find);
	}
    }
    if (!@nodes) {
	Trace::debug('Cannot find %s in object base', $find) if Trace::check(Trace::DEBUG3);
	return;
    }
    return wantarray ? @nodes : $nodes[0];
}

# populateNode - Discover the children for a node or group of nodes,
# optionally `recursive' and with an `action' that will be called upon
# completion. `force' to populate an already populated node.

sub populateNode {
    my $base = shift;
    my %parms = @_;

    if (!defined($parms{'node'}) && !defined($parms{'group'})) {
	Trace::mandatory('node||group');
	return;
    }
    my $recurse = defined($parms{'recursive'}) ? $parms{'recursive'} : 0;
    my $action = $parms{'action'};
    my $force = defined($parms{'force'}) ? $parms{'force'} : 0;

    my $nodes;
    if (defined($parms{'group'})) {
	$nodes = $parms{'group'};
    }
    else {
	$nodes = [ $parms{'node'} ];
    }

    my $id = $base->{'populateQid'}++;
    $base->{'populateQ'}->{ $id } = { id => $id, nodes => $nodes, action => $action, recurse => $recurse, force => $force };

    if (Trace::performance) {
	$base->{'populateQ'}->{ $id }->{'time'} = [Time::HiRes::gettimeofday()];
    }

    map { Trace::debug('Population request %d for: %s, r -> %d, f -> %d', $id, $_->desc, $recurse, $force) if Trace::check(Trace::DEBUG6); } @{ $nodes };

    return $base->_populateCheck;
}

sub _populateCheck {
    my $base = shift;

    PENDING: foreach my $pending (sort { $a->{'id'} <=> $b->{'id'} } values(%{ $base->{'populateQ'} })) {

	my @nodes = @{ $pending->{'nodes'} };
	if ($pending->{'recurse'}) {
	    my @children = @nodes;
	    while (@children = map { if ($_->populated) { $_->children; } else { (); } } @children) {
		push @nodes, @children;
	    }
	}

	my $queryPending = 0;
	NODE: foreach my $node (@nodes) {
	    next if ($node->leaf);
	    next if (!$pending->{'force'} && $node->populated);
	    $queryPending++;
	    foreach my $query (@{ $base->{'populateStack'} }) {
		if ($query->{'node'} == $node) {
		    Trace::debug('Query pending for: %s', $node->desc) if Trace::check(Trace::DEBUG3);
		    next NODE;
		}
	    }
	    if ($#{ $base->{'populateStack'} } > 4) {
		Trace::debug('No room to query: %s', $node->desc) if Trace::check(Trace::DEBUG6);
		next PENDING;
	    }
	    $pending->{'force'} = 0;
	    if ($base->_populateNode($node)) {
		$queryPending--;
	    }
	}
	next PENDING if ($queryPending);

	# Finished, remove and call any outstanding action
	Trace::debug('Population request %d finished', $pending->{'id'}) if Trace::check(Trace::DEBUG6);
    	delete $base->{'populateQ'}->{ $pending->{'id'} };
	if (Trace::performance) {
	    Trace::debug('Query took: %f', Time::HiRes::tv_interval($pending->{'time'}));
	    $pending->{'time'} = [Time::HiRes::gettimeofday()];
	}
	if (defined($pending->{'action'})) {
	    &{ $pending->{'action'} };
	    if (Trace::performance) {
		Trace::debug('Action took: %f', Time::HiRes::tv_interval($pending->{'time'}));
	    }
	}
    }
}

sub _populateNode {
    my $base = shift;
    my $node = shift;

    if ($node->type eq Repo) {
	Thing::context->message(name => 'query', to => $node->what, parms => { type => 'artists' });
    }
    elsif ($node->type eq Player) {
	if (!Thing::context->message(name => Player::State, to => $node->what) ||
	    !Thing::context->message(name => Player::Spool, to => $node->what)) {
	    return 0;
	}
    }
    elsif ($node->type eq Cdrom) {
	if (!Thing::context->message(name => Cdrom::State, to => $node->what)) {
	    return 0;
	}
	Thing::context->message(name => Controller::RipState, to => $base->{'controller'});
    }
    elsif ($node->type eq Artist) {
	Thing::context->message(name => 'query', to => $base->{'repo'}->what, parms => { type => 'albums', artist => $node->what });
    }
    elsif ($node->type eq Album) {
	Thing::context->message(name => 'query', to => $base->{'repo'}->what, parms => { type => 'tracks', album => $node->what });
    }
    elsif ($node->type eq Lists) {
	$base->listsPopulate;
	return 1;
    }
    elsif ($node->type eq List) {
	if ($base->listPopulate($node)) {
	    return 1;
	}
    }
    else {
	# Root and Scratchpad are fixed.
	Trace::error('Nothing to populate for this node');
	return 1;
    }

    my $busyId = $base->busyStart(timeout => 30, msg => 'Querying '.$node->desc);
    push @{ $base->{'populateStack'} }, { node => $node, busy => $busyId };

    Trace::debug('Populating: %s', $node->desc) if Trace::check(Trace::DEBUG6);
    return 0;
}

sub _populateReturn {
    my $base = shift;
    my $node = shift;

    my $pos = 0;
    foreach my $query (@{ $base->{'populateStack'} }) {
	last if ($query->{'node'} == $node);
	$pos++;
    }

    if ($pos > $#{ $base->{'populateStack'} }) {
	Trace::debug('No query pending for: %s', $node->desc) if (Trace::check(Trace::DEBUG3));
	return;
    }
    my $query = $base->{'populateStack'}->[ $pos ];

    splice @{ $base->{'populateStack'} }, $pos, 1;
    $base->busyStop($query->{'busy'});

    $base->_populateCheck;
    return 1;
}

# depopulateNode - Purge the children for a node or group of
# nodes. Triggers a delete for each child and their children and so
# on.

sub depopulateNode {
    my $base = shift;
    my %parms = @_;

    if (!defined($parms{'node'}) && !defined($parms{'group'})) {
	Trace::mandatory('node||group');
	return;
    }

    my @nodes;
    if (defined($parms{'group'})) {
	@nodes = @{ $parms{'group'}; }
    }
    else {
	@nodes = ( $parms{'node'} );
    }

    # Node type specifics

    map {
	if ($_->type eq Cdrom) {
	    delete $_->gm_subspace->{'status'};
	    delete $_->gm_subspace->{'ripStatus'};
	    delete $_->gm_subspace->{'cd'};
	}
	elsif ($_->type eq Player) {
	    delete $_->gm_subspace->{'status'};
	}
    } @nodes;

    # Flatten out all children

    my @flat;
    my @children = @nodes;
    while (@children = map { if ($_->populated) { $_->children; } else { (); } } @children) {
	push @flat, @children;
    }
    @children = @flat;

    foreach my $node (@children) {
	$base->_nodeDelete($node);
    }
    foreach my $node (@children) {
	$node->parent->gm_removeChild(child => $node);
    }

    map { $_->gm_populated(0) } @children;
    map { $_->gm_populated(0) } @nodes;
    return 1;
}

# refreshNode - Purge a node and repopulate it.

sub refreshNode {
    my $base = shift;

    $base->depopulateNode(@_);
    $base->populateNode(@_);
    return 1;
}

# REPO - Handle repo responses

sub repo { $_[0]->{'repo'} }

# _repoInfo - Info reply to Repo query

sub _repoInfo {
    my $base = shift;
    my %tParms = @_;
    my $from = $tParms{'from'};
    my $response = $tParms{'parms'};

    Trace::debug('Repo query response of type: %s', $response->{'type'}) if Trace::check(Trace::DEBUG4);

    if (defined($response->{'_list'})) {
	return $base->_listInfo($response);
    }

    my @parents;
    my @objects;
    if ($response->{'type'} eq 'artists') {
	@parents = ( $base->{'repo'} );
	@objects = defined($response->{'artists'}) ? @{ $response->{'artists'} } : ();
    }
    elsif ($response->{'type'} eq 'albums') {
	Trace::debug('Albums response for: %s', $response->{'artist'}->name) if Trace::check(Trace::DEBUG4);
	@parents = $base->findNode(what => $response->{'artist'}->monitor);
	@objects = defined($response->{'albums'}) ? @{ $response->{'albums'} } : ();
    }
    elsif ($response->{'type'} eq 'tracks') {
	Trace::debug('Tracks response for: %s', $response->{'album'}->name) if Trace::check(Trace::DEBUG4);
	@parents = $base->findNode(what => $response->{'album'}->monitor);
	@objects = defined($response->{'tracks'}) ? @{ $response->{'tracks'} } : ();
    }
    else {
	Trace::debug('Unhandled query response') if Trace::check(Trace::DEBUG4);
	return;
    }

    if (!@parents) {
	Trace::debug('Repo query response not in tree') if Trace::check(Trace::DEBUG4);
	return;
    }

    foreach my $parent (@parents) {

	EXISTING: foreach my $existing ($parent->children) {
	    foreach my $object (@objects) {
		# Existing in repsonse
		next EXISTING if ($existing->what->ioid eq $object->ioid);
	    }
	    
	    # Existing child not in response, delete it
	    
	    $base->_nodeDelete($existing);
	    $parent->gm_removeChild(child => $existing);
	}
	
	my @new = ();
	RESPONSE: foreach my $object (@objects) {
	    foreach my $existing ($parent->children) {
		# Response exists
		next RESPONSE if ($existing->what->ioid eq $object->ioid);
	    }

	    # Response is new, add to the parent 
	    push @new, $object;
	}
	
	foreach my $object (@new) {
	    $object = $object->monitor;
	    my $node;
	    if ($response->{'type'} eq 'artists') {
		$node = Gui::Node->gm_new(parent => $parent, what => $object, leaf => 0, refresh => 1, homogeneous => 1, childType => Album);
	    }
	    elsif ($response->{'type'} eq 'albums') {
		$node = Gui::Node->gm_new(parent => $parent, what => $object, leaf => 0, refresh => 1, homogeneous => 1, childType => Track);
	    }
	    elsif (($response->{'type'} eq 'tracks') || ($response->{'type'} eq 'track')) {
		$node = Gui::Node->gm_new(parent => $parent, what => $object, leaf => 1, refresh => 1);
	    }
	    $parent->gm_addChild(child => $node);
	    $base->_nodeAdd($node);
	}

	$parent->gm_populated(1); # Force population if empty response

	# Inform anything waiting for it that it is done
	
	$base->_populateReturn($parent);
    }

    return 1;
}

# _listInfo - Track information response for lists. 

sub _listInfo {
    my $base = shift;
    my $response = shift;

    my $list = $base->findNode(what => $response->{'_list'}->{'list'}->monitor);
    my $at = $response->{'_list'}->{'at'};
    
    foreach my $track (@{ $response->{'tracks'} }) {
	my $node = Gui::Node->gm_new(parent => $list, what => $track->monitor, leaf => 1, refresh => 1);
	$list->gm_addChild(child => $node, at => $at);
	if (defined($at)) {
	    $at++;
	}
	$base->_nodeAdd($node);
    }

    $list->gm_populated(1); # Force population if empty response
    $base->_populateReturn($list);
    return 1;
}

# SCRATCHPAD

sub scratchpad { $_[0]->{'scratchpad'}; }

# scratchpadAppend - append nodes to the end of the scratchpad

sub scratchpadAppend {
    my $base = shift;
    return $base->scratchpadInsert(@_);
}

# scratchpadInsert - insert nodes at position

sub scratchpadInsert {
    my $base = shift;
    my $scratchpad = $base->scratchpad;
    my %parms = @_;

    if (!defined($parms{'tracks'})) {
	Trace::mandatory('tracks');
	return;
    }
    my $tracks = $parms{'tracks'};
    my @children = $scratchpad->children;
    my $at = $parms{'at'};
    if (defined($at) && ($at > $#children)) {
	$at = undef;
    }

    map {
	my $node = Gui::Node->gm_new(parent => $scratchpad, what => $_, leaf => 1, refresh => 1);
	$scratchpad->gm_addChild(child => $node, at => $at);
	$base->_nodeAdd($node);
    } @{ $tracks };

    return 1;
}

# scratchpadRemove - remove nodes from the scratchpad

sub scratchpadRemove {
    my $base = shift;
    my $scratchpad = shift;
    my %parms = @_;

    if (!defined($parms{'nodes'})) {
	Trace::mandatory('nodes');
	return;
    }
    my $nodes = $parms{'nodes'};

    map {
	$base->_nodeDelete($_);
	$scratchpad->gm_removeChild(child => $_);
    } @{ $nodes };

    return 1;
}

# LISTS

sub lists { $_[0]->{'lists'} }

sub listsPopulate {
    my $base = shift;

    my @lists = sort { $a->name cmp $b->name } Info::List->find(owner => $base->repo->what);

    if (!@lists || ($#lists == -1)) {
	$base->{'lists'}->gm_populated(1);
    }
    else {
	map {
	    my $node = Gui::Node->gm_new(parent => $base->{'lists'}, what => $_->monitor, leaf => 0, refresh => 1, homogeneous => 1, childType => Track);
	    $base->{'lists'}->gm_addChild(child => $node);
	    $base->_nodeAdd($node);
	} @lists;
    }

    return 1;
}

# LIST

sub listPopulate {
    my $base = shift;
    my $list = shift;
    my @tracks = map { Info::Track->find(owner => $base->repo->what, ioid => $_ ) } @{ $list->what->track_ioids };

    if (!@tracks || ($#tracks == -1)) {
	$list->gm_populated(1);
    }
    else {
	Thing::context->message(name => 'query', to => $base->{'repo'}->what, 
				parms => { type => 'track', tracks => \@tracks, _list => { list => $list->what } });
	return 0;
    }

    return 1;
}

# listAdd - Browser request to add tracks to the list

sub listAdd {
    my $base = shift;
    my $list = shift;
    my %parms = @_;

    if (!defined($parms{'tracks'})) {
	Trace::mandatory('tracks');
	return;
    }

    my @add = map { $_->ioid } @{ $parms{'tracks'} };

    $list->what->addTracks(trackIoids => \@add, at => $parms{'at'});
    return 1;
}

# _listAdded - Object update to indicate tracks added to a list

sub _listAdded {
    my $base = shift;
    my $list = shift;
    my $ioids = shift;
    my $at = shift;

    my @tracks = map { Info::Track->find(owner => $base->repo->what, ioid => $_ ) } @{ $ioids };

    if (!@tracks || ($#tracks == -1)) {
	$list->gm_populated(1);
    }
    else {
	Thing::context->message(name => 'query', to => $base->{'repo'}->what, 
				parms => { type => 'track', tracks => \@tracks, _list => { list => $list->what, at => $at } });
	return 0;
    }
    return 1;
}

# listRemove - Browser request to remove nodes from the list

sub listRemove {
    my $base = shift;
    my $list = shift;
    my %parms = @_;

    if (!defined($parms{'nodes'})) {
	Trace::mandatory('nodes');
	return;
    }

    my @remove = map { $_->position } @{ $parms{'nodes'} };

    $list->what->deleteTracks(trackPositions => \@remove);
    return 1;
}

# _listRemoved - Object update to indicate tracks removed from a list

sub _listRemoved {
    my $base = shift;
    my $list = shift;
    my $positions = shift;

    my @tracks = $list->children;
    my @remove = sort { $a <=> $b } @{ $positions };

    while (@remove) {
	my $rem = pop @remove;
	if ($#tracks >= $rem) {
	    my $child = $tracks[$rem];
	    $base->_nodeDelete($child);
	    $list->gm_removeChild(child => $child);
	}
    }
    return 1;
}

# CDROM

sub cdroms { grep { $_->type eq Cdrom } $_[0]->rootNode->children };

# _cdromStatus

sub _cdromStatus {
    my $base = shift;
    my %tParms = @_;
    my $from = $tParms{'from'};
    my $parms = $tParms{'parms'};

    my $found = 0;
    my $cdrom_bn;
    foreach my $existing ($base->cdroms) {
	next if (!Thing::Handle::cmpNameHost($from, $existing->what));
	$cdrom_bn = $existing;
	$found = 1;
	last;
    }
    return if (!$found);

    my $status = $cdrom_bn->gm_subspace->{'status'} = $parms;
    my $attr = $status->{'attributes'};

    # Update track listing

    my $cd = $attr->{'cd'};
    if (defined($cd)) {
	$cd = $cd->monitor;
	if (!defined($cdrom_bn->cd) || ($cd->ioid ne $cdrom_bn->cd->ioid)) {

	    Trace::debug('New CD') if Trace::check(Trace::DEBUG2);

	    my @tracks = $cd->tracks;
	    shift @tracks; pop @tracks; # Lead & MSTN
	    @tracks = map { $_->monitor } @tracks;

	    my @current = $cdrom_bn->children;
	    foreach my $child (@current) {
		# Remove existing
		$base->_nodeDelete($child);
		$cdrom_bn->gm_removeChild(child => $child);
	    }
	    foreach my $track (@tracks) {
		# Add new
		my $track_bn = Gui::Node->gm_new(parent => $cdrom_bn, what => $track, leaf => 1, refresh => 1);
		$cdrom_bn->gm_addChild(child => $track_bn);
		$base->_nodeAdd($track_bn);
	    }

	    # Handle remote lookup activity
	    $base->cdromObjectUpdate($cd);
	}
    }
    elsif (defined($cdrom_bn->cd)) {

	Trace::debug('No CD') if Trace::check(Trace::DEBUG2);

	my @current = $cdrom_bn->children;
	foreach my $child (@current) {
	    # Remove existing
	    $base->_nodeDelete($child);
	    $cdrom_bn->gm_removeChild(child => $child);
	}
    }
    if (defined($cd)) {
	$cdrom_bn->gm_subspace->{'cd'} = $cd;
    }
    else {
	delete $cdrom_bn->gm_subspace->{'ripStatus'};
	delete $cdrom_bn->gm_subspace->{'cd'};
    }

    $base->cdromStatusDiagnostic($status, $from);

    # Possibly thru a populate request
    $cdrom_bn->gm_populated(1);
    $base->_populateReturn($cdrom_bn);

    $base->_nodeUpdate($cdrom_bn);
}

# cdromStatusDiagnostic

sub cdromStatusDiagnostic {
    my $base = shift;
    my $status = shift;
    my $from = shift;

    return if ($status->{'cause'} eq Cdrom::None);
    return if ($status->{'cause'} eq Cdrom::Message);
    return if ($status->{'cause'} eq Cdrom::End);

    my $str = '';
    if ($status->{'cause'} eq Cdrom::Ignored) {
	$str = 'Message ignored';
    }
    elsif ($status->{'cause'} eq Cdrom::Invalid) {
	$str = 'Message had invalid parameters';
    }
    elsif ($status->{'cause'} eq Cdrom::Error) {
	$str = 'The cdrom had an error';
    }
    elsif ($status->{'cause'} eq Cdrom::RipError) {
	$str = 'Ripping had an error';
    }
    elsif ($status->{'cause'} eq Cdrom::RipStuck) {
	$str = 'Ripping took too long and was cancelled';
    }
    else {
	$str = 'Unknown cause';
    }

    $base->gui->diagnostic->add(type => 'Cdrom',
				name => $from->key,
				summary => $str,
				diagnostic => $status->{'attributes'}->{'diagnostic'});
    return 1;
}

# _cdromRipStatus

sub _cdromRipStatus {
    my $base = shift;
    my %tParms = @_;
    my $from = $tParms{'from'};
    my $parms = $tParms{'parms'};

    my $found = 0;
    my $cdrom_bn;
    foreach my $ripper (@{ $parms }) {
	foreach my $existing ($base->cdroms) {
	    next if (!Thing::Handle::cmpNameHost($ripper->{'handle'}, $existing->what));
	    $cdrom_bn = $existing;
	    $found = 1;
	    last;
	}
	next if (!$found);

	# Only update changes...
	if (!defined($cdrom_bn->gm_subspace->{'ripStatus'}) || !defined($cdrom_bn->gm_subspace->{'ripStatus'}->{'tracks'})) {
	    $cdrom_bn->gm_subspace->{'ripStatus'} = $ripper;
	    map { $base->_nodeUpdate($_) } $cdrom_bn->children;
	}
	else {
	    my @tracks = $cdrom_bn->children;
	    my $prev = $cdrom_bn->gm_subspace->{'ripStatus'}->{'tracks'};
	    $cdrom_bn->gm_subspace->{'ripStatus'} = $ripper;
	    foreach my $track (values(%{ $ripper->{'tracks'} })) {
		if (!defined($prev->{ $track->{'index'} }) ||
		    (($track->{'want'} ne $prev->{ $track->{'index'} }->{'want'}) ||
		     ($track->{'encoded'} ne $prev->{ $track->{'index'} }->{'encoded'}) ||
		     ($track->{'ripped'} ne $prev->{ $track->{'index'} }->{'ripped'}))) {
		    if (defined($tracks[ $track->{'index'} - 1 ])) {
			$base->_nodeUpdate($tracks[ $track->{'index'} - 1 ]);
		    }
		}
	    }
	}
    }
}

# cdromObjectUpdate

sub cdromObjectUpdate {
    my $base = shift;
    my $cd = shift;

    $cd = $cd->monitor; # In case of artist change following remote lookup.

    my $previous = $cd->{'previousRemoteResult'};
    my $current = $cd->remoteResult;

    if ((!defined($previous) && ($current->{'status'} eq Info::Cd::Underway)) ||
	(defined($previous) && ($current->{'status'} eq Info::Cd::Underway) && ($previous->{'status'} ne Info::Cd::Underway))) {
	$cd->{'busyId'} = $base->busyStart(timeout => 30, msg => $current->{'message'});
    }
    elsif ((!defined($previous) && ($current->{'status'} eq Info::Cd::Success)) ||
	   (defined($previous) && ($current->{'status'} eq Info::Cd::Success) && ($previous->{'status'} eq Info::Cd::Underway))) {
	$base->busyStop($cd->{'busyId'}) if (defined($cd->{'busyId'}));
	GtkEx::Utility::ackPopup('msg' => $current->{'message'}, truncate => 200) if (!$cd->known);
    }
    elsif ((!defined($previous) && ($current->{'status'} eq Info::Cd::Failure)) ||
	   (defined($previous) && ($current->{'status'} eq Info::Cd::Failure) && ($previous->{'status'} eq Info::Cd::Underway))) {
	$base->busyStop($cd->{'busyId'}) if (defined($cd->{'busyId'}));
	GtkEx::Utility::ackPopup('msg' => 'Remote operation failed:'."\n".$current->{'message'}, truncate => 200);
    }
    elsif ((!defined($previous) && ($current->{'status'} eq Info::Cd::Multiple)) ||
	   (defined($previous) && ($current->{'status'} eq Info::Cd::Multiple) && ($previous->{'status'} eq Info::Cd::Underway))) {
	$base->busyStop($cd->{'busyId'}) if (defined($cd->{'busyId'}));
	$base->cdromChoicePopup($cd);
    }

    my %copy = %{ $current };
    $cd->{'previousRemoteResult'} = \%copy;

    foreach my $cdrom ($base->cdroms) {
	if (defined($cdrom->cd) && ($cdrom->cd->ioid eq $cd->ioid)) {
	    $base->_nodeUpdate($cdrom);
	}
    }
    return 1;
}

sub cdromChoicePopup {
    my $base = shift;
    my $cd = shift;

    # Multiple or inexact choice

    my $box = Gtk::VBox->new(0, 0);
    $box->border_width(0);
    $box->show();

    my $label = Gtk::Label->new('');
    $label->set_justify('left');
    $label->set_line_wrap(1);
    $label->set_text($cd->remoteResult->{'message'}.".\n".'Choose the correct CD from the list below and click OK. If none of the matches are acceptable click Cancel.');
    $label->show();
    $box->pack_start($label, 0, 0, 0);

    my $scroll = Gtk::ScrolledWindow->new(undef, undef);
    $scroll->set_policy('automatic', 'automatic');
    $scroll->show();

    my $list = Gtk::CList->new_with_titles('Name', 'Artist', 'Category');
    $list->column_titles_passive();
    my $height = $base->gui->rowHeight;
    $list->set_row_height($base->gui->rowHeight);
    foreach my $match (@{ $cd->remoteResult->{'matches'} }) {
	$list->append($match->{'name'}, $match->{'artist'}, $match->{'category'});
	$height += $base->gui->rowHeight;
    }
    $list->columns_autosize;
    my $width = 0;
    foreach my $col (0, 1) {
	my $cwidth = $list->optimal_column_width($col);
	if ($cwidth > 200) {
	    $cwidth = 200;
	    $list->set_column_width($col, 200);
	}
	$width += $cwidth;
    }
    $list->set_usize($width + 40, $height + 40);
    $list->show();
    $scroll->add($list);
    $box->pack_start($scroll, 1, 1, 0);

    GtkEx::Utility::choicePopup(widget => $box, title => 'Multiple or inexact matches', okFunc => sub { $base->cdromChoiceOk($cd, $list) });
    return 1;
}

sub cdromChoiceOk {
    my $base = shift;
    my $cd = shift;
    my $list = shift;

    my @sel = $list->selection;
    return 1 if (!@sel || ($#sel == -1));
    my $row = $sel[0];

    my $matches = $cd->remoteResult->{'matches'};
    return 1 if (!defined($matches->[$row]));

    $cd->remoteChoose(cddb_id => $matches->[$row]->{'cddb_id'}, category => $matches->[$row]->{'category'});
    return 1;
}

# PLAYER

sub players { grep { $_->type eq Player } $_[0]->rootNode->children };

# _playerStatus

sub _playerStatus {
    my $base = shift;
    my %tParms = @_;
    my $from = $tParms{'from'};
    my $parms = $tParms{'parms'};

    my $found = 0;
    my $player_bn;
    foreach my $existing ($base->players) {
	next if (!Thing::Handle::cmpNameHost($from, $existing->what));
	$player_bn = $existing;
	$found = 1;
	last;
    }
    return if (!$found);

    $player_bn->gm_subspace->{'status'} = $parms;

    $base->playerStatusDiagnostic($parms, $from);

    $base->_nodeUpdate($player_bn);
}

# _playerSpooled

sub _playerSpooled {
    my $base = shift;
    my %tParms = @_;
    my $from = $tParms{'from'};
    my $parms = $tParms{'parms'};

    my $found = 0;
    my $player_bn;
    foreach my $existing ($base->players) {
	next if (!Thing::Handle::cmpNameHost($from, $existing->what));
	$player_bn = $existing;
	$found = 1;
	last;
    }
    return if (!$found);

    # Players children should be the same as the spool and in the same
    # order.

    my @tracks = map { $_->monitor } @{ $parms->{'tracks'} };
    my @current = $player_bn->children;

    my $diff = 0;
    if (@current && @tracks && ($#current == $#tracks)) {
	for (my $i = 0; $i <= $#current; $i++) {
	    next if ($current[$i]->what->ioid eq $tracks[$i]->ioid);
	    $diff = 1;
	    last;
	}
    }
    else {
	$diff = 1;
    }

    my $last = ($#current > $#tracks) ? $#current : $#tracks;
    for (my $i = 0; $i <= $last; $i++) {
	if (!defined($tracks[$i])) {
	    # Remove existing
	    $base->_nodeDelete($current[$i]);
	    $player_bn->gm_removeChild(child => $current[$i]);
	}
	elsif (!defined($current[$i]) || ($current[$i]->what->ioid ne $tracks[$i]->ioid)) {
	    # Add new
	    my $track_bn = Gui::Node->gm_new(parent => $player_bn, what => $tracks[$i], leaf => 1, refresh => 1);
	    $player_bn->gm_addChild(child => $track_bn, at => $i);
	    @current = $player_bn->children;
	    $base->_nodeAdd($track_bn);
	}
    }

    if (!@tracks) {
	$player_bn->gm_populated(1);
    }
    $player_bn->gm_subspace->{'spool'} = \@tracks;

    # Possibly thru a populate request
    $base->_populateReturn($player_bn);

    return 1;
}

# playerStatusDiagnostic

sub playerStatusDiagnostic {
    my $base = shift;
    my $status = shift;
    my $from = shift;

    return if ($status->{'cause'} eq Player::None);
    return if ($status->{'cause'} eq Player::Message);
    return if ($status->{'cause'} eq Player::End);

    my $str = '';
    if ($status->{'cause'} eq Player::Ignored) {
	$str = 'Message ignored';
    }
    elsif ($status->{'cause'} eq Player::Invalid) {
	$str = 'Message had invalid parameters';
    }
    elsif ($status->{'cause'} eq Player::Error) {
	$str = 'The player had an error';
    }
    else {
	$str = 'Unknown cause';
    }

    $base->gui->diagnostic->add(type => 'Player',
				name => $from->key,
				summary => $str,
				diagnostic => $status->{'attributes'}->{'diagnostic'});
    return 1;
}

# Internal

sub _nodeAdd    { _nodeTrigger(shift, shift, 'add', @_); }
sub _nodeDelete { _nodeTrigger(shift, shift, 'delete', @_); }
sub _nodeUpdate { _nodeTrigger(shift, shift, 'update', @_); }
sub _nodeMove   { _nodeTrigger(shift, shift, 'move', @_); }

sub _nodeTrigger {
    my $base = shift;
    my $node = shift;
    my $what = shift;
    my @parms = @_;

    map { &{ $_->{ $what } }($node, @parms) } grep { defined( $_->{ $what } ) } values(%{ $base->{'registry'} });
}

sub _dumpTree {
    my $base = shift;
    my $node = shift || $base->rootNode;
    my $level = shift || 0;

    my $t;
    $t = sprintf('%40.40s', $node);
    for (my $i = 0; $i < $level; $i++) {
	$t .= ' ';
    }
    $t .= $node->desc.' --> '."\n";
    $t .= join('', map { $base->_dumpTree($_, $level + 2) } $node->children);
    return $t;
}

sub busyStart {
    my $base = shift;
    my @p = @_;

    my @ids = map { { busyId => $_->{'window'}->busy->start(@p), regId => $_->{'id'} } } values(%{ $base->{'registry'} });
    return \@ids;
}

sub busyStop {
    my $base = shift;
    my $ids = shift;

    foreach my $id (@{ $ids }) {
	$base->{'registry'}->{ $id->{'regId'} }->{'window'}->busy->stop(id => $id->{'busyId'});
    }
    return 1;
}

1;
