# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.3 $

package Gui::Node;

# Data class for elements (nodes) in the user interface hierarchy. 

# Attributes are:

#   parent - parent node or undef if none.

#   children - a list of this nodes children, empty if it has none.

#   populated - the children of this node have been discovered.

#   leaf - boolean, this node may have children.

#   refresh - boolean, this node can be refreshed.

#   homogeneous - boolean, the children of this node are all of the
#   same type (childType).

#   childType - The type of homogeneous children.

#   what - the object this node encapsulates.

#   type - the type of object encapsulated, dictates the type returned
#   by what.

#     type                    what
#     ----                    ----
#     Root                    undef
#     Repo                    undef
#     Artist                  Info::Artist
#     Album                   Info::Album
#     Track                   Info::Track
#     Lists                   undef
#     List                    Info::List
#     Player                  Thing::Handle
#     Cdrom                   Thing::Handle
#     Scratchpad              undef

#   widget - this node has its own builtin widget creator, call with
#   create => 1 to create a widget.

#   desc - a useful way to describe a node in error messages and debug.

use Thing;
use Thing::ServiceH;
use Trace;
use strict;

# Gui internal methods

# gm_new - Instantiate a new node.

sub gm_new {
    my $proto = shift;
    my $type = ref($proto) || $proto;
    my %parms = @_;

    # The nodes are held in an array
    my $node = [];
    bless $node, $type;

    $node->[0] = $parms{'parent'};
    $node->[1] = defined($parms{'leaf'}) ? $parms{'leaf'} : 0;
    $node->[2] = defined($parms{'refresh'}) ? $parms{'refresh'} : 0;
    $node->[3] = defined($parms{'homogeneous'}) ? $parms{'homogeneous'} : 0;
    $node->[4] = defined($parms{'widget'}) ? $parms{'widget'} : 0;
    $node->[5] = $parms{'what'};
    $node->[6] = defined($parms{'type'}) ? $parms{'type'} : $node->_type;
    $node->[7] = $parms{'children'};
    $node->[8] = defined($parms{'desc'}) ? $parms{'desc'} : undef;
    $node->[9] = $parms{'childType'};
    #$node->[10] = {}; # Subspace

    if ($node->homogeneous && !defined($node->childType)) {
	Trace::error('A childType is required for homogeneous nodes');
	return;
    }

    return $node;
}

# gm_children - Reassign the children of this node.

sub gm_children { $_[0]->[7] = $_[1]; return @{ $_[0]->[7] }; }

# gm_populated - Set the populated status of this node.

sub gm_populated { 
    my $node = shift;
    my $pop = shift;
    if (!$pop) {
	$node->[7] = undef;
    }
    elsif ($pop && !defined($node->[7])) {
	$node->[7] = [];
    }
    return $pop;
}

# gm_addChild - Add a child `at' an optional position or append it.

sub gm_addChild {
    my $node = shift;
    my %parms = @_;

    if (!defined($parms{'child'})) {
	Trace::mandatory('child');
	return;
    }
    my $child = $parms{'child'};

    my $at;
    if (defined($parms{'at'})) {
	$at = $parms{'at'};
	if ($at && (!defined($node->[7]) || ($#{ $node->[7] } < $at - 1))) {
	    Trace::error('Position (%d) is outside the bounds of this node', $at);
	    return;
	}
    }
    if (!defined($node->[7])) {
	$node->[7] = [];
    }

    Trace::debug('Adding: %s, to: %s, at: %s', $child->desc, $node->desc, defined($at) ? $at : 'end') if Trace::check(Trace::DEBUG6);

    if (!defined($at)) {
	push @{ $node->[7] }, $child;
    }
    else {
	splice @{ $node->[7] }, $at, 0, $child;
    }
}

# gm_removeChild - Remove a child.

sub gm_removeChild {
    my $node = shift;
    my %parms = @_;

    if (!defined($parms{'child'})) {
	Trace::mandatory('child');
	return;
    }
    my $child = $parms{'child'};

    Trace::debug('Removing: %s, from: %s', $child->desc, $node->desc) if Trace::check(Trace::DEBUG6);

    my $pos = 0;
    if (defined($node->[7])) {
	foreach my $check (@{ $node->[7] }) {
	    if ($child == $check) {
		splice @{ $node->[7] }, $pos, 1;
		return 1;
	    }
	    $pos++;
	}
    }
    
    Trace::error('This is not a child of this node');
    return;
}

# gm_moveChild - Move a child from a position `from' inserting it before a position `to'.

sub gm_moveChild {
    my $node = shift;
    my %parms = @_;

    if (!defined($parms{'from'})) {
	Trace::mandatory('from');
	return;
    }
    my $to = $parms{'to'};
    my $from = $parms{'from'};

    Trace::debug('Moving from %d to %s, in: %s', $from, defined($to) ? $to : 'end', $node->desc) if Trace::check(Trace::DEBUG6);

    my $pos = 0;
    if (defined($node->[7]) && defined($node->[7]->[ $from ])) {
	my $hold = $node->[7]->[ $from ];
	splice @{ $node->[7] }, $from, 1;
	if (!defined($to)) {
	    push @{ $node->[7] }, $hold;
	}
	else {
	    if ($to > $from) {
		$to--;
	    }
	    splice @{ $node->[7] }, $to, 0, $hold;
	}
	return 1;
    }
    
    Trace::error('This is not a child of this node');
    return;
}

# gm_subspace - A hash ref where sub classes can store data

sub gm_subspace {
    if (!defined($_[0]->[10])) {
	$_[0]->[10] = {};
    }
    return $_[0]->[10];
}

# Attributes

sub parent       { $_[0]->[0]; }
sub children     { defined($_[0]->[7]) ? @{ $_[0]->[7] } : (); }
sub populated    { defined($_[0]->[7]) ? 1 : 0; }
sub leaf         { $_[0]->[1]; }
sub refresh      { $_[0]->[2]; }
sub homogeneous  { $_[0]->[3]; }
sub childType    { $_[0]->[9]; }
sub what         { $_[0]->[5]; }
sub type         { $_[0]->[6]; }
sub desc         { if (defined($_[0]->[8])) { $_[0]->[8]; } else { $_[0]->[8] = $_[0]->_desc; } }

sub widget {
    my $node = shift;
    my %parms = @_;

    if (defined($parms{'create'}) && $parms{'create'}) {
	if ($node->[4]) {
	    delete $parms{'create'};
	    return &{ $node->[4] }(%parms);
	}
	else {
	    Trace::error('No widget creator for create');
	    return undef;
	}
    }

    return $node->[6] ? 1 : 0;
}

# child - return the child following `after'.

sub child {
    my $node = shift;
    my %parms = @_;

    return if (!$node->populated);
    if (!defined($parms{'after'})) {
	Trace::mandatory('after');
	return;
    }
    my @children = $node->children;
    my $pos = $#children;
    foreach my $child (reverse @children) {
	last if ($child == $parms{'after'});
	$pos--;
    }
    if ($pos > $#children) {
	Trace::error('Not in this node');
	return;
    }
    return if ($pos == $#children);
    return $children[$pos + 1];
}

# position - where we stand within our parents children

sub position {
    my $node = shift;
    my $parent = $node->parent || return;

    my @children = $parent->children;
    my $pos = 0;
    foreach my $child (@children) {
	last if ($child == $node);
	$pos++;
    }
    return if ($pos > $#children);
    return $pos;
}

sub count { defined($_[0]->[7]) ? $#{ $_[0]->[7] } + 1 : undef }

# Internal

# _type - Discover the type for a node that was created without one specified.

sub _type {
    my $node = shift;

    if (UNIVERSAL::isa($node->[5], 'Info::Artist')) {
	return Gui::Base::Artist;
    }
    elsif (UNIVERSAL::isa($node->[5], 'Info::Album')) {
	return Gui::Base::Album;
    }
    elsif (UNIVERSAL::isa($node->[5], 'Info::Track')) {
	return Gui::Base::Track;
    }
    elsif (UNIVERSAL::isa($node->[5], 'Info::List')) {
	return Gui::Base::List;
    }
    else {
	Trace::error('Need to specify the type for this');
	return undef;
    }
}

# _desc - Discover a good descriptor for this node, used mainly for debug
# messages.

sub _desc {
    my $node = shift;

    if (defined($node->what) && 
	(UNIVERSAL::isa($node->what, 'Info::Artist') || UNIVERSAL::isa($node->what, 'Info::Album') || UNIVERSAL::isa($node->what, 'Info::Track') || UNIVERSAL::isa($node->what, 'Info::List'))) {
	return $node->type.' '.$node->what->name;
    }
    elsif (defined($node->what) && UNIVERSAL::isa($node->what, 'Thing::Handle')) {
	return $node->what->key;
    }
    else {
	return $node->type;
    }
}

1;
