# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.13.2.1 $

package Gui::Page::Browser;

use Gui::Page;
@ISA = qw( Gui::Page );
use Gtk;
use Gtk::Gdk;
use GtkEx::CList;
use GtkEx::Button;
use PlayerH;
use CdromH;
use Format::Mp3;
use Format::Wav;
use Info::Artist;
use Info::Album;
use Info::Track;
use Gui::Base;
use Gui::Node;
use Gui::Node::Player;
use Gui::Node::Cdrom;
use Trace;
use strict;

sub new {
  my $proto = shift;
  my $type = ref($proto) || $proto;
  my %parms = @_;

  my $vbox = Gtk::VBox->new(0, 0);
  $vbox->border_width(0);
  $vbox->show();

  my $browser = $type->SUPER::new(@_, title => 'Music', widget => $vbox);

  $vbox->signal_connect('event', sub { $browser->event(@_); }) if ($browser->gui->eventDebug);

  # Pane

  my $pane = $browser->{'pane'} = Gtk::HPaned->new();
  $pane->signal_connect('event', sub { $browser->event(@_); }) if ($browser->gui->eventDebug);
  $pane->set_gutter_size(4);
  $pane->set_handle_size(4);
  $pane->show();
  $browser->widget->pack_start($pane, 1, 1, 0);

  # Tree

  my $scroll = $browser->{'treeScroll'} = Gtk::ScrolledWindow->new(undef, undef);
  $scroll->set_policy('automatic', 'automatic');
  $scroll->signal_connect('event', sub { $browser->event(@_); }) if ($browser->gui->eventDebug);
  $scroll->set_style($browser->gui->style);
  $scroll->show();
  $browser->{'pane'}->add1($scroll);

  my $tree = $browser->{'tree'} = Gtk::CTree->new(1, 0);
  $tree->set_selection_mode('browse');
  $tree->show();
  $browser->{'treeScroll'}->add($tree);

  $browser->{'treeSelectId'} = 0; # Order outstanding selections

  # Display

  $browser->{'display'} = {};
  $browser->{'display'}->{'history'} = [];
  $browser->{'display'}->{'historyLimit'} = 10;

  $browser->{'display'}->{'container'} = my $display = Gtk::VBox->new(0, 0);
  $display->border_width(0);
  $display->show();
  $pane->add2($display);

  # Display - Info

  my $infoLabel = $browser->{'display'}->{'infoLabel'} = Gtk::Label->new('');
  $infoLabel->set_justify('left');
  $infoLabel->hide;
  $display->pack_start($infoLabel, 0, 0, 0);

  # Display - Scroll for list

  $browser->{'display'}->{'scroll'} = $scroll = Gtk::ScrolledWindow->new(undef, undef);
  $scroll->signal_connect('event', sub { $browser->event(@_); }) if ($browser->gui->eventDebug);
  $scroll->set_policy('automatic', 'automatic');
  $scroll->set_style($browser->gui->style);
  $scroll->show();
  $display->pack_start($scroll, 1, 1, 0);

  # Display - Progress bar

  $browser->{'display'}->{'progress'} = { users => {} };
  $browser->{'display'}->{'progress'}->{'bar'} = my $progress = Gtk::ProgressBar->new();
  $browser->{'display'}->{'progress'}->{'event'} = my $event = Gtk::EventBox->new();
  $event->signal_connect('button-press-event', sub { $browser->progressButtonPressEvent(@_); });
  $progress->set_show_text(1);
  $progress->set_format_string('%p%%');
  $progress->show;
  $event->show;
  $event->add($progress);
  $display->pack_start($event, 0, 0, 0);

  # Display - Buttons

  $browser->{'display'}->{'buttons'} = { };
  $browser->{'display'}->{'buttons'}->{'box'} = my $buttonBox = Gtk::VBox->new(0, 0);
  $buttonBox->border_width(0);
  $buttonBox->show;
  $display->pack_start($buttonBox, 0, 0, 0);

  $browser->buttonsCreate;

  # Menu

  $browser->menuCreate;

  $browser->{'baseId'} = $parms{'baseId'};
  $browser->{'base'} = $browser->gui->base;

  return $browser;
}

sub base                  { $_[0]->{'base'}; }
# Too see most of everything in the Browser window when looking at
# a player or scratchpad we need at least 800.
sub default_width         { 800; }
sub default_height        { 400; }

sub attach {
  my $browser = shift;

  $browser->SUPER::attach(@_);

  # Pixmaps, done here as they need a window

  $browser->{'pixmaps'} = {};
  my ($pixmap, $mask) = 
    Gtk::Gdk::Pixmap->create_from_xpm($browser->window->window->window,
				      $browser->gui->style->bg('normal'),
				      $browser->gui->pixmapDir.'/folder.xpm');
  $browser->{'pixmaps'}->{ Gui::Base::Root } = {};
  $browser->{'pixmaps'}->{ Gui::Base::Root }->{'closed'} = [$pixmap, $mask];
  $browser->{'pixmaps'}->{ Gui::Base::Repo } = {};
  $browser->{'pixmaps'}->{ Gui::Base::Repo }->{'closed'} = [$pixmap, $mask];
  $browser->{'pixmaps'}->{ Gui::Base::Scratchpad } = {};
  $browser->{'pixmaps'}->{ Gui::Base::Scratchpad }->{'closed'} = [$pixmap, $mask];
  $browser->{'pixmaps'}->{ Gui::Base::Lists } = {};
  $browser->{'pixmaps'}->{ Gui::Base::Lists }->{'closed'} = [$pixmap, $mask];
  $browser->{'pixmaps'}->{ Gui::Base::Artist } = {};
  $browser->{'pixmaps'}->{ Gui::Base::Artist }->{'closed'} = [$pixmap, $mask];

  ($pixmap, $mask) = 
    Gtk::Gdk::Pixmap->create_from_xpm($browser->window->window->window,
				      $browser->gui->style->bg('normal'),
				      $browser->gui->pixmapDir.'/folder_open.xpm');
  $browser->{'pixmaps'}->{ Gui::Base::Root }->{'open'} = [$pixmap, $mask];
  $browser->{'pixmaps'}->{ Gui::Base::Repo }->{'open'} = [$pixmap, $mask];
  $browser->{'pixmaps'}->{ Gui::Base::Scratchpad }->{'open'} = [$pixmap, $mask];
  $browser->{'pixmaps'}->{ Gui::Base::Lists }->{'open'} = [$pixmap, $mask];
  $browser->{'pixmaps'}->{ Gui::Base::Artist }->{'open'} = [$pixmap, $mask];

  ($pixmap, $mask) = 
    Gtk::Gdk::Pixmap->create_from_xpm($browser->window->window->window,
				      $browser->gui->style->bg('normal'),
				      $browser->gui->pixmapDir.'/album.xpm');
  $browser->{'pixmaps'}->{ Gui::Base::Album } = {};
  $browser->{'pixmaps'}->{ Gui::Base::Album }->{'closed'} = [$pixmap, $mask];
  $browser->{'pixmaps'}->{ Gui::Base::Album }->{'open'} = [$pixmap, $mask];
  $browser->{'pixmaps'}->{ Gui::Base::Cdrom } = {};
  $browser->{'pixmaps'}->{ Gui::Base::Cdrom }->{'closed'} = [$pixmap, $mask];
  $browser->{'pixmaps'}->{ Gui::Base::Cdrom }->{'open'} = [$pixmap, $mask];
  $browser->{'pixmaps'}->{ Gui::Base::List } = {};
  $browser->{'pixmaps'}->{ Gui::Base::List }->{'closed'} = [$pixmap, $mask];
  $browser->{'pixmaps'}->{ Gui::Base::List }->{'open'} = [$pixmap, $mask];

  ($pixmap, $mask) = 
    Gtk::Gdk::Pixmap->create_from_xpm($browser->window->window->window,
				      $browser->gui->style->bg('normal'),
				      $browser->gui->pixmapDir.'/track.xpm');
  $browser->{'pixmaps'}->{ Gui::Base::Track } = {};
  $browser->{'pixmaps'}->{ Gui::Base::Track }->{'closed'} = [$pixmap, $mask];
  $browser->{'pixmaps'}->{ Gui::Base::Track }->{'open'} = [$pixmap, $mask];

  ($pixmap, $mask) = 
    Gtk::Gdk::Pixmap->create_from_xpm($browser->window->window->window,
				      $browser->gui->style->bg('normal'),
				      $browser->gui->pixmapDir.'/rate.xpm');
  $browser->{'pixmaps'}->{ Gui::Base::Player } = {};
  $browser->{'pixmaps'}->{ Gui::Base::Player }->{'closed'} = [$pixmap, $mask];
  $browser->{'pixmaps'}->{ Gui::Base::Player }->{'open'} = [$pixmap, $mask];

  # Accel, also need a window

  if (0) { # TODO

  my $accelGroup = $browser->window->accelGroup;

  # [pP] - Play / Pause
  $accelGroup->add(0x070, 'release-mask', [], $browser->{'buttons'}->{'Play'}, 'clicked');
  $accelGroup->add(0x050, 'release-mask', [], $browser->{'buttons'}->{'Play'}, 'clicked');

  # [sS] - Stop
  $accelGroup->add(0x073, 'release-mask', [], $browser->{'buttons'}->{'Stop'}, 'clicked');
  $accelGroup->add(0x053, 'release-mask', [], $browser->{'buttons'}->{'Stop'}, 'clicked');

  # [eEcC] - Eject/Close
  $accelGroup->add(0x065, 'release-mask', [], $browser->{'buttons'}->{'Eject'}, 'clicked');
  $accelGroup->add(0x045, 'release-mask', [], $browser->{'buttons'}->{'Eject'}, 'clicked');
  $accelGroup->add(0x063, 'release-mask', [], $browser->{'buttons'}->{'Eject'}, 'clicked');
  $accelGroup->add(0x043, 'release-mask', [], $browser->{'buttons'}->{'Eject'}, 'clicked');
  }

  # Register with Gui::Base

  $browser->base->register( id => $browser->{'baseId'},
			    add => sub { $browser->baseNodeAdd(@_); },
			    delete => sub { $browser->baseNodeDelete(@_); },
			    update => sub { $browser->baseNodeUpdate(@_); },
			    move => sub { $browser->baseNodeMove(@_); },
			    window => $browser->window );

  $browser->{'pane'}->set_position($browser->window->width * 0.33);

  $browser->treeDraw;		# Kick off

  $browser->aspect;
  return 1;
}

sub aspect {
  my $browser = shift;
  return 1;
}

# TREE - The LHS

# treeDraw

sub treeDraw {
  my $browser = shift;

  my $tree = $browser->{'tree'};
  $tree->freeze;

  # Create the top level tree
    
  my $root_bn = $browser->base->rootNode;
  $browser->{'treeRootNode'} = my $root_tn = $tree->insert_node(undef, undef, [ $browser->treeLabelNode($root_bn) ], 5,
								@{ $browser->{'pixmaps'}->{ $root_bn->type }->{'closed'} },
								@{ $browser->{'pixmaps'}->{ $root_bn->type }->{'open'} },
								0, 1);
  $tree->node_set_row_data($root_tn, { basenode => $root_bn });

  # Build the existing tree

  foreach my $basenode ($root_bn->children) {
    $browser->treeDrawNode($basenode, $root_tn);
  }

  $browser->{'selectSignalId'} = $tree->signal_connect('tree-select-row', sub { $browser->treeSelectSignal(@_) });
  $tree->signal_connect('event', sub { $browser->treeEvent(@_); } );

  $tree->drag_source_set(['button1_mask'], ['copy', 'move',], {'target' => "STRING", 'flags' => 0, 'info' => 0});
  $tree->signal_connect('drag_data_get', sub { $browser->dndDragDataGet('tree', @_); });
  $tree->drag_dest_set('all', ['copy', 'move',], {'target' => "STRING", 'flags' => 0, 'info' => 0});
  $tree->signal_connect('drag_data_received', sub { $browser->treeDragDataReceived(@_); });

  # View the root

  $tree->select_row(0, 0);

  $browser->{'treeHidden'} = 0;
  $tree->thaw;

  return 1;
}

sub treeDrawNode {
  my $browser = shift;
  my $node = shift;
  my $parent_tn = shift;
  my $tree = $browser->{'tree'};

  return 1 if ($node->leaf);

  my $treenode = $tree->insert_node($parent_tn, undef, [ $browser->treeLabelNode($node) ], 5, 
				    @{ $browser->{'pixmaps'}->{ $node->type }->{'closed'} },
				    @{ $browser->{'pixmaps'}->{ $node->type }->{'open'} },
				    $browser->treeLeafNode($node), 0);
  $tree->node_set_row_data($treenode, { basenode => $node });
  map {
    $browser->treeDrawNode($_, $treenode);
  } $node->children;

  return 1;
}

# treeRootNode

sub treeRootNode {
  my $browser = shift;
  return $browser->{'treeRootNode'};
}

# treeFindNode

sub treeFindNode {
  my $browser = shift;
  my $tree = $browser->{'tree'};
  my $basenode = shift;

  my $treenode = $browser->treeRootNode;
  do {
    my $nd = $tree->node_get_row_data($treenode);
    return $treenode if ($nd->{'basenode'} && ($nd->{'basenode'} == $basenode));

    if ($treenode->row->children) {
      $treenode = $treenode->row->children;
    }
    elsif ($treenode->row->sibling) {
      $treenode = $treenode->row->sibling;
    }
    else {
      do {
	$treenode = $treenode->row->parent;
      } while ($treenode && !$treenode->row->sibling);
      $treenode = $treenode->row->sibling if ($treenode);
    }
  } while ($treenode);

  return;
}
 
# treeLeafNode - Is this base node a leaf in the browser tree?

sub treeLeafNode {
  my $browser = shift;
  my $node = shift;

  my $type = $node->type;
  if (($type eq 'Player') || ($type eq 'Cdrom') || ($type eq 'Scratchpad') || ($type eq 'Album') || ($type eq 'List')) {
    return 1;
  }
  return 0;
}

# treeLabelNode - Derive a good label for a base node.

sub treeLabelNode {
  my $browser = shift;
  my $basenode = shift;

  my $type = $basenode->type;
  my $what = $basenode->what;
  if ($type eq Gui::Base::Root) {
    return 'Music';
  }
  elsif ($type eq Gui::Base::Repo) {
    return 'Artists';
  }
  elsif ($type eq Gui::Base::Scratchpad) {
    return 'Scratchpad';
  }
  elsif ($type eq Gui::Base::Lists) {
    return 'Lists';
  }
  elsif (defined($what) && (UNIVERSAL::isa($what, 'Info::Artist') || UNIVERSAL::isa($what, 'Info::Album') || 
			    UNIVERSAL::isa($what, 'Info::Track') || UNIVERSAL::isa($what, 'Info::List'))) {
    return $what->name;
  }
  elsif (defined($what) && UNIVERSAL::isa($what, 'Thing::Handle')) {
    # Players and Cdroms
    return $what->key;
  }
  Trace::error('Cannot determin good label for node: %s', $basenode->desc);
  return;
}

# treeEvent

sub treeEvent {
  my $browser = shift;
  my $tree = shift;
  my $event = shift;

  return 0 if (!defined($event));

  if ($browser->gui->eventDebug) {
    Trace::debug(ref($tree).'::'.$event->{'type'}) if Trace::check(Trace::DEBUG3);
    if (0) {
      require 'dumpvar.pl';
      main::dumpValue($event);
      $| = 1;
    }
  }

  # Button 3 brings up a menu

  if (($event->{'type'} eq 'button_press') && ($event->{'button'} == 3)) {
    # Used to select at point...
    #(my $row, my $col) = $tree->get_selection_info($event->{'x'}, $event->{'y'});
    #$tree->select_row($row, $col) if (defined($row) && defined($col));
    $browser->menuDraw('tree');
    $browser->menu->popup(0, 0, $event->{'button'}, $event->{'time'});
    return 1;
  }

  return 0;
}

# treeSelectSignal - Signal for tree node selected. 

sub treeSelectSignal {
  my $browser = shift;
  my $tree = shift;
  my $treenode = shift;

  my $nd = $tree->node_get_row_data($treenode);
  my $basenode = $nd->{'basenode'};
  my $selectId = ++$browser->{'treeSelectId'};

  # Populate the base and possibly draw the display. Populating the
  # base will trigger base node add events which builds the tree.

  $browser->base->populateNode(node => $basenode,
			       action => sub {
				 if ($selectId == $browser->{'treeSelectId'}) {
				   $browser->displayDraw($basenode);
				 }
			       });
  return 1;
}

# treeSelect - Function to select a base nodes corresponding tree node.

sub treeSelect {
  my $browser = shift;
  my $tree = $browser->{'tree'};
  my $basenode = shift;

  # Check the base node still exists

  if (!$browser->base->findNode(node => $basenode)) {
    Trace::debug('This node is no longer in the tree: %s', $browser->treeLabelNode($basenode)) if Trace::check(Trace::DEBUG6);
    return 1;
  }

  my $treenode = $browser->treeFindNode($basenode);
  return if (!$treenode);

  # Expand back up the tree

  my $parent_tn = $treenode->row->parent;
  while ($parent_tn) {
    $tree->expand($parent_tn);
    $parent_tn = $parent_tn->row->parent;
  }

  # Make sure it is in view

  my $row = $tree->get_node_position($treenode);
  if ($tree->row_is_visible($row) ne 'full') {
    $tree->moveto($row, 0, 0.5, 0);
  }

  # Grab the focus TODO

  $tree->select($treenode);	# Select it, generating the signal
  return 1;
}

# treeHide

sub treeHide {
  my $browser = shift;

  $browser->{'treeScroll'}->ref;
  $browser->{'pane'}->remove($browser->{'treeScroll'});
  $browser->{'display'}->{'container'}->ref;
  $browser->{'pane'}->remove($browser->{'display'}->{'container'});
  $browser->widget->remove($browser->{'pane'});
  $browser->{'pane'}->destroy;
  delete $browser->{'pane'};

  $browser->widget->pack_start($browser->{'display'}->{'container'}, 1, 1, 0);
  $browser->{'display'}->{'container'}->unref;
  $browser->{'treeHidden'} = 1;
  return 1;
}

# treeShow

sub treeShow {
  my $browser = shift;

  $browser->{'display'}->{'container'}->ref;
  $browser->widget->remove($browser->{'display'}->{'container'});

  my $pane = $browser->{'pane'} = Gtk::HPaned->new();
  $pane->signal_connect('event', sub { $browser->event(@_); }) if ($browser->gui->eventDebug);
  $pane->set_gutter_size(4);
  $pane->set_handle_size(4);
  $pane->set_position($browser->window->width * 0.33);
  $pane->show();
  $browser->widget->pack_start($pane, 1, 1, 0);
  $pane->add1($browser->{'treeScroll'});
  $pane->add2($browser->{'display'}->{'container'});

  $browser->{'display'}->{'container'}->unref;
  $browser->{'treeScroll'}->unref;

  $browser->{'treeHidden'} = 0;
  return 1;
}

# treeHidden

sub treeHidden {
  my $browser = shift;
  return $browser->{'treeHidden'};
}

# DND

# dndDragDataGet - Drag from a tree node

sub dndDragDataGet {
  my $browser = shift;
  my $source = shift;
  my $widget = shift;
  my ($context, $data, $info, $time) = @_;

  my $list = $browser->{'display'}->{'list'};
  my $tree = $browser->{'tree'};

  my @basenodes;
  if ($source eq 'display') {
    @basenodes = $browser->displaySelection;
  }
  elsif ($source eq 'tree') {
    my $treenode = $tree->selection;
    if (defined($treenode)) {
      @basenodes = ( $tree->node_get_row_data($treenode)->{'basenode'} );
    }
  }

  return 1 if (!@basenodes);

  map {
    Trace::debug('Grabbed from %s', $browser->treeLabelNode($_)) if Trace::check(Trace::DEBUG4);
  } @basenodes;

  $browser->gui->dragData(id => 666, data => \@basenodes);

  $data->set($data->target, 888, 666);
  return 1;
}

# treeDragDataReceived - Drop on a tree

sub treeDragDataReceived {
  my $browser = shift;
  my $tree = shift;
  my ($context, $x, $y, $data, $info, $time) = @_;

  (my $row, my $col) = $tree->get_selection_info($x, $y);
  my $treenode = $tree->node_nth($row);
  return 1 if (!$treenode);

  my $onto_bn = $tree->node_get_row_data($treenode)->{'basenode'};

  Trace::debug('Dropped onto %s', $browser->treeLabelNode($onto_bn)) if Trace::check(Trace::DEBUG4);

  my $from_bns = $browser->gui->dragData(id => $data->data);
  if ($onto_bn->type eq Gui::Base::Player) {
    $browser->base->populateNode(group => $from_bns, 
				 recursive => 1,
				 action => sub { $browser->playerAppendNodes($onto_bn, $from_bns); });
  }
  elsif ($onto_bn->type eq Gui::Base::Scratchpad) {
    $browser->base->populateNode(group => $from_bns, 
				 recursive => 1,
				 action => sub { $browser->scratchpadAppendNodes($from_bns); });
  }
  elsif ($onto_bn->type eq Gui::Base::List) {
    $browser->base->populateNode(group => $from_bns, 
				 recursive => 1,
				 action => sub { $browser->base->listAdd($onto_bn, tracks => [ $browser->getTracks($from_bns) ]); });
  }
  else {
    foreach my $from_bn (@{ $from_bns }) {
      if (($from_bn->type eq Gui::Base::Album) && ($onto_bn->type eq Gui::Base::Artist)) {
	$from_bn->what->reparent($onto_bn->what);
      }
    }
  }

  return 1;
}

# displayDragDataReceived - Drop on a display

sub displayDragDataReceived {
  my $browser = shift;
  my $list = shift;
  my ($context, $x, $y, $data, $info, $time) = @_;

  my $ld = $browser->{'display'}->{'list'}->get_user_data;

  # row derived from x/y list->get_selection_info seems one row too large.
  my $row;
  $y -= $browser->gui->rowHeight;
  if ($y < $browser->gui->rowHeight) {
    $y = 0;
  }

  ($row, my $col) = $list->get_selection_info($x, $y);

  my $into_bn = $ld->{'parent_bn'};
  my $onto_bn = defined($row) && ($#{ $ld->{'basenodes'} } >= $row) ? $ld->{'basenodes'}->[ $row ] : undef;

  Trace::debug('Dropped at (%d, %d), into %s, row %s, at %s', $x, $y, $browser->treeLabelNode($into_bn), 
	       defined($row) ? $row : 'none',
	       defined($onto_bn) ? $browser->treeLabelNode($onto_bn) : 'end') if Trace::check(Trace::DEBUG4);

  my $from_bns = $browser->gui->dragData(id => $data->data);

  if (defined($onto_bn) && ($onto_bn->type eq Gui::Base::Player)) {
    $browser->base->populateNode(group => $from_bns, 
				 recursive => 1,
				 action => sub { $browser->playerAppendNodes($onto_bn, $from_bns); });
  }
  elsif ($into_bn->type eq Gui::Base::Player) {
    if ($from_bns->[0]->parent eq $into_bn) { # TODO - Multiwindow
      # Source and dest are both players and the same window so no affect - no movement allowed in player
    }
    else {
      $browser->base->populateNode(group => $from_bns, 
				   recursive => 1,
				   action => sub { $browser->playerInsertNodes($into_bn, $from_bns, $row); });
    }
  }
  elsif (defined($onto_bn) && ($onto_bn->type eq Gui::Base::Scratchpad)) {
    $browser->base->populateNode(group => $from_bns, 
				 recursive => 1,
				 action => sub { $browser->scratchpadAppendNodes($from_bns); });
  }
  elsif ($into_bn->type eq Gui::Base::Scratchpad) {
    if ($from_bns->[0]->parent eq $into_bn) { # TODO - Multiwindow
      # Source and dest are both the scratchpad and the same window so move
      $browser->displayMove($row);
      $list->unselect_all;	# Clear selection for ease of list reorder with movement.
    }
    else {
      $browser->base->populateNode(group => $from_bns, 
				   recursive => 1,
				   action => sub { $browser->scratchpadInsertNodes($from_bns, $row); });
    }
  }
  elsif (defined($onto_bn) && ($onto_bn->type eq Gui::Base::List)) {
    # Dropped onto a list
    $browser->base->populateNode(group => $from_bns, 
				 recursive => 1,
				 action => sub { 
				   $browser->base->listAdd($onto_bn, tracks => [ $browser->getTracks($from_bns) ]); });
  }
  elsif ($into_bn->type eq Gui::Base::List) {
    # Dropped into a list
    if ($from_bns->[0]->parent eq $into_bn) { # TODO - Multiwindow
      # Source and dest are both the list and the same window so move
      $browser->displayMove($row);
      $list->unselect_all;	# Clear selection for ease of list reorder with movement.
    }
    else {
      $browser->base->populateNode(group => $from_bns, 
				   recursive => 1,
				   action => sub { 
				     $browser->base->listAdd($into_bn, tracks => [ $browser->getTracks($from_bns) ], at => $row); });
    }
  }
  else {
    foreach my $from_bn (@{ $from_bns }) {
      if ($from_bn->type eq Gui::Base::Album) {
	if ($into_bn->type eq Gui::Base::Artist) {
	  $from_bn->what->reparent($into_bn->what);
	}
	elsif (defined($onto_bn) && ($onto_bn->type eq Gui::Base::Artist)) {
	  $from_bn->what->reparent($onto_bn->what);
	}
      }
    }
  }

  return 1;
}

# BASE

# baseNodeAdd 

sub baseNodeAdd {
  my $browser = shift;
  my $basenode = shift;
  my $parent = $basenode->parent;

  Trace::debug('Add for: %s', $browser->treeLabelNode($basenode)) if Trace::check(Trace::DEBUG3);

  my $sibling_bn = defined($parent) ? $parent->child(after => $basenode) : undef;

  # See if it should be in our visible tree and if so add it

  if (!$basenode->leaf) {
    my $tree = $browser->{'tree'};
    _freezeThaw($tree);
    my $visible = 0;
    my $parent_tn;
    if (defined($parent) && ($parent == $browser->base->rootNode)) {
      # Toplevel items are always in view
      $visible = 1;
      $parent_tn = $browser->treeRootNode;
    }
    elsif (defined($parent)) {
      # Parent is in view
      $parent_tn = $browser->treeFindNode($parent);
      if (!$parent_tn) {
	Trace::debug('Not adding %s to tree, parent is not visible', $browser->treeLabelNode($basenode)) if Trace::check(Trace::DEBUG3);
      }
      $visible = 1;
    }
	
    if ($visible) {
      Trace::debug('Adding: %s, to: %s, after: %s', $browser->treeLabelNode($basenode), 
		   $browser->treeLabelNode($parent), defined($sibling_bn) ? $browser->treeLabelNode($sibling_bn) : 'none') if Trace::check(Trace::DEBUG6);
      my $sibling_tn;
      if ($sibling_bn) {
	$sibling_tn = $browser->treeFindNode($sibling_bn);
      }
      my $treenode = $tree->insert_node($parent_tn, $sibling_tn, [ $browser->treeLabelNode($basenode) ], 5, 
					@{ $browser->{'pixmaps'}->{ $basenode->type }->{'closed'} },
					@{ $browser->{'pixmaps'}->{ $basenode->type }->{'open'} },
					$browser->treeLeafNode($basenode), 0);
      $tree->node_set_row_data($treenode, { basenode => $basenode });
    }
  }

  # See if it is in any display list and if so insert or append it

  foreach my $hist (@{ $browser->{'display'}->{'history'} }) {
    if ($hist->{'basenode'} == $parent) {
      _freezeThaw($hist->{'list'});
      my $pixmap = $browser->{'pixmaps'}->{ $basenode->type }->{'closed'};
      my $ld = $hist->{'list'}->get_user_data;
      if (defined($sibling_bn)) {
	my $row = 0;
	foreach my $search_bn (@{ $ld->{'basenodes'} }) {
	  last if ($search_bn == $sibling_bn);
	  $row++;
	}
	$hist->{'list'}->insert($row, '', map { defined($_) ? $_ : '' } $browser->displayValues($basenode));
	$hist->{'list'}->set_pixmap($row, 0, $pixmap->[0], $pixmap->[1]);
	splice @{ $ld->{'basenodes'} }, $row, 0, $basenode;
      }
      else {
	$hist->{'list'}->append('', map { defined($_) ? $_ : '' } $browser->displayValues($basenode));
	$hist->{'list'}->set_pixmap($hist->{'list'}->rows - 1, 0, $pixmap->[0], $pixmap->[1]);
	push @{ $ld->{'basenodes'} }, $basenode;
      }
      $browser->displayUpdateColumns($hist->{'list'});
      $browser->displayInvalidateSort($hist->{'list'});
    }
  }

  # Buttons may need update
  $browser->buttonsDraw;
  return 1;
}

# baseNodeDelete 

sub baseNodeDelete {
  my $browser = shift;
  my $basenode = shift;
    
  Trace::debug('Delete for: %s', $browser->treeLabelNode($basenode)) if Trace::check(Trace::DEBUG3);

  # See if it is in our tree and if so delete it

  if (!$basenode->leaf) {
    my $treenode = $browser->treeFindNode($basenode);
    if ($treenode) {
      _freezeThaw($browser->{'tree'});
      $browser->{'tree'}->remove_node($treenode);
    }
    else {
      Trace::debug('Deleted base node not in tree', $browser->treeLabelNode($basenode)) if Trace::check(Trace::DEBUG3);
    }
  }

  # See if it is a display list and if so delete it

  my $hrow = 0;
  foreach my $hist (@{ $browser->{'display'}->{'history'} }) {
    my $parent_bn = $hist->{'list'}->get_user_data->{'parent_bn'};
    if ($parent_bn == $basenode) {
      splice @{ $browser->{'display'}->{'history'} }, $hrow, 1;
      last;
    }
    $hrow++;
  }

  # See if it is in any display list and if so delete it

  foreach my $hist (@{ $browser->{'display'}->{'history'} }) {
    my $row = 0;
    my @list_bns = @{ $hist->{'list'}->get_user_data->{'basenodes'} };
    foreach my $list_bn (@list_bns) {
      if ($list_bn == $basenode) {
	_freezeThaw($hist->{'list'});
	$hist->{'list'}->remove($row);
	splice @list_bns, $row, 1;
	$hist->{'list'}->get_user_data->{'basenodes'} = \@list_bns;
	last;
      }
      $row++;
    }
  }

  # Buttons may need update
  $browser->buttonsDraw;

  return 1;
}

# baseNodeUpdate

sub baseNodeUpdate {
  my $browser = shift;
  my $basenode = shift;

  Trace::debug('Update for: %s', $browser->treeLabelNode($basenode)) if Trace::check(Trace::DEBUG3);

  # See if it is in our tree and if so update it

  if (!$basenode->leaf) {
    my $treenode = $browser->treeFindNode($basenode);
    if ($treenode) {
      my $tree = $browser->{'tree'};
      _freezeThaw($tree);
      my @pt = $tree->node_get_pixtext($treenode, 0);
      $pt[0] = $browser->treeLabelNode($basenode);
      $tree->node_set_pixtext($treenode, 0, @pt);
    }
    else {
      Trace::debug('Updated base node not in tree', $browser->treeLabelNode($basenode)) if Trace::check(Trace::DEBUG3);
    }
  }

  # See if it is in any display list and if so modify it

  foreach my $hist (@{ $browser->{'display'}->{'history'} }) {
    my $row = 0;
    foreach my $list_bn (@{ $hist->{'list'}->get_user_data->{'basenodes'} }) {
      if ($list_bn == $basenode) {
	_freezeThaw($hist->{'list'});
	my @values = map { defined($_) ? $_ : '' } $browser->displayValues($basenode);
	for (my $col = 1; $col <= $#values + 1; $col++) {
	  $hist->{'list'}->set_text($row, $col, $values[ $col - 1 ]);
	}
	$browser->displayUpdateColumns($hist->{'list'});
	$browser->displayInvalidateSort($hist->{'list'});
	last;
      }
      $row++;
    }
  }

  # Changes in Cdrom or Player status

  if (($basenode->type eq Gui::Base::Player) || ($basenode->type eq Gui::Base::Cdrom)) {
    # Progress bar will need redraw
    delete $browser->{'display'}->{'progress'}->{'users'}->{ $basenode->what->unique };
    my $list = $browser->{'display'}->{'list'};
    if (defined($list) && ($list->get_user_data->{'parent_bn'} == $basenode)) {
      # Redraw the display
      if (($basenode->what->previousState eq Thing::Handle::Dead) && ($basenode->what->currentState eq Thing::Handle::Alive)) {
	# It has switched to alive so it needs population first
	$browser->base->populateNode(node => $basenode, action => sub { $browser->displayDraw($basenode); });
      }
      else {
	$browser->displayDraw($basenode);
      }
    }
  }

  # Buttons may need update
  $browser->buttonsDraw;

  return 1;
}

# baseNodeMove

sub baseNodeMove {
  my $browser = shift;
  my $basenode = shift;
  my %parms = @_;

  Trace::debug('Move for: %s', $browser->treeLabelNode($basenode)) if Trace::check(Trace::DEBUG3);

  # See if it is a display list and if so move the row

  foreach my $hist (@{ $browser->{'display'}->{'history'} }) {
    my $parent_bn = $hist->{'list'}->get_user_data->{'parent_bn'};
    if ($parent_bn == $basenode) {
      $hist->{'list'}->get_user_data->{'basenodes'} = [ $basenode->children ]; # new order from updated base
      my $dest;
      if (!defined($parms{'to'})) {
	$dest = $hist->{'list'}->rows - 1;
      }
      elsif ($parms{'to'} > $parms{'from'}) {
	$dest = $parms{'to'} - 1;
      }
      else {
	$dest = $parms{'to'};
      }
      $hist->{'list'}->row_move($parms{'from'}, $dest);
      $browser->displayInvalidateSort($hist->{'list'});
    }
  }

  # Buttons may need update
  $browser->buttonsDraw;

  return 1;
}

# getTracks - Figure out recursively the tracks underneath a list of
# base nodes, first use the order of any display and then that of the base.

sub getTracks {
  my $browser = shift;
  my $basenodes = shift;

  my @tracks = ();
  foreach my $basenode (@{ $basenodes }) {
    my @ctracks = $browser->_getTracks($basenode);
    if (@ctracks) {
      push @tracks, @ctracks;
    }
  }
  return @tracks;
}

# _getTracks

sub _getTracks {
  my $browser = shift;
  my $basenode = shift;

  if ($basenode->type eq Gui::Base::Track) {
    return ($basenode->what);
  }
  elsif (!$basenode->leaf) {
    my @tracks = ();
    my @children = $basenode->children;
    # Check display history
    foreach my $hist (@{ $browser->{'display'}->{'history'} }) {
      if ($hist->{'basenode'} == $basenode) {
	my $ld = $hist->{'list'}->get_user_data;
	@children = @{ $ld->{'basenodes'} };
	last;
      }
    }
    foreach my $child (@children) {
      my @ctracks = $browser->_getTracks($child);
      if (@ctracks) {
	push @tracks, @ctracks;
      }
    }
    return @tracks;
  }
  return;
}

# DISPLAY - The RHS

# displayDraw - Draw the display list for this node.

sub displayDraw {
  my $browser = shift;
  my $basenode = shift;
  my $force = shift || 0;

  Trace::debug('Display draw request for: %s', $browser->treeLabelNode($basenode)) if Trace::check(Trace::DEBUG4);

  # Check the base node still exists

  if (!$browser->base->findNode(node => $basenode)) {
    Trace::debug('This node is no longer in the tree: %s', $browser->treeLabelNode($basenode)) if Trace::check(Trace::DEBUG6);
    return 0;
  }

  # Sanity checks

  if ($basenode->leaf) {
    Trace::debug('This node has no children: %s', $browser->treeLabelNode($basenode)) if Trace::check(Trace::DEBUG6);
    return 0;
  }

  if (!$basenode->populated) {
    Trace::debug('This node is not populated: %s', $browser->treeLabelNode($basenode)) if Trace::check(Trace::DEBUG6);
    return 0;
  }

  # Create or reuse the CList

  my $list;

  # Reuse from the history

  if (!$force) {
    my $pos = 0;
    foreach my $hist (@{ $browser->{'display'}->{'history'} }) {
      if ($hist->{'basenode'} == $basenode) {
	$list = $hist->{'list'};
	Trace::debug('Reusing an existing list') if Trace::check(Trace::DEBUG6);
	# Put it back on the front
	splice @{ $browser->{'display'}->{'history'} }, $pos, 1;
	unshift @{ $browser->{'display'}->{'history'} }, $hist;
	last;
      }
      $pos++;
    }
  }

  # Create from scratch

  if (!defined($list) || $force) {

    Trace::debug('Creating a list from scratch') if Trace::check(Trace::DEBUG6);

    my @columns = $browser->displayLabels($basenode);
    my @rows = ();
    my @pixmaps = ();
    my @basenodes = ();
    foreach my $child ($basenode->children) {
      push @rows, [ $browser->displayValues($child) ];
      push @pixmaps, $browser->{'pixmaps'}->{ $child->type }->{'closed'};
      push @basenodes, $child;
    }

    $list = Gtk::CList->new_with_titles('', (map { $_->{'title'} } @columns));
    $list->signal_connect('button_press_event', sub { $browser->displayEvent(@_); });
    $list->signal_connect('button_release_event', sub { $browser->displayEvent(@_); });
    $list->signal_connect('key_press_event', sub { $browser->displayEvent(@_); });
    my $selectRowSid = $list->signal_connect('select_row', sub { $browser->displaySelectRowSignal(@_); });
    my $unselectRowSid = $list->signal_connect('unselect_row', sub { $browser->displayUnselectRowSignal(@_); });
    my $clickColSid = $list->signal_connect('click_column', sub { $browser->displayClickColumnSignal(@_); });
    $list->set_row_height($browser->gui->rowHeight);
    $list->set_user_data({ parent_bn => $basenode, basenodes => \@basenodes, columns => \@columns, 
			   clickColSid => $clickColSid, selectRowSid => $selectRowSid, unselectRowSid => $unselectRowSid });
    $list->drag_source_set(['button1_mask'], ['copy', 'move',], {'target' => "STRING", 'flags' => 0, 'info' => 0});
    $list->signal_connect('drag_data_get', sub { $browser->dndDragDataGet('display', @_); });
    $list->drag_dest_set('all', ['copy', 'move',], {'target' => "STRING", 'flags' => 0, 'info' => 0});
    $list->signal_connect('drag_data_received', sub { $browser->displayDragDataReceived(@_); });

    # Set the sort manner for the list.

    $list->column_titles_passive();
    my $col = 1;
    foreach my $column (@columns) {
      $list->column_title_active($col) if (defined($column->{'sortFunc'}) || defined($column->{'clickFunc'}));
      $col++;
    }

    # Set tooltips as appropriate.

    $col = 1;
    foreach my $column (@columns) {
      if (defined($column->{'tooltip'}) || defined($column->{'sortFunc'})) {
	my $label = Gtk::Label->new($column->{'title'});
	$label->set_justify('left');
	$label->show();
	$column->{'label'} = $label;
	my $event = Gtk::EventBox->new;
	$event->show;
	if (defined($column->{'tooltip'}) && ref($column->{'tooltip'}) && UNIVERSAL::isa($column->{'tooltip'}, 'CODE')) {
	  # Dynamic tip generation
	  my $te = $event; my $tc = $col;
	  $event->signal_connect('event', sub { if ($_[1]->{'type'} eq 'enter_notify') { $browser->displaySetTip($te, $tc) } });
	}
	else {
	  # Static
	  my $tip = '';
	  if (defined($column->{'tooltip'})) {
	    $tip = $column->{'tooltip'};
	  }
	  if (defined($column->{'sortFunc'})) {
	    $tip .= (($tip ne '') ? ' ' : '' ).'Click to sort by this column.';
	  }
	  $browser->gui->tooltips->set_tip($event, $tip, '');
	}
	my $box = Gtk::HBox->new(0, 0);
	$box->border_width(0);
	$box->show();
	$box->pack_start($label, 0, 0, 0);
	$event->add($box);
	$list->set_column_widget($col, $event);
      }
      $col++;
    }

    # Fill the list. 

    for (my $row = 0; $row <= $#rows; $row++) {
      my $pixmap = $pixmaps[$row];
      $list->append('', map { defined($_) ? $_ : '' } @{ $rows[$row] });
      $list->set_pixmap($row, 0, $pixmap->[0], $pixmap->[1]);
    }
    $browser->displayUpdateColumns($list);

    # Save in history

    unshift @{ $browser->{'display'}->{'history'} }, { basenode => $basenode, list => $list };
    $list->ref;			# For when we are in the history but not displayed
    if ($#{ $browser->{'display'}->{'history'} } == $browser->{'display'}->{'historyLimit'}) {
      my $remove = pop @{ $browser->{'display'}->{'history'} };
      $remove->{'list'}->unref;
    }

    # Selection method

    my $mode;
    if (($basenode->type eq Gui::Base::Player) || ($basenode->type eq Gui::Base::Cdrom)) {
      $mode = 'single';
    }
    else {
      $mode = 'multiple';
    }

    $list->set_selection_mode($mode);
  }

  # Parent specific list drawing options

  if (($basenode->type eq Gui::Base::Player) || ($basenode->type eq Gui::Base::Cdrom)) {

    # Select position

    if (defined($basenode->status) && defined($basenode->status->{'attributes'})) {
      my $position;
      if (($basenode->type eq Gui::Base::Player) && defined($basenode->status->{'attributes'}->{'position'})) {
	$position = $basenode->status->{'attributes'}->{'position'};
      }
      elsif (($basenode->type eq Gui::Base::Cdrom) && defined($basenode->status->{'attributes'}->{'trackIndex'})) {
	$position = $basenode->status->{'attributes'}->{'trackIndex'} - 1;
      }
      else {
	$position = 0;
      }
      my @sel = $list->selection;
      if (! grep { $position == $_ } @sel) {
	$list->select_row($position, 0);
      }
    }
  }

  if (!defined($browser->{'display'}->{'list'}) || ($browser->{'display'}->{'list'} != $list)) {
    $browser->{'display'}->{'list'}->hide if (defined($browser->{'display'}->{'list'}));
    my @children = $browser->{'display'}->{'scroll'}->children;
    map { $browser->{'display'}->{'scroll'}->remove($_) } @children;
    $browser->{'display'}->{'scroll'}->add($list);
    $browser->{'display'}->{'list'} = $list;
    $list->show;
  }

  $browser->progressDraw;
  $browser->buttonsDraw;
  $browser->infoDraw;
  $browser->window->set_title($browser->treeLabelNode($basenode));

  return 1;
}

# displayUpdateColumns - List mod implies possible need to update the
# width of the columns, invalidates existing sort.

sub displayUpdateColumns {
  my $browser = shift;
  my $list = shift;

  $list->set_column_auto_resize(0, 1); # Pixmap
  my $ld = $list->get_user_data;
  my @columns = @{ $ld->{'columns'} };
  my $num = 1;
  foreach my $column (@columns) {
    if ($column->{'optimize'}) {
      $list->set_column_width($num, $list->optimal_column_width($num));
    }
    elsif ($column->{'width'}) {
      $list->set_column_width($num, $column->{'width'});
    }
    $num++;
  }

  return 1;
}

# displayInvalidateSort - Modification, addition, movement of list
# elements implies that the list sort method is not useful for reverse
# sorting.

sub displayInvalidateSort {
  my $browser = shift;
  my $list = shift;

  my $ld = $list->get_user_data;
  if (defined($ld->{'sortCol'})) {
    $ld->{'sortInvalid'} = 1;
  }
  return 1;
}

# displayEvent

sub displayEvent {
  my $browser = shift;
  return 0 if ($#_ != 1);
  my $widget = shift;
  my $event = shift;

  return 0 if (!defined($event));

  if ($browser->gui->eventDebug) {
    Trace::debug(ref($widget).'::'.$event->{'type'}) if (Trace::check(Trace::DEBUG3));
    if ($browser->gui->eventDebug eq 'v') {
      require 'dumpvar.pl';
      main::dumpValue($event);
      $| = 1;
    }
  }

  # Button 3 brings up a menu

  if (($event->{'type'} eq 'button_press') && ($event->{'button'} == 3)) {
    $browser->menuDraw('display');
    $browser->menu->popup(0, 0, $event->{'button'}, $event->{'time'});
    return 0;
  }

  # Double click descends a node if it has children...

  if ($event->{'type'} eq '2button_press') {
    (my $row, my $col) = $browser->{'display'}->{'list'}->get_selection_info($event->{'x'}, $event->{'y'});
    if (defined($row)) {
      Trace::debug('Row double click selection %d', $row) if Trace::check(Trace::DEBUG6);
      my $basenodes = $browser->{'display'}->{'list'}->get_user_data->{'basenodes'};
      $browser->treeSelect($basenodes->[$row]);
    }
    return 0;
  }

  if (($event->{'type'} eq 'button_press') && ($event->{'button'} == 1)) {
    # Record button press for kludge below.
    $browser->{'_resizeKludge'} = $event;
  }

  if ($event->{'type'} eq 'button_release') {

    # Kludge: Press on column resize gives co-ordinates in this range,
    # so we sacrifice this space so that resize doesn't cause track seek etc.
    if (!defined($browser->{'_resizeKludge'}) ||
	(($browser->{'_resizeKludge'}->{'x'} <= 7) && ($browser->{'_resizeKludge'}->{'y'} <= 22))) {
      delete $browser->{'_resizeKludge'};
      return 0;
    }

    (my $row, my $col) = $browser->{'display'}->{'list'}->get_selection_info($event->{'x'}, $event->{'y'});

    Trace::debug('Button release at %d, %d (%d, %d) for %s', defined($row) ? $row : -1, defined($col) ? $col : -1, 
		 $event->{'x'}, $event->{'y'}, $browser->{'display'}->{'list'}->selection_mode) if Trace::check(Trace::DEBUG6);

    # Cdrom: Switching rip desire

    my $ld = $browser->{'display'}->{'list'}->get_user_data;
    if (defined($col) && ($col == 4) && ($ld->{'parent_bn'}->type eq Gui::Base::Cdrom)) {

      return 0 if !defined($ld->{'parent_bn'}->ripStatus);

      my $cdrom_bn = $ld->{'parent_bn'};
      my $cdrom = $cdrom_bn->what;
      $browser->gui->message(name => $cdrom_bn->ripStatus->{'tracks'}->{ $row + 1 }->{'want'} ? Controller::RipperRemove : Controller::RipperAdd,
			     to => $browser->base->controller,
			     parms =>
			     {
			      handle => $cdrom,
			      trackIndices => [ $row + 1 ],
			     });
      return 0;
    }

    # Middle mouse button in Cdrom or Player does an offset seek, First button switches track

    my $offset = 0.0;
    if ($event->{'button'} == 2) {
      my $dims = [ $event->{'window'}->get_size ];
      if (($#{ $dims } == 1) && defined($dims->[1])) {
	$offset = $event->{'x'} / $dims->[1];
      }
    }

    if ($ld->{'parent_bn'} && ($ld->{'parent_bn'}->type eq Gui::Base::Player)) {

      # Player - Click to seek
      my $player_bn = $ld->{'parent_bn'};
      my $player = $player_bn->what;
      $browser->gui->message(name => Player::Seek, to => $player, parms => { position => $row, seek => $offset * 100.0 });
      return 0;
    }

    if ($ld->{'parent_bn'} && ($ld->{'parent_bn'}->type eq Gui::Base::Cdrom)) {

      # Cdrom - Click to seek
      my $cdrom_bn = $ld->{'parent_bn'};
      my $cdrom = $cdrom_bn->what;
      if ($offset && $ld->{'parent_bn'}->cd && $ld->{'parent_bn'}->cd->track($row + 1)) {
	$offset = $ld->{'parent_bn'}->cd->track($row + 1)->duration * $offset;
      }
      $browser->gui->message(name => Cdrom::Seek, to => $cdrom, parms => { trackIndex => $row + 1, offset => $offset });
      return 0;
    }
  }

  return 0;
}

# displaySelectRowSignal

sub displaySelectRowSignal {
  my $browser = shift;
  my $list = shift;
  my $row = shift;
  my $col = shift;
  my $event = shift;

  if ($browser->gui->eventDebug && defined($event)) {
    Trace::debug(ref($list).'::'.$event->{'type'}) if (Trace::check(Trace::DEBUG3));
    if ($browser->gui->eventDebug eq 'v') {
      require 'dumpvar.pl';
      main::dumpValue($event);
      $| = 1;
    }
  }

  if ($list->row_is_visible($row) ne 'full') {
    $list->moveto($row, 0, 0.5, 1);
  }

  # Buttons change depending on the selection, but do it asynch in
  # case this select was signaled as a result of a next or prev
  # button.
  Gtk->timeout_add(1, sub { $browser->buttonsDraw; return 0;});

  return 0;
}

# displayUnselectRowSignal

sub displayUnselectRowSignal {
  my $browser = shift;
  my $list = shift;
  my $row = shift;
  my $col = shift;
  my $event = shift;

  if ($browser->gui->eventDebug && defined($event)) {
    Trace::debug(ref($list).'::'.$event->{'type'}) if (Trace::check(Trace::DEBUG3));
    if ($browser->gui->eventDebug eq 'v') {
      require 'dumpvar.pl';
      main::dumpValue($event);
      $| = 1;
    }
  }

  # As above selection changes can affect buttons.

  Gtk->timeout_add(1, sub { $browser->buttonsDraw; return 0;});

  return 0;
}

# displayClickColumnSignal - Primarily sorting...

sub displayClickColumnSignal {
  my $browser = shift;
  my $list = shift;
  my $col = shift;

  my $ld = $list->get_user_data;

  if (defined($ld->{'columns'}->[ $col - 1 ]->{'clickFunc'})) {
    &{ $ld->{'columns'}->[ $col - 1 ]->{'clickFunc'} };
  }

  return if (!defined($ld->{'columns'}->[ $col - 1 ]->{'sortFunc'}));

  # Sort of column changed, rearrange the objects accordingly.

  my $column = $ld->{'columns'}->[ $col - 1 ];
  my $sortFunc = $column->{'sortFunc'};
  my $sortField = $column->{'sortField'};
  my $parent_bn = $ld->{'parent_bn'};

  my $dir = 1;			# Ascending
  my $reverse = 0;
  if (defined($ld->{'sortCol'}) && ($ld->{'sortCol'} == $col)) {
    if (defined($ld->{'sortDir'}) && ($ld->{'sortDir'} == 1)) {
      $dir = -1;		# Descending
    }
    # Reverse
    $reverse = 1;
  }
  $ld->{'sortDir'} = $dir;

  if (defined($ld->{'sortCol'})) {
    if (defined($ld->{'columns'}->[ $ld->{'sortCol'} - 1 ]->{'label'})) {
      my $title = $ld->{'columns'}->[ $ld->{'sortCol'} - 1 ]->{'label'}->get;
      if ($title =~ /(.*)[\+\-]/) {
	$title = $1;
	$ld->{'columns'}->[ $ld->{'sortCol'} - 1 ]->{'label'}->set_text($title);
      }
    }
    else {
      my $title = $list->get_column_title($ld->{'sortCol'});
      if ($title =~ /(.*)[\+\-]/) {
	$title = $1;
	$list->set_column_title($ld->{'sortCol'}, $title);
      }
    }
  }

  if (defined($column->{'label'})) {
    my $title = $column->{'label'}->get;
    $title .= ($dir == 1) ? '+' : '-';
    $column->{'label'}->set_text($title);
  }
  else {
    my $title = $list->get_column_title($col);
    $title .= ($dir == 1) ? '+' : '-';
    $list->set_column_title($col, $title);
  }

  $ld->{'sortCol'} = $col;

  Trace::debug('Sorting by %d %s', $col, 
	       ($reverse && ! defined($ld->{'sortInvalid'}) ) ? 'reverse' : ($dir == 1 ? 'ascending' : 'descending')) if Trace::check(Trace::DEBUG4);
    
  $list->signal_handler_block($ld->{'clickColSid'});

  my $basenodes = $ld->{'basenodes'};
  if (!$reverse || defined($ld->{'sortInvalid'})) {
    my @sortData = ();
    my $row = 0;
    while ($row < $list->rows) {

      # Get the data from the list unless it needs to be pulled
      # from the basenode object e.g. when the list is showing
      # modified more humanly readable data.

      if (!defined($sortField)) {
	push @sortData, $list->get_text($row, $col);
      }
      else {
	if (UNIVERSAL::isa($sortField, 'CODE')) {
	  push @sortData, &{ $sortField }($basenodes->[ $row ]->what);
	}
	else {
	  # PERL 5.00X which can't handle $basenodes->[ $row ]->what->$sortField; so we'll rely on the above
	  # push @sortData, $basenodes->[ $row ]->what->$sortField;
	  die;
	}
      }
      $row++;
    }

    # sortOrder - Array order indicates desired position, value indicates current position
    my @sortOrder = sort { &{ $sortFunc }($sortData[$a], $sortData[$b]) } ( 0..$#sortData );
    @sortOrder = reverse @sortOrder if ($dir == -1);

    #Trace::debug('New order: %s', join(' ', map { $sortData[$_] } @sortOrder)) if Trace::check(Trace::DEBUG6);
    #Trace::debug('Map: %s', join(' ', @sortOrder)) if Trace::check(Trace::DEBUG6);
    # Feel like I have to change these inline, could be bollix
    my @new = ();
    map { $new[ $_ ] = $basenodes->[ $sortOrder[ $_ ] ] } ( 0..$#sortData );
    map { $basenodes->[ $_ ] = $new[ $_ ] } ( 0..$#sortData );

    if ($parent_bn->type eq Gui::Base::List) {
      my @newTrackIoids = map { $_->what->ioid; } @new;
      # Remove existing 
      $parent_bn->what->deleteTracks(trackPositions => [ ( 0..$#sortData ) ]);
      if (!$browser->gui->wait(event => sub { $#{ $parent_bn->what->track_ioids } == -1 }, timeout => 10)) {
	GtkEx::Utility::ackPopup(msg => 'Timed out waiting for tracks to change');
      }
      else {
	# Add in new order
	$parent_bn->what->addTracks(trackIoids => \@newTrackIoids);
      }
    }
    else {
      $list->freeze;
      map { $list->row_move( $sortOrder[ $_ ], $_ ); 
	    # Current position is changed by the row move
	    my $from = $sortOrder[ $_ ]; my $to = $_; 
	    if ($from > $to) {
	      # Push down position values of everything between to and from
	      map { $sortOrder[ $_ ] = $sortOrder[ $_ ] + 1 if (($sortOrder[ $_ ] < $from) && ($sortOrder[ $_ ] >= $to)) } (0..$#sortOrder);
	    }
	    elsif ($to > $from) {
	      # Cannot happen but the logic hurts
	      die;
	    }
	    $sortOrder[ $_ ] = $_; 
	  } ( 0..$#sortData );
      $list->thaw;
    }
    delete $ld->{'sortInvalid'};
  }
  else {
    # Reversal
    if ($parent_bn->type eq Gui::Base::List) {
      my @newTrackIoids = reverse map { $_->what->ioid; } @{ $basenodes };
      # Remove existing 
      $parent_bn->what->deleteTracks(trackPositions => [ ( 0..$#newTrackIoids ) ]);
      if (!$browser->gui->wait(event => sub { $#{ $parent_bn->what->track_ioids } == -1 }, timeout => 10)) {
	GtkEx::Utility::ackPopup(msg => 'Timed out waiting for tracks to change');
      }
      else {
	# Add in new order
	$parent_bn->what->addTracks(trackIoids => \@newTrackIoids);
      }
    }
    else {
      # Swap
      my @range = (0..(POSIX::floor($list->rows / 2) - 1));
      $list->freeze;
      map { $list->swap_rows($_, $list->rows - 1 - $_) } @range;
      map { my $hold = $basenodes->[ $_ ]; 
	    $basenodes->[ $_ ] = $basenodes->[ $#{ $basenodes } - $_ ]; 
	    $basenodes->[ $#{ $basenodes } - $_ ] = $hold } @range;
      $list->thaw;
    }
  }

  $list->signal_handler_unblock($ld->{'clickColSid'});

  return 1;
}

# displayMoveUp - Move the display selection one row up.

sub displayMoveUp {
  my $browser = shift;

  my $list = $browser->{'display'}->{'list'};
  my @sel = sort { $a <=> $b } $list->selection;
  return if (!@sel);
  return if ($sel[0] == 0);	# At top
  return $browser->displayMove($sel[0] - 1);
}

# displayMoveDown - Move the display selection one row down.

sub displayMoveDown {
  my $browser = shift;

  my $list = $browser->{'display'}->{'list'};
  my @sel = sort { $a <=> $b } $list->selection;
  return if (!@sel);
  return if ($sel[ $#sel ] == $list->rows - 1);  # At bottom
  my $to;
  if ($sel[ $#sel ] != $list->rows - 2) {
    $to = $sel[ $#sel ] + 2;
  }
  return $browser->displayMove($to);
}

# displayMove - Move the display selection to the point before row.

sub displayMove {
  my $browser = shift;
  my $row = shift;

  my $list = $browser->{'display'}->{'list'};
  my @sel = sort { $a <=> $b } $list->selection;

  my $ld = $list->get_user_data;
  my $basenodes = $ld->{'basenodes'};

  # Bounds check

  if (defined($row) && ($row >= $sel[0]) && ($row <= $sel[ $#sel ])) {
    return 1;
  }
  $list->freeze;

  my $modifier;

  # Before - move lowest to highest, incrementing point
  if (defined($row) && ($row < $sel[0])) {
    $modifier = 1;
  }
  else {
    # After - move  highest to lowest, decrementing point
    $modifier = -1;
    @sel = reverse @sel;
  }

  while (@sel) {
    my $from = shift @sel;
    my $to = $row;

    if ($ld->{'parent_bn'}->type ne Gui::Base::List) {
      $list->row_move($from, defined($to) ? ($to > $from ? $to - 1 : $to) : $list->rows - 1);
      if (!defined($to)) {
	push @{ $basenodes }, $basenodes->[ $from ];
      }
      else {
	splice @{ $basenodes }, $to, 0, $basenodes->[ $from ];
	if ($from >= $to) {
	  $from++;
	}
      }
      splice @{ $basenodes }, $from, 1;
    }
    else {
      $ld->{'parent_bn'}->what->moveTrack(from => $from, to => $to);
    }

    if (!defined($row)) {
      $row = $list->rows - 1;
    }
    else {
      $row += $modifier;
    }
  }

  $list->thaw;

  $browser->displayInvalidateSort($list);
  return 1;
}

# displayRemove - Remove the selected tracks from the display, only
# called for players, lists and the scratchpad.

sub displayRemove {
  my $browser = shift;
    
  my $list = $browser->{'display'}->{'list'};
  my @sel = sort { $a <=> $b } $list->selection;
  return 1 if (!@sel);
    
  my $ld = $list->get_user_data;
  my $basenodes = $ld->{'basenodes'};

  if ($ld->{'parent_bn'}->type eq Gui::Base::Scratchpad) {
    $browser->base->scratchpadRemove($ld->{'parent_bn'}, nodes => [ map { $basenodes->[$_] } @sel ]);
  }
  elsif ($ld->{'parent_bn'}->type eq Gui::Base::List) {
    $browser->base->listRemove($ld->{'parent_bn'}, nodes => [ map { $basenodes->[$_] } @sel ]);
  }
  elsif ($ld->{'parent_bn'}->type eq Gui::Base::Player) {
    $browser->gui->message(name => Player::Clear, to => $ld->{'parent_bn'}->what, parms => { rows => \@sel });
  }

  return 1;
}

# displayClear - Clear the display, only called for players and the
# scratchpad.

sub displayClear {
  my $browser = shift;
     
  my $list = $browser->{'display'}->{'list'};
  my $ld = $list->get_user_data;
  my $basenodes = $ld->{'basenodes'};

  if ($ld->{'parent_bn'}->type eq Gui::Base::Scratchpad) {
    $browser->base->scratchpadRemove($ld->{'parent_bn'}, nodes => $basenodes);
  }
  elsif ($ld->{'parent_bn'}->type eq Gui::Base::Player) {
    $browser->gui->message(name => Player::Clear, to => $ld->{'parent_bn'}->what);
  }

  return 1;
}

# displayLabels - Derive column labels, sizes and sortability to use
# when displaying the children of this base node.

sub displayLabels {
  my $browser = shift;
  my $basenode = shift;

  # Labels can depend on both the node and its children.

  if (!$basenode->homogeneous) {

    # Non homogeneous children, one column with the name.
    return ( { title => 'Name' } );
  }

  # Homogeneous children

  my $childType = $basenode->childType;

  my $alphaSort = sub { 
    return ($_[0] cmp $_[1]) if (defined($_[0]) && defined($_[1]));
    return 0 if (!defined($_[0]) && !defined($_[1]));
    return -1 if (!defined($_[0]));
    return 1;
  };

  my $numSort = sub { 
    return ($_[0] <=> $_[1]) if (defined($_[0]) && defined($_[1]));
    return 0 if (!defined($_[0]) && !defined($_[1]));
    return -1 if (!defined($_[0]));
    return 1;
  };

  my $totDuration = sub {
    # $_[0] -- browser
    # $_[1] -- parent_bn
    my $data = $_[2];
    my $tot = 0;
    if (@{ $data }) {
      map { $tot += $_ } @{ $data };
    }
    return sprintf('Total time is %d:%02d.', int($tot / 60), int($tot % 60));
  };

  # Column widths are optimized to take up as much room as they need
  # or fixed.

  my $artistW = my $albumW = my $trackW = my $listW = 150;

  if ($childType eq Gui::Base::Artist) {
    return ({ title => 'Artist',      width => $artistW, sortFunc => $alphaSort },
	    { title => 'Last Played', optimize => 1, sortFunc => $alphaSort });
  }
  elsif (($childType eq Gui::Base::Album) && ($basenode->type eq Gui::Base::Artist)) {
    # Artist visible in tree don't show in list
    return ({ title => 'Album',       width => $albumW, sortFunc => $alphaSort },
	    { title => '',            optimize => 1, sortFunc => $numSort,
	      tooltip => 'The number of times this album has been played.'   },
	    { title => 'Last Played', optimize => 1, sortFunc => $alphaSort });
  }
  elsif ($childType eq Gui::Base::Album) {
    return ({ title => 'Album',       width => $albumW, sortFunc => $alphaSort },
	    { title => 'Artist',      width => $artistW, sortFunc => $alphaSort },
	    { title => '',            optimize => 1, sortFunc => $numSort,
	      tooltip => 'The number of times this album has been played.'   },
	    { title => 'Last Played', optimize => 1, sortFunc => $alphaSort });
  }
  elsif ($childType eq Gui::Base::List) {
    return ({ title => 'Name',        width => $listW, sortFunc => $alphaSort },
	    { title => '#',           optimize => 1, sortFunc => $numSort,
	      tooltip => 'The number of tracks in this play list.'   });
  }
  elsif ($childType eq Gui::Base::Track) {

    my @cols =  ({ title => '',            optimize => 1, sortFunc => $numSort,  sortField => sub { $_[0]->index } }, 
		 { title => '',            optimize => 1, sortFunc => $numSort,  sortField => sub { $_[0]->duration },
		   tooltip => $totDuration },
		 { title => 'Track',       width => $trackW, sortFunc => $alphaSort },
		 { title => 'Album',       width => $albumW, sortFunc => $alphaSort },
		 { title => 'Artist',      width => $artistW, sortFunc => $alphaSort },
		 { title => '',            optimize => 1, sortFunc => $numSort,
		   tooltip => 'The number of times this track has been played.'   },
		 { title => 'Last Played', optimize => 1, sortFunc => $alphaSort });

    # No sort for Players or Cdroms
    if (($basenode->type eq Gui::Base::Cdrom) || ($basenode->type eq Gui::Base::Player)) {
      map { delete $_->{'sortFunc'}; } @cols;
    }

    # Artist/Album visible in tree (or otherwise) don't show in display
    if (($basenode->type eq Gui::Base::Cdrom) || 
	(($basenode->type eq Gui::Base::Album) && $basenode->parent && ($basenode->parent->type eq Gui::Base::Artist))) {
      splice @cols, 3, 2;
    }

    # Extra fields for Cdrom
    if ($basenode->type eq Gui::Base::Cdrom) {
      splice @cols, 3, 0, ({ title => 'R?', optimize => 1, clickFunc => sub { $browser->cdromClickColumn($basenode); },
			     tooltip => 'Click to enable or disable tracks for ripping.' },
			   { title => 'RE', optimize => 1,
			     tooltip => '\'R\' - This track is ripped. \'E\' - This track is encoded.' });
    }

    return @cols;
  }
}

# displayValues - Return the values to display for this base node.

sub displayValues {
  my $browser = shift;
  my $basenode = shift;

  my $parent = $basenode->parent;

  # Values can depend on both the node and its parent.

  if (!$parent->homogeneous) {
    # Non homogeneous children i.e. Root
    return ( $browser->treeLabelNode($basenode) );
  }

  my $type = $basenode->type;
  my $what = $basenode->what;
  my $parentType = defined($parent) ? $parent->type : undef;

  # Homogeneous which for now will be Artists, Albums, Tracks and Lists

  if ($type eq Gui::Base::Artist) {
    return ($what->name, $what->play_date);
  }
  elsif (($type eq Gui::Base::Album) && $parent && ($parentType eq Gui::Base::Artist)) {
    # Artist visible in tree don't show in list
    return ($what->name, $what->play_count, $what->play_date);
  }
  elsif ($type eq Gui::Base::Album) {
    return ($what->name, $what->artist->name, $what->play_count, $what->play_date);
  }
  elsif ($type eq Gui::Base::List) {
    return ($what->name, defined($what->track_ioids) ? $#{ $what->track_ioids } + 1 : 0);
  }
  elsif ($type eq Gui::Base::Track) {
    my $duration = sprintf('[%d:%02d]', int($what->duration / 60), int($what->duration % 60));
    my @values;

    # Artist/Album visible in tree (or otherwise) don't show in display
    if ($parent &&
	(($parentType eq Gui::Base::Cdrom) || 
	 (($parentType eq Gui::Base::Album) && $parent->parent && ($parent->parent->type eq Gui::Base::Artist)))) {
      @values = ( $what->index.'.', $duration, $what->name, $what->play_count, $what->play_date );
    }
    else {
      @values = ( $what->index.'.', $duration, $what->name, $what->album->name, $what->album->artist->name, 
		  $what->play_count, $what->play_date );
    }

    # Extra fields for Cdrom
    if ($parentType eq Gui::Base::Cdrom) {
      my @xvals;
      if (defined($parent->ripStatus) && defined($parent->ripStatus->{'tracks'}) && defined($parent->ripStatus->{'tracks'}->{ $what->index })) {
	my $ripStatus = $parent->ripStatus->{'tracks'}->{ $what->index };
	@xvals = ($ripStatus->{'want'} ? 'Y' : 'N', defined($ripStatus->{'ripped'}) && $ripStatus->{'ripped'} ? 'R' : ''.
		  defined($ripStatus->{'encoded'}) && $ripStatus->{'encoded'} ? 'E' : 'N');
      }
      else {
	@xvals = ('?', '?');
      }
      splice @values, 3, 0, @xvals;
    }
    return @values;
  }
}

# displaySelection - What is selected in the display list.

sub displaySelection {
  my $browser = shift;
  my $row = shift;

  my @sel;
  if (!defined($row)) {
    @sel = $browser->{'display'}->{'list'}->selection;
  }
  else {
    @sel = ( $row );
  }
  return () if (!@sel);

  my $basenodes = $browser->{'display'}->{'list'}->get_user_data->{'basenodes'};
  my @sel_bn = ();
  foreach my $row (sort { $a <=> $b } @sel) {
    push @sel_bn, $basenodes->[$row];
  }
  return @sel_bn;
}

# displaySetTip - Dynamically set a tooltip for a column based on the
# list parent and/or the column data.

sub displaySetTip {
  my $browser = shift;
  my $event = shift;
  my $col = shift;

  my $list = $browser->{'display'}->{'list'};
  my $ld = $list->get_user_data;
  my $column = $ld->{'columns'}->[ $col - 1 ];
  my $sortField = $column->{'sortField'};
  my $parent_bn = $ld->{'parent_bn'};
  my $basenodes = $ld->{'basenodes'};
    
  my @data;
  my $row = 0;
  while ($row < $list->rows) {

    # Similiar to sorting get the data from the list unless it
    # needs to be pulled from the basenode object e.g. when the
    # list is showing modified more humanly readable data.

    if (!defined($sortField)) {
      push @data, $list->get_text($row, $col);
    }
    else {
      if (UNIVERSAL::isa($sortField, 'CODE')) {
	push @data, &{ $sortField }($basenodes->[ $row ]->what);
      }
      else {
	# PERL 5.00X which can't handle $basenodes->[ $row ]->what->$sortField; so we'll rely on the above
	# push @sortData, $basenodes->[ $row ]->what->$sortField;
	die;
      }
    }
    $row++;
  }

  my $tip = &{ $column->{'tooltip'} }($browser, $parent_bn, \@data);
  if (defined($column->{'sortFunc'})) {
    $tip .= ((defined($tip) && ($tip ne '')) ? ' ' : '' ).'Click to sort by this column.';
  }
  $browser->gui->tooltips->set_tip($event, $tip, '');
  return 0;
}

# displayParent(type) - The parent of the list, undef if it is not type.

sub displayParent {
  my $browser = shift;
  my $type = shift;

  my $list = $browser->{'display'}->{'list'};
  return undef if !$list;

  my $ld = $list->get_user_data;
  return undef if !$ld;
  return undef if !$ld->{'parent_bn'};

  return $ld->{'parent_bn'} if !$type;

  return undef if ($type ne $ld->{'parent_bn'}->type);
  return $ld->{'parent_bn'};
}

# PROGRESS

# progressDraw - Show the progress bar state pertinent to the current display window.

sub progressDraw {
  my $browser = shift;
  my $progressH = $browser->{'display'}->{'progress'}->{'users'};
  my $bar = $browser->{'display'}->{'progress'}->{'bar'};
  my $timeout = $browser->{'display'}->{'progress'}->{'timeout'};
  my $basenode = $browser->{'display'}->{'list'}->get_user_data->{'parent_bn'};
  my $progress;

  if ($basenode->type eq Gui::Base::Player) {
    if (defined($basenode->status) && defined($basenode->status->{'state'}) && ($basenode->status->{'state'} eq Player::Playing)) {
      if (!defined($progressH->{ $basenode->what->unique })) {
	# Create new active progress bar
	$progress = $progressH->{ $basenode->what->unique } = {};
	$progress->{'progress'} = $basenode->status->{'attributes'}->{'seek'} / 100.0;
	$progress->{'lastProgressTime'} = $basenode->status->{'attributes'}->{'time'};
	$progress->{'endEstimateTime'} = $basenode->status->{'attributes'}->{'endTime'};
	if (!defined($basenode->status->{'attributes'}->{'direction'}) ||
	    ($basenode->status->{'attributes'}->{'direction'} eq Player::Forward)) {
	  $progress->{'direction'} = 'forward';
	}
	else {
	  $progress->{'direction'} = 'reverse';
	}
      }
      else {
	$progress = $progressH->{ $basenode->what->unique };
      }
    }
    elsif (defined($basenode->status) && defined($basenode->status->{'attributes'}) && defined($basenode->status->{'attributes'}->{'seek'})) {
      my $value = $basenode->status->{'attributes'}->{'seek'} / 100.0;
      $value = 1.0 if ($value > 1.0);
      $value = 0.0 if ($value < 0);
      $bar->update($value);
      delete $progressH->{ $basenode->what->unique };
    }
    else {
      $bar->update(0);
      delete $progressH->{ $basenode->what->unique };
    }
    $bar->show;
  }
  elsif ($basenode->type eq Gui::Base::Cdrom) {
    if (defined($basenode->status) && defined($basenode->status->{'state'}) && ($basenode->status->{'state'} eq Cdrom::Playing)) {
      if (!defined($progressH->{ $basenode->what->unique })) {
	# Create new active progress bar
	$progress = $progressH->{ $basenode->what->unique } = {};
	my $duration = $basenode->cd->track( $basenode->status->{'attributes'}->{'trackIndex'} )->duration;
	my $left = $basenode->status->{'attributes'}->{'left'};
	$progress->{'progress'} = 1.0 - ($left / $duration);
	$progress->{'lastProgressTime'} = $basenode->status->{'attributes'}->{'time'};
	$progress->{'endEstimateTime'} = $basenode->status->{'attributes'}->{'time'} + $left;
	$progress->{'direction'} = 'forward';
      }
      else {
	$progress = $progressH->{ $basenode->what->unique };
      }
    }
    elsif (defined($basenode->status) && defined($basenode->status->{'state'}) && ($basenode->status->{'state'} eq Cdrom::Ripping)) {
      my $value = $basenode->status->{'attributes'}->{'percentageComplete'} / 100;
      if ($value >= 1.0) {
	$value = 1.0;
      }
      $bar->update($value);
      delete $progressH->{ $basenode->what->unique };
    }
    elsif (defined($basenode->status) && defined($basenode->status->{'attributes'}) && defined($basenode->status->{'attributes'}->{'trackIndex'}) &&
	   defined($basenode->status->{'attributes'}->{'left'})) {
      my $duration = $basenode->cd->track( $basenode->status->{'attributes'}->{'trackIndex'} )->duration;
      my $left = $basenode->status->{'attributes'}->{'left'};
      my $value = 1.0 - ($left / $duration);
      $value = 1.0 if ($value > 1.0);
      $value = 0.0 if ($value < 0);
      $bar->update($value);
      delete $progressH->{ $basenode->what->unique };
    }
    else {
      $bar->update(0);
      delete $progressH->{ $basenode->what->unique };
    }
    $bar->show;
  }
  else {
    $bar->hide;
  }

  if (defined($progress)) {
    $browser->progressTimeout; # Initial tick
    if (!defined($timeout)) {
      $browser->{'display'}->{'progress'}->{'timeout'} = Gtk->timeout_add(1000, sub { $browser->progressTimeout });
    }
  }

  return 1;
}

sub progressTimeout {
  my $browser = shift;
  my $progressH = $browser->{'display'}->{'progress'}->{'users'};
  my $bar = $browser->{'display'}->{'progress'}->{'bar'};
  my $basenode = $browser->{'display'}->{'list'}->get_user_data->{'parent_bn'};

  if (($basenode->type eq Gui::Base::Player) || ($basenode->type eq Gui::Base::Cdrom)) {
    my $progress = $progressH->{ $basenode->what->unique };
    if (defined($progress)) {
      my $currentTime = time;
      my $value;
      if ($progress->{'lastProgressTime'} < $progress->{'endEstimateTime'}) {
	my $delta = ( ( ( $currentTime - $progress->{'lastProgressTime'} ) / 
			( $progress->{'endEstimateTime'} - $progress->{'lastProgressTime'} ) ) *
		      ( ($progress->{'direction'} eq 'forward') ? 1.0 - $progress->{'progress'} : $progress->{'progress'}) );
	if ($progress->{'direction'} eq 'forward') {
	  $value = $progress->{'progress'} + $delta;
	}
	else {
	  $value = $progress->{'progress'} - $delta;
	}
      }
      else {
	$value = ($progress->{'direction'} eq 'forward') ? 1.0 : 0.0;
      }
      if ($value > 1.0) {
	$value = 1.0;
      }
      elsif ($value < 0.0) {
	$value = 0.0;
      }
      $bar->update($value);
      return 1;
    }
    delete $browser->{'display'}->{'progress'}->{'timeout'};
    return 0;
  }
  delete $browser->{'display'}->{'progress'}->{'timeout'};
  return 0;
}

sub progressButtonPressEvent {
  my $browser = shift;
  my $widget = shift;
  my $event = shift;

  return 0 if (!defined($event) || !defined($event->{'x'}) || !defined($event->{'window'}));

  my $dims = [$event->{'window'}->get_size];
  return 0 if (($#{ $dims } != 1) || !defined($dims->[1]));
  my $seek = $event->{'x'} / $dims->[1];

  my $basenode = $browser->{'display'}->{'list'}->get_user_data->{'parent_bn'};
  if ($basenode->type eq Gui::Base::Player) {
    $browser->gui->message(name => Player::Seek, to => $basenode->what, parms => { seek => $seek * 100.0 });
  }
  elsif (($basenode->type eq Gui::Base::Cdrom) && defined($basenode->status) && defined($basenode->status->{'attributes'})) {
    my $ti = $basenode->status->{'attributes'}->{'trackIndex'};
    if (defined($ti) && $basenode->cd->track($ti)) {
      $seek = $basenode->cd->track($ti)->duration * $seek;
      $browser->gui->message(name => Cdrom::Seek, to => $basenode->what, parms => { trackIndex => $ti, offset => $seek });
    }
  }

  return 0;
}

# PLAYER

sub playerPlay {
  my $browser = shift;
  my $player_bn = $browser->displayParent(Gui::Base::Player) || return 1;
  my $player = $player_bn->what || return 1;;
  $browser->gui->message(name => Player::Play, to => $player);
  return 1;
}

sub playerPause {
  my $browser = shift;
  my $player_bn = $browser->displayParent(Gui::Base::Player) || return 1;
  my $player = $player_bn->what || return 1;;
  $browser->gui->message(name => Player::Pause, to => $player);
  return 1;
}

sub playerStop {
  my $browser = shift;
  my $player_bn = $browser->displayParent(Gui::Base::Player) || return 1;
  my $player = $player_bn->what || return 1;;
  $browser->gui->message(name => Player::Stop, to => $player);
  return 1;
}

sub playerSearch {
  my $browser = shift;
  my $parms = shift;
  my $player_bn = $browser->displayParent(Gui::Base::Player) || return 1;
  my $player = $player_bn->what || return 1;;
  $browser->gui->message(name => Player::Search, to => $player, parms => $parms);
  return 1;
}

sub playerClear {
  my $browser = shift;
  my $player_bn = $browser->displayParent(Gui::Base::Player) || return 1;
  my $player = $player_bn->what || return 1;;
  $browser->gui->message(name => Player::Clear, to => $player);
  return 1;
}

sub playerPlayNodes {
  my $browser = shift;
  my $player_bn = shift;
  my $player = $player_bn->what;
  my $basenodes = shift;

  my @tracks = $browser->getTracks($basenodes);
  return 1 if (!@tracks);

  $browser->gui->message(name => Player::Stop, to => $player);
  $browser->gui->message(name => Player::Clear, to => $player);
  $browser->gui->message(name => Player::Spool, to => $player, parms => { tracks => \@tracks } );
  $browser->gui->message(name => Player::Play, to => $player);
  return 1;
}

sub playerAppendNodes {
  my $browser = shift;
  my $player_bn = shift;
  my $player = $player_bn->what;
  my $basenodes = shift;

  my @tracks = $browser->getTracks($basenodes);
  return 1 if (!@tracks);

  $browser->gui->message(name => Player::Spool, to => $player, parms => { at => 'append', tracks => \@tracks } );
  return 1;
}

sub playerInsertNodes {
  my $browser = shift;
  my $player_bn = shift;
  my $player = $player_bn->what;
  my $basenodes = shift;
  my $row = shift;

  my @tracks = $browser->getTracks($basenodes);
  return 1 if (!@tracks);

  my @at = defined($row) ? (at => 'insert', position => $row) : (at => 'append');
  $browser->gui->message(name => Player::Spool, to => $player, parms => { @at, tracks => \@tracks } );
  return 1;
}

# playerNext

sub playerNext {
  my $browser = shift;
  my $player_bn = $browser->displayParent(Gui::Base::Player) || return 1;
  my $player = $player_bn->what || return 1;;

  return 1 if !defined($player_bn->status);
  return 1 if !defined($player_bn->spool);
  return 1 if !defined($player_bn->status->{'attributes'}->{'position'});

  my $position = $player_bn->status->{'attributes'}->{'position'};
  my $rows = $#{ $player_bn->spool } + 1;

  my $newPos;
  if (!defined($position)) {
    $newPos = 0;
  }
  else {
    $newPos = $position + 1;
  }
  if ($newPos >= $rows) {
    $newPos = 0;
  }
  $browser->gui->message(name => Player::Seek, to => $player, parms => { position => $newPos });
  return 1;
}

# playerPrev

sub playerPrev {
  my $browser = shift;
  my $player_bn = $browser->displayParent(Gui::Base::Player) || return 1;
  my $player = $player_bn->what || return 1;;

  return 1 if !defined($player_bn->status);
  return 1 if !defined($player_bn->spool);
  return 1 if !defined($player_bn->status->{'attributes'}->{'position'});

  my $position = $player_bn->status->{'attributes'}->{'position'};
  my $rows = $#{ $player_bn->spool } + 1;

  my $newPos;
  if (!defined($position)) {
    $newPos = $rows - 1;
  }
  else {
    $newPos = $position - 1;
  }
  if ($newPos < 0) {
    $newPos = $rows - 1;
  }

  $browser->gui->message(name => Player::Seek, to => $player, parms => { position => $newPos });
  return 1;
}

# CDROM

sub cdromPlay {
  my $browser = shift;
  my $cdrom_bn = $browser->displayParent(Gui::Base::Cdrom) || return 1;
  my $cdrom = $cdrom_bn->what || return 1;;
  $browser->gui->message(name => Cdrom::Play, to => $cdrom);
  return 1;
}

sub cdromPause {
  my $browser = shift;
  my $cdrom_bn = $browser->displayParent(Gui::Base::Cdrom) || return 1;
  my $cdrom = $cdrom_bn->what || return 1;;
  $browser->gui->message(name => Cdrom::Pause, to => $cdrom);
  return 1;
}

sub cdromRip {
  my $browser = shift;
  my $cdrom_bn = $browser->displayParent(Gui::Base::Cdrom) || return 1;
  my $cdrom = $cdrom_bn->what || return 1;;
  $browser->gui->message(name => Controller::RipperRip, to => $browser->base->controller, parms => { handle => $cdrom, }, );
  return 1;
}

sub cdromStop {
  my $browser = shift;
  my $cdrom_bn = $browser->displayParent(Gui::Base::Cdrom) || return 1;
  my $cdrom = $cdrom_bn->what || return 1;;
  $browser->gui->message(name => Cdrom::Stop, to => $cdrom);
  return 1;
}

sub cdromEject {
  my $browser = shift;
  my $cdrom_bn = $browser->displayParent(Gui::Base::Cdrom) || return 1;
  my $cdrom = $cdrom_bn->what || return 1;;
  $browser->gui->message(name => Cdrom::Eject, to => $cdrom);
  $browser->{'display'}->{'infoLabel'}->set_text('Ejecting...');
  return 1;
}

sub cdromClose {
  my $browser = shift;
  my $cdrom_bn = $browser->displayParent(Gui::Base::Cdrom) || return 1;
  my $cdrom = $cdrom_bn->what || return 1;;
  $browser->gui->message(name => Cdrom::Close, to => $cdrom);
  $browser->{'display'}->{'infoLabel'}->set_text('Closing...');
  return 1;
}

# cdromNext

sub cdromNext {
  my $browser = shift;
  my $cdrom_bn = $browser->displayParent(Gui::Base::Cdrom) || return 1;
  my $cdrom = $cdrom_bn->what || return 1;;

  return 1 if !defined($cdrom_bn->status);
  return 1 if !defined($cdrom_bn->cd);
  return 1 if !defined($cdrom_bn->status->{'attributes'}->{'trackIndex'});

  my $position = $cdrom_bn->status->{'attributes'}->{'trackIndex'} - 1;
  my $rows = $cdrom_bn->cd->track_count;

  my $newPos;
  if (!defined($position)) {
    $newPos = 0;
  }
  else {
    $newPos = $position + 1;
  }
  if ($newPos >= $rows) {
    $newPos = 0;
  }
  $browser->gui->message(name => Cdrom::Seek, to => $cdrom, parms => { trackIndex => $newPos + 1 });
  return 1;
}

# cdromPrev

sub cdromPrev {
  my $browser = shift;
  my $cdrom_bn = $browser->displayParent(Gui::Base::Cdrom) || return 1;
  my $cdrom = $cdrom_bn->what || return 1;;

  return 1 if !defined($cdrom_bn->status);
  return 1 if !defined($cdrom_bn->cd);
  return 1 if !defined($cdrom_bn->status->{'attributes'}->{'trackIndex'});

  my $position = $cdrom_bn->status->{'attributes'}->{'trackIndex'} - 1;
  my $rows = $cdrom_bn->cd->track_count;

  my $newPos;
  if (!defined($position)) {
    $newPos = $rows - 1;
  }
  else {
    $newPos = $position - 1;
  }
  if ($newPos < 0) {
    $newPos = $rows - 1;
  }

  $browser->gui->message(name => Cdrom::Seek, to => $cdrom, parms => { trackIndex => $newPos + 1 });
  return 1;
}

sub cdromClickColumn {
  my $browser = shift;
  my $cdrom_bn = shift;
  my $cdrom = $cdrom_bn->what;

  return 1 if (!defined($cdrom_bn->ripStatus));

  my $wantAllOff = 1;
  foreach my $track (values(%{ $cdrom_bn->ripStatus->{'tracks'} })) {
    if ($track->{'want'}) {
      $wantAllOff = 0;
      last;
    }
  }
  $browser->gui->message(name => $wantAllOff ? Controller::RipperAdd : Controller::RipperRemove,
			 to => $browser->base->controller,
			 parms => 
			 { 
			  handle => $cdrom,
			 });
  return 1;
}

sub cdromPropertyPopup {
  my $browser = shift;
  my $cdrom_bn = shift;

  return 1 if (!defined($cdrom_bn->cd));
  my $cd = $cdrom_bn->cd;
  my $props = {};

  # Edit table

  my $table = GtkEx::TableEdit->new(colSpacing => 5, rowSpacing => 5);
  $table->show();

  my $entry = $props->{'editArtist'} = Gtk::Entry->new(1024);
  $entry->show;
  $entry->set_text($cd->artist->name);
  $entry->set_usize(400, 20);
  $table->appendRow(title => 'Artist:', widget => $entry);

  $entry = $props->{'editAlbum'} = Gtk::Entry->new(1024);
  $entry->show;
  $entry->set_text($cd->name);
  $entry->set_usize(400, 20);
  $table->appendRow(title => 'Album:', widget => $entry);

  my $combo = $props->{'editCategory'} = Gtk::Combo->new();
  $combo->show;
  $combo->set_popdown_strings('blues', 'classical', 'country', 'data', 'folk', 'jazz', 
			      'misc', 'newage', 'reggae', 'rock', 'soundtrack');
  if (defined($cd->category)) {
    $combo->entry->set_text($cd->category);
  }
  else {
    $combo->entry->set_text('Choose a category');
  }
  $combo->set_usize(400, 20);
  $table->appendRow(title => 'Category:', widget => $combo);

  $props->{'editRows'} = [];

  my @tracks = $cd->tracks;
  shift @tracks; pop @tracks;	# Lead & MSTN
  foreach my $track (@tracks) {
    $entry = Gtk::Entry->new(1024);
    $entry->show;
    $table->appendRow(title => $track->index.':', widget => $entry);
    $entry->set_text($track->name);
    $entry->set_usize(400, 20);
    push @{ $props->{'editRows'} }, $entry;
  }

  GtkEx::Utility::choicePopup(widget => $table, title => 'CD Properties', 
			      okFunc => sub { $browser->cdromPropertiesOk($props, $cd); });

  return 1;
}

# cdromPropertiesOk - Make the user changes to the CDs properties.

sub cdromPropertiesOk {
  my $browser = shift;
  my $props = shift;
  my $cd = shift;

  my $remotemod = 0;
    
  # Check for bad vals
  map {
    if (!defined($_) || ($_ eq '')) {
      GtkEx::Utility::ackPopup(msg => 'All values must be entered');
      return 0;
    }
    if ($_ eq 'Unknown') {
      GtkEx::Utility::ackPopup(msg => 'The special value Unknown is not allowed');
      return 0;
    }
  } ( $props->{'editArtist'}->get_text, $props->{'editAlbum'}->get_text, $props->{'editCategory'}->entry->get_text, 
      map { $_->get_text } @{ $props->{'editRows'} } );

  my $category = $props->{'editCategory'}->entry->get_text;
  if ($category eq 'Choose a category') {
    GtkEx::Utility::ackPopup(msg => 'You must choose a category');
    return 0;
  }

  my $artistname = $props->{'editArtist'}->get_text;
  my $artist = $cd->artist;
  if ($artistname ne $artist->name) {
    if ($artist->name eq 'Unknown') {
      # If the former name was 'Unknown' allow creation of a new artist.
      my $newartist = Info::Artist->find(owner => $cd->owner, name => $artistname, create => 1);
      if (!$newartist) {
	GtkEx::Utility::ackPopup(msg => Trace::lastError);
	return 0;
      }
      $cd->reparent($newartist);
      if (!$browser->gui->wait(event => sub { $cd->artist_ioid eq $newartist->ioid }, timeout => 10)) {
	GtkEx::Utility::ackPopup('System Error: Timed out waiting for reparent');
	return 0;
      }
    }
    else {
      # If the new artist name exists it is a reparent, otherwise an artist name change
      my $newartist = Info::Artist->find(owner => $cd->owner, name => $artistname, create => 0);
      if ($newartist) {
	$cd->reparent($newartist);
	if (!$browser->gui->wait(event => sub { $cd->artist_ioid eq $newartist->ioid }, timeout => 10)) {
	  GtkEx::Utility::ackPopup('System Error: Timed out waiting for reparent');
	  return 0;
	}
      }
      else {
	if (!$artist->change_name($artistname)) {
	  GtkEx::Utility::ackPopup(msg => Trace::lastError);
	  return 0;
	}
	if (!$browser->gui->wait(event => sub { $artist->name eq $artistname }, timeout => 10)) {
	  GtkEx::Utility::ackPopup('System Error: Timed out waiting for name to change');
	  return 0;
	}
      }
    }
    $remotemod = 1;
  }

  my $albumname = $props->{'editAlbum'}->get_text;
  if ($albumname ne $cd->name) {
    if (!$cd->change_name($albumname)) {
      GtkEx::Utility::ackPopup(msg => Trace::lastError);
      return 0;
    }
    if (!$browser->gui->wait(event => sub { $cd->name eq $albumname }, timeout => 10)) {
      GtkEx::Utility::ackPopup('System Error: Timed out waiting for name to change');
      return 0;
    }
    $remotemod = 1;
  }

  if ($cd->category ne $category) {
    $cd->change_category($category);
    if (!$browser->gui->wait(event => sub { $cd->category eq $category }, timeout => 10)) {
      GtkEx::Utility::ackPopup('System Error: Timed out waiting for name to change');
      return 0;
    }
    $remotemod = 1;
  }

  my $index = 1;
  foreach my $entry (@{ $props->{'editRows'} }) {
    my $name = $entry->get_text;
    my $track = $cd->track($index);
    if ($name ne $track->name) {
      if (!$track->change_name($name)) {
	GtkEx::Utility::ackPopup(msg => Trace::lastError);
	return 0;
      }
      if (!$browser->gui->wait(event => sub { $track->name eq $name }, timeout => 10)) {
	GtkEx::Utility::ackPopup('System Error: Timed out waiting for track name to change');
	return 0;
      }
      $remotemod = 1;
    }
    $index++;
  }

  if ($remotemod) {
    GtkEx::Utility::choicePopup(title => 'Remote Database Update', msg => 'Do you wish to send these changes to the remote database?', 
				okFunc => sub { if (!$cd->remoteSave) { GtkEx::Utility::ackPopup(msg => Trace::lastError, truncate => 200); } 1; });
  }

  return 1;
}

# SCRATCH

sub scratchpadAppendNodes {
  my $browser = shift;
  my $basenodes = shift;

  my @tracks = $browser->getTracks($basenodes);
  return 1 if (!@tracks);

  $browser->base->scratchpadAppend(tracks => \@tracks);
  return 1;
}

sub scratchpadInsertNodes {
  my $browser = shift;
  my $basenodes = shift;
  my $row = shift;

  my @tracks = $browser->getTracks($basenodes);
  return 1 if (!@tracks);

  $browser->base->scratchpadInsert(tracks => \@tracks, at => $row);
  return 1;
}

# LISTS

sub listsNewPopup {
  my $browser = shift;
  my $basenodes = shift;

  my $table = GtkEx::TableEdit->new(colSpacing => 5, rowSpacing => 5);
  $table->show();

  my $name = Gtk::Entry->new(1024);
  $name->show;
  $table->appendRow(title => 'Name:', widget => $name);

  GtkEx::Utility::choicePopup(widget => $table, title => 'Create a new list', 
			      okFunc => sub { $browser->listsNewOk($name, $basenodes); });

  return 1;
}

sub listsNewOk {
  my $browser = shift;
  my $name = shift;
  my $basenodes = shift;

  my $listname = $name->get_text;

  if (!defined($listname) || ($listname eq '')) {
    GtkEx::Utility::ackPopup(msg => 'Please specify a list name');
    return 1;
  }
  if (Info::List->find(owner => $browser->base->repo->what, name => $listname)) {
    GtkEx::Utility::ackPopup(msg => 'A list with this name already exists. Please use a unique list name.');
    return 1;
  }

  if (defined($basenodes) && ($#{ $basenodes } != -1)) {
    $browser->base->populateNode(group => $basenodes, 
				 recursive => 1,
				 action => sub { my @tracks = $browser->getTracks($basenodes);
						 my $track_ioids = [];
						 if (@tracks) {
						   my @tl = map { $_->ioid } @tracks;
						   $track_ioids = \@tl;
						 }
						 $browser->listsCreate($listname, $track_ioids);
						 return 1; });
  }
  else {
    $browser->listsCreate($listname, []);
  }

  return 1;
}

sub listsCreate {
  my $browser = shift;
  my $name = shift;
  my $track_ioids = shift;

  if (Info::List->create(owner => $browser->base->repo->what, name => $name, track_ioids => $track_ioids)) {
    $browser->base->refreshNode(node => $browser->base->lists);
  }
  else {
    GtkEx::Utility::ackPopup(msg => Trace::lastError);
    return 1;
  }
  return 1;
}

# MENU 

# menu

sub menu { $_[0]->{'menu'}->{'root'}; }

# menuCreate

sub menuCreate {
  my $browser = shift;    

  my $menuItem;
  my $sep;

  my @menuItems = (
		   {
		    title => 'Go',
		   },
		   {
		    title => 'Play',
		   },
		   {
		    title => 'Append',
		   },
		   {
		    title => 'Rip',
		   },
		   {
		    separator => 1,
		   },
		   {
		    title => 'Delete...',
		   },
		   {
		    title => 'Properties...',
		   },
		   {
		    separator => 1,
		   },
		   {
		    title => 'Move Up',
		   },
		   {
		    title => 'Move Down',
		   },
		   {
		    title => 'Remove',
		   },
		   {
		    title => 'Clear',
		   },
		   {
		    title => 'Refresh',
		   },
		   {
		    title => 'Select All',
		   },
		   {
		    title => 'Deselect All',
		   },
		  );

  $browser->{'menu'} = {};
  $browser->{'menu'}->{'root'} = my $menu = Gtk::Menu->new();
  $menu->border_width(1);
  foreach my $item (@menuItems) {
    if (defined($item->{'title'})) {
      $menuItem = Gtk::MenuItem->new($item->{'title'});
    }
    else {
      $menuItem = Gtk::MenuItem->new();
    }
    if ($item->{'separator'}) {
      $sep = Gtk::HSeparator->new;
      $sep->show;
      $menuItem->add($sep);
      $menuItem->set_sensitive(0);
    }
    $menuItem->show();
    $menu->append($menuItem);
    if ($item->{'title'}) {
      $browser->{'menu'}->{ $item->{'title'} } = $menuItem;
    }
  }
  return 1;
}

# menuDraw - Build the menu options depending on where it is displayed
# (tree or display) and what is displayed, or selected.

sub menuDraw {
  my $browser = shift;
  my $source = shift;

  my $list = $browser->{'display'}->{'list'};
  my $tree = $browser->{'tree'};
  my $ld = $list->get_user_data;

  my $sid;
  my $menuItem;

  my @basenodes;
  my $parent_bn;
  my $treenode;
  if ($source eq 'display') {
    @basenodes = $browser->displaySelection;
    $parent_bn = $ld->{'parent_bn'};
  }
  elsif ($source eq 'tree') {
    $treenode = $tree->selection;
    if (defined($treenode)) {
      @basenodes = ( $tree->node_get_row_data($treenode)->{'basenode'} );
    }
  }

  # Some frequent checks

  my $displayAndCdromWithCd = sub
    {
      return (($source eq 'display') && $parent_bn && ($parent_bn->type eq Gui::Base::Cdrom) && 
	      $parent_bn->status && ($parent_bn->status->{'state'} ne Cdrom::NoCd));
    };
  $displayAndCdromWithCd = &$displayAndCdromWithCd;

  my $displayIs = sub
    {
      return (defined($ld->{'parent_bn'}) && ($ld->{'parent_bn'}->type eq $_[0]));
    };

  my $selectionAreCdromsWithCd = sub 
    {
      return 0 if (!@basenodes);
      foreach my $basenode (@basenodes) {
	return 0 if ($basenode->type ne Gui::Base::Cdrom);
	return 0 if (!defined($basenode->status) || ($basenode->status->{'state'} eq Cdrom::NoCd));
      }
      return 1;
    };
  $selectionAreCdromsWithCd = &$selectionAreCdromsWithCd;

  my $selectionAreArtistAlbumTrack = sub 
    {
      return 0 if (!@basenodes);
      foreach my $basenode (@basenodes) {
	return 0 if (($basenode->type ne Gui::Base::Artist) && ($basenode->type ne Gui::Base::Album) && ($basenode->type ne Gui::Base::Track));
      }
      return 1;
    };
  $selectionAreArtistAlbumTrack = &$selectionAreArtistAlbumTrack;

  my $selectionAreList = sub 
    {
      return 0 if (!@basenodes);
      foreach my $basenode (@basenodes) {
	return 0 if ($basenode->type ne Gui::Base::List);
      }
      return 1;
    };
  $selectionAreList = &$selectionAreList;

  # Disconnect all previous signal handlers

  if (defined($browser->{'menu'}->{'sids'}) && @{ $browser->{'menu'}->{'sids'} }) {
    map { $_->{'object'}->signal_disconnect($_->{'sid'}) } @{ $browser->{'menu'}->{'sids'} };
  }
  my @sids = ();

  # Go, build a submenu from the history
 
  my $go = $browser->{'menu'}->{'Go'};
  $go->set_sensitive(0);
  $go->remove_submenu;
 
  if (@{ $browser->{'display'}->{'history'} }) {
    my $goSubMenu = $browser->menuCreateGoSubmenu;
    $go->set_submenu($goSubMenu);
    $go->set_sensitive(1);
  }
 
  # Play - Means play selection recursively on a choosen player.

  my $play = $browser->{'menu'}->{'Play'};
  $play->set_sensitive(0);
  $play->remove_submenu;
  my @players = $browser->base->players;

  if (!(# Not a case of...
	# Tree source and nothing selected
	(($source eq 'tree') && !@basenodes) ||
	# No players
	!@players
       )) {

    # In the display window nothing selected translates to everything

    my @all = @basenodes;
    if (($source eq 'display') && !@all) {
      @all = @{ $ld->{'basenodes'} };
    }

    # In the tree window the play means everything under this node
    # but using the sort of the display, so we can use the current
    # display if it displays what is selected in the tree.

    if (($source eq 'tree') && defined($ld->{'parent_bn'}) && ($ld->{'parent_bn'} == $basenodes[0])) {
      @all = @{ $ld->{'basenodes'} };
    }

    if (@all &&
	# No Cdroms or Players selected
	!(grep { ($_->type eq Gui::Base::Cdrom) || ($_->type eq Gui::Base::Player) } @all)) {

      if ($#players == 0) {
	$sid = $play->signal_connect('activate', sub {
				       $browser->base->populateNode(group => \@all, 
								    recursive => 1,
								    action => sub { $browser->playerPlayNodes($players[0], \@all); } ); });
	push @sids, { object => $play, sid => $sid };
      }
      else {
	my $playSubMenu = Gtk::Menu->new();
	foreach my $player (@players) {
	  $menuItem = Gtk::MenuItem->new(ucfirst $player->what->key);
	  $menuItem->signal_connect('activate', sub {
				      $browser->base->populateNode(group => \@all, 
								   recursive => 1,
								   action => sub { $browser->playerPlayNodes($player, \@all); } ); });
	  $menuItem->show();
	  $playSubMenu->append($menuItem);
	}
	$play->set_submenu($playSubMenu);
      }
      $play->set_sensitive(1);
    }
    elsif (@all &&
	   # Just Players selected
	   !(grep { ($_->type ne Gui::Base::Player) } @all)) {

      # Want to allow (hopefully) simultaneous play.
      $sid = $play->signal_connect('activate', sub { map { $browser->gui->message(name => Player::Play, to => $_->what); } @players });
      push @sids, { object => $play, sid => $sid };
      $play->set_sensitive(1);
    }
  }

  # Append - Means append selection recursively to a choosen player or scratchpad.

  my $append = $browser->{'menu'}->{'Append'};
  $append->set_sensitive(0);
  $append->remove_submenu;

  if (!(# Not a case of...
	# Tree source and nothing selected
	(($source eq 'tree') && !@basenodes)
       )) {

    # In the display window nothing selected translates to everything

    my @all = @basenodes;
    if (($source eq 'display') && !@all) {
      @all = @{ $ld->{'basenodes'} };
    }

    # In the tree window the append means everything under this node
    # but using the sort of the display, so we can use the current
    # display if it displays what is selected in the tree.

    if (($source eq 'tree') && defined($ld->{'parent_bn'}) && ($ld->{'parent_bn'} == $basenodes[0])) {
      @all = @{ $ld->{'basenodes'} };
    }

    if (@all &&
	# No Cdroms selected
	!(grep { $_->type eq Gui::Base::Cdrom } @all)) {

      if ($#players == -1) {
	$sid = $append->signal_connect('activate', sub {
					 $browser->base->populateNode(group => \@all, 
								      recursive => 1,
								      action => sub { $browser->scratchpadAppendNodes(\@all); } ); });
	push @sids, { object => $append, sid => $sid };
      }
      else {
	my $appendSubMenu = Gtk::Menu->new();
	$menuItem = Gtk::MenuItem->new('Scratchpad');
	$menuItem->signal_connect('activate', sub {
				    $browser->base->populateNode(group => \@all, 
								 recursive => 1,
								 action => sub { $browser->scratchpadAppendNodes(\@all); } ); });
	$menuItem->show();
	$appendSubMenu->append($menuItem);
	foreach my $player (@players) {
	  $menuItem = Gtk::MenuItem->new(ucfirst $player->what->key);
	  $menuItem->signal_connect('activate', sub {
				      $browser->base->populateNode(group => \@all, 
								   recursive => 1,
								   action => sub { $browser->playerAppendNodes($player, \@all); } ); });
	  $menuItem->show();
	  $appendSubMenu->append($menuItem);
	}
	$append->set_submenu($appendSubMenu);
      }
      $append->set_sensitive(1);
    }
  }

  # Rip

  my $rip = $browser->{'menu'}->{'Rip'};
  if ($displayAndCdromWithCd || $selectionAreCdromsWithCd) {
    if ($displayAndCdromWithCd) {
      $sid = $rip->signal_connect('activate', sub { $browser->cdromRip($parent_bn) });
    }
    else {
      $sid = $rip->signal_connect('activate', sub { map { $browser->cdromRip($_) } @basenodes });
    }
    push @sids, { object => $rip, sid => $sid };
    $rip->set_sensitive(1);
  }
  else {
    $rip->set_sensitive(0);
  }

  # Delete... 

  my $delete = $browser->{'menu'}->{'Delete...'};
  if (($selectionAreArtistAlbumTrack && ($basenodes[0]->parent->type ne Gui::Base::List)) || $selectionAreList) {
    $sid = $delete->signal_connect('activate', sub { $browser->deletePopup(@basenodes) });
    push @sids, { object => $delete, sid => $sid };
    $delete->set_sensitive(1);
  }
  else {
    $delete->set_sensitive(0);
  }

  # Properties...

  my $prop = $browser->{'menu'}->{'Properties...'};
  $prop->set_sensitive(0);
  if ($displayAndCdromWithCd || $selectionAreCdromsWithCd || $selectionAreArtistAlbumTrack || $selectionAreList) {
    if ($displayAndCdromWithCd) {
      $sid = $prop->signal_connect('activate', sub { $browser->cdromPropertyPopup($parent_bn); });
    }
    elsif ($selectionAreCdromsWithCd) {
      $sid = $prop->signal_connect('activate', sub { map { $browser->cdromPropertyPopup($_) } @basenodes });
    }
    else {
      $sid = $prop->signal_connect('activate', sub { map { $browser->propertyPopup($_); } @basenodes });
    }
    push @sids, { object => $prop, sid => $sid };
    $prop->set_sensitive(1);
  }

  # Move up

  my $moveUp = $browser->{'menu'}->{'Move Up'};
  if ($source eq 'display') {
    $moveUp->show;
    if (@basenodes && $selectionAreArtistAlbumTrack && !&$displayIs(Gui::Base::Player) && !&$displayIs(Gui::Base::Cdrom)) {
      my $sid = $moveUp->signal_connect('activate', sub { $browser->displayMoveUp; } );
      push @sids, { object => $moveUp, sid => $sid };
      $moveUp->set_sensitive(1);
    }
    else {
      $moveUp->set_sensitive(0);
    }
  }
  else {
    $moveUp->hide;
    $moveUp->set_sensitive(0);
  }

  # Move down

  my $moveDown = $browser->{'menu'}->{'Move Down'};
  if ($source eq 'display') {
    $moveDown->show;
    if (@basenodes && $selectionAreArtistAlbumTrack && !&$displayIs(Gui::Base::Player) && !&$displayIs(Gui::Base::Cdrom)) {
      my $sid = $moveDown->signal_connect('activate', sub { $browser->displayMoveDown; } );
      push @sids, { object => $moveDown, sid => $sid };
      $moveDown->set_sensitive(1);
    }
    else {
      $moveDown->set_sensitive(0);
    }
  }
  else {
    $moveDown->hide;
    $moveDown->set_sensitive(0);
  }

  # Remove - Can be done to anything selected in the display of a scratchpad or list

  my $remove = $browser->{'menu'}->{'Remove'};
  if ($source eq 'display') {
    $remove->show;
    if (@basenodes && (($parent_bn->type eq Gui::Base::Scratchpad) || ($parent_bn->type eq Gui::Base::List))) {
      $sid = $remove->signal_connect('activate', sub { $browser->displayRemove; } );
      push @sids, { object => $remove, sid => $sid };
      $remove->set_sensitive(1);
    }
    else {
      $remove->set_sensitive(0);
    }
  }
  else {
    $remove->hide;
    $remove->set_sensitive(0);
  }

  # Clear - Can be done to the display of a scratchpad or a player

  my $clear = $browser->{'menu'}->{'Clear'};
  if ($source eq 'display') {
    $clear->show;
    if (($parent_bn->type eq Gui::Base::Player) || ($parent_bn->type eq Gui::Base::Scratchpad)) {
      $sid = $clear->signal_connect('activate', sub { $browser->displayClear; } );
      push @sids, { object => $clear, sid => $sid };
      $clear->set_sensitive(1);
    }
    else {
      $clear->set_sensitive(0);
    }
  }
  else {
    $clear->hide;
    $clear->set_sensitive(0);
  }

  # Refresh

  my $refresh = $browser->{'menu'}->{'Refresh'};
  my $what;
  if ($source eq 'display') {
    $what = $parent_bn;
  }
  elsif (($source eq 'tree') && @basenodes) {
    $what = $basenodes[0];
  }
  if ($what && $what->refresh) {
    $sid = $refresh->signal_connect('activate', sub { $browser->base->refreshNode(node => $what); } );
    push @sids, { object => $refresh, sid => $sid };
    $refresh->set_sensitive(1);
  }
  else {
    $refresh->set_sensitive(0);
  }

  # Select/Deselect All

  my $selectAll = $browser->{'menu'}->{'Select All'};
  my $deselectAll = $browser->{'menu'}->{'Deselect All'};
  if ($source eq 'display') {
    $selectAll->show;
    $deselectAll->show;
    if (($parent_bn->type ne Gui::Base::Player) || ($parent_bn->type ne Gui::Base::Cdrom)) {
      $sid = $selectAll->signal_connect('activate', sub { $list->select_all; } );
      push @sids, { object => $selectAll, sid => $sid };
      $selectAll->set_sensitive(1);
      $sid = $deselectAll->signal_connect('activate', sub { $list->unselect_all; } );
      push @sids, { object => $deselectAll, sid => $sid };
      $deselectAll->set_sensitive(1);
    }
    else {
      $selectAll->set_sensitive(0);
      $deselectAll->set_sensitive(0);
    }
  }
  else {
    $selectAll->hide;
    $selectAll->set_sensitive(0);
    $deselectAll->hide;
    $deselectAll->set_sensitive(0);
  }

  $browser->{'menu'}->{'sids'} = \@sids;
  return 1;
}

# menuCreateGoSubmenu

sub menuCreateGoSubmenu {
  my $browser = shift;
  my $goSubMenu;
  if ($#_ == 0) {
    $goSubMenu = shift;
  }
  else {
    $goSubMenu = Gtk::Menu->new();
  }

  my $current_bn = $browser->{'display'}->{'list'}->get_user_data->{'parent_bn'};
  my $up_bn = $current_bn->parent;

  # Top
  my $menuItem = Gtk::MenuItem->new('Top');
  $menuItem->signal_connect('activate', sub { $browser->treeSelect($browser->base->rootNode); } );
  $menuItem->show();
  $goSubMenu->append($menuItem);

  # Up
  if (defined($up_bn) && ($up_bn != $browser->base->rootNode)) {
    $menuItem = Gtk::MenuItem->new('Up - '.$browser->treeLabelNode($up_bn));
    $menuItem->signal_connect('activate', sub { $browser->treeSelect($up_bn); } );
  }
  else {
    $menuItem = Gtk::MenuItem->new('Up');
    $menuItem->set_sensitive(0);
  }
  $menuItem->show();
  $goSubMenu->append($menuItem);

  $menuItem = Gtk::MenuItem->new;
  my $sep = Gtk::HSeparator->new;
  $sep->show;
  $menuItem->add($sep);
  $menuItem->set_sensitive(0);
  $menuItem->show();
  $goSubMenu->append($menuItem);

  foreach my $hist (@{ $browser->{'display'}->{'history'} }) {
    next if ($hist->{'basenode'} == $browser->base->rootNode);
    my $menuItem = Gtk::MenuItem->new($browser->treeLabelNode($hist->{'basenode'}));
    $menuItem->signal_connect('activate', sub { $browser->treeSelect($hist->{'basenode'}); } );
    $menuItem->show();
    $goSubMenu->append($menuItem);
  }
  return $goSubMenu;
}

# menuCreateListsSubmenu

sub menuCreateListsSubmenu {
  my $browser = shift;
  my $listsSubMenu;
  if ($#_ == 0) {
    $listsSubMenu = shift;
  }
  else {
    $listsSubMenu = Gtk::Menu->new();
  }

  my $list = $browser->{'display'}->{'list'};
  my @basenodes = $browser->displaySelection;
  my $parent_bn = $list->get_user_data->{'parent_bn'};

  # New from selection
  my $menuItem = Gtk::MenuItem->new('Create List From Selection...');
  $menuItem->signal_connect('activate', sub { $browser->listsNewPopup(\@basenodes); } );
  $menuItem->show();
  $menuItem->set_sensitive(0) if (!Info::Persist::configured);
  $listsSubMenu->append($menuItem);

  # New empty list
  $menuItem = Gtk::MenuItem->new('Create Empty List...');
  $menuItem->signal_connect('activate', sub { $browser->listsNewPopup([]); } );
  $menuItem->show();
  $menuItem->set_sensitive(0) if (!Info::Persist::configured);
  $listsSubMenu->append($menuItem);

  return $listsSubMenu;
}

# BUTTONS

sub buttonsCreate {
  my $browser = shift;

  my @playerButtonDescs =
    (
     { label => 'Play', accel => '_   ',  signals => [ [ 'clicked', sub { $browser->playerPlay } ] ] },
     { label => 'Pause', hide => 1, accel => '_   ', signals => [ [ 'clicked', sub { $browser->playerPause } ] ] },
     { label => 'Stop', accel => '_   ',  signals => [ [ 'clicked', sub { $browser->playerStop } ] ] },
     { label => 'Next', signals => [ [ 'clicked', sub { $browser->playerNext } ] ]},
     { label => 'Prev', signals => [ [ 'clicked', sub { $browser->playerPrev } ] ]},

     { label => 'Ffwd', 
       signals => [ [ 'pressed', sub { $browser->playerSearch({ search => 1 }) } ],
		    [ 'released', sub { $browser->playerSearch({ search => 0 }) } ], ]},
     { label => 'Rwnd', 
       signals => [ [ 'pressed', sub { $browser->playerSearch({ search => 1, direction => Player::Reverse }) } ],
		    [ 'released', sub { $browser->playerSearch({ search => 0 }) } ], ]},

     { label => 'Clear', signals => [ [ 'clicked', sub { $browser->playerClear } ] ]},
    );

  my @cdromButtonDescs =
    (
     { label => 'Play', accel => '_   ', signals => [ [ 'clicked', sub { $browser->cdromPlay } ] ] },
     { label => 'Pause', hide => 1, accel => '_   ', signals => [ [ 'clicked', sub { $browser->cdromPause } ] ] },
     { label => 'Stop', accel => '_   ', signals => [ [ 'clicked', sub { $browser->cdromStop } ] ] },
     { label => 'Next', signals => [ [ 'clicked', sub { $browser->cdromNext } ] ] },
     { label => 'Prev', signals => [ [ 'clicked', sub { $browser->cdromPrev } ] ] },
     { label => 'Eject', accel => '_    ', signals => [ [ 'clicked', sub { $browser->cdromEject } ] ] },
     { label => 'Close', hide => 1, accel => '_    ', signals => [ [ 'clicked', sub { $browser->cdromClose } ] ] },
     { label => 'Rip', signals => [ [ 'clicked', sub { $browser->cdromRip } ] ] },
    );

  my @otherButtonDescs =
    (
     { label => 'Play', accel => '_   ', signals => [] },
     { label => 'Stop', accel => '_   ', signals => [] },
    );

  my $box = $browser->{'display'}->{'buttons'}->{'box'};
  my $modes = $browser->{'display'}->{'buttons'}->{'modes'} = {};

  foreach my $modeDesc ([ Gui::Base::Player, \@playerButtonDescs ], 
			[ Gui::Base::Cdrom, \@cdromButtonDescs ], 
			[ 'OTHER', \@otherButtonDescs ]) {

    my $mode = $modes->{ $modeDesc->[0] } = {};

    $mode->{'box'} = my $buttonBox = Gtk::HButtonBox->new();
    $buttonBox->hide;
    $buttonBox->set_layout($modeDesc->[0] eq 'OTHER' ? 'spread' : 'edge');
    $buttonBox->set_spacing($browser->gui->spacing);
    $buttonBox->set_child_ipadding(15, 0);
    $buttonBox->set_child_size(0, 0);
    $box->pack_start($buttonBox, 0, 0, 0);

    my $buttons = $mode->{'buttons'} = {};

    foreach my $desc (@{ $modeDesc->[1] }) {

      $buttons->{ $desc->{'label'} } = my $button = GtkEx::Button->new($desc->{'label'});

      if (defined($desc->{'hide'}) && $desc->{'hide'}) {
	$button->hide;
      }
      else {
	$button->show;
      }
      if (defined($desc->{'sensitive'}) && !$desc->{'sensitive'}) {
	$button->set_sensitive(0);
      }
      else {
	$button->set_sensitive(1);
      }
      $button->label->set_justify('left');
      if (defined($desc->{'accel'})) {
	$button->label->set_pattern($desc->{'accel'});
      }
      $buttonBox->add($button);

      foreach my $signal (@{ $desc->{'signals'} }) {
	$button->signal_connect(@{ $signal });
      }
    }
  }

  return 1;
}

# buttonsDraw - Activate the buttons depending on the current display list.

sub buttonsDraw {
  my $browser = shift;

  # Node addition/deletion/update can cause a new meaning for the
  # buttons. However as these events can come in batches
  # (e.g. artist list) this should be done in a delayed manner to
  # avoid making changes for a transient state. Similiar to
  # _freezeThaw method.

  if (defined($browser->{'buttonsDrawPending'})) {
    Gtk->timeout_remove($browser->{'buttonsDrawPending'});
  }
  $browser->{'buttonsDrawPending'} = Gtk->timeout_add(100, sub { delete $browser->{'buttonsDrawPending'}; 
								 $browser->_buttonsDraw; return 0; });
  return 1;
}

sub _buttonsDraw {
  my $browser = shift;
  my $list = $browser->{'display'}->{'list'};
  my $parent_bn = $list->get_user_data->{'parent_bn'};

  my $modes = $browser->{'display'}->{'buttons'}->{'modes'};

  # Hide others

  if ($parent_bn->type eq Gui::Base::Player) {
    $modes->{ Gui::Base::Player }->{'box'}->show;
    $modes->{ Gui::Base::Cdrom }->{'box'}->hide;
    $modes->{ 'OTHER' }->{'box'}->hide;
  }
  elsif ($parent_bn->type eq Gui::Base::Cdrom) {
    $modes->{ Gui::Base::Player }->{'box'}->hide;
    $modes->{ Gui::Base::Cdrom }->{'box'}->show;
    $modes->{ 'OTHER' }->{'box'}->hide;
  }
  else {
    $modes->{ Gui::Base::Player }->{'box'}->hide;
    $modes->{ Gui::Base::Cdrom }->{'box'}->hide;
    $modes->{ 'OTHER' }->{'box'}->show;
  }

  # Type specifics

  if ($parent_bn->type eq Gui::Base::Player) {

    $browser->{'display'}->{'buttons'}->{'activePlayer'} = $parent_bn;
    my $buttons = $modes->{ Gui::Base::Player }->{'buttons'};

    if (defined($parent_bn->status) && defined($parent_bn->status->{'state'})) {

      if (defined($parent_bn->spool) && @{ $parent_bn->spool }) {

	$buttons->{'Play'}->set_sensitive(1);
	$buttons->{'Pause'}->set_sensitive(1);
	$buttons->{'Next'}->set_sensitive(1);
	$buttons->{'Prev'}->set_sensitive(1);
	$buttons->{'Clear'}->set_sensitive(1);

	if ($parent_bn->status->{'state'} eq Player::Playing) {
	  $buttons->{'Play'}->hide;
	  $buttons->{'Pause'}->show;
	  $buttons->{'Stop'}->set_sensitive(1);
	  $buttons->{'Ffwd'}->set_sensitive(1);
	  $buttons->{'Rwnd'}->set_sensitive(1);
	}
	else {
	  $buttons->{'Play'}->show;
	  $buttons->{'Pause'}->hide;
	  $buttons->{'Stop'}->set_sensitive(0);
	  $buttons->{'Ffwd'}->set_sensitive(0);
	  $buttons->{'Rwnd'}->set_sensitive(0);
	}
      }
      else {

	# Empty spool
	$buttons->{'Play'}->set_sensitive(0);
	$buttons->{'Pause'}->set_sensitive(0);
	$buttons->{'Next'}->set_sensitive(0);
	$buttons->{'Prev'}->set_sensitive(0);
	$buttons->{'Clear'}->set_sensitive(0);
	$buttons->{'Stop'}->set_sensitive(0);
	$buttons->{'Ffwd'}->set_sensitive(0);
	$buttons->{'Rwnd'}->set_sensitive(0);
      }

    }
  }
  elsif ($parent_bn->type eq Gui::Base::Cdrom) {

    my $buttons = $modes->{ Gui::Base::Cdrom }->{'buttons'};

    if (defined($parent_bn->status) && defined($parent_bn->status->{'state'})) {

      if ($parent_bn->status->{'state'} ne Cdrom::NoCd) {

	$buttons->{'Play'}->set_sensitive(1);
	$buttons->{'Pause'}->set_sensitive(1);
	$buttons->{'Next'}->set_sensitive(1);
	$buttons->{'Prev'}->set_sensitive(1);

	if ($parent_bn->status->{'state'} eq Cdrom::Playing) {
	  $buttons->{'Play'}->hide;
	  $buttons->{'Pause'}->show;
	  $buttons->{'Stop'}->set_sensitive(1);
	  $buttons->{'Rip'}->set_sensitive(1);
	}
	elsif ($parent_bn->status->{'state'} eq Cdrom::Ripping) {
	  $buttons->{'Play'}->show;
	  $buttons->{'Pause'}->hide;
	  $buttons->{'Stop'}->set_sensitive(1);
	  $buttons->{'Rip'}->set_sensitive(0);
	}
	else {
	  $buttons->{'Play'}->show;
	  $buttons->{'Pause'}->hide;
	  $buttons->{'Stop'}->set_sensitive(0);
	  $buttons->{'Rip'}->set_sensitive(1);
	}

	$buttons->{'Close'}->hide;
	$buttons->{'Eject'}->show;
      }
      else {

	# No CD

	$buttons->{'Play'}->set_sensitive(0);
	$buttons->{'Pause'}->set_sensitive(0);
	$buttons->{'Rip'}->set_sensitive(0);
	$buttons->{'Next'}->set_sensitive(0);
	$buttons->{'Prev'}->set_sensitive(0);

	if (defined($parent_bn->status->{'attributes'}->{'tray'}) && 
	    ($parent_bn->status->{'attributes'}->{'tray'} eq Cdrom::TrayOpen)) {
	  $buttons->{'Close'}->show;
	  $buttons->{'Eject'}->hide;
	}
	else {
	  $buttons->{'Close'}->hide;
	  $buttons->{'Eject'}->show;
	}
      }
    }
  }
  else {

    my @basenodes = $browser->displaySelection;
    my @players = $browser->base->players;

    my $buttons = $modes->{ 'OTHER' }->{'buttons'};

    # Similiar to menuDraw choices

    $buttons->{'Play'}->set_sensitive(0);
    $buttons->{'Stop'}->set_sensitive(0);

    if (@players) {

      if (defined($buttons->{'sids'})) {
	map { $_->{'object'}->signal_disconnect($_->{'sid'}) } @{ $buttons->{'sids'} };
      }
      my @sids;
      my $sid;

      if (!defined($browser->{'display'}->{'buttons'}->{'activePlayer'})) {
	$browser->{'display'}->{'buttons'}->{'activePlayer'} = $players[0];
      }
      my $player = $browser->{'display'}->{'buttons'}->{'activePlayer'};

      if (!@basenodes) {
	@basenodes = @{ $list->get_user_data->{'basenodes'} };
      }

      my $playerOrCdromInList = 0;
      foreach my $bn (@basenodes) {
	my $type = $bn->type;
	if (($type eq Gui::Base::Cdrom) || ($type eq Gui::Base::Player)) {
	  $playerOrCdromInList = 1;
	  last;
	}
      }
      if (!$playerOrCdromInList) {
	$buttons->{'Play'}->set_sensitive(1);
	$sid = $buttons->{'Play'}->signal_connect('clicked', sub { 
		    $browser->base->populateNode(group => \@basenodes, 
						 recursive => 1,
						 action => sub { $browser->playerPlayNodes($player, \@basenodes); } ); });
	push @sids, { object => $buttons->{'Play'}, sid => $sid };

      }

      if (defined($player->status) && ($player->status->{'state'} eq Player::Playing)) {
	$buttons->{'Stop'}->set_sensitive(1);
	$sid = $buttons->{'Stop'}->signal_connect('clicked', sub { 
						    $browser->gui->message(name => Player::Stop, to => $player->what); });
	push @sids, { object => $buttons->{'Stop'}, sid => $sid };
      }

      $buttons->{'sids'} = \@sids;
    }
  }

  return 1;
}

# INFO BAR

# infoDraw - Change the info label depending on the currently displayed list.

sub infoDraw {
  my $browser = shift;
  my $list = $browser->{'display'}->{'list'};
  my $ld = $list->get_user_data;
  my $parent_bn = $ld->{'parent_bn'};
  my $label = $browser->{'display'}->{'infoLabel'};

  if ((($parent_bn->type eq Gui::Base::Cdrom) || ($parent_bn->type eq Gui::Base::Player)) &&
      ($parent_bn->what->currentState eq Thing::Handle::Dead)) {
    $label->set_text(ucfirst $parent_bn->what->key.' is dead');
    $label->show;
    return;
  }

  if (($parent_bn->type ne Gui::Base::Cdrom) || !defined($parent_bn->status)) {
    $label->hide;
    return;
  }

  my $info;
  if (defined($parent_bn->status->{'attributes'}->{'configureError'})) {
    $info = $parent_bn->status->{'attributes'}->{'configureError'};
  }
  elsif (defined($parent_bn->cd)) {
    $info = $parent_bn->cd->artist->name.' - '.$parent_bn->cd->name;
    $info .= ' [';
    if ($parent_bn->status->{'state'} eq Cdrom::Ripping) {
      $info .= 'Ripping';
    }
    elsif ($parent_bn->status->{'state'} eq Cdrom::Stopped) {
      $info .= 'Stopped';
    }
    elsif ($parent_bn->status->{'state'} eq Cdrom::Paused) {
      $info .= 'Paused';
    }
    elsif ($parent_bn->status->{'state'} eq Cdrom::Playing) {
      $info .= 'Playing';
    }
    if (defined($parent_bn->ripStatus) && ($parent_bn->ripStatus->{'state'} ne Controller::Halted) &&
	defined($parent_bn->ripStatus->{'reason'}) && ($parent_bn->ripStatus->{'reason'} ne '')) {
      $info .=': '.$parent_bn->ripStatus->{'reason'};
    }
    $info .= ']';
  }
  elsif (defined($parent_bn->status->{'attributes'}->{'tray'}) && ($parent_bn->status->{'attributes'}->{'tray'} eq Cdrom::TrayClosed)){
    $info = 'No CD in drive';
  }
  else {
    $info = 'Drive open';
  }

  $label->set_text($info);
  $label->show;
  return 1;
}

# PROPERTIES

# propertyPopup() 

sub propertyPopup {
  my $browser = shift;
  my $basenode = shift;

  my $title = 'Edit Properties';

  my $vbox = Gtk::VBox->new(0, 0);
  $vbox->border_width(0);
  $vbox->show();

  # Top label

  my $label = Gtk::Label->new($browser->propertyTitle($basenode));
  $label->set_justify('left');
  $label->set_line_wrap(1);
  $label->show();
  $vbox->pack_start($label, 0, 0, 0);
  my $hsep = Gtk::HSeparator->new;
  $hsep->show;
  $vbox->pack_start($hsep, 0, 0, 5);

  # Middle fields

  my $fbox = Gtk::VBox->new(0, 0);
  $fbox->border_width(0);
  $fbox->show();
  $vbox->pack_start($fbox, 1, 1, 0);

  my $fields = [];
  my @labels = $browser->propertyLabels($basenode);
  my @values = $browser->propertyValues($basenode);
  my @edit = $browser->propertyEditables($basenode);

  my $table = GtkEx::TableEdit->new(rowSpacing => 5, colSpacing => 5);
  for (my $i = 0; $i <= $#labels; $i++) {
    my $widget;
    if ($edit[$i]) {
      $widget = Gtk::Entry->new(1024);
      $widget->set_text($values[$i]);
      $widget->show;
    }
    else {
      $widget = Gtk::Label->new($values[$i]);
      $widget->set_justify('left');
      $widget->set_line_wrap(1);
      $widget->show;
    }
    $table->appendRow(title => $labels[$i], widget => $widget);
    push @{ $fields }, $widget;
  }
  $table->show;
  $fbox->pack_start($table, 0, 0, 0);

  GtkEx::Utility::choicePopup(widget => $vbox, title => $title, okFunc => sub { $browser->propertyOk($basenode, @{ $fields }); });
  return 1;
}

sub propertyLabels {
  my $browser = shift;
  my $basenode = shift;

  if ($basenode->type eq Gui::Base::Artist) {
    return ('Artist');
  }
  elsif ($basenode->type eq Gui::Base::Album) {
    return ('Album');
  }
  elsif ($basenode->type eq Gui::Base::Track) {
    return ('Index', 'Duration', 'Track');
  }
  elsif ($basenode->type eq Gui::Base::List) {
    return ('Name');
  }
}

sub propertyValues {
  my $browser = shift;
  my $basenode = shift;

  if (($basenode->type eq Gui::Base::Artist) || ($basenode->type eq Gui::Base::Album) || ($basenode->type eq Gui::Base::List)) {
    return ($basenode->what->name);
  }
  elsif ($basenode->type eq Gui::Base::Track) {
    my $duration = sprintf('%d:%02d', int($basenode->what->duration / 60), int($basenode->what->duration % 60));
    return ($basenode->what->index, $duration, $basenode->what->name);
  }
}

sub propertyEditables {
  my $browser = shift;
  my $basenode = shift;

  if (($basenode->type eq Gui::Base::Artist) || ($basenode->type eq Gui::Base::Album) || ($basenode->type eq Gui::Base::List)) {
    return (1);
  }
  elsif ($basenode->type eq Gui::Base::Track) {
    return (0, 0, 1);
  }
}

sub propertyTitle {
  my $browser = shift;
  my $basenode = shift;

  return $basenode->what->name;
}

sub propertyOk {
  my $browser = shift;
  my $basenode = shift;
  my @widgets = @_;

  my $name;
  if (($basenode->type eq Gui::Base::Artist) || ($basenode->type eq Gui::Base::Album) || ($basenode->type eq Gui::Base::List)) {
    $name = $widgets[0]->get_text;
  }
  elsif ($basenode->type eq Gui::Base::Track) {
    $name = $widgets[2]->get_text;
  }

  if ($name ne $basenode->what->name) {
    if (!$basenode->what->change_name($name)) {
      GtkEx::Utility::ackPopup(msg => Trace::lastError);
      return 0;
    }
  }
  return 1;
}

# DELETE

# deletePopup() 

sub deletePopup {
  my $browser = shift;
  my @basenodes = @_;

  my @objects = map { $_->what } @basenodes;

  my $msg = 'Are you sure you want to delete ';
  if ($#objects == 0) {
    if ($objects[0]->isa('Info::Artist')) {
      $msg .= 'the artist '.$objects[0]->name.' and all of its albums and tracks';
    }
    elsif ($objects[0]->isa('Info::Album')) {
      $msg .= 'the album '.$objects[0]->name.' and all of its tracks';
    }
    elsif ($objects[0]->isa('Info::Track')) {
      $msg .= 'the track '.$objects[0]->name;
    }
    elsif ($objects[0]->isa('Info::List')) {
      $msg .= 'the list '.$objects[0]->name;
    }
    else {
      die;
    }
  }
  elsif ($#objects > 0) {
    # Homogenous objects assumption
    if ($objects[0]->isa('Info::Artist')) {
      $msg .= 'the artists '.join(', ', map { $_->name } @objects).' and all of their albums and tracks';
    }
    elsif ($objects[0]->isa('Info::Album')) {
      $msg .= 'the albums '.join(', ', map { $_->name } @objects).' and all of their tracks';
    }
    elsif ($objects[0]->isa('Info::Track')) {
      $msg .= 'the tracks '.join(', ', map { $_->name } @objects);
    }
    elsif ($objects[0]->isa('Info::List')) {
      $msg .= 'the lists '.join(', ', map { $_->name } @objects);
    }
    else {
      die;
    }
  }
  else {
    return 1;
  }
  $msg .= '?';

  my $okFunc;
  if ($objects[0]->isa('Info::Artist')) {
    $okFunc = sub { map { $_->delete } @objects; $browser->base->refreshNode(node => $browser->base->repo); 1; };
  }
  elsif ($objects[0]->isa('Info::List')) {
    $okFunc = sub { map { $_->delete } @objects; $browser->base->refreshNode(node => $browser->base->lists); 1; };
  }
  else {
    $okFunc = sub { map { $_->delete } @objects; 1; };
  }

  GtkEx::Utility::choicePopup(msg => $msg, okFunc => $okFunc);

  return 1;
}

# FREEZE-THAW - The tree and lists are typically batch
# updated/created/deleted. Without the use of freeze/thaw this causes
# flickering and screen lock. This can be avoided with a delayed freeze
# thaw action.

sub _freezeThaw {
  my $object = shift;

  if (defined($object->{'_freezeThaw'})) {
    Gtk->timeout_remove($object->{'_freezeThaw'}->{'timeout'});
    delete $object->{'_freezeThaw'};
  }
  else {
    $object->freeze;
  }
  $object->{'_freezeThaw'} = {};
  $object->{'_freezeThaw'}->{'timeout'} = Gtk->timeout_add(250, sub { delete $object->{'_freezeThaw'}; $object->thaw; return 0; });
  return 1;
}

1;
