# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.5 $

package Gui::Page::Config;

# Configuration.

# Tool to create emptytree seedy configuration. Exists as a Gui
# page. Takes care of starting newly configured things when service is
# running in trusted mode.

# Maintains the following:

# page->

#   hosts - A hash array keyed by hostname of all the hosts
#   configured. 

# page->{ hosts }->{ <name> }->

#   name - the hostname

#   page - a hash array of the various widgets that make up the
#   notebook page for this host.

#   things - a hash array keyed by thing name of all the things
#   configured for this host.

# page->{ hosts }->{ <name> }->{ things }->{ <name> }->

#   handle - Thing:Handle for the thing.

#   config - Current displayed config.

#   saved - Is this the same as the saved config.

#   applied - Is this the same as the applied (running) config.

#   state - Where in the Creation/Deletion/Application phase is this
#           thing. One of the states below.

# page->{ hosts }->{ <name> }->{ things }->{ <name> }->{ config }

#   active - Is this for a thing that will typically be
#   running. Allows old config to remain arround and be reused if
#   needed.

#   boot - Should this thing be always started upon system boot.

# Thing Config States

sub Startup { '__Startup'; } # Possibly previously configured thing,
# the GUI is starting, the state of this thing is not yet known. It
# will go to either Created or Configured.

sub Created       { '__Created'; }
sub Starting      { '__Starting'; }
sub Started       { '__Started'; }
sub Configuring   { '__Configuring'; }
sub Configured    { '__Configured'; }
sub ShuttingDown  { '__ShuttingDown'; }

use Gui::Page;
@ISA = qw( Gui::Page );
use Thing::ServiceH;
use ControllerH;
use RepoH;
use PlayerH;
use Gtk;
use GtkEx::CList;
use GtkEx::TableEdit;
use GtkEx::EntryViaFileSelection;
use GtkEx::Utility;
use Command;
use Format::Mp3;
use Format::Wav;
use Cddb;
use Utility;
use Socket;
use Trace;
use strict;

sub new {
    my $proto = shift;
    my $type = ref($proto) || $proto;
    my %parms = @_;

    my $vbox = Gtk::VBox->new(0, $parms{'gui'}->spacing);
    $vbox->border_width($parms{'gui'}->border);
    $vbox->show();

    my $page = $type->SUPER::new(@_, title => 'Configuration', widget => $vbox);
    $page->{'box'} = $vbox;

    # The config page consists of various stackable `screens'. The
    # main screen is a notebook which does the config and sits at the
    # bottom of the stack.

    $page->{'stack'} = [];

    # Notebook

    my $notebook = Gtk::Notebook->new;
    $notebook->set_tab_pos('top');
    $notebook->set_show_border(1);
    $notebook->border_width(0);
    $notebook->show();
    $vbox->pack_start($notebook, 1, 1, 0);

    $page->{'hosts'} = {};
    $page->{'saving'} = 0;
    $page->{'applying'} = 0;
    $page->{'configing'} = 0;
    $page->{'notebook'} = $notebook;
    push @{ $page->{'stack'} }, $notebook;

    # First page - `General' Tab

    my $scroll = Gtk::ScrolledWindow->new(undef, undef);
    $scroll->set_policy('automatic', 'automatic');
    $scroll->border_width($page->gui->border);
    $scroll->show();

    my $label = Gtk::Label->new('General');
    $label->show();

    $notebook->append_page($scroll, $label);

    my $pageOneBox = Gtk::VBox->new(0, $page->gui->spacing);
    $pageOneBox->border_width(1);
    $pageOneBox->show();
    $scroll->add_with_viewport($pageOneBox);
    $pageOneBox->parent->set_shadow_type('none');
    $pageOneBox->parent->border_width(0);

    # 'Hosts' frame

    $page->{'hostsList'} = my $hostList = GtkEx::CList->new(titles => ['Name'], tooltips => $page->gui->tooltips,
       tips => [ 'The hostname of a machine that runs the emptytree-seedy system. This machine must have executed \'emptytree-seedy --boot\'.' ]);
    $hostList->setEdit(edit => 1);
    $hostList->setEditColumn(column => 0, edit => 1);
    my $h = sub { 
	my $l = shift; 
	my %p = @_; 
	return lc $p{'entry'}->get_text;
    };
    $hostList->setEditColumnGet(column => 0, handler => $h);
    $hostList->setEditRowAdd(handler => sub { my $l = shift; $page->addHost(@_); });
    $hostList->setEditRowDelete(handler => sub { my $l = shift; $page->deleteHost(@_); });
    $hostList->setEditRowModify(handler => sub { my $l = shift; $page->modifyHost(@_); });
    $hostList->show;
    my $edit = $hostList->getEdit();
    $edit->{'container'}->border_width($page->gui->border);
    $edit->{'container'}->show;

    my $frame = Gtk::Frame->new('Hosts');
    $frame->add($edit->{'container'});
    $frame->border_width(0);
    $frame->show;
    $pageOneBox->pack_start($frame, 0, 1, 0);

    # 'Music Store' 

    my $fs = GtkEx::EntryViaFileSelection->new(1024, 'Please choose a directory for music files.');
    $fs->show;
    $page->{'musicStoresRef'} = [];
    $page->{'musicStoresList'} = my $musicList = GtkEx::CList->new(titles => ['Directory'], tooltips => $page->gui->tooltips,
				   tips => [ 'A directory where your music will be placed. Must be exported to each host and have the same path on each host. Should be used solely by emptytree-seedy.' ]);
    $musicList->setEdit(edit => 1);
    $musicList->setEditColumn(column => 0, edit => 1, container => $fs->container, entry => $fs);
    $musicList->setEditRowAdd(handler => sub { my $l = shift; $page->addMusicStore(@_); });
    $musicList->setEditRowDelete(handler => sub { my $l = shift; $page->deleteMusicStore(@_); });
    $musicList->setEditRowModify(handler => sub { my $l = shift; $page->modifyMusicStore(@_); });
    $musicList->show;
    $edit = $musicList->getEdit();
    $edit->{'container'}->border_width($page->gui->border);
    $edit->{'container'}->show;

    $frame = Gtk::Frame->new('Music directories');
    $frame->add($edit->{'container'});
    $frame->border_width(0);
    $frame->show;
    $pageOneBox->pack_start($frame, 0, 1, 0);

    # 'Temp Area' - Since 1.3.0 these are built by Repo::musicDirsToTmpDirs
    
    $page->{'tmpDirsRef'} = [];

    # Encoding rate

    my $rates = Gtk::Combo->new();
    $rates->set_popdown_strings(Format::Mp3::rates);
    $rates->show;
    $page->{'ratesRef'} = [];
    $page->{'ratesList'} = my $rateList = GtkEx::CList->new(titles => ['Rate'], tooltips => $page->gui->tooltips,
				   tips => [ 'A rate to encode ripped CD tracks to.' ]);
    $rateList->setEdit(edit => 1);
    $rateList->setEditColumn(column => 0, edit => 1, container => $rates, entry => $rates->entry);
    $rateList->setEditRowAdd(handler => sub { my $l = shift; $page->addRate(@_); });
    $rateList->setEditRowDelete(handler => sub { my $l = shift; $page->deleteRate(@_); });
    $rateList->setEditRowModify(handler => sub { my $l = shift; $page->modifyRate(@_); });
    $rateList->show;
    $edit = $rateList->getEdit();
    $edit->{'container'}->border_width($page->gui->border);
    $edit->{'container'}->show;

    $frame = Gtk::Frame->new('Encoding rates');
    $frame->add($edit->{'container'});
    $frame->border_width(0);
    $frame->show;
    $pageOneBox->pack_start($frame, 0, 1, 0);

    # Second Page - 'Database' Tab

    $scroll = Gtk::ScrolledWindow->new(undef, undef);
    $scroll->set_policy('automatic', 'automatic');
    $scroll->border_width($page->gui->border);
    $scroll->show();

    $label = Gtk::Label->new('Database');
    $label->show();

    $notebook->append_page($scroll, $label);

    my $pageTwoBox = Gtk::VBox->new(0, $page->gui->spacing);
    $pageTwoBox->border_width(1);
    $pageTwoBox->show();
    $scroll->add_with_viewport($pageTwoBox);
    $pageTwoBox->parent->set_shadow_type('none');
    $pageTwoBox->parent->border_width(0);

    # Local

    $frame = Gtk::Frame->new('Local database');
    $frame->border_width(0);
    $frame->show;

    $pageTwoBox->pack_start($frame, 0, 1, 0);

    my $table = GtkEx::TableEdit->new(colSpacing => $page->gui->spacing, rowSpacing => $page->gui->spacing, tooltips => $page->gui->tooltips);
    $table->border_width($page->gui->border);
    $table->show();
    $frame->add($table);

    $page->{'dbRef'} = { local => { type => 'None', name => 'emptytree-seedy' }, remotes => [ { 'URL' => 'None', } ] };
    $page->{'repoRef'} = Thing::Handle->new(name => 'Repo', host => 'n/a');

    # Type

    $page->{'dbType'} = my $dbType = Gtk::Combo->new();
    $dbType->set_popdown_strings('None', 'Postgresql');
    $dbType->entry->set_text('None');
    $dbType->set_usize(100, 20); # Minimum width
    $dbType->show;
    $table->appendRow(title => 'Type:', 
		      tip => 'The type of database that you are running.',
		      widget => $dbType);
    # Change via edit
    $dbType->entry->signal_connect_after('key_press_event', sub { $page->dbTypeChanged(@_); });
    # Change through combo choice list
    $dbType->entry->signal_connect('changed', sub { $page->dbTypeChanged(@_); });
    
    # Name

    $page->{'dbName'} = my $dbName = Gtk::Entry->new(1024);
    $dbName->set_text('emptytree-seedy');
    $dbName->show;
    $table->appendRow(title => 'Name:', 
		      tip => 'The name for this database.',
		      widget => $dbName);
    # Change via edit
    $dbName->signal_connect_after('key_press_event', sub { $page->dbNameChanged(@_); });

    # Host

    $page->{'dbHost'} = my $dbHost = Gtk::Entry->new(1024);
    $dbHost->show;
    $table->appendRow(title => 'Host:', 
		      tip => 'The host computer the database server is running on.',
		      widget => $dbHost);
    # Change via edit
    $dbHost->signal_connect_after('key_press_event', sub { $page->dbHostChanged(@_); });

    # Port

    $page->{'dbPort'} = my $dbPort = Gtk::Entry->new(1024);
    $dbPort->show;
    $dbPort->set_usize(80, 20); # Minimum width
    $table->appendRow(title => 'Port:', 
		      tip => 'The port the database server is listening too. Leave blank for the default.',
		      widget => $dbPort);
    # Change via edit
    $dbPort->signal_connect_after('key_press_event', sub { $page->dbPortChanged(@_); });

    # User

    $page->{'dbUser'} = my $dbUser = Gtk::Entry->new(1024);
    $dbUser->show;
    $table->appendRow(title => 'User:', 
		      tip => 'The user name you use to connect to the database.',
		      widget => $dbUser);
    # Change via edit
    $dbUser->signal_connect_after('key_press_event', sub { $page->dbUserChanged(@_); });

    # Password

    $page->{'dbPassword'} = my $dbPassword = Gtk::Entry->new(1024);
    $dbPassword->set_visibility(0);
    $dbPassword->show;
    $table->appendRow(title => 'Password:', 
		      tip => 'The password you use to connect to the database.',
		      widget => $dbPassword);
    # Change via edit
    $dbPassword->signal_connect_after('key_press_event', sub { $page->dbPasswordChanged(@_); });

    # CDDB

    $frame = Gtk::Frame->new('Remote CDDBP database');
    $frame->border_width(0);
    $frame->show;
    $pageTwoBox->pack_start($frame, 0, 1, 0);

    my $frameBox = Gtk::VBox->new(0, 0);
    $frameBox->border_width(0);
    $frameBox->show;
    $frame->add($frameBox);

    $table = GtkEx::TableEdit->new(colSpacing => $page->gui->spacing, rowSpacing => $page->gui->spacing, 
				   tooltips => $page->gui->tooltips);
    $table->border_width($page->gui->border);
    $table->show();
    $frameBox->pack_start($table, 0, 1, 0);

    # URL

    $page->{'cddbUrl'} = my $cddbUrl = Gtk::Combo->new();
    $cddbUrl->set_popdown_strings('None', 
				  'http://freedb.freedb.org/~cddb/cddb.cgi',
				  'http://ca.freedb.org/~cddb/cddb.cgi',
				  'http://cz.freedb.org/~cddb/cddb.cgi',
				  'http://de.freedb.org/~cddb/cddb.cgi',
				  'http://uk.freedb.org/~cddb/cddb.cgi', );
    $cddbUrl->set_usize(320, 20); # Minimum width
    $cddbUrl->show;
    $table->appendRow(title => 'URL:', 
		      tip => 'The URL of the database server to connect to.',
		      widget => $cddbUrl);
    # Change via edit
    $cddbUrl->entry->signal_connect_after('key_press_event', sub { $page->cddbUrlChanged(@_); });
    # Change via popdown
    $cddbUrl->entry->signal_connect('changed', sub { $page->cddbUrlChanged(@_); });

    # User

    $page->{'cddbUser'} = my $cddbUser = Gtk::Entry->new(1024);
    $cddbUser->show;
    $table->appendRow(title => 'User:',
		      tip => 'The user part of a user@host email address sent to the database. Required for submissions.', 
		      widget => $cddbUser);
    # Change via edit
    $cddbUser->signal_connect_after('key_press_event', sub { $page->cddbUserChanged(@_); });

    # Host

    $page->{'cddbHost'} = my $cddbHost = Gtk::Entry->new(1024);
    $cddbHost->show;
    $table->appendRow(title => 'Host:', 
		      tip => 'The host part of a user@host email address sent to the database. Required for submissions.', 
		      widget => $cddbHost);
    # Change via edit
    $cddbHost->signal_connect_after('key_press_event', sub { $page->cddbHostChanged(@_); });

    # Separate table for longer names

    $table = GtkEx::TableEdit->new(colSpacing => $page->gui->spacing, rowSpacing => $page->gui->spacing, 
				   tooltips => $page->gui->tooltips);
    $table->border_width($page->gui->border);
    $table->show();
    $frameBox->pack_start($table, 0, 1, 0);

    # Cddb Cleanup

    $page->{'cddbWhitespace'} = my $whitespace = Gtk::CheckButton->new;
    $whitespace->show;
    $table->appendRow(title => 'Remove Spaces:', 
		      tip => 'Remove extra spacing in the Cd lookup reply.', 
		      widget => $whitespace);
    $whitespace->signal_connect('button_release_event', sub { $page->cddbWhitespaceChanged(@_); });

    $page->{'cddbUcwords'} = my $ucwords = Gtk::CheckButton->new;
    $ucwords->show;
    $table->appendRow(title => 'Uppercase Words:', 
		      tip => 'Uppercase the words in the Cd lookup reply.', 
		      widget => $ucwords);
    $ucwords->signal_connect('button_release_event', sub { $page->cddbUppercaseChanged(@_); });

    # Refresh URLs

    my $button = Gtk::Button->new('Go');
    $button->signal_connect('clicked', sub { $page->cddbUrlList(@_); });
    $button->show();
    $table->appendRow(title => 'Refresh URL List:', 
		      tip => 'Load an up to date list of CDDBP URLs.', 
		      widget => $button);

    # CDROMs & Encoders for controller config

    $page->{'cdromsRef'} = [];
    $page->{'encodersRef'} = [];

    # Buttons

    $page->{'buttonBox'} = my $buttonBox = Gtk::HButtonBox->new();
    $buttonBox->show;
    $buttonBox->set_layout('edge');
    $buttonBox->set_spacing($page->gui->spacing);
    $buttonBox->set_child_ipadding(15, 0);
    $buttonBox->set_child_size(0, 0);

    $page->{'applyButton'} = $button = Gtk::Button->new('Apply');
    $button->signal_connect('clicked', sub { $page->apply if ($page->applyCheck); });
    $button->show();
    $buttonBox->add($button);
    $page->{'saveButton'} = $button = Gtk::Button->new('Save');
    $button->signal_connect('clicked', sub { $page->save(@_); });
    $button->show();
    $buttonBox->add($button);
    $page->{'revertButton'} = $button = Gtk::Button->new('Revert');
    $button->signal_connect('clicked', sub { $page->revert(@_); });
    $button->show();
    $button->set_sensitive(0); # TODO
    $buttonBox->add($button);
    $button = Gtk::Button->new('Done');
    $button->signal_connect('clicked', sub { $page->window->destroy; });
    $button->show();
    $buttonBox->add($button);
    $vbox->pack_end($buttonBox, 0, 0, 0);

    # OK Messages

    $page->{'messageBox'} = my $messageBox = Gtk::VBox->new(0, $page->gui->spacing);
    $messageBox->border_width($page->gui->border);
    $messageBox->show();

    $page->{'messageLabel'} = $label = Gtk::Label->new(' ');
    $label->set_line_wrap(1);
    $label->set_justify('left');
    $label->show();
    my $style = Gtk::Style->new();
    $label->set_style($style);
    $messageBox->pack_start($label, 1, 1, 0);

    $buttonBox = Gtk::HButtonBox->new();
    $buttonBox->show;
    $buttonBox->set_layout('edge');
    $buttonBox->set_spacing($page->gui->spacing);
    $buttonBox->set_child_ipadding(15, 0);
    $buttonBox->set_child_size(0, 0);

    my $hsep = Gtk::HSeparator->new;
    $hsep->show;
    $messageBox->pack_start($hsep, 0, 0, $page->gui->spacing);

    $button = Gtk::Button->new('OK');
    $button->signal_connect('clicked', $type.'::messageOk', $page);
    $button->show();
    $buttonBox->add($button);
    $messageBox->pack_start($buttonBox, 0, 0, 0);

    $vbox->pack_start($messageBox, 1, 1, 0);
    $messageBox->hide;
    $page->{'messages'} = [];

    # Busy Messages

    $page->{'busy'} = { work => 0 };
    $page->{'busy'}->{'box'} = my $busyBox = Gtk::VBox->new(0, 0);
    $busyBox->border_width(0);
    $page->{'busy'}->{'label'} = $label = Gtk::Label->new(' ');
    $label->set_line_wrap(1);
    $label->set_justify('left');
    $label->show();
    $style = Gtk::Style->new();
    $label->set_style($style);
    $busyBox->pack_start($label, 1, 1, 0);
    $page->{'busy'}->{'progressBar'} = my $progress = Gtk::ProgressBar->new();
    $progress->show();
    $busyBox->pack_start($progress, 0, 0, 0);
    $vbox->pack_start($busyBox, 1, 1, 0);
    $busyBox->hide;
    $page->gui->register(name => '_delayedBusyClear', 
			 handler => sub { my $g = shift; $g->config->delayedBusyClear(); });
    $page->gui->register(name => '_applyTimeout', 
			 handler => sub { my $g = shift; $g->config->applyTimeout(@_); });

    # Handler registration

    $page->gui->register(name => Thing::Service::ConfigLoadResponse, 
			 handler => sub { my $g = shift; $g->config->configLoadResponse(@_); });
    $page->gui->register(name => Thing::Service::ConfigSaveResponse, 
			 handler => sub { my $g = shift; $g->config->configSaveResponse(@_); });
    
    $page->gui->register(name => Thing::Service::StartResponse, 
			 handler => sub { my $g = shift; $g->config->startResponse(@_); });
    
    $page->gui->register(name => Thing::ShutdownResponse, 
			 handler => sub { my $g = shift; $g->config->shutdownResponse(@_); });
    $page->gui->register(name => Thing::ConfigureResponse, 
			 handler => sub { my $g = shift; $g->config->configureResponse(@_); });

    # Find configured hosts, read from config file rather than query
    # the Service to discover trusted mode.

    my $cfg = Thing::Service::readConfig;
    if (!defined($cfg)) {
	$page->displayOkMessage(Trace::lastError);
	$page->{'trusted'} = 0;
    }
    else {
	$page->{'trusted'} =  $cfg->{'trusted'};
    }

    if ($page->{'trusted'}) {
	# See what the service thinks
	$page->gui->message(name => Thing::Service::ConfigLoadRequest,
			    to => Thing::Service::Handle,
			    parms =>
			    {
				handle => Thing::Service::Handle,
				_tag => 'hostDiscovery',
			    });
    }
    else {
	# Direct from mistrusted config file
	$cfg = Thing::Service::readConfig(Thing::Service::configPath().'/mistrust-configs/'.$page->gui->host.'.pl');

	if (!defined($cfg)) {
	    $page->displayOkMessage(Trace::lastError);
	}

	$page->displayOkMessage('
You are running in mistrust mode. This is the safest mode for executing emptytree-seedy. However it requires some user intervention to do configuration.'."\n".
'If this is not the mode you wish to operate under, exit this application and execute:'."\n".
"\n".
'  emptytree-seedy --kill'."\n".
'  emptytree-seedy --trusted'."\n".
'  emptytree-seedy --boot'."\n".
'  emptytree-seedy');

	$page->configuredHosts(defined($cfg) ? $cfg->{'config'}->{ Thing::Service::NAME } : undef);
    }

    return $page;
}

sub default_width         { 500; }
sub default_height        { 660; }

# initialStart - First time invocation for the system, undef when still unknown.

sub initialStart { $_[0]->{'initialStart'}; }

sub config {
    my $page = shift;
    return $page->{'hosts'};
}

# save - Save configuration changes.

sub save {
    my $page = shift;
    my $widget = shift;

    if ($page->{'trusted'}) {

	my @allThingsUnsaved = grep { !$_->{'saved'} } map { values(%{ $_->{'things'} }) } values(%{ $page->{'hosts'} });

	$page->beginBusySequence('Saving configuration', $#allThingsUnsaved + 1);

	foreach my $thing (@allThingsUnsaved) {
	    $page->gui->message(name => Thing::Service::ConfigSaveRequest,
				to => Thing::Service::Handle,
				parms => 
				{ 
				    handle => $thing->{'handle'},
				    config => $thing->{'config'},
				});
	    $page->{'saving'}++;
	}
    }
    else {

	# Mistrusted environment, save to CONFPATH and tell the user about it.

	if (!defined(Thing::Service::configPath())) {
	    $page->displayOkMessage(Thing::Service::configPathError());
	    return;
	}

	my $configdir = Thing::Service::configPath().'/mistrust-configs';
	if ((! -e  $configdir) && system("mkdir -p $configdir")) {
	    Trace::error('Could not create configuration directory: %s', $configdir);
	    $page->displayOkMessage(Trace::lastError);
	    return;
	}

	foreach my $host (values(%{ $page->{'hosts'} })) {
	    my $configfile = $configdir.'/'.$host->{'name'}.'.pl';
	    my %configs = map { $_->{'handle'}->name => $_->{'config'} } values(%{ $host->{'things'} });
	    if (!Thing::Service::writeConfig({ version => Thing::VERSION(), trusted => 0, config => \%configs }, $configfile)) {
		$page->displayOkMessage(Trace::lastError);
		next;
	    }
	    map { $_->{'saved'} = 1 } values(%{ $host->{'things'} });
	}
	
	# Instructions

	$page->displayOkMessage('
You are operating in mistrust mode so you will need to manually install these configuration changes.  A configuration file for each host is saved under: '."\n".
"\n".
'  '.Thing::Service::configPath().'/mistrust-configs/<hostname>.pl'."\n".
"\n".
'Copy these files to the appropriate host and on each host execute:'."\n".
"\n".
'  emptytree-seedy --install_config=<config file>'."\n".
"\n".
'To apply these changes execute the following on each host:'."\n".
"\n".
'  emptytree-seedy --kill'."\n".
'  emptytree-seedy --boot');

	$page->activateButtons();
    }
}

# configSaveResponse() - 

sub configSaveResponse {
    my $page = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};

    $page->{'saving'}--;
    $page->incBusySequence();
    if ($page->{'saving'} == 0) {
	$page->endBusySequence;
    }

    if (defined($parms->{'error'})) {
	$page->displayOkMessage('Error saving configuration for '.$parms->{'handle'}->key.":\n".$parms->{'error'});
	return;
    }
    return if (!defined($parms->{'handle'}));

    my $host = $page->{'hosts'}->{ $parms->{'handle'}->host };
    my $thing = $host->{'things'}->{ $parms->{'handle'}->name };
    $thing->{'saved'} = 1;

    $page->activateButtons();
}

# apply - Apply configuration changes.

sub applyCheck {
    my $page = shift;

    Trace::debug('Value check') if Trace::check(Trace::DEBUG4);

    # If the Gui can't configure the database, there is no point
    # continuing with the rest of the startup/configuration

    if (!Info::Persist::configure(db => $page->{'dbRef'}->{'local'}) || !Info::Cd::configure(remotes => $page->{'dbRef'}->{'remotes'})) {
	GtkEx::Utility::ackPopup(msg => 'Error configuring database:'."\n".Trace::lastError);
	return 0;
    }
    return 1;
}

sub apply {
    my $page = shift;    

    Trace::debug('Application') if Trace::check(Trace::DEBUG4);

    my @allThings = map { values(%{ $_->{'things'} }) } values(%{ $page->{'hosts'} });

    my @list1 = grep { $_->{'config'}->{'active'} &&
			   (($_->{'config'}->{'type'} eq 'Encoder') ||
			    ($_->{'config'}->{'type'} eq 'Player') ||
			    ($_->{'config'}->{'type'} eq 'Cdrom') ||
			    ($_->{'config'}->{'type'} eq 'Repo')) } @allThings;

    my @list2 = grep { $_->{'config'}->{'active'} && 
			   ($_->{'config'}->{'type'} eq 'Controller') } @allThings;

    my @list3 = grep { $_->{'config'}->{'active'} && 
			   ($_->{'config'}->{'type'} eq 'Repo') } @allThings;

    my @list4 = grep { $_->{'config'}->{'active'} && 
			   ($_->{'config'}->{'type'} eq 'Thing::Service') } @allThings;

    # Working on the assumption that a controller/repo pair on the
    # same host are active or none at all, if this is not true follow
    # drastic measures...

    if (@list2 && @list3 &&
	(($#list2 > 0) || ($#list3 > 0) || 
	 ($#list2 != $#list3) ||
	 (($#list2 == 0) && ($list2[0]->{'handle'}->host ne $list3[0]->{'handle'}->host)))) {
	Trace::error('Bad controller/repo config, resorting to none configured');
	foreach my $thing (@list2, @list3) {
	    $thing->{'config'}->{'active'} = 0;
	    $thing->{'config'}->{'boot'} = 0;
	}
	foreach my $host (values(%{ $page->{'hosts'} })) {
	    if (defined($host->{'page'})) {
		$host->{'page'}->{'main'}->set_active(0);
	    }
	}
    }

    my @shutdownJobs = grep {!$_->{'config'}->{'active'} && ($_->{'state'} ne Created) } @allThings;
    
    my @startJobsC = grep { !(($_->{'state'} eq Started) || ($_->{'state'} eq Configured) ||
			      ($_->{'state'} eq Configuring)) } @list2;
    my @startJobsNonC = grep { !(($_->{'state'} eq Started) || ($_->{'state'} eq Configured) ||
				 ($_->{'state'} eq Configuring)) } @list1;

    my @configureJobs = ((grep { !$_->{'applied'} } (@list1, @list4)), @list2);

    if (!$page->{'applying'}) {
	$page->{'applying'} = time;
	$page->beginBusySequence('Applying configuration changes', 
				 $#shutdownJobs + $#startJobsC + $#startJobsNonC + $#configureJobs + 4);
	$page->activateButtons;
	$page->gui->at(name => '_applyTimeout', to => $page->gui, time => time + 60, 
		       parms => { applying => $page->{'applying'} } );
    }

    # Shutdown

    Trace::debug('Begin shutdown stage') if Trace::check(Trace::DEBUG4);

    my $continue = 1;
    foreach my $thing (@shutdownJobs) {
	Trace::debug('Shutdown for %s in state %s', $thing->{'handle'}->key, $thing->{'state'}) if Trace::check(Trace::DEBUG4);
	if ($page->gui->ismonitor($thing->{'handle'})) {
	    if ($thing->{'handle'}->currentState eq Thing::Handle::Unknown) {
		$continue = 0;
	    }
	    else {
		if (($thing->{'state'} eq Started) || ($thing->{'state'} eq Configured)) {
		    if (!$page->gui->message(name => Thing::ShutdownRequest, to => $thing->{'handle'})) {
			Trace::debug('Should not be dead already') if Trace::check(Trace::DEBUG4);
			$thing->{'state'} = Created;
		    }
		    else {
			$thing->{'state'} = ShuttingDown;
			$continue = 0;
		    }
		}
		elsif (($thing->{'state'} eq ShuttingDown) || ($thing->{'state'} eq Startup)) {
		    $continue = 0;
		}
		elsif (($thing->{'state'} eq Starting) || ($thing->{'state'} eq Configuring)) {
		    Trace::debug('Should not be in this state') if Trace::check(Trace::DEBUG4);
		    $continue = 0;
		}
	    }
	}
	else {
	    Trace::error('Non monitored thing in application: %s', $thing->{'handle'}->key);
	    $thing->{'handle'} = $page->gui->monitor($thing->{'handle'});
	    $continue = 0;
	}
    }
    return 1 if (!$continue);

    # Monitor or start all encoders, players, cdroms and the repo. If
    # they are all started move on to the controller.

    Trace::debug('Begin start stage') if Trace::check(Trace::DEBUG4);

    $continue = 1;

    foreach my $list (\@startJobsNonC, \@startJobsC) {
	foreach my $thing (@{ $list }) {
	    Trace::debug('Start for %s in state %s', $thing->{'handle'}->key, $thing->{'state'}) if Trace::check(Trace::DEBUG4);
	    if ($page->gui->ismonitor($thing->{'handle'})) {
		if ($thing->{'handle'}->currentState eq Thing::Handle::Unknown) {
		    $continue = 0;
		}
		else {
		    if (($thing->{'state'} eq Starting) || ($thing->{'state'} eq Startup)) {
			$continue = 0;
		    }
		    elsif ($thing->{'state'} eq Created) {
			$page->gui->message(name => Thing::Service::StartRequest,
					    to => Thing::Service::Handle,
					    parms => 
					    { 
						type => $thing->{'config'}->{'type'},
						name => $thing->{'config'}->{'name'},
						host => $thing->{'handle'}->host,
						trace => $thing->{'config'}->{'trace'},
						stdio => $thing->{'config'}->{'stdio'},
					    });			
			$thing->{'applied'} = 0;
			$thing->{'state'} = Starting;
			$continue = 0;
		    }
		    elsif ($thing->{'state'} eq ShuttingDown) {
			Trace::debug('Should not be in this state') if Trace::check(Trace::DEBUG4);
			$continue = 0;
		    }
		}
	    }
	    else {
		Trace::error('Non monitored thing in application: %s', $thing->{'handle'}->key);
		$thing->{'handle'} = $page->gui->monitor($thing->{'handle'});
		$continue = 0;
	    }
	}

	return 1 if (!$continue);
    }

    # Configure the un applied configs for active things and always
    # the controller (to kick start the auto process).

    Trace::debug('Begin configuration stage') if Trace::check(Trace::DEBUG4);

    $page->{'configing'} = 0;
    foreach my $thing ((grep { !$_->{'applied'} } (@list1, @list4)), @list2) {
	if (($thing->{'state'} eq Started) || ($thing->{'state'} eq Configured)) {
	    Trace::debug('Configure for %s in state %s', $thing->{'handle'}->key, $thing->{'state'}) if Trace::check(Trace::DEBUG4);
	    my %parms = %{ $thing->{'config'} };
	    if ($page->gui->message(name => Thing::ConfigureRequest, to => $thing->{'handle'}, parms => \%parms)) {
		$page->{'configing'}++;
	    }
	    else {
		Trace::debug('Should not be dead') if Trace::check(Trace::DEBUG4);
	    }
	}
	else {
	    Trace::debug('Should not be in this state') if Trace::check(Trace::DEBUG4);
	}
    }
    if (!$page->{'configing'}) {
	$page->endBusySequence;
	$page->{'applying'} = 0;
    }
    $page->activateButtons();
    return 1;
}

sub applyTimeout {
    my $page = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};

    return if ($page->{'applying'} != $parms->{'applying'});

    $page->endBusySequence;
    $page->{'applying'} = 0;
    $page->displayOkMessage('Application of config timed out');
}

# monitorAddedThing - Monitor something that has just been added, it may already be monitored.

sub monitorAddedThing {
    my $page = shift;
    my $thing = shift;

    $thing->{'handle'} = $page->gui->monitor($thing->{'handle'});
    if ($thing->{'handle'}->currentState eq Thing::Handle::Unknown) {
	$thing->{'state'} = Startup;
    }
    elsif ($thing->{'handle'}->currentState eq Thing::Handle::Alive) {
	$thing->{'state'} = Started;
    }
    elsif ($thing->{'handle'}->currentState eq Thing::Handle::Dead) {
	$thing->{'state'} = Created;
    }
}

sub monitorChange {
    my $page = shift;
    my $handle = shift;

    Trace::debug('Monitor change for %s to %s', $handle->key, $handle->currentState) if Trace::check(Trace::DEBUG3);

    my $host = $page->{'hosts'}->{ $handle->host };
    my $thing = $host->{'things'}->{ $handle->name } if (defined($host));
    if (!defined($host) || !defined($thing)) {
	Trace::error('Monitor change for unknown handle: %s', $handle->key);
	$page->gui->demonitor($handle);
	return;
    }

    if ($thing->{'state'} eq Startup) {
	# Now we know
	if ($handle->currentState eq Thing::Handle::Alive) {
	    $thing->{'state'} = Started;
	}
	else {
	    $thing->{'state'} = Created;
	}
    }
    elsif (($thing->{'state'} eq Created) && ($handle->currentState eq Thing::Handle::Alive)) {
	$thing->{'state'} = Started;
    }

    if ($handle->name eq Thing::Service::NAME) {
	$page->updateHost($handle->host);
    }

    if (($handle->name eq Thing::Service::NAME) && ($handle->currentState eq Thing::Handle::Alive) && 
	(scalar keys %{ $host->{'things'} } == 1) && $page->{'trusted'}) {
	# Service has come alive and we don't know about its things, do a discovery
	my $id = $page->window->busy->start(msg => 'Discovering config for '.$handle->host) if (defined($page->window));
	$page->gui->message(name => Thing::Service::ConfigLoadRequest,
			    to => Thing::Service::Handle,
			    parms => { host => $handle->host, _tag => 'thingDiscovery', _busyId => $id, });
    }

    if ($page->{'applying'}) {
	$page->apply;
    }
}

sub startResponse {
    my $page = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};
    my $from = $tParms{'from'};

    my $host = $page->{'hosts'}->{ $parms->{'host'} };
    my $thing = $host->{'things'}->{ $parms->{'name'} } if (defined($host));
    if (!defined($host) || !defined($thing)) {
	Trace::error('Start response for unknown: %s@%s', $parms->{'name'}, $parms->{'host'});
	return;
    }

    Trace::debug('Start response for %s in state %s', $thing->{'handle'}->key, $thing->{'state'}) if Trace::check(Trace::DEBUG4);

    if (defined($parms->{'error'})) {
	$page->displayOkMessage('Error starting: '.$parms->{'name'}.'@'.$parms->{'host'}.":\n".$parms->{'error'});
	$page->endBusySequence;
	$page->{'applying'} = 0;
	$page->activateButtons();
	$thing->{'state'} = Created;
	return;
    }
    else {
	$thing->{'state'} = Started;
    }
    if ($page->{'applying'}) {
	$page->apply;
    }
    $page->incBusySequence();
    $page->activateButtons();
}

sub shutdownResponse {
    my $page = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};
    my $from = $tParms{'from'};

    my $host = $page->{'hosts'}->{ $from->host };
    my $thing = $host->{'things'}->{ $from->name } if (defined($host));

    if (!defined($host) || !defined($thing)) {
	Trace::error('Shutdown response for unknown: %s', $from->key);
	$page->gui->demonitor($from);
	return;
    }

    Trace::debug('Shutdown response for %s in state %s', $thing->{'handle'}->key, $thing->{'state'}) if Trace::check(Trace::DEBUG4);

    $thing->{'state'} = Created;

    if ($page->{'applying'}) {
	$page->apply;
    }
    $page->incBusySequence();
    $page->activateButtons();
}

sub configureResponse {
    my $page = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};
    my $from = $tParms{'from'};

    my $host = $page->{'hosts'}->{ $from->host };
    my $thing = $host->{'things'}->{ $from->name } if (defined($host));

    if (!defined($host) || !defined($thing)) {
	Trace::error('Configure response for unknown: %s', $from->key);
	$page->gui->demonitor($from);
	return;
    }

    Trace::debug('Configure response for %s in state %s', $thing->{'handle'}->key, $thing->{'state'}) if Trace::check(Trace::DEBUG4);

    $page->incBusySequence();
    $page->{'configing'}--;
    if (!$page->{'configing'}) {
	$page->endBusySequence;
	$page->{'applying'} = 0;
    }

    if (defined($parms->{'error'})) {
	$page->displayOkMessage('Error configuring: '.$from->key.":\n".$parms->{'error'});
	$thing->{'applied'} = 0;
	$thing->{'state'} = Started;
    }
    else {
	$thing->{'applied'} = 1;
	$thing->{'state'} = Configured;
    }
    $page->activateButtons();
}

sub revert {
    my $page = shift;
    my $widget = shift;

    foreach my $hostname (keys(%{ $page->{'hosts'} })) {
	my $host = $page->{'hosts'}->{ $hostname };
	foreach my $thing (values(%{ $host->{'things'} })) {
	    if (!$thing->{'saved'}) {
		$page->gui->message(name => Thing::Service::ConfigLoadRequest,
				    to => Thing::Service::Handle,
				    parms => 
				    { 
					handle => $thing->{'handle'},
				    });
	    }
	}
    }
}

# configLoadResponse - Response from service with config.

sub configLoadResponse {
    my $page = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};

    if (defined($parms->{'_tag'}) && ($parms->{'_tag'} eq 'thingDiscovery')) {
	$page->window->busy->stop(id => $parms->{'_busyId'}) if (defined($page->window) && defined($parms->{'_busyId'}));
    }

    if (defined($parms->{'error'})) {
	$page->displayOkMessage('Error loading configuration for '.$parms->{'host'}.":\n".
				($parms->{'error'} eq Thing::Service::HostUnreachable) ? 'Host is unreachable' :
				($parms->{'error'} eq Thing::Service::TimedOut) ? 'Host is not responding' : $parms->{'error'});
	return;
    }

    if (defined($parms->{'_tag'}) && ($parms->{'_tag'} eq 'hostDiscovery')) {
	$page->configuredHosts($parms->{'config'});
    }
    elsif (defined($parms->{'_tag'}) && ($parms->{'_tag'} eq 'thingDiscovery')) {
	$page->configuredHost($parms->{'host'}, $parms->{'configs'});
    }
}

# configuredHosts - Handle configured list of hosts in the emptytree system.

sub configuredHosts {
    my $page = shift;
    my $config = shift;
    
    if (!defined($config) || !defined($config->{'hosts'}) || ($#{ $config->{'hosts'} } == -1)) {

	# Initial start, no config, create active Controller and Repo
	$page->initHost($page->gui->host);
	my $host = $page->{'hosts'}->{ $page->gui->host };
	$page->setHostMain($host);
	$page->drawHost($page->gui->host);
	$page->updateHost($page->gui->host);
	$page->{'hostsList'}->append($page->gui->host);
	$host->{'active'} = 1;
	$page->{'initialStart'} = 1;

	# Create dummy Cdrom to hold database config
	my $dummy = $host->{'things'}->{ 'Cdrom' } = { config => {}, applied => 1, saved => 0, state => Created };
	$dummy->{'config'} = { name => 'Cdrom', type => 'Cdrom', db => $page->{'dbRef'}, active => 0, boot => 0 };
	$dummy->{'handle'} = Thing::Handle->new(name => 'Cdrom', host => $host->{'name'});
	$page->updateDbSense;
	$page->updateCddbSense;
    }
    else {
	$page->{'initialStart'} = 0;
	foreach my $host (@{ $config->{'hosts'} }) {
	    $page->initHost($host);
	    $page->drawHost($host);
	    $page->updateHost($host);
	    $page->{'hostsList'}->append($host);
	    $page->{'hosts'}->{ $host }->{'active'} = 1;
	}
    }
    
    my @activeHosts = grep { $page->{'hosts'}->{ $_ }->{'active'} } keys(%{ $page->{'hosts'} });
    my @serviceThings = map { $page->{'hosts'}->{ $_ }->{'things'}->{ Thing::Service::NAME } } keys(%{ $page->{'hosts'} });
    foreach my $serviceThing (@serviceThings) {
	
	# Make sure each service knows about the other
	    
	if (!defined($serviceThing->{'config'}->{'hosts'}) || 
	    !Utility::Array::same(\@activeHosts, $serviceThing->{'config'}->{'hosts'})) {
	    $serviceThing->{'config'}->{'hosts'} = \@activeHosts;
	    $serviceThing->{'saved'} = 0;
	    $serviceThing->{'applied'} = 0;	    
	}
    }

    if (!$page->{'trusted'}) {
	$page->mistrustLoadConfigs;
    }
}

# mistrustLoadConfigs - In an mistrusted environment the configs are
# loaded from configPath/mistrust-configs/<hostname>.pl.

sub mistrustLoadConfigs {
    my $page = shift;

    if (!defined(Thing::Service::configPath())) {
	$page->displayOkMessage(Thing::Service::configPathError());
    }

    foreach my $host (values(%{ $page->{'hosts'} })) {
	my $cfg = Thing::Service::readConfig(Thing::Service::configPath().'/mistrust-configs/'.$host->{'name'}.'.pl');
	if (!defined($cfg)) {
	    $page->displayOkMessage(Trace::lastError);
	    next;
	}
	$page->configuredHost($host->{'name'}, $cfg->{'config'});
    }
}

# configuredHost - Handle configuration for a specific host.

sub configuredHost {
    my $page = shift;
    my $hostname = shift;
    my $configs = shift;

    my $host = $page->{'hosts'}->{ $hostname };

    if (defined($configs) && values(%{ $configs })) {
	
	foreach my $thingName (keys(%{ $configs })) {
	  
	    my $thingConfig = $configs->{ $thingName };
	    my $thing = $host->{'things'}->{ $thingName }  = {};
	    $thing->{'config'} = $thingConfig;
	    $thing->{'handle'} = Thing::Handle->new(name => $thingName, host => $host->{'name'});		
	    $thing->{'saved'} = 1;
	    $thing->{'applied'} = 1;
	    if ($thing->{'config'}->{'active'}) {
		$thing->{'handle'} = $page->gui->monitor($thing->{'handle'});
		if ($thing->{'handle'}->currentState eq Thing::Handle::Unknown) {
		    $thing->{'state'} = Startup;
		}
		elsif ($thing->{'handle'}->currentState eq Thing::Handle::Alive) {
		    $thing->{'state'} = Configured;
		}
		elsif ($thing->{'handle'}->currentState eq Thing::Handle::Dead) {
		    $thing->{'state'} = Created;
		}
	    }
	    else {
		$thing->{'state'} = Created;
	    }
	    
	    # Inactivate the Controller/Repo pair if we already have a main host
	    
	    if (($thing->{'config'}->{'type'} eq 'Repo') || ($thing->{'config'}->{'type'} eq 'Controller')) {
		
		my @allThings = map { values(%{ $_->{'things'} }) } values(%{ $page->{'hosts'} });
		
		my @active;
		if ($thing->{'config'}->{'type'} eq 'Controller') {
		    @active = grep { $_->{'config'}->{'active'} && 
					 ($_->{'config'}->{'type'} eq 'Controller') } @allThings;
		}
		else {
		    @active = grep { $_->{'config'}->{'active'} && 
					 ($_->{'config'}->{'type'} eq 'Repo') } @allThings;
		}
		if ($#active == 1) {
		    $thing->{'config'}->{'active'} = 0;
		    $thing->{'config'}->{'boot'} = 0;
		    $thing->{'saved'} = 1;
		    $thing->{'applied'} = 1;
		}
	    }
	    
	    # Assign and build refs

	    if ($thing->{'config'}->{'type'} eq 'Repo') {
		if ($thing->{'config'}->{'active'}) {
		    map { push @{ $page->{'musicStoresRef'} }, $_ } @{ $thing->{'config'}->{'dirs'} };
		    Repo::musicDirsToTmpDirs($page->{'musicStoresRef'}, $page->{'tmpDirsRef'});
		    $page->{'repoRef'}->host($thing->{'handle'}->host);
		}
		$thing->{'config'}->{'db'} = $page->{'dbRef'};
		$thing->{'config'}->{'dirs'} = $page->{'musicStoresRef'};
	    }
	    elsif ($thing->{'config'}->{'type'} eq 'Controller') {
		if ($thing->{'config'}->{'active'}) {
		    map { push @{ $page->{'ratesRef'} }, $_ }  @{ $thing->{'config'}->{'rates'} };
		    map { push @{ $page->{'cdromsRef'} }, $_ } @{ $thing->{'config'}->{'cdroms'} };
		    map { push @{ $page->{'encodersRef'} }, $_ } @{ $thing->{'config'}->{'encoders'} };
		}
		$thing->{'config'}->{'tmpdirs'} = $page->{'tmpDirsRef'};
		$thing->{'config'}->{'rates'} = $page->{'ratesRef'};
		$thing->{'config'}->{'cdroms'} = $page->{'cdromsRef'};
		$thing->{'config'}->{'encoders'} = $page->{'encodersRef'};
		$thing->{'config'}->{'db'} = $page->{'dbRef'};
		$thing->{'config'}->{'repo'} = $page->{'repoRef'};
	    }
	    elsif ($thing->{'config'}->{'type'} eq 'Cdrom') {
		$page->{'dbRef'}->{'local'} = $thing->{'config'}->{'db'}->{'local'};
		$page->{'dbRef'}->{'remotes'} = $thing->{'config'}->{'db'}->{'remotes'};
		$thing->{'config'}->{'db'} = $page->{'dbRef'};
		$thing->{'config'}->{'repo'} = $page->{'repoRef'};
	    }
	    elsif ($thing->{'config'}->{'type'} eq 'Player') {
		$thing->{'config'}->{'db'} = $page->{'dbRef'};
	    }
	}
    }

    $page->updateMusicStores;
    $page->updateRates;
    $page->updatePlayers(host => $host);
    $page->updateCdroms;
    $page->updateEncoders(host => $host);
    $page->updateDb;
    $page->updateCddb;
    
    $page->updateHost($host->{'name'});
    $page->activateButtons();
}

sub initHost {
    my $page = shift;
    my $hostname = shift;
    my $host = $page->{'hosts'}->{ $hostname } = {};

    $host->{'name'} = $hostname;
    $host->{'things'} = {};

    my @activeHosts = grep { $page->{'hosts'}->{ $_ }->{'active'} } keys(%{ $page->{'hosts'} });

    # Create default service 

    my $serviceThing = $host->{'things'}->{ Thing::Service::NAME } = {};
    $serviceThing->{'handle'} = Thing::Handle->new(name => Thing::Service::NAME, host => $host->{'name'}, port => Thing::Service::PORT);
    $serviceThing->{'config'} = 
    { 
	name => Thing::Service::NAME,
	type => 'Thing::Service',
	hosts => \@activeHosts,
	active => 1,
	boot => 1,
    };
    $serviceThing->{'saved'} = 0;
    $serviceThing->{'applied'} = 0;
    $serviceThing->{'state'} = Started; # Always
}

# drawHost - Draw a host page

sub drawHost {
    my $page = shift;
    my $hostname = shift;
    my $host = $page->{'hosts'}->{ $hostname };

    Trace::debug('Drawing host: %s', $hostname) if Trace::check(Trace::DEBUG3);

    $host->{'page'} = {};

    # 'Host' tab

    $host->{'page'}->{'container'} = my $scroll = Gtk::ScrolledWindow->new(undef, undef);
    $scroll->set_policy('automatic', 'automatic');
    $scroll->border_width($page->gui->border);
    $scroll->show();
    
    my $label = Gtk::Label->new(ucfirst $hostname);
    $label->show();
    
    $page->{'notebook'}->append_page($scroll, $label);

    my $box = Gtk::VBox->new(0, $page->gui->spacing);
    $box->border_width(1);
    $box->show();
    $scroll->add_with_viewport($box);
    $box->parent->set_shadow_type('none');
    $box->parent->border_width(0);

    # Frame and Table for connect method, main and player
    
    my $frame = Gtk::Frame->new();
    $frame->border_width(0);
    $frame->show;
    $box->pack_start($frame, 0, 1, 0);

    my $table = GtkEx::TableEdit->new(colSpacing => $page->gui->spacing, rowSpacing => $page->gui->spacing, tooltips => $page->gui->tooltips);
    $table->border_width($page->gui->border);
    $table->show();
    $frame->add($table);
    
    # Main 
    
    $host->{'page'}->{'main'} = my $check = Gtk::CheckButton->new;
    $check->set_active(0);
    $check->show;
    $table->appendRow(title => 'Main:', widget => $check, tip => 'One host needs to be the \'main\' host. If possible this should be a host that is usually left running and that serves the music directories.');

    # PLAYERS

    my $using = Gtk::Combo->new();
    $using->set_popdown_strings('Decoder', 'OSS Driver');
    $using->show;

    $host->{'page'}->{'playersList'} = 
      my $playerList = 
	GtkEx::CList->new(titles => ['Name', 'Play Using', 'Decode Buffer', 'Play Buffer', 'Device', 'Command Line'],
			  tooltips => $page->gui->tooltips,
			  tips => [ 'A host unique, user friendly name to describe the player for your soundcard. e.g Soundblaster.',
				    'What plays the decoded audio, the decoder or a Linux OSS driver.',
				    'Seconds of encoded audio to buffer from files. May improve play when files are accessed over a congested network. Defaults to 5 seconds.',
				    'Seconds of decoded audio to buffer. Higher values may improve performance on machines doing other tasks. Defaults to 5 seconds.',
				    'The unix device for this player. e.g. /dev/dsp.',
				    'Provides the ability to use another command line decoder other than mpg123 or pass other options to mpg123. The decoder must accept input from the STDIN. When using the OSS driver the decoder must output to the STDOUT. e.g. mympg123 -qs -', ]);

    $playerList->setEdit(edit => 1);

    $playerList->setEditColumn(column => 0, edit => 1);
    $playerList->setEditColumnDefault(column => 0, handler => sub { my $l = shift; my %p = @_; $p{'entry'}->set_text('Player'); } );

    $playerList->setEditColumn(column => 1, edit => 1, container => $using, entry => $using->entry);
    $playerList->setEditColumnDefault(column => 1, handler => sub { my $l = shift; my %p = @_; $p{'entry'}->set_text('Decoder'); } );

    $playerList->setEditColumn(column => 2, edit => 1);

    $playerList->setEditColumn(column => 3, edit => 1);

    $playerList->setEditColumn(column => 4, edit => 1);

    $playerList->setEditColumn(column => 5, edit => 1);

    $playerList->setEditRowAdd(handler => sub { my $l = shift; $page->addPlayer(@_); }, host => $host);
    $playerList->setEditRowDelete(handler => sub { my $l = shift; $page->deletePlayer(@_); }, host => $host);
    $playerList->setEditRowModify(handler => sub { my $l = shift; $page->modifyPlayer(@_); }, host => $host);
    $playerList->show;
    my $edit = $playerList->getEdit();
    $edit->{'container'}->border_width($page->gui->border);
    $edit->{'container'}->show;

    $frame = Gtk::Frame->new('Players');
    $frame->add($edit->{'container'});
    $frame->border_width(0);
    $frame->show;
    $box->pack_start($frame, 0, 1, 0);

    # Cdroms
    
    my $lock = Gtk::CheckButton->new;
    $lock->show;
    my $autoDetect = Gtk::CheckButton->new;
    $autoDetect->show;

    my $autoRip = Gtk::CheckButton->new;
    $autoRip->show;
    my $autoPlay = Gtk::CheckButton->new;
    $autoPlay->show;
    
    $host->{'page'}->{'cdromsList'} = my $cdromList = GtkEx::CList->new(titles => ['Name', 'Device', 'Lock', 'Auto Detect', 'Auto Rip', 'Auto Play'], 
									tooltips => $page->gui->tooltips,
				   tips => [ 'A host unique, user friendly name to describe your Cdrom. e.g Top Cdrom.',
					     'The unix device for this cdrom. e.g. /dev/cdrom.',
					     'Stop other programs using this device.',
					     'Automatically detect manually inserted CDs.',
					     'Automatically rip non ripped nor encoded CDs.',
					     'Automatically play newly inserted CDs.', ]);
    $cdromList->setEdit(edit => 1);

    $cdromList->setEditColumn(column => 0, edit => 1);
    $cdromList->setEditColumn(column => 1, edit => 1);

    $cdromList->setEditColumn(column => 2, edit => 1, container => $lock);
    my $getH = sub { 
	my $l = shift; 
	my %p = @_; 
	return $p{'container'}->get_active ? 'Yes' : 'No';
    };
    my $setH = sub { 
	my $l = shift; 
	my %p = @_; 
	$p{'container'}->set_active(($p{'value'} eq 'Yes') ? 1 : 0);
    };
    my $defaultHActive = sub { 
	my $l = shift; 
	my %p = @_; 
	$p{'container'}->set_active(1);
    };
    my $defaultHInactive = sub { 
	my $l = shift; 
	my %p = @_; 
	$p{'container'}->set_active(0);
    };

    $cdromList->setEditColumnDefault(column => 0, handler => sub { my $l = shift; my %p = @_; $p{'entry'}->set_text('Cdrom'); } );
    $cdromList->setEditColumnDefault(column => 1, handler => sub { my $l = shift; my %p = @_; $p{'entry'}->set_text('/dev/cdrom'); } );

    $cdromList->setEditColumnGet(column => 2, handler => $getH);
    $cdromList->setEditColumnSet(column => 2, handler => $setH);
    $cdromList->setEditColumnDefault(column => 2, handler => $defaultHInactive);
    $cdromList->setEditColumn(column => 3, edit => 1, container => $autoDetect);
    $cdromList->setEditColumnGet(column => 3, handler => $getH);
    $cdromList->setEditColumnSet(column => 3, handler => $setH);
    $cdromList->setEditColumnDefault(column => 3, handler => $defaultHActive);

    $cdromList->setEditColumn(column => 4, edit => 1, container => $autoRip);
    $cdromList->setEditColumnGet(column => 4, handler => $getH);
    $cdromList->setEditColumnSet(column => 4, handler => $setH);
    $cdromList->setEditColumnDefault(column => 4, handler => $defaultHInactive);
    $cdromList->setEditColumn(column => 5, edit => 1, container => $autoPlay);
    $cdromList->setEditColumnGet(column => 5, handler => $getH);
    $cdromList->setEditColumnSet(column => 5, handler => $setH);
    $cdromList->setEditColumnDefault(column => 5, handler => $defaultHInactive);

    $cdromList->setEditRowAdd(handler => sub { my $l = shift; $page->addCdrom(@_); }, host => $host);
    $cdromList->setEditRowDelete(handler => sub { my $l = shift; $page->deleteCdrom(@_); }, host => $host);
    $cdromList->setEditRowModify(handler => sub { my $l = shift; $page->modifyCdrom(@_); }, host => $host);
    $cdromList->show;
    $edit = $cdromList->getEdit();
    $edit->{'container'}->border_width($page->gui->border);
    $edit->{'container'}->show;

    $frame = Gtk::Frame->new('Cdroms');
    $frame->add($edit->{'container'});
    $frame->border_width(0);
    $frame->show;
    $box->pack_start($frame, 0, 1, 0);

    # Encoders
    
    my $encoders = Gtk::Combo->new();
    $encoders->set_popdown_strings('lame', '8hz-mp3');
    $encoders->show;
    
    $host->{'page'}->{'encodersList'} = 
      my $encoderList = 
	GtkEx::CList->new(titles => ['Name', 'Type', 'Executable'], 
			  tooltips => $page->gui->tooltips,
			  tips => [ 'A host unique, user friendly name to describe your Encoder. e.g Encoder #1.',
				    'The type of encoder you want to use.',
				    'A path to specify your encoder executable if it is non standard.' ]);
    $encoderList->setEdit(edit => 1);
    $encoderList->setEditColumn(column => 0, edit => 1);
    $encoderList->setEditColumn(column => 1, edit => 1, entry => $encoders->entry, container => $encoders);
    $encoderList->setEditColumn(column => 2, edit => 1);
    $encoderList->setEditColumnDefault(column => 0, handler => sub { my $l = shift; my %p = @_; $p{'entry'}->set_text('Encoder'); } );
    $encoderList->setEditColumnDefault(column => 1, handler => sub { my $l = shift; my %p = @_; $p{'entry'}->set_text('lame'); } );
    $encoderList->setEditRowAdd(handler => sub { my $l = shift; $page->addEncoder(@_); }, host => $host);
    $encoderList->setEditRowDelete(handler => sub { my $l = shift; $page->deleteEncoder(@_); }, host => $host);
    $encoderList->setEditRowModify(handler => sub { my $l = shift; $page->modifyEncoder(@_); }, host => $host);
    $encoderList->show;
    $edit = $encoderList->getEdit();
    $edit->{'container'}->border_width($page->gui->border);
    $edit->{'container'}->show;

    $frame = Gtk::Frame->new('Encoders');
    $frame->add($edit->{'container'});
    $frame->border_width(0);
    $frame->show;
    $box->pack_start($frame, 0, 1, 0);

    # Attach signals

    my $widget = $host->{'page'}->{'main'};
    $widget->signal_connect('button_release_event', sub { $page->mainChanged($host, @_); });
}

# updateHost() - Update a hosts display

sub updateHost {
    my $page = shift;
    my $hostname = shift;
    my $host = $page->{'hosts'}->{ $hostname };

    # Ensure monitoring

    my $serviceHandle = $host->{'things'}->{ Thing::Service::NAME }->{'handle'} = 
	$page->gui->monitor($host->{'things'}->{ Thing::Service::NAME }->{'handle'});

    # Activate / Deactivate the page if necessary

    if (defined($host) && defined($host->{'page'}) && defined($host->{'page'}->{'container'})) {
	if (!$page->{'trusted'} || ($serviceHandle->currentState eq Thing::Handle::Alive)) {
	    $host->{'page'}->{'container'}->set_sensitive(1);
	}
	else {
	    $host->{'page'}->{'container'}->set_sensitive(0);
	}
    }

    # Set main/player selects
    
    if (defined($host->{'things'}->{ Controller::NAME }) && $host->{'things'}->{ Controller::NAME }->{'config'}->{'active'}) {
	$host->{'page'}->{'main'}->set_active(1);
    }
    else {
	$host->{'page'}->{'main'}->set_active(0);
    }
}

sub activateButtons {
    my $page = shift;

    my @allThings = map { values(%{ $_->{'things'} }) } values(%{ $page->{'hosts'} });

    my $apply = $page->{'trusted'} && 
	grep { ($_->{'config'}->{'active'} && !$_->{'applied'}) ||
               (!$_->{'config'}->{'active'} && (($_->{'state'} eq Started) || ($_->{'state'} eq Configured))) } @allThings;
    my $save = grep { !$_->{'saved'} } @allThings;

    $page->{'applyButton'}->set_sensitive(!$page->{'applying'} && !$page->{'saving'} && $apply);
    $page->{'saveButton'}->set_sensitive(!$page->{'applying'} && !$page->{'saving'} && $save);
}

# GENERAL LEVEL CONFIG

sub addHost {
    my $page = shift;
    my %parms = @_;

    my $hostname = $parms{'values'}->[0];

    if (!defined($hostname) || ($hostname =~ /^\s*$/)) {
	$page->displayOkMessage('Please enter a host name to add.');
	return 0;
    }

    if ($hostname =~ /\s+/) {
	$page->displayOkMessage('"'.$hostname.'" is not a valid host name.');
	return 0;
    }

    if (defined($page->{'hosts'}->{ $hostname }) && $page->{'hosts'}->{ $hostname }->{'active'}) {
	$page->displayOkMessage(ucfirst $hostname.' is already added.');
	return 0;
    }

    my $ehost;
    my $pos = 0;
    while ($ehost = $page->{'hostsList'}->get_text($pos, 0)) {
	if ($ehost eq $hostname) {
	    $page->displayOkMessage(ucfirst $hostname.' is already added.');
	    return 0;
	}
	$pos++;
    }

    Trace::debug('Adding host: %s', $hostname) if Trace::check(Trace::DEBUG2);

    if (!defined($page->{'hosts'}->{ $hostname })) {
	$page->initHost($hostname);
    }
    $page->drawHost($hostname);
    $page->updateHost($hostname);
    $page->{'hosts'}->{ $hostname }->{'active'} = 1;

    # Add to service

    my @activeHosts = grep { $page->{'hosts'}->{ $_ }->{'active'} } keys(%{ $page->{'hosts'} });
    foreach my $serviceThing (map { $_->{'things'}->{ Thing::Service::NAME } } values(%{ $page->{'hosts'} })) {
	$serviceThing->{'config'}->{'hosts'} = \@activeHosts;
    	$serviceThing->{'saved'} = 0;
    	$serviceThing->{'applied'} = 0;
    }

    $page->activateButtons();
    return 1;
}

sub modifyHost {
    my $page = shift;
    my %parms = @_;

    my $from = $parms{'oldValues'}->[0];
    my $to = $parms{'values'}->[0];;

    Trace::debug('Modifying host: %s, to: %s', $from, $to) if Trace::check(Trace::DEBUG2);

    return if ($from eq $to);

    # Essentially adding a new host

    if ($page->deleteHost(values => $parms{'oldValues'})) {
	return $page->addHost(values => $parms{'values'});
    }
    return 0;
}

sub deleteHost {
    my $page = shift;
    my %parms = @_;

    my $hostname = $parms{'values'}->[0] || return 1;
    my $host = $page->{'hosts'}->{ $hostname } || return 1;

    Trace::debug('Removing host: %s', $host->{'name'}) if Trace::check(Trace::DEBUG2);

    $host->{'active'} = 0;

    # Set configs inactive

    foreach my $thing (values(%{ $host->{'things'} })) {
	next if ($thing->{'handle'}->name eq Thing::Service::NAME);
	$thing->{'config'}->{'active'} = 0;
	$thing->{'config'}->{'boot'} = 0;
	$thing->{'saved'} = 0;
	$thing->{'applied'} = 0;
    }

    # Remove from service

    my @activeHosts = grep { $page->{'hosts'}->{ $_ }->{'active'} } keys(%{ $page->{'hosts'} });
    foreach my $serviceThing (map { $_->{'things'}->{ Thing::Service::NAME } } values(%{ $page->{'hosts'} })) {
	$serviceThing->{'config'}->{'hosts'} = \@activeHosts;
    	$serviceThing->{'saved'} = 0;
    	$serviceThing->{'applied'} = 0;
    }

    # Remove encoders and cdroms from controller config

    if (grep { $_->{'handle'}->host eq $hostname } @{ $page->{'encodersRef'} }) {
	my @encoders = grep { $_->{'handle'}->host ne $hostname } @{ $page->{'encodersRef'} };
	map { pop @{ $page->{'encodersRef'} }; } @{ $page->{'encodersRef'} };
	map { push @{ $page->{'encodersRef'} }, $_ } @encoders;
	$page->configChange(type => 'Controller');
    }
    if (grep { $_->{'handle'}->host eq $hostname } @{ $page->{'cdromsRef'} }) {
	my @cdroms = grep { $_->{'handle'}->host ne $hostname } @{ $page->{'cdromsRef'} };
	map { pop @{ $page->{'cdromsRef'} }; } @{ $page->{'cdromsRef'} };
	map { push @{ $page->{'cdromsRef'} }, $_ } @cdroms;
	$page->configChange(type => 'Controller');
    }

    # Remove notebook page

    my @children = $page->{'notebook'}->children;
    foreach my $child (@children) {
	if (lc $page->{'notebook'}->get_tab_label($child->child)->get eq $host->{'name'}) {
	    $page->{'notebook'}->remove($child->child);
	    $child->child->destroy;
	    last;
	}
    }

    # Demonitor handles

    foreach my $thing (values(%{ $host->{'things'} })) {
	$page->gui->demonitor($thing->{'handle'});
    }

    delete $host->{'page'};

    $page->activateButtons();

    return 1;
}

sub addMusicStore {
    my $page = shift;
    my %parms = @_;

    my $dir = $parms{'values'}->[0];

    if (!defined($dir) || ($dir =~ /^\s*$/)) {
	$page->displayOkMessage('Please enter a directory to add.');
	return 0;
    }

    my $edir;
    my $pos = 0;
    while ($edir = $page->{'musicStoresList'}->get_text($pos, 0)) {
	if ($edir eq $dir) {
	    $page->displayOkMessage($dir.' is already added.');
	    return 0;
	}
	$pos++;
    }

    if (!$page->_test('e', 'd', 'r', 'w', $dir)) {
	return 0;
    }
    
    Trace::debug('Adding music store: %s', $dir) if Trace::check(Trace::DEBUG2);

    push @{ $page->{'musicStoresRef'} }, $dir;
    $page->configChange(type => 'Repo');
    Repo::musicDirsToTmpDirs($page->{'musicStoresRef'}, $page->{'tmpDirsRef'});
    $page->configChange(type => 'Controller');
    $page->activateButtons();
    return 1;
}

sub modifyMusicStore {
    my $page = shift;
    my %parms = @_;

    my $from = $parms{'oldValues'}->[0];
    my $to = $parms{'values'}->[0];;

    Trace::debug('Modifying music store: %s, to: %s', $from, $to) if Trace::check(Trace::DEBUG2);

    return if ($from eq $to);

    if ($page->deleteMusicStore(values => $parms{'oldValues'})) {
	return $page->addMusicStore(values => $parms{'values'});
    }
    return 0;
}

sub deleteMusicStore {
    my $page = shift;
    my %parms = @_;

    my $dir = $parms{'values'}->[0];

    Trace::debug('Deleting music store: %s', $dir) if Trace::check(Trace::DEBUG2);

    my $pos = 0;
    foreach my $edir (@{ $page->{'musicStoresRef'} }) {
	if ($edir eq $dir) {
	    splice @{ $page->{'musicStoresRef'} }, $pos, 1;
	}
	$pos++;
    }
    $page->configChange(type => 'Repo');
    Repo::musicDirsToTmpDirs($page->{'musicStoresRef'}, $page->{'tmpDirsRef'});
    $page->configChange(type => 'Controller');
    $page->activateButtons();

    return 1;
}

sub updateMusicStores {
    my $page = shift;

    $page->{'musicStoresList'}->freeze;
    $page->{'musicStoresList'}->clear;

    foreach my $dir (@{ $page->{'musicStoresRef'} }) {
	$page->{'musicStoresList'}->append($dir);
    }

    $page->{'musicStoresList'}->thaw;
}

sub addRate {
    my $page = shift;
    my %parms = @_;

    my $rate = $parms{'values'}->[0];

    if (!defined($rate) || ($rate =~ /^\s*$/)) {
	$page->displayOkMessage('Please enter a rate to add.');
	return 0;
    }

    my $erate;
    my $pos = 0;
    while ($erate = $page->{'ratesList'}->get_text($pos, 0)) {
	if ($erate eq $rate) {
	    $page->displayOkMessage($rate.' is already added.');
	    return 0;
	}
	$pos++;
    }

    Trace::debug('Adding rate: %s', $rate) if Trace::check(Trace::DEBUG2);

    push @{ $page->{'ratesRef'} }, $rate;
    $page->configChange(type => 'Controller');
    $page->activateButtons();
    return 1;
}

sub modifyRate {
    my $page = shift;
    my %parms = @_;

    my $from = $parms{'oldValues'}->[0];
    my $to = $parms{'values'}->[0];;

    Trace::debug('Modifying temp rate: %s, to: %s', $from, $to) if Trace::check(Trace::DEBUG2);

    return if ($from eq $to);

    if ($page->deleteRate(values => $parms{'oldValues'})) {
	return $page->addRate(values => $parms{'values'});
    }
    return 0;
}

sub deleteRate {
    my $page = shift;
    my %parms = @_;

    my $rate = $parms{'values'}->[0];

    Trace::debug('Deleting tmp rate: %s', $rate) if Trace::check(Trace::DEBUG2);

    my $pos = 0;
    foreach my $erate (@{ $page->{'ratesRef'} }) {
	if ($erate eq $rate) {
	    splice @{ $page->{'ratesRef'} }, $pos, 1;
	}
	$pos++;
    }

    $page->configChange(type => 'Controller');
    $page->activateButtons();

    return 1;
}

sub updateRates {
    my $page = shift;

    $page->{'ratesList'}->freeze;
    $page->{'ratesList'}->clear;

    foreach my $rate (@{ $page->{'ratesRef'} }) {
	$page->{'ratesList'}->append($rate);
    }

    $page->{'ratesList'}->thaw;
}

sub mainChanged {
    my $page = shift;
    my $host = shift;    
    my $widget = shift;

    Trace::debug($widget->get_active ? 'enabled' : 'disabled') if Trace::check(Trace::DEBUG3);

    # OK, I can't seem to get the button to remain unchanged, i.e. the
    # default handler is getting called no matter what I do, so at
    # this point its about to change so I remember that if I need to
    # revert to current state.

    my $desiredState = !$widget->get_active;

    if ($desiredState) {

	# Only one main host at a time

	my @allThings = map { values(%{ $_->{'things'} }) } values(%{ $page->{'hosts'} });
	my @activeCR = grep { $_->{'config'}->{'active'} && 
				  (($_->{'config'}->{'type'} eq 'Controller') || 
				   ($_->{'config'}->{'type'} eq 'Repo')) } @allThings;

	foreach my $activeThing (@activeCR) {
	    $activeThing->{'config'}->{'active'} = 0;
	    $activeThing->{'config'}->{'boot'} = 0;
	    $activeThing->{'applied'} = 0;
	    $activeThing->{'saved'} = 0;
	    my $ahost = $page->{'hosts'}->{ $activeThing->{'handle'}->host };
	    if (defined($ahost->{'page'})) {
		$ahost->{'page'}->{'main'}->set_active(0);
	    }
	}

	# Create configs

	$page->setHostMain($host);
    }
    else {
	$page->clearHostMain($host);
    }

    $page->activateButtons();
    return 1;
}

sub setHostMain {
    my $page = shift;
    my $host = shift;

    my $repo = $host->{'things'}->{ Repo::NAME };
    if (!defined($repo)) {
	$repo = $host->{'things'}->{ Repo::NAME } = {};
	$repo->{'handle'} = Thing::Handle->new(host => $host->{'name'}, name => Repo::NAME );
	$repo->{'config'} = { name => Repo::NAME, type => 'Repo', dirs => $page->{'musicStoresRef'}, db => $page->{'dbRef'} };
    }
    $page->monitorAddedThing($repo);
    $repo->{'state'} = Startup;
    $repo->{'saved'} = 0;
    $repo->{'applied'} = 0;
    $repo->{'config'}->{'active'} = 1;
    $repo->{'config'}->{'boot'} = 1; # Needed by most
    $page->{'repoRef'}->host($repo->{'handle'}->host);

    $page->configChange(type => 'Cdrom'); # Due to repo change

    my $controller = $host->{'things'}->{ Controller::NAME };
    if (!defined($controller)) {
	$controller = $host->{'things'}->{ Controller::NAME } = {};
	$controller->{'handle'} = Thing::Handle->new(host => $host->{'name'}, name => Controller::NAME);
	$controller->{'config'} = { name => Controller::NAME,
				    type => 'Controller',
				    repo => $page->{'repoRef'}, 
				    tmpdirs => $page->{'tmpDirsRef'}, 
				    cdroms => $page->{'cdromsRef'},
				    encoders => $page->{'encodersRef'},
				    rates => $page->{'ratesRef'}, 
				    db => $page->{'dbRef'} };
    }
    $page->monitorAddedThing($controller);
    $controller->{'state'} = Startup;
    $controller->{'saved'} = 0;
    $controller->{'applied'} = 0;
    $controller->{'config'}->{'active'} = 1;
    $controller->{'config'}->{'boot'} = 3;
}

sub clearHostMain {
    my $page = shift;
    my $host = shift;

    my $controller = $host->{'things'}->{ Controller::NAME };
    $controller->{'saved'} = 0;
    $controller->{'applied'} = 0;
    $controller->{'config'}->{'active'} = 0;
    $controller->{'config'}->{'boot'} = 0;

    my $repo = $host->{'things'}->{ Repo::NAME };
    $repo->{'saved'} = 0;
    $repo->{'applied'} = 0;
    $repo->{'config'}->{'active'} = 0;
    $repo->{'config'}->{'boot'} = 0;
}

sub playerChanged {
    my $page = shift;
    my $host = shift;    
    my $widget = shift;

    Trace::debug('Player '.((!$widget->get_active) ? 'enabled' : 'disabled')) if Trace::check(Trace::DEBUG3);

    # OK, I can't seem to get the button to remain unchanged, i.e. the
    # default handler is getting called no matter what I do, so at
    # this point its about to change so I remember that if I need to
    # revert to current state.

    my $desiredState = !$widget->get_active;

    if ($desiredState) {
	my $player = $host->{'things'}->{ Player::NAME };
	if (!defined($player)) {
	    $player = $host->{'things'}->{ Player::NAME } = {};
	    $player->{'handle'} = Thing::Handle->new(host => $host->{'name'}, name => Player::NAME);
	    $player->{'config'} = { name => Player::NAME, type => 'Player', db => $page->{'dbRef'} };
	}
	$page->monitorAddedThing($player);
	$player->{'saved'} = 0;
	$player->{'applied'} = 0;
	$player->{'config'}->{'active'} = 1;
	$player->{'config'}->{'boot'} = 2;
    }
    else {
	my $player = $host->{'things'}->{ Player::NAME };
	$player->{'saved'} = 0;
	$player->{'applied'} = 0;
	$player->{'config'}->{'active'} = 0;
	$player->{'config'}->{'boot'} = 0;
    }

    $page->activateButtons();
    return 1;
}

# DATABASE CONFIG

sub dbTypeChanged {
    my $page = shift;
    my $widget = shift;

    my $desiredValue = $widget->get_text;
    my $local = $page->{'dbRef'}->{'local'};
    return 1 if (defined($local->{'type'}) && defined($desiredValue) && ($local->{'type'} eq $desiredValue));

    if ($desiredValue eq 'None') {
	$local->{'type'} = 'None';
    }
    elsif ($desiredValue eq 'Postgresql') {
	$local->{'type'} = 'Postgresql';
	if (!defined($page->{'dbName'}->get_text) || ($page->{'dbName'}->get_text eq '')) {
	    $page->{'dbName'}->set_text('emptytree-seedy');
	    $local->{'name'} = 'emptytree-seedy';
	}
    }
    else {
	$page->displayOkMessage('Only no database or Postgresql is currently supported.');
	return 0;
    }
    $page->configChange(type => 'Cdrom');
    $page->configChange(type => 'Player');
    $page->configChange(type => 'Repo');
    $page->configChange(type => 'Controller');
    $page->updateDbSense;
    $page->activateButtons();
    return 1;
}

sub dbNameChanged {
    my $page = shift;
    my $widget = shift;

    my $name = $widget->get_text;

    if (!defined($name) || ($name =~ /^\s*$/)) {
	$page->displayOkMessage('Please enter a name for this database.');
	return 0;
    }

    my $local = $page->{'dbRef'}->{'local'};
    return 1 if (defined($local->{'name'}) && defined($name) && ($local->{'name'} eq $name));

    $local->{'name'} = $name;
    $page->configChange(type => 'Cdrom');
    $page->configChange(type => 'Player');
    $page->configChange(type => 'Repo');
    $page->configChange(type => 'Controller');
    $page->activateButtons();
    return 1;
}

sub dbHostChanged {
    my $page = shift;
    my $widget = shift;

    my $host = $widget->get_text;

    if ($host =~ /^\s*$/) {
	$host = undef;
    }

    my $local = $page->{'dbRef'}->{'local'};
    return 1 if (defined($local->{'host'}) && defined($host) && ($local->{'host'} eq $host));

    $local->{'host'} = $host;
    $page->configChange(type => 'Cdrom');
    $page->configChange(type => 'Player');
    $page->configChange(type => 'Repo');
    $page->configChange(type => 'Controller');
    $page->activateButtons();
    return 1;
}

sub dbPortChanged {
    my $page = shift;
    my $widget = shift;

    my $port = $widget->get_text;

    if ($port =~ /^\s*$/) {
	$port = undef;
    }

    my $local = $page->{'dbRef'}->{'local'};
    return 1 if (defined($local->{'port'}) && defined($port) && ($local->{'port'} eq $port));

    $local->{'port'} = $port;
    $page->configChange(type => 'Cdrom');
    $page->configChange(type => 'Player');
    $page->configChange(type => 'Repo');
    $page->configChange(type => 'Controller');
    $page->activateButtons();
    return 1;
}

sub dbUserChanged {
    my $page = shift;
    my $widget = shift;

    my $user = $widget->get_text;

    if ($user =~ /^\s*$/) {
	$user = undef;
    }

    my $local = $page->{'dbRef'}->{'local'};
    return 1 if (defined($local->{'user'}) && defined($user) && ($local->{'user'} eq $user));

    $local->{'user'} = $user;
    $page->configChange(type => 'Cdrom');
    $page->configChange(type => 'Player');
    $page->configChange(type => 'Repo');
    $page->configChange(type => 'Controller');
    $page->activateButtons();
    return 1;
}

sub dbPasswordChanged {
    my $page = shift;
    my $widget = shift;

    my $password = $widget->get_text;

    if ($password =~ /^\s*$/) {
	$password = undef;
    }

    my $local = $page->{'dbRef'}->{'local'};
    return 1 if (defined($local->{'password'}) && defined($password) && ($local->{'password'} eq $password));

    $local->{'password'} = $password;
    $page->configChange(type => 'Cdrom');
    $page->configChange(type => 'Player');
    $page->configChange(type => 'Repo');
    $page->configChange(type => 'Controller');
    $page->activateButtons();
    return 1;
}

sub updateDb {
    my $page = shift;
    
    my $local = $page->{'dbRef'}->{'local'};
    
    if (defined($local->{'type'})) {
	$page->{'dbType'}->entry->set_text($local->{'type'});
    }
    else {
	$page->{'dbType'}->entry->set_text('');
    }

    $page->updateDbSense;

    if (defined($local->{'name'})) {
	$page->{'dbName'}->set_text($local->{'name'});
    }
    else {
	$page->{'dbName'}->set_text('');
    }

    if (defined($local->{'host'})) {
	$page->{'dbHost'}->set_text($local->{'host'});
    }
    else {
	$page->{'dbHost'}->set_text('');
    }

    if (defined($local->{'port'})) {
	$page->{'dbPort'}->set_text($local->{'port'});
    }
    else {
	$page->{'dbPort'}->set_text('');
    }

    if (defined($local->{'user'})) {
	$page->{'dbUser'}->set_text($local->{'user'});
    }
    else {
	$page->{'dbUser'}->set_text('');
    }

    if (defined($local->{'password'})) {
	$page->{'dbPassword'}->set_text($local->{'password'});
    }
    else {
	$page->{'dbPassword'}->set_text('');
    }
}

sub updateDbSense {
    my $page = shift;    
    my $local = $page->{'dbRef'}->{'local'};
    my $sense = ($local->{'type'} ne 'None');
    $page->{'dbName'}->set_sensitive($sense);
    $page->{'dbHost'}->set_sensitive($sense);
    $page->{'dbPort'}->set_sensitive($sense);
    $page->{'dbUser'}->set_sensitive($sense);
    $page->{'dbPassword'}->set_sensitive($sense);
}

sub cddbUrlChanged {
    my $page = shift;
    my $widget = shift;

    my $url = $widget->get_text;

    if ($url =~ /^(\S+)\s+\S+/) {
	$url = $1;
    }

    my $remote = $page->{'dbRef'}->{'remotes'}->[0];
    if (!defined($remote)) {
	$remote = $page->{'dbRef'}->{'remotes'}->[0] = {};
    }
    return 1 if (defined($remote->{'URL'}) && defined($url) && ($remote->{'URL'} eq $url));

    $remote->{'URL'} = $url;
    $page->updateCddbSense;
    $page->configChange(type => 'Cdrom');
    $page->configChange(type => 'Player');
    $page->configChange(type => 'Repo');
    $page->configChange(type => 'Controller');
    $page->activateButtons();
    return 1;
}

sub cddbUrlList {
    my $page = shift;

    # If the user has yet to configure a server use freedb.org

    if (!Cddb::configured) {
	if (!Cddb::configure(URL => 'http://freedb.freedb.org/~cddb/cddb.cgi',
			     User => 'anonymous',
			     Host => 'anonymous',
			     Program => 'emptytree-seedy',
			     Version => Thing::VERSION)) {
	    GtkEx::Utility::ackPopup(msg => 'Could not get list'."\n".Trace::lastError);
	    return 1;
	}
    }

    my @list = Cddb::sites;
    if (!@list) {
	GtkEx::Utility::ackPopup(msg => 'Could not get list'."\n".Trace::lastError);
	return 1;
    }

    $page->{'cddbUrl'}->set_popdown_strings('None', map { $_->{'URL'}.' '.$_->{'Desc'} } @list);
    $page->activateButtons();
    return 1;
}

sub cddbUserChanged {
    my $page = shift;
    my $widget = shift;

    my $user = $widget->get_text;

    if ($user =~ /^\s*$/) {
	$user = undef;
    }

    my $remote = $page->{'dbRef'}->{'remotes'}->[0];
    if (!defined($remote)) {
	$remote = $page->{'dbRef'}->{'remotes'}->[0] = {};
    }
    return 1 if (defined($remote->{'user'}) && defined($user) && ($remote->{'user'} eq $user));

    $remote->{'user'} = $user;
    $page->configChange(type => 'Cdrom');
    $page->configChange(type => 'Player');
    $page->configChange(type => 'Repo');
    $page->configChange(type => 'Controller');
    $page->activateButtons();
    return 1;
}

sub cddbHostChanged {
    my $page = shift;
    my $widget = shift;

    my $host = $widget->get_text;

    if ($host =~ /^\s*$/) {
	$host = undef;
    }

    my $remote = $page->{'dbRef'}->{'remotes'}->[0];
    if (!defined($remote)) {
	$remote = $page->{'dbRef'}->{'remotes'}->[0] = {};
    }
    return 1 if (defined($remote->{'host'}) && defined($host) && ($remote->{'host'} eq $host));

    $remote->{'host'} = $host;
    $page->configChange(type => 'Cdrom');
    $page->configChange(type => 'Player');
    $page->configChange(type => 'Repo');
    $page->configChange(type => 'Controller');
    $page->activateButtons();
    return 1;
}

sub cddbWhitespaceChanged {
    my $page = shift;
    my $widget = shift;

    Trace::debug('Whitespace '.((!$widget->get_active) ? 'enabled' : 'disabled')) if Trace::check(Trace::DEBUG3);

    my $desiredState = !$widget->get_active ? 1 : 0;

    my $remote = $page->{'dbRef'}->{'remotes'}->[0];
    if (!defined($remote)) {
	$remote = $page->{'dbRef'}->{'remotes'}->[0] = {};
    }
    $remote->{'removeWhitespace'} = $desiredState;

    $page->configChange(type => 'Cdrom');
    $page->configChange(type => 'Player');
    $page->configChange(type => 'Repo');
    $page->configChange(type => 'Controller');
    $page->activateButtons();
    return 1;
}

sub cddbUppercaseChanged {
    my $page = shift;
    my $widget = shift;

    Trace::debug('Uppercase '.((!$widget->get_active) ? 'enabled' : 'disabled')) if Trace::check(Trace::DEBUG3);

    my $desiredState = !$widget->get_active ? 1 : 0;

    my $remote = $page->{'dbRef'}->{'remotes'}->[0];
    if (!defined($remote)) {
	$remote = $page->{'dbRef'}->{'remotes'}->[0] = {};
    }
    $remote->{'uppercaseWords'} = $desiredState;

    $page->configChange(type => 'Cdrom');
    $page->configChange(type => 'Player');
    $page->configChange(type => 'Repo');
    $page->configChange(type => 'Controller');
    $page->activateButtons();
    return 1;
}

sub updateCddb {
    my $page = shift;
    
    my $remote = $page->{'dbRef'}->{'remotes'}->[0];
    
    if (defined($remote->{'URL'})) {
	$page->{'cddbUrl'}->entry->set_text($remote->{'URL'});
    }
    else {
	$page->{'cddbUrl'}->entry->set_text('');
    }
    $page->updateCddbSense;

    if (defined($remote->{'user'})) {
	$page->{'cddbUser'}->set_text($remote->{'user'});
    }
    else {
	$page->{'cddbUser'}->set_text('');
    }

    if (defined($remote->{'host'})) {
	$page->{'cddbHost'}->set_text($remote->{'host'});
    }
    else {
	$page->{'cddbHost'}->set_text('');
    }

    if (defined($remote->{'removeWhitespace'})) {
	$page->{'cddbWhitespace'}->set_active($remote->{'removeWhitespace'});
    }
    else {
	$page->{'cddbWhitespace'}->set_active(1);
    }

    if (defined($remote->{'uppercaseWords'})) {
	$page->{'cddbUcwords'}->set_active($remote->{'uppercaseWords'});
    }
    else {
	$page->{'cddbUcwords'}->set_active(1);
    }
}

sub updateCddbSense {
    my $page = shift;
    my $remote = $page->{'dbRef'}->{'remotes'}->[0];
    my $sense = ($remote->{'URL'} ne 'None');
    $page->{'cddbUser'}->set_sensitive($sense);
    $page->{'cddbHost'}->set_sensitive($sense);
}

# HOST LEVEL CONFIG

# PLAYERS

sub addPlayer {
    my $page = shift;
    my %parms = @_;

    my $host = $parms{'host'};

    return if (!$page->testPlayerValues(host => $parms{'host'}, values => $parms{'values'}));

    my $name = $parms{'values'}->[0];
    my $playUsing = ($parms{'values'}->[1] eq 'OSS Driver') ? 'oss' : undef;
    my $decodeBufferSecs = ($parms{'values'}->[2] eq '') ? undef : $parms{'values'}->[2];
    my $playBufferSecs = ($parms{'values'}->[3] eq '') ? undef : $parms{'values'}->[3];
    my $device = ($parms{'values'}->[4] eq '') ? undef : $parms{'values'}->[4];
    my $decoderCommandLine = ($parms{'values'}->[5] eq '') ? undef : $parms{'values'}->[5];

    if (defined($host->{'things'}->{ $name }) &&
	$host->{'things'}->{ $name }->{'config'}->{'active'}) {
	$page->displayOkMessage($name.' is already in the list.');
	return 0;
    }

    my @activePlayers = grep { $_->{'config'}->{'active'} &&
			      ($_->{'config'}->{'type'} eq 'Player') } values(%{ $host->{'things'} });
    if (defined($device)) {
      my @thisDevice = grep { $_->{'config'}->{'device'} eq $device } @activePlayers;
      if ($#thisDevice >= 0) {
	$page->displayOkMessage('A player using '.$device.' is already in the list.');
	return 0;
      }
    }

    my $player = $host->{'things'}->{ $name };
    if (!defined($player)) {
	$player = $host->{'things'}->{ $name } = {};
	$player->{'handle'} = Thing::Handle->new(name => $name, host => $host->{'name'});
    }
    $page->monitorAddedThing($player);
    $player->{'config'} = { name => $name, type => 'Player', playUsing => $playUsing, 
			    decodeBufferSecs => $decodeBufferSecs, playBufferSecs => $playBufferSecs,
			    device => $device, db => $page->{'dbRef'}, decoderCommandLine => $decoderCommandLine,
			    active => 1, boot => 2 };
    $player->{'saved'} = 0;
    $player->{'applied'} = 0;

    Trace::debug('Adding player: %s', $player->{'handle'}->key) if Trace::check(Trace::DEBUG2);

    $page->activateButtons();
    return 1;
}

sub modifyPlayer {
    my $page = shift;
    my %parms = @_;

    return 0 if (!$page->testPlayerValues(host => $parms{'host'}, values => $parms{'values'}));

    if ($page->deletePlayer(host => $parms{'host'}, values => $parms{'oldValues'})) {
      return $page->addPlayer(host => $parms{'host'}, values => $parms{'values'});
    }
    return 0;
}

sub testPlayerValues {
    my $page = shift;
    my %parms = @_;

    my $name = $parms{'values'}->[0];

    if (!defined($name) || ($name =~ /^\s*$/)) {
	$page->displayOkMessage('Please enter a name for this player.');
	return 0;
    }

    my $playUsing = $parms{'values'}->[1];

    if (!defined($playUsing) || (($playUsing ne 'OSS Driver') && ($playUsing ne 'Decoder'))) {
	$page->displayOkMessage('Play Using must be either Decoder or OSS Driver.');
	return 0;
    }

    return 1;
}

sub deletePlayer {
    my $page = shift;
    my %parms = @_;

    my $name = $parms{'values'}->[0];
    my $host = $parms{'host'};

    Trace::debug('Deleting player: %s', $name) if Trace::check(Trace::DEBUG2);

    $host->{'things'}->{ $name }->{'config'}->{'active'} = 0;
    $host->{'things'}->{ $name }->{'config'}->{'boot'} = 0;
    $host->{'things'}->{ $name }->{'saved'} = 0;
    $host->{'things'}->{ $name }->{'applied'} = 0;

    $page->activateButtons();

    return 1;
}

sub updatePlayers {
  my $page = shift;
  my %parms = @_;

  my $host = $parms{'host'};

  $host->{'page'}->{'playersList'}->freeze;
  $host->{'page'}->{'playersList'}->clear;

  foreach my $player (grep { ($_->{'config'}->{'type'} eq 'Player') &&
				($_->{'config'}->{'active'}) } values(%{ $host->{'things'} })) {

    my $thing = $host->{'things'}->{ $player->{'handle'}->name };
    next if (!defined($thing));
    my $playUsing = (defined($thing->{'config'}->{'playUsing'}) && ($thing->{'config'}->{'playUsing'} eq 'oss')) 
      ? 'OSS Driver' : 'Decoder';
    my $decodeBufferSecs = defined($thing->{'config'}->{'decodeBufferSecs'})
      ? $thing->{'config'}->{'decodeBufferSecs'} : '';
    my $playBufferSecs = defined($thing->{'config'}->{'playBufferSecs'})
      ? $thing->{'config'}->{'playBufferSecs'} : '';
    my $device = defined($thing->{'config'}->{'device'})
      ? $thing->{'config'}->{'device'} : '';
    my $decoderCommandLine = defined($thing->{'config'}->{'decoderCommandLine'})
      ? $thing->{'config'}->{'decoderCommandLine'} : '';
    $host->{'page'}->{'playersList'}->append($player->{'handle'}->name,
					     $playUsing, $decodeBufferSecs, $playBufferSecs, $device, $decoderCommandLine,
					    );
  }
  $host->{'page'}->{'playersList'}->thaw;
}

# CDROMS

sub addCdrom {
    my $page = shift;
    my %parms = @_;

    my $host = $parms{'host'};

    return if (!$page->testCdromValues(host => $parms{'host'}, values => $parms{'values'}));
    
    my $name = $parms{'values'}->[0];
    my $device = $parms{'values'}->[1];
    my $lock = ($parms{'values'}->[2] eq 'Yes') ? 1 : 0;
    my $autoDetect = ($parms{'values'}->[3] eq 'Yes') ? 1 : 0;
    my $autoRip = ($parms{'values'}->[4] eq 'Yes') ? 1 : 0;
    my $autoPlay = ($parms{'values'}->[5] eq 'Yes') ? 1 : 0;

    if (defined($host->{'things'}->{ $name }) &&
	$host->{'things'}->{ $name }->{'config'}->{'active'}) {
	$page->displayOkMessage($name.' is already in the list.');
	return 0;
    }

    my @activeCdroms = grep { $_->{'config'}->{'active'} &&
			      ($_->{'config'}->{'type'} eq 'Cdrom') } values(%{ $host->{'things'} });
    my @thisDevice = grep { $_->{'config'}->{'device'} eq $device } @activeCdroms;
    if ($#thisDevice >= 0) {
	$page->displayOkMessage($device.' is already in the list.');
	return 0;
    }

    my $cdrom = $host->{'things'}->{ $name };
    if (!defined($cdrom)) {
	$cdrom = $host->{'things'}->{ $name } = {};
	$cdrom->{'handle'} = Thing::Handle->new(name => $name, host => $host->{'name'});
    }
    $page->monitorAddedThing($cdrom);
    $cdrom->{'config'} = { name => $name, type => 'Cdrom', device => $device, db => $page->{'dbRef'}, repo => $page->{'repoRef'},
			   lock => $lock, autoDetect => $autoDetect, active => 1, boot => 2 };
    $cdrom->{'saved'} = 0;
    $cdrom->{'applied'} = 0;

    Trace::debug('Adding cdrom: %s', $cdrom->{'handle'}->key) if Trace::check(Trace::DEBUG2);

    push @{ $page->{'cdromsRef'} }, { handle => $cdrom->{'handle'}, autoRip => $autoRip, autoPlay => $autoPlay };

    $page->configChange(type => 'Controller');
    $page->activateButtons();
    return 1;
}

sub modifyCdrom {
    my $page = shift;
    my %parms = @_;

    my $name = $parms{'values'}->[0];
    my $device = $parms{'values'}->[1];

    my $host = $parms{'host'};

    if (($name eq $parms{'oldValues'}->[0]) && ($device eq $parms{'oldValues'}->[1])) {
    
	Trace::debug('Modifying cdrom: %s', $name) if Trace::check(Trace::DEBUG2);

	my $lock = ($parms{'values'}->[2] eq 'Yes') ? 1 : 0;
	my $autoDetect = ($parms{'values'}->[3] eq 'Yes') ? 1 : 0;
	my $autoRip = ($parms{'values'}->[4] eq 'Yes') ? 1 : 0;
	my $autoPlay = ($parms{'values'}->[5] eq 'Yes') ? 1 : 0;

	my $cdrom = $host->{'things'}->{ $name };
	$cdrom->{'config'}->{'lock'} = $lock;
	$cdrom->{'config'}->{'autoDetect'} = $autoDetect;
	$cdrom->{'saved'} = 0;
	$cdrom->{'applied'} = 0;
	
	my @thisCdrom = grep { $_->{'handle'}->name eq $name } @{ $page->{'cdromsRef'} };
	$thisCdrom[0]->{'autoRip'} = $autoRip;
	$thisCdrom[0]->{'autoPlay'} = $autoPlay;

	$page->configChange(type => 'Controller');
	$page->activateButtons();

	return 1;
    }
    else {

	return 0 if (!$page->testCdromValues(host => $parms{'host'}, values => $parms{'values'}));

	if ($page->deleteCdrom(host => $parms{'host'}, values => $parms{'oldValues'})) {
	    return $page->addCdrom(host => $parms{'host'}, values => $parms{'values'});
	}
	return 0;
    }
}

sub testCdromValues {
    my $page = shift;
    my %parms = @_;

    my $host = $parms{'host'};

    my $name = $parms{'values'}->[0];

    if (!defined($name) || ($name =~ /^\s*$/)) {
	$page->displayOkMessage('Please enter a name for this cdrom.');
	return 0;
    }

    my $device = $parms{'values'}->[1];

    if (!defined($device) || ($device =~ /^\s*$/)) {
	$page->displayOkMessage('Please enter a device for this cdrom.');
	return 0;
    }

    return 1;
}

sub deleteCdrom {
    my $page = shift;
    my %parms = @_;

    my $name = $parms{'values'}->[0];
    my $host = $parms{'host'};

    Trace::debug('Deleting cdrom: %s', $name) if Trace::check(Trace::DEBUG2);

    $host->{'things'}->{ $name }->{'config'}->{'active'} = 0;
    $host->{'things'}->{ $name }->{'config'}->{'boot'} = 0;
    $host->{'things'}->{ $name }->{'saved'} = 0;
    $host->{'things'}->{ $name }->{'applied'} = 0;

    my $pos = 0;
    foreach my $ecdrom (@{ $page->{'cdromsRef'} }) {
	if (Thing::Handle::cmpNameHost($ecdrom->{'handle'}, $host->{'things'}->{ $name }->{'handle'})) {
	    splice @{ $page->{'cdromsRef'} }, $pos, 1;
	}
	$pos++;
    }

    $page->configChange(type => 'Controller');
    $page->activateButtons();

    return 1;
}

sub updateCdroms {
    my $page = shift;

    foreach my $host (values(%{ $page->{'hosts'} })) {

	next if (!defined($host->{'page'}->{'cdromsList'}));

	$host->{'page'}->{'cdromsList'}->freeze;
	$host->{'page'}->{'cdromsList'}->clear;

	foreach my $cdrom (grep { $_->{'handle'}->host eq $host->{'name'} } @{ $page->{'cdromsRef'} }) {
	    my $thing = $host->{'things'}->{ $cdrom->{'handle'}->name };
	    next if (!defined($thing));
	    # lock and autoDetect added 1.2.0, create defaults
	    $thing->{'config'}->{'lock'} = defined($thing->{'config'}->{'lock'}) ? $thing->{'config'}->{'lock'} : 0;
	    $thing->{'config'}->{'autoDetect'} = defined($thing->{'config'}->{'autoDetect'}) ? $thing->{'config'}->{'autoDetect'} : 1;
	    $host->{'page'}->{'cdromsList'}->append($cdrom->{'handle'}->name, 
						    $thing->{'config'}->{'device'},
						    $thing->{'config'}->{'lock'} ? 'Yes' : 'No',
						    $thing->{'config'}->{'autoDetect'} ? 'Yes' : 'No',
						    $cdrom->{'autoRip'} ? 'Yes' : 'No',
						    $cdrom->{'autoPlay'} ? 'Yes' : 'No',);
	}
	$host->{'page'}->{'cdromsList'}->thaw;
    }
}

# ENCODERS

sub addEncoder {
    my $page = shift;
    my %parms = @_;

    my $host = $parms{'host'};

    my $name = $parms{'values'}->[0];

    if (!defined($name) || ($name =~ /^\s*$/)) {
	$page->displayOkMessage('Please enter a name for this encoder.');
	return 0;
    }

    if (defined($host->{'things'}->{ $name }) &&
	$host->{'things'}->{ $name }->{'config'}->{'active'}) {
	$page->displayOkMessage($name.' is already in the list.');
	return 0;
    }

    my $type = $parms{'values'}->[1];

    if (!defined($type) || ($type =~ /^\s*$/)) {
	$page->displayOkMessage('Please choose a type for this encoder.');
	return 0;
    }

    my $executable = $parms{'values'}->[2];

    my $encoder = $host->{'things'}->{ $name };
    if (!defined($encoder)) {
	$encoder = $host->{'things'}->{ $name } = {};
	$encoder->{'handle'} = Thing::Handle->new(name => $name, host => $host->{'name'});
    }
    $page->monitorAddedThing($encoder);
    $encoder->{'config'} = { name => $name, type => 'Encoder', encoder => $type, 
			     exe => $executable, active => 1, boot => 2 };
    $encoder->{'saved'} = 0;
    $encoder->{'applied'} = 0;

    Trace::debug('Adding encoder: %s', $encoder->{'handle'}->key) if Trace::check(Trace::DEBUG2);

    push @{ $page->{'encodersRef'} }, { handle => $encoder->{'handle'} };

    $page->configChange(type => 'Controller');
    $page->activateButtons();
    return 1;
}

sub modifyEncoder {
    my $page = shift;
    my %parms = @_;

    my $from = $parms{'oldValues'}->[0];
    my $to = $parms{'values'}->[0];

    Trace::debug('Modifying encoder: %s, to: %s', $from, $to) if Trace::check(Trace::DEBUG2);

    if ($page->deleteEncoder(host => $parms{'host'}, values => $parms{'oldValues'})) {
	return $page->addEncoder(host => $parms{'host'}, values => $parms{'values'});
    }
    return 0;
}

sub deleteEncoder {
    my $page = shift;
    my %parms = @_;

    my $name = $parms{'values'}->[0];
    my $host = $parms{'host'};

    Trace::debug('Deleting encoder: %s', $name) if Trace::check(Trace::DEBUG2);

    my $pos = 0;
    foreach my $eencoder (@{ $page->{'encodersRef'} }) {
	if (Thing::Handle::cmpNameHost($eencoder->{'handle'}, $host->{'things'}->{ $name }->{'handle'})) {
	    splice @{ $page->{'encodersRef'} }, $pos, 1;
	}
	$pos++;
    }
    $host->{'things'}->{ $name }->{'config'}->{'active'} = 0;
    $host->{'things'}->{ $name }->{'config'}->{'boot'} = 0;
    $host->{'things'}->{ $name }->{'saved'} = 0;
    $host->{'things'}->{ $name }->{'applied'} = 0;

    $page->configChange(type => 'Controller');
    $page->activateButtons();

    return 1;
}

sub updateEncoders {
    my $page = shift;
    my %parms = @_;

    my $host = $parms{'host'};

    $host->{'page'}->{'encodersList'}->freeze;
    $host->{'page'}->{'encodersList'}->clear;

    foreach my $encoder (grep { ($_->{'config'}->{'type'} eq 'Encoder') &&
				    ($_->{'config'}->{'active'}) } values(%{ $host->{'things'} })) {
	$host->{'page'}->{'encodersList'}->append($encoder->{'handle'}->name, 
						  $encoder->{'config'}->{'encoder'},
						  $encoder->{'config'}->{'exe'},);
    }

    $host->{'page'}->{'encodersList'}->thaw;
}

# configChange() - Set a config change for all things of a particular type

sub configChange {
    my $page = shift;
    my %parms = @_;

    Trace::debug('Setting config changed for all things of type: %s', $parms{'type'}) if Trace::check(Trace::DEBUG6);

    my @allThings = map { values(%{ $_->{'things'} }) } values(%{ $page->{'hosts'} });
    my @change = grep { $_->{'config'}->{'type'} eq $parms{'type'} } @allThings;

    foreach my $thing (@change) {
	$thing->{'saved'} = 0;
	$thing->{'applied'} = 0;
    }
    $page->activateButtons();
}

sub displayOkMessage {
    my $page = shift;
    my $message = shift;

    push @{ $page->{'messages'} }, $message;
    if ($#{ $page->{'messages'} } == 0 ) {
	$page->{'messageLabel'}->set_text($message);	
    }
    $page->display;
}

sub messageOk {
    my $widget = shift;
    my $page = shift;

    shift @{ $page->{'messages'} };
    if ($#{ $page->{'messages'} } == -1) {
	$page->{'messageBox'}->hide;
    }
    else {
	my $message = $page->{'messages'}->[0];
	$page->{'messageLabel'}->set_text($message);	
    }
    $page->display;
    return 1;
}

sub beginBusySequence {
    my $page = shift;
    my $message = shift;
    my $work = shift;

    Trace::debug('Beginning busy sequence with %d iterations', $work) if Trace::check(Trace::DEBUG6);

    $page->{'busy'}->{'progressBar'}->set_value(0);
    $page->{'busy'}->{'count'} = 0;
    $page->{'busy'}->{'work'} = $work;
    $page->{'busy'}->{'message'} = $message;
    $page->{'busy'}->{'label'}->set_text($message.'....');
    $page->display;
}

sub endBusySequence {
    my $page = shift;

    Trace::debug('Finished busy sequence after %d iterations', $page->{'busy'}->{'count'}) if Trace::check(Trace::DEBUG6);

    $page->{'busy'}->{'work'} = 0;
    $page->{'busy'}->{'label'}->set_text($page->{'busy'}->{'message'}.'....done.');
    $page->gui->at(to => $page->gui, name => '_delayedBusyClear', time => time + 2);
}

sub incBusySequence {
    my $page = shift;

    Trace::debug('One busy iteration') if Trace::check(Trace::DEBUG6);

    $page->{'busy'}->{'count'}++;
    if ($page->{'busy'}->{'work'}) {
	$page->{'busy'}->{'progressBar'}->set_value(100.0 * ($page->{'busy'}->{'count'} / $page->{'busy'}->{'work'}));
    }
}

sub delayedBusyClear {
    my $page = shift;
    $page->display;
}

sub display {
    # Display the proper widget order busy, ok, notebook
    my $page = shift;

    if ($page->{'busy'}->{'work'}) {
	$page->{'notebook'}->hide;
	$page->{'messageBox'}->hide;
	$page->{'busy'}->{'box'}->show;
    }
    elsif ($#{ $page->{'messages'} } != -1) {
	$page->{'notebook'}->hide;
	$page->{'messageBox'}->show;
	$page->{'busy'}->{'box'}->hide;
    }
    else {
	$page->{'notebook'}->show;
	$page->{'messageBox'}->hide;
	$page->{'busy'}->{'box'}->hide;
    }
}

sub _test {
    my $page = shift;
    my $file = $_[ $#_ ];
    pop @_;

    foreach my $test (@_) {

	if (($test eq 'e') && (! -e $file)) {
	    $page->displayOkMessage(
			    $file.' does not exist.'."\n"
			    .'You can use the Choose... button to create it.');
	    return 0;
	}

	if (($test eq 'd') && (! -d $file)) {
	    $page->displayOkMessage(
		    $file.' is not a directory.');
	    return 0;
	}

	if (($test eq 'r') && (! -r $file)) {
	    $page->displayOkMessage(
		    'You do not have permission to read '.$file);
	    return 0;
	}

	if (($test eq 'w') && (! -w $file)) {
	    $page->displayOkMessage(
		    'You do not have permission to write to '.$file);
	    return 0;
	}
    }
    return 1;
}

1;
