# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.7 $

package Info::Album;

sub table { 'albums'; };

use Thing::Info;
@ISA = qw( Thing::Info);
use Info::Persist;
use Info::Artist;
use Info::Track;
use RepoH;
use Trace;
use strict;

sub find {
    my $proto = shift;
    my $type = ref($proto) || $proto;
    my %parms = @_;

    if (!defined($parms{'owner'}) || ($parms{'owner'}->name ne Repo::NAME)) {
	Trace::error('Object must be owned by Repo');
	return undef;
    }

    if (Info::Persist::configured) {

	# Local database in use for persistence

	my @selects;
	if (defined($parms{'ioid'})) {
	    push @selects, 'ioid='.$parms{'ioid'};
	}
	if (defined($parms{'artist_ioid'})) {
	    push @selects, 'artist_ioid='.$parms{'artist_ioid'};
	}
	elsif (defined($parms{'artist'})) {
	    push @selects, 'artist_ioid='.$parms{'artist'}->ioid;
	}
	if (defined($parms{'name'})) {
	    push @selects, 'name='.Info::Persist::quote($parms{'name'});
	}
	my $select = Info::Persist::buildSelect(table => 'albums', wheres => \@selects);

	my $dbh = Info::Persist::connect;
	return if (!defined($dbh));
    
	my $sth;
	if (!defined($sth = $dbh->prepare($select)) || !defined($sth->execute)) {
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return;
	}
	
	my $data = $sth->fetchall_arrayref({});
	$sth->finish;
	if (defined($data) && ($#{ $data } >= 0)) {
	    Trace::debug('Found %d albums', $#{ $data } + 1) if Trace::check(Trace::DEBUG6);
	    if (!wantarray) {
		my $album = $type->new(owner => $parms{'owner'}, ioid => $data->[0]->{'ioid'}, data => $data->[0]);
		Trace::debug('Found album: %s: %s', $album->name, $album->ioid) if Trace::check(Trace::DEBUG4);
		if (defined($parms{'artist'})) {
		    $album->get->{'artist'} = $parms{'artist'};
		}
		return $album;
	    }
	    else {
		my @albums = map { $type->new(owner => $parms{'owner'}, ioid => $_->{'ioid'}, data => $_) } @{ $data };
		if (defined($parms{'artist'})) {
		    map { $_->get->{'artist'} = $parms{'artist'} } @albums;
		}
		return @albums;
	    }
	}
    }

    Trace::debug('No matching locally persistent album') if Trace::check(Trace::DEBUG4);

    if ($parms{'create'}) {
	$parms{'_bypassfind'} = 1;
	delete $parms{'create'};
	return $type->create(%parms);
    }
    return;
}

sub create {
    my $proto = shift;
    my $type = ref($proto) || $proto;
    my %parms = @_;

    if (!defined($parms{'owner'}) || ($parms{'owner'}->name ne Repo::NAME)) {
	Trace::error('Object must be owned by Repo');
	return undef;
    }

    my $album;
    my $artist_ioid;
    my $artist = $parms{'artist'};

    if (!defined($parms{'name'})) {
	Trace::debug('Missing album name for create') if Trace::check(Trace::DEBUG4);
	return undef;
    }

    if (Info::Persist::configured) {

	# Local database in use for persistence

	$artist_ioid = defined($parms{'artist_ioid'}) ? $parms{'artist_ioid'} : defined($parms{'artist'}) ? $parms{'artist'}->ioid : undef;
	if (!defined($artist_ioid)) {
	    Trace::debug('Missing required artist IOID') if Trace::check(Trace::DEBUG4);
	    return undef;
	}

	if (!$parms{'_bypassfind'}) {
	    my $album = $type->find(owner => $parms{'owner'}, artist_ioid => $artist_ioid, name => $parms{'name'});
	    return $album if (defined($album));
	}

	# Force uniqueness of album name by artist
	
	$artist = $parms{'artist'} || Info::Artist->find(owner => $parms{'owner'}, ioid => $artist_ioid);
	return if (!defined($artist));
	AGAIN: foreach my $album (Info::Album->find(owner => $parms{'owner'}, artist_ioid => $artist_ioid)) {
	    if ($parms{'name'} eq $album->name) {
		Trace::debug('Album exists with the same name (%s) for %s', $parms{'name'}, $artist->name) if Trace::check(Trace::DEBUG4);
		$parms{'name'} = $parms{'name'}.' ('.(int(rand(99999))).')';
		redo AGAIN;
	    }
	}

	my $dbh = Info::Persist::connect;
	return undef if (!defined($dbh));
    
	my $cmd = 'INSERT INTO albums ';
	my @fields;
	my @values;
	
	push @fields, 'artist_ioid';
	push @values, $artist_ioid;
	push @fields, 'name';
	push @values, Info::Persist::quote($parms{'name'});
	push @fields, 'category';
	push @values, Info::Persist::quote(defined($parms{'category'}) ? $parms{'category'} : 'Unknown');
	if (defined($parms{'cddb_id'})) {
	    push @fields, 'cddb_id';
	    push @values, Info::Persist::quote($parms{'cddb_id'});
	}
	if (defined($parms{'running_time'})) {
	    push @fields, 'running_time';
	    push @values, $parms{'running_time'};
	}
	if (defined($parms{'track_count'})) {
	    push @fields, 'track_count';
	    push @values, $parms{'track_count'};
	}
	
	$cmd .= '('.join(', ', @fields).')';
	$cmd .= ' VALUES ('.join(', ', @values).');';
	
	if (!$dbh->do($cmd)) {
	    Trace::error($cmd);
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return undef;
	}

	# Find the assigned ioid

	my @albums = $type->find(owner => $parms{'owner'}, artist => $artist, name => $parms{'name'});
	$album = shift @albums; # Take the lowest ioid
	if ($#albums >= 0) {
	    # Race, something else created this album
	    map { $_->dbDelete } @albums;
	}
    }
    else {
	if (!defined($artist)) {
	    Trace::debug('Missing artist for create') if Trace::check(Trace::DEBUG4);
	    return undef;
	}
	$album = $type->new(owner => $parms{'owner'}, 
			    data =>
			    {
				artist => $artist,
				artist_ioid => $artist->ioid,
				name => $parms{'name'},
				category => $parms{'category'},
				cddb_id => $parms{'cddb_id'},
				running_time => $parms{'running_time'},
				track_count => $parms{'track_count'},
			    },
			    );
    }
    Trace::debug('Created album: %s: %s', $album->name, $album->ioid) if Trace::check(Trace::DEBUG4);
    return $album;
}

# save() - Save modified (getN) values.

sub save {
    my $album = shift;

    return 1 if (!Info::Persist::configured);

    Trace::debug('Update album: %s: %s', $album->name, $album->ioid) if Trace::check(Trace::DEBUG3);
	
    my $dbh = Info::Persist::connect;
    return if (!defined($dbh));

    # Still have to force uniqueness of album name by artist (one might sneak in through Info::Cd::remoteChoose)

    if ($album->saveCheck('name')) {
	my $newname = $album->getN->{'name'};
	my $checkartist;
	if (defined($album->getN->{'artist_ioid'}) && ($album->get->{'artist_ioid'} != $album->getN->{'artist_ioid'})) {
	    $checkartist = Info::Artist->find(owner => $album->owner, ioid => $album->getN->{'artist_ioid'});
	}
	else {
	    $checkartist = Info::Artist->find(owner => $album->owner, ioid => $album->get->{'artist_ioid'});
	}
	AGAIN: foreach my $checkalbum (Info::Album->find(owner => $album->owner, artist_ioid => $checkartist->ioid)) {
	    next AGAIN if ($checkalbum->ioid == $album->ioid);
	    if ($newname eq $checkalbum->name) {
		Trace::debug('Album exists with the same name (%s) for %s', $newname, $checkartist->name) if Trace::check(Trace::DEBUG4);
		$newname = $newname.' ('.(int(rand(99999))).')';
		redo AGAIN;
	    }
	}
	$album->getN->{'name'} = $newname;
    }
    
    my $cmd = 'UPDATE albums SET ';
    my @sets;
    if ($album->saveCheck('name')) {
	push @sets, 'name='.Info::Persist::quote($album->getN->{'name'});
    }
    if ($album->saveCheck('artist_ioid')) {
	push @sets, 'artist_ioid='.$album->getN->{'artist_ioid'};
    }
    if ($album->saveCheck('category')) {
	push @sets, 'category='.Info::Persist::quote($album->getN->{'category'});
    }
    if ($album->saveCheck('cddb_id')) {
	push @sets, 'cddb_id='.Info::Persist::quote($album->getN->{'cddb_id'});
    }
    if ($album->saveCheck('running_time')) {
	push @sets, 'running_time='.$album->getN->{'running_time'};
    }
    if ($album->saveCheck('track_count')) {
	push @sets, 'track_count='.$album->getN->{'track_count'};
    }
    if ($album->saveCheck('play_count')) {
	push @sets, 'play_count='.$album->getN->{'play_count'};
    }
    if ($album->saveCheck('play_date')) {
	push @sets, 'play_date='.Info::Persist::quote($album->getN->{'play_date'});
    }
    return 1 if (!@sets);

    $cmd .= join(', ', @sets);
    $cmd .= ' WHERE ioid='.$album->ioid.';';

    if (!$dbh->do($cmd)) {
      Trace::error($DBI::errstr);
      Info::Persist::disconnect;
      return;
    }
    return 1;
}

sub delete {
    my $album = shift;

    Trace::debug('Delete album: %s: %s', $album->name, $album->ioid) if Trace::check(Trace::DEBUG3);

    # The repo does the work
    Thing::context->message(name => Repo::Delete, to => $album->owner, parms => { album => $album });
    return 1;
}

sub dbDelete {
    my $album = shift;

    Trace::debug('Delete album (database part): %s: %s', $album->name, $album->ioid) if Trace::check(Trace::DEBUG3);

    my $dbh = Info::Persist::connect;
    return if (!defined($dbh));

    if (!$dbh->do('DELETE FROM albums WHERE ioid='.$album->ioid.';')) {
	Trace::error($DBI::errstr);
	Info::Persist::disconnect;
	return 0;
    }

    if (!$dbh->do('DELETE FROM tracks WHERE album_ioid='.$album->ioid.';')) {
	Trace::error($DBI::errstr);
	Info::Persist::disconnect;
	return 0;
    }

    return 1;
}

sub artist_ioid { $_[0]->get->{'artist_ioid'}; }

sub name { $_[0]->get->{'name'}; }

sub running_time { $_[0]->get->{'running_time'}; }

sub track_count { $_[0]->get->{'track_count'}; }

sub cddb_id { $_[0]->get->{'cddb_id'}; }

sub category { $_[0]->get->{'category'}; }

sub play_date { $_[0]->get->{'play_date'}; }

sub play_count { $_[0]->get->{'play_count'}; }

sub artist {
    my $album = shift;
    return $album->get->{'artist'} if defined($album->get->{'artist'});
    if (!Info::Persist::configured) {
	Trace::error('Artist must be specified at creation for non persistence');
	return;
    }
    # Albums are allowed to change artist so this must be kept in the sync space
    $album->get->{'artist'} = Info::Artist->find(ioid => $album->artist_ioid, owner => $album->owner);
    if ($album->ismonitor) {
	$album->get->{'artist'} = $album->get->{'artist'}->monitor;
    }
    return $album->get->{'artist'};
}

# discover_play_date() - Find the play date for an album, based on the
# date of the least recently played track.

sub discover_play_date {
    my $album = shift;
    if (!Info::Persist::configured) {
	Trace::debug('Cannot determin album play date with non persistence') if (Trace::check(Trace::DEBUG3));
	return;
    }
    return _discover_play_date($album->owner, $album->ioid);
}

sub _discover_play_date {
    my $owner = shift;
    my $album_ioid = shift;

    my @tracks = Info::Track->find(owner => $owner, album_ioid => $album_ioid);
    my $date;
    foreach my $track (@tracks) {
	next if ($track->index == Info::Track::LEADOUTTN);
	next if ($track->index == Info::Track::MSTN);
	if (!defined($track->play_date)) {
	    return;
	}
	if (!defined($date) || ($date gt $track->play_date)) {
	    $date = $track->play_date;
	}
	Trace::debug('Checking date: %s', $track->play_date) if Trace::check(Trace::DEBUG6);
    }
    Trace::debug('Least recent date: %s', $date) if Trace::check(Trace::DEBUG6);
    return $date;
}

# discover_play_count() - Find the play count for an album, based on the
# count of the least played track.

sub discover_play_count {
    my $album = shift;
    if (!Info::Persist::configured) {
	Trace::debug('Cannot determin album play count with non persistence') if (Trace::check(Trace::DEBUG3));
	return;
    }
    return _discover_play_count($album->owner, $album->ioid);
}

sub _discover_play_count {
    my $owner = shift;
    my $album_ioid = shift;

    my @tracks = Info::Track->find(owner => $owner, album_ioid => $album_ioid);
    my $count;
    foreach my $track (@tracks) {
	next if ($track->index == Info::Track::LEADOUTTN);
	next if ($track->index == Info::Track::MSTN);
	if (!defined($track->play_count)) {
	    return;
	}
	if (!defined($count) || ($count > $track->play_count)) {
	    $count = $track->play_count;
	}
    }
    return $count;
}

sub change_category {
    my $album = shift;
    my $category = shift;

    return 1 if ($category eq $album->category);
    $album->setN;
    $album->getN->{'category'} = $category;
    $album->save;
    $album->sync;
    return 1;
}

sub change_name {
    my $album = shift;
    my $name = shift;

    return 1 if ($name eq $album->name);

    if (Info::Persist::configured) {

	# Cannot use the name of an existing album

	my $exists = Info::Album->find(owner => $album->owner, name => $name, artist_ioid => $album->artist_ioid, create => 0);
	if ($exists) {
	    Trace::error('There is already an album named %s for the artist %s', $name, $album->artist->name);
	    return 0;
	}
    }

    # The repo does the work

    Thing::context->message(name => Repo::Change, to => $album->owner, parms => { object => $album, name => $name });
    return 1;
}

sub reparent {
    my $album = shift;
    my $newartist = shift;

    return 1 if ($album->artist->ioid eq $newartist->ioid);

    Thing::context->message(name => Repo::Reparent, to => $album->owner, parms => { object => $album, parent => $newartist });
    return 1;
}

sub monitor {
    my $album = shift;
    $album = Thing::Info::monitor($album, @_);
    if (defined($album->get->{'artist'})) {
	$album->get->{'artist'} = $album->get->{'artist'}->monitor;
    }
    return $album;
}

1;
