# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.7 $

package Info::Artist;

sub table { 'artists'; };

use Thing::Info;
@ISA = qw( Thing::Info );
use Info::Persist;
use Info::Album;
use Info::Track;
use RepoH;
use Trace;
use strict;

sub find {
    my $proto = shift;
    my $type = ref($proto) || $proto;
    my %parms = @_;

    if (!defined($parms{'owner'}) || ($parms{'owner'}->name ne Repo::NAME)) {
	Trace::error('Object must be owned by Repo');
	return undef;
    }

    if (Info::Persist::configured) {

	# Local database in use for persistence

	my @selects;
	if (defined($parms{'ioid'})) {
	    push @selects, 'ioid='.$parms{'ioid'};
	}
	if (defined($parms{'name'})) {
	    push @selects, 'name='.Info::Persist::quote($parms{'name'});
	}
	my $select = Info::Persist::buildSelect(table => 'artists', wheres => \@selects);

	my $dbh = Info::Persist::connect;
	return if (!defined($dbh));
	
	my $sth;
	if (!defined($sth = $dbh->prepare($select)) || !defined($sth->execute)) {
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return;
	}

	my $data = $sth->fetchall_arrayref({});
	$sth->finish;
	if (defined($data) && ($#{ $data } >= 0)) {
	    Trace::debug('Found %d artists', $#{ $data } + 1) if Trace::check(Trace::DEBUG6);
	    if (!wantarray) {
		my $artist = $type->new(owner => $parms{'owner'}, ioid => $data->[0]->{'ioid'}, data => $data->[0]);
		Trace::debug('Found artist: %s: %s', $artist->name, $artist->ioid) if Trace::check(Trace::DEBUG4);
		return $artist;
	    }
	    else {
		return map { $type->new(owner => $parms{'owner'}, ioid => $_->{'ioid'}, data => $_) } @{ $data };
	    }
	}
    }

    Trace::debug('No matching locally persistent artist') if Trace::check(Trace::DEBUG4);

    if ($parms{'create'}) {
	$parms{'_bypassfind'} = 1;
	delete $parms{'create'};
	return $type->create(%parms);
    }
    return;
}

sub create {
    my $proto = shift;
    my $type = ref($proto) || $proto;
    my %parms = @_;

    if (!defined($parms{'owner'}) || ($parms{'owner'}->name ne Repo::NAME)) {
	Trace::error('Object must be owned by Repo');
	return undef;
    }

    if (!defined($parms{'name'})) {
	Trace::debug('Missing name for create') if Trace::check(Trace::DEBUG4);
	return undef;
    }

    my $artist;
    if (Info::Persist::configured) {

	# Local database in use for persistence

	if (!defined($parms{'_bypassfind'})) {
	    my $artist = $type->find(owner => $parms{'owner'}, name => $parms{'name'});
	    return $artist if (defined($artist));
	}

	my $dbh = Info::Persist::connect;
	return undef if (!defined($dbh));
	
	my $cmd = 'INSERT INTO artists ';
	my @fields;
	my @values;
	
	push @fields, 'name';
	push @values, Info::Persist::quote($parms{'name'});
	
	$cmd .= '('.join(', ', @fields).')';
	$cmd .= ' VALUES ('.join(', ', @values).');';
	
	if (!$dbh->do($cmd)) {
	    Trace::error($cmd);
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return undef;
	}
	
	# Find the assigned ioid
	
	my @artists = $type->find(owner => $parms{'owner'}, name => $parms{'name'});
	$artist = shift @artists; # Take the lowest ioid
	if ($#artists >= 0) {
	    # Race, something else created this artist
	    map { $_->dbDelete } @artists;
	}
    }
    else {
	$artist = $type->new(owner => $parms{'owner'}, 
			     data =>
			     {
				 name => $parms{'name'},
			     }
			     );
    }
    Trace::debug('Created artist: %s: %s', $artist->name, $artist->ioid) if Trace::check(Trace::DEBUG4);
    return $artist;
}

# save() - Save modified (getN) values.

sub save {
    my $artist = shift;

    return 1 if (!Info::Persist::configured);

    Trace::debug('Update artist: %s: %s', $artist->name, $artist->ioid) if Trace::check(Trace::DEBUG3);
	
    my $dbh = Info::Persist::connect;
    return if (!defined($dbh));

    my $cmd = 'UPDATE artists SET ';
    my @sets;
    if ($artist->saveCheck('name')) {
	push @sets, 'name='.Info::Persist::quote($artist->getN->{'name'});
    }
    if ($artist->saveCheck('play_date')) {
	push @sets, 'play_date='.Info::Persist::quote($artist->getN->{'play_date'});
    }
    return 1 if (!@sets);

    $cmd .= join(', ', @sets);
    $cmd .= ' WHERE ioid='.$artist->ioid.';';

    if (!$dbh->do($cmd)) {
      Trace::error($DBI::errstr);
      Info::Persist::disconnect;
      return;
    }

    return 1;
}

sub delete {
    my $artist = shift;

    Trace::debug('Delete artist: %s: %s', $artist->name, $artist->ioid) if Trace::check(Trace::DEBUG3);

    # The repo does the work
    Thing::context->message(name => Repo::Delete, to => $artist->owner, parms => { artist => $artist });
    return 1;
}

sub dbDelete {
    my $artist = shift;

    Trace::debug('Delete artist (database part): %s: %s', $artist->name, $artist->ioid) if Trace::check(Trace::DEBUG3);

    my $dbh = Info::Persist::connect;
    return if (!defined($dbh));

    if (!$dbh->do('DELETE FROM tracks WHERE 
                   artists.ioid='.$artist->ioid.' AND
                   albums.artist_ioid=artists.ioid AND
                   tracks.album_ioid=albums.ioid;')) {
	Trace::error($DBI::errstr);
	Info::Persist::disconnect;
	return 0;
    }

    if (!$dbh->do('DELETE FROM albums WHERE artist_ioid='.$artist->ioid.';')) {
	Trace::error($DBI::errstr);
	Info::Persist::disconnect;
	return 0;
    }

    if (!$dbh->do('DELETE FROM artists WHERE ioid='.$artist->ioid.';')) {
	Trace::error($DBI::errstr);
	Info::Persist::disconnect;
	return 0;
    }

    return 1;
}

sub name { $_[0]->get->{'name'}; }

sub play_date { $_[0]->get->{'play_date'}; }

# discover_play_date() - Find the play date for an artist, based on the
# most recently played track.

sub discover_play_date {
    my $artist = shift;
    if (!Info::Persist::configured) {
	Trace::debug('Cannot determin artist play date with non persistence') if (Trace::check(Trace::DEBUG3));
	return;
    }

    my @tracks = Info::Persist::find(owner => $artist->owner, type => 'Info::Track', 
				     select => 'SELECT * FROM tracks WHERE 
                                                artists.ioid='.$artist->ioid.' AND
                                                albums.artist_ioid=artists.ioid AND
                                                tracks.album_ioid=albums.ioid;' );
    return if (!@tracks);

    my $date;
    foreach my $track (@tracks) {
	next if ($track->index == Info::Track::LEADOUTTN);
	next if ($track->index == Info::Track::MSTN);
	next if (!defined($track->play_date));
	if (!defined($date) || ($date lt $track->play_date)) {
	    $date = $track->play_date;
	}
    }
    return $date;
}

sub change_name {
    my $artist = shift;
    my $name = shift;

    return 1 if ($name eq $artist->name);

    if (Info::Persist::configured) {

	# Cannot use the name of an existing artist

	my $exists = Info::Artist->find(owner => $artist->owner, name => $name, create => 0);
	if ($exists) {
	    Trace::error('An artist named %s already exists', $name);
	    return 0;
	}
    }

    # The repo does the work

    Thing::context->message(name => Repo::Change, to => $artist->owner, parms => { object => $artist, name => $name });
    return 1;
}

1;
