# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.7 $

package Info::Cd;
use Trace;
use Info::Album;
@ISA = qw( Info::Album );
use RepoH;
use Cddb;
use Thing;
use POSIX qw( _exit );
use Info::Track;
use Info::Artist;
use strict;

# Remote lookup status

sub Never     { 'Never'; }
sub Underway  { 'Underway'; }  # Current operation.
sub Success   { 'Success'; }
sub Failure   { 'Failure'; }
sub Multiple  { 'Multiple'; }  # Multiple responses were returned.

use vars qw ( %_remoteConnectArgs );

# configure() - Configure the remote database connection methods.

sub configure {
    my %parms = @_;

    if (defined($parms{'remotes'}) && ($#{ $parms{'remotes'} } != -1)) {
	%_remoteConnectArgs = %{ $parms{'remotes'}->[0] };
	# Added post 1.2.0, default to on if not in config
	if (!defined($_remoteConnectArgs{'removeWhitespace'})) {
	    $_remoteConnectArgs{'removeWhitespace'} = 1;
	}
	if (!defined($_remoteConnectArgs{'uppercaseWords'})) {
	    $_remoteConnectArgs{'uppercaseWords'} = 1;
	}
    }
    return _cddbInit();
}

# find() - Find a CD from its track information.

sub find {
    my $proto = shift;
    my $type = ref($proto) || $proto;
    
    my %parms = @_;

    if (!defined($parms{'owner'}) || ($parms{'owner'}->name ne Repo::NAME)) {
	Trace::error('Object must be owned by Repo');
	return undef;
    }

    if (!defined($parms{'trackinfo'})) {
	return $type->SUPER::find(@_);
    }

    # Work out the cddb_id, running time, track count

    my $tracks = $parms{'trackinfo'};
    my $firstIndex;
    my $lastIndex;
    my @audioIndices = ();
    foreach my $index (keys %{ $tracks }) {
	next if ($index == Info::Track::LEADOUTTN);
	next if ($index == Info::Track::MSTN);
	if (!$tracks->{ $index }->{'data'}) {
	    push @audioIndices, $index;
	}
	else {
	    Trace::debug('Data track: %d', $index) if Trace::check(Trace::DEBUG6);
	}
	if (!defined($firstIndex)) {
	    $firstIndex = $index;
	    $lastIndex = $index;
	    next;
	}
	if ($firstIndex > $index) {
	    $firstIndex = $index;
	}
	elsif ($lastIndex < $index) {
	    $lastIndex = $index;
	}
    }
    my $cddb_id;
    my $trackCount = 0;
    my $runningTime = $tracks->{ $firstIndex }->{'lba'} / 75;
    my $cddbN = 0;
    my $cddbT = 0;

    for (my $index = $lastIndex; $index >= $firstIndex; $index--) {

	my $track = $tracks->{ $index };
	my $nextTrack;
	if ($index == $lastIndex) {
	    $nextTrack = $tracks->{ Info::Track::LEADOUTTN };
	}
	else {
	    $nextTrack = $tracks->{ $index + 1 };
	}

	$track->{'duration'} = ($nextTrack->{'lba'} - $track->{'lba'}) / 75;
	$runningTime += $track->{'duration'};

	# Any data track takes space from the previous audio track if
	# its within the `multisession'. Don't know why but this is
	# what I've figured from experimentation and cdparanoia code.
	
	my $correction = 0;
	if ($tracks->{ Info::Track::MSTN }->{'data'} &&
	    $nextTrack->{'data'} &&
	    !$track->{'data'} &&
	    ($nextTrack->{'lba'} > 
	     ($tracks->{ Info::Track::MSTN }->{'lba'} - 14400))) {
	    $correction = 11400 / 75;
	}
	$track->{'duration'} = int($track->{'duration'} - $correction);

	my $cddbSum = int($track->{'lba'} / 75);
	while ($cddbSum > 0) {
	    $cddbN += ($cddbSum % 10);
	    $cddbSum = int($cddbSum / 10);
	}
	$trackCount++;
    }

    $runningTime = int($runningTime);

    $cddbT = (int($tracks->{ Info::Track::LEADOUTTN }->{'lba'} / 75) - int($tracks->{ $firstIndex }->{'lba'} / 75));
    $cddb_id = sprintf("%08x", (($cddbN % 0xff) << 24) | ($cddbT << 8) | $trackCount);

    my $cd;
    if (Info::Persist::configured) {

	# Local database in use for persistence

	my $dbh = Info::Persist::connect;
	return undef if (!defined($dbh));
	
	my $sth;
	if (!defined($sth = $dbh->prepare('SELECT * FROM albums WHERE (cddb_id='.Info::Persist::quote($cddb_id).');')) ||
	    !defined($sth->execute)) {
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return undef;
	}

	my $cds = $sth->fetchall_arrayref({});
	$sth->finish;
	
	Trace::debug('Found %d matching cds for cddb id: %s', $#{ $cds } + 1, $cddb_id) if Trace::check(Trace::DEBUG4);
    
	# Match these against the tracks
	
	CD: for (my $cdidx = 0; $cdidx <= $#{ $cds }; $cdidx++) {
	    if (!defined($sth = $dbh->prepare('SELECT index, lba FROM tracks WHERE (album_ioid='.$cds->[$cdidx]->{'ioid'}.');')) || 
		!defined($sth->execute)) {
		Trace::error($DBI::errstr);
		Info::Persist::disconnect;
		return undef;
	    }
	    my $matchtracks = $sth->fetchall_arrayref({});
	    $sth->finish;
	    
	    if (!defined($matchtracks) || ($#{ $matchtracks } == -1)) {
		Trace::debug('Cd %d:%s does not match', $cds->[$cdidx]->{'ioid'}, $cds->[$cdidx]->{'name'}) if Trace::check(Trace::DEBUG4);
		next CD;
	    }
	    
	    foreach my $track (values(%{ $tracks })) {
		if (!grep { ($track->{'index'} == $_->{'index'}) && ($track->{'lba'} == $_->{'lba'}) } @{ $matchtracks }) {
		    Trace::debug('Mismatch on track %d', $track->{'index'}) if Trace::check(Trace::DEBUG4);
		    Trace::debug('Cd %d:%s does not match', $cds->[$cdidx]->{'ioid'}, $cds->[$cdidx]->{'name'}) if Trace::check(Trace::DEBUG4);
		    next CD;
		}
	    }	    
	    
	    # This CD is a match
	    $cd = $type->new(owner => $parms{'owner'}, ioid => $cds->[$cdidx]->{'ioid'}, data => $cds->[$cdidx]);
	    Trace::debug('Found matching cd: %d:%s', $cd->ioid, $cd->get->{'name'}) if Trace::check(Trace::DEBUG4);
	    last CD;
	}
    }

    if (!defined($cd)) {
	Trace::debug('No matching locally persistent cd found') if Trace::check(Trace::DEBUG4);
	if (defined($parms{'create'})) {
	    my $artist = Info::Artist->find(owner => $parms{'owner'}, name => 'Unknown', create => 1);
	    return if (!defined($artist));
	    $cd = $type->create(owner => $parms{'owner'}, name => 'Unknown', artist => $artist, cddb_id => $cddb_id, 
				category => 'Unknown', track_count => $trackCount, running_time => $runningTime, _bypassfind => 1);
	    return if (!defined($cd));
	    $cd->{'tracksH'} = {};
	    foreach my $track (values %{ $tracks }) {
		$cd->{'tracksH'}->{ $track->{'index'} } = Info::Track->create(owner => $parms{'owner'}, name => 'Unknown', 
									      index => $track->{'index'}, album => $cd, 
									      lba => $track->{'lba'}, data => $track->{'data'}, 
									      duration => $track->{'duration'}, _bypassfind => 1);
	    }
	}
    }

    $cd->{'firstTrackIndex'} = $firstIndex;
    $cd->{'lastTrackIndex'} = $lastIndex;
    @audioIndices = sort { $a <=> $b } @audioIndices;
    $cd->{'audioIndices'} = \@audioIndices;
    $cd->{'firstAudioTrackIndex'} = $cd->{'audioIndices'}->[0];
    $cd->{'lastAudioTrackIndex'} = $cd->{'audioIndices'}->[ $#{ $cd->{'audioIndices'} } ];

    $cd->setN;
    $cd->getN->{'remoteResult'} = { status => Never };

    $cd->sync; # Sync so that monitors will get this data

    return $cd;
}

# remoteLookup() - Discovers more information about a CD from a remote
# database. The CD is updated asynchronously.

sub remoteLookup {
    my $cd = shift;
    my %parms = @_;
    Trace::point;

    if (defined($cd->get->{'remoteResult'}) && ($cd->get->{'remoteResult'}->{'status'} eq Underway)) {
	Trace::error('Remote operation is already in progress');
	return;
    }
    
    my $result = { status => Underway, message => 'Doing remote lookup' };
    $cd->setN;
    $cd->getN->{'remoteResult'} = $result;
    $cd->sync;

    if (!Cddb::configured) {
	$cd->getN->{'remoteResult'}->{'status'} = Failure;
	$cd->getN->{'remoteResult'}->{'message'} = 'Remote database is not configured';
	$cd->sync;
	return;
    }

    if ($parms{'forking'}) {
	Info::Persist::disconnect;
	my $fork = Thing->context->fork;
	return 1 if ($fork != 0);
	$cd->_remoteLookup(@_);
	$cd->sync;
	_exit(1);
    }
    else {
	$cd->_remoteLookup(@_);
	$cd->sync;
	return;
    }
}

sub _remoteLookup {
    my $cd = shift;
    my %parms = @_;
    Trace::point;

    my $matches = [];
    my $result = $cd->getN->{'remoteResult'};
    $result->{'matches'} = $matches;
    $result->{'status'} = Failure;

    # Work out the frame offsets

    my @tracks = $cd->tracks;
    my $frameOffsets = '';
    foreach my $track (@tracks) {
	next if ($track->index == Info::Track::LEADOUTTN);
	next if ($track->index == Info::Track::MSTN);
	$frameOffsets .= $track->lba;
	$frameOffsets .= ' ';
    }

    # Match the band and CD title, this can be tricky so rather than get
    # too tangled in regular expressions lets take a step by step approach.

    # This assumes the CDDB recommendations at www.cddb.org are followed,
    # however erroneous entries are also handled.

    # In general band name and title should be seperated by ' / '

    if (!_cddbInit()) {
	$result->{'message'} = 'Problems with CDDB configuration:'."\n".Trace::lastError;
	Trace::debug($result->{'message'}) if Trace::check(Trace::DEBUG4);
	return;
    }

    my $cddb_id = $cd->cddb_id;
    my $str = Cddb::command(Command => sprintf('cddb query %s %d %s%d', $cd->cddb_id, $cd->track_count, $frameOffsets, $cd->running_time));
    if (!defined($str)) {
	$result->{'message'} = Trace::lastError;
	return;
    }
    elsif ($str =~ m/^200\s(\w+)\s$cddb_id\s(.*)$/) {
	Trace::debug('Single match from remote') if Trace::check(Trace::DEBUG6);
	$matches->[0] = { category => $1, cddb_id => $cd->cddb_id, data => $2, str => $str };
    }
    elsif (($str =~ m/^210/) || ($str =~ m/^211/)) {
	Trace::debug('Multiple or inexact match from remote') if Trace::check(Trace::DEBUG6);
	$result->{'status'} = Multiple;
	my @strs = split(/\n/, $str);
	shift @strs;
	foreach my $str (@strs) {
	    if ($str =~ m/^(\w+)\s(\w+)\s(.*)$/) {
		push @{ $matches }, { category => $1, cddb_id => $2, data => $3, str => $str };
	    }
	}
    }
    elsif ($str =~ m/^202/) {
	$result->{'status'} = Success;
	$result->{'message'} = 'No match found in the remote database';
	Trace::debug($result->{'message'}) if Trace::check(Trace::DEBUG4);
	return 1;
    }
    else {
	Trace::error($str) if defined($str);
	$result->{'message'} = 'Unacceptable response from CDDB:'."\n".$str;
	return 1;
    }

    if ($#{ $matches } == -1) {
	$result->{'message'} = 'Unacceptable response from CDDB:'."\n".$str;
	Trace::debug($result->{'message'}) if Trace::check(Trace::DEBUG4);
	return 1;
    }

    foreach my $match (@{ $matches }) {
	my @strs = split(/ \/ /, $match->{'data'});
	my $band;
	my $title;
    
	if ($#strs == 1) {
	    $band = $strs[0];
	    $title = $strs[1];
	}
	elsif ($#strs > 1) {
	
	    # However a ' / ' may exist in the band name or title, if its in
	    # the band name we are SOL and I'm assuming that CDDB actively
	    # clean up such entries. We'll treat extra ' / ' as part of the
	    # title.
	
	    $band = $strs[0];
	    splice(@strs, 0, 1);
	    $title = join(' / ', @strs);
	}
	else {
	
	    # No ' / ' in the string can be because the band and title match
	    # and the entry follows the CDDB recommendation for such
	    # things. More likely however it is because the / was not
	    # surrounded by the recommeded spaces.
	
	    if ($match->{'data'} !~ m/\//) {
		$band = $match->{'data'};
		$title = $match->{'data'};
	    }
	    else {
	    
		@strs = split(/\//, $match->{'data'});
	    
		if ($#strs == 1) {
		    $band = $strs[0];
		    $title = $strs[1];
		}
		elsif ($#strs > 1) {
		    $band = $strs[0];
		    splice(@strs, 0, 1);
		    $title = join('/', @strs);
		}
	    }
	}
    
	if (defined($band) && defined($title)) {
	    
	    # Do various clean ups of the CDDB data
	    foreach $str ($band, $title) {
		if ($_remoteConnectArgs{'removeWhitespace'}) {
		    # Remove proceeding and trailing whitespace
		    $str = _removeWhitespace($str);
		    $str = _removeDoubleSpace($str);
		}
		if ($_remoteConnectArgs{'uppercaseWords'}) {
		    # Uppercase all words
		    $str = _uppercaseWords($str);
		}
	    }
	    
	    # Some CDDB entries have band names with the article at the
	    # end. i.e. Fall, The. For consistency I want it au natural.
	    $band = _unSwapArticle($band);
	}
	else {
	    Trace::debug('Incorrect CD listing from CDDB') if (Trace::check(Trace::DEBUG3));
	    Trace::debug($str) if (Trace::check(Trace::DEBUG3));
	    $result->{'message'} = 'Unacceptable response from CDDB:'."\n".$match->{'str'};
	    return 1;
	}

	$match->{'name'} = $title;
	$match->{'artist'} = $band;
	my $mtracks = $match->{'tracks'} = {};

	$str = Cddb::command(Command => 'cddb read '.$match->{'category'}.' '.$match->{'cddb_id'});
	if (!defined($str)) {
	    $result->{'message'} = 'Unacceptable track response from CDDB';
	    return 1;
	}

	Trace::debug('Cddb lookup:') if Trace::check(Trace::DEBUG6);
	Trace::dump(Trace::DEBUG6, $str);

	my @lines = split(/\n/, $str);
	foreach my $line (@lines) {
	    if ($line =~ m/^TTITLE(\d+)=(.*)$/) {
		# Sanity check that the CDDB matches CD
		my $track = grep { $_->index == $1 + 1 } @tracks;
		if (defined($track)) {
		    my $title = $2;
		    # Do various clean ups of the CDDB data
		    # Remove proceeding and trailing whitespace
		    $title = _removeWhitespace($title);
		    # Uppercase all words
		    $title = _uppercaseWords($title);
		    # I don't like double spaces
		    $title = _removeDoubleSpace($title);
		    $mtracks->{ $1 + 1 }->{'name'} = $title;
		    $mtracks->{ $1 + 1 }->{'index'} = $1 + 1;
		}
		else {
		    Trace::error("Incorrect track listing from CDDB");
		    Trace::error($str);
		    $result->{'message'} = 'Unacceptable track response from CDDB:'."\n".$str;
		    return 1;
		}
	    }
	}
	shift @lines; # cmd...
	pop @lines;   # terminating .
	$match->{'cddb read data'} = join("\n", @lines); # Save for possible submission
    }

    if ($#{ $matches } == 0) {
	if ($cd->cddb_id eq $matches->[0]->{'cddb_id'}) {
	    $result->{'message'} = 'Remote lookup was successful';
	    $result->{'status'} = Success;
	    $cd->_remoteChooseMatch($matches->[0]) if ($parms{'chooseExact'}); # Choose when one exact match
	}
	else {
	    $result->{'message'} = 'Inexact match found in remote database';
	}
    }
    else {
	$result->{'message'} = 'Multiple CDs to choose from';
    }
    return 1;
}

sub remoteChoose {
    my $cd = shift;
    my %parms = @_;

    my $cddb_id = $parms{'cddb_id'};
    my $category = $parms{'category'};
    my $matches = $cd->get->{'remoteResult'}->{'matches'};
    foreach my $match (@{ $matches }) {
	if (($cddb_id eq $match->{'cddb_id'}) && (!defined($category) || ($category eq $match->{'category'}))) {
	    $cd->setN;
	    $cd->_remoteChooseMatch($match);
	    $cd->sync;
	    return 1;
	}
    }
    return 0;
}

sub _remoteChooseMatch {
    my $cd = shift;
    my $match = shift;

    my @tracks = $cd->tracks;

    $cd->getN->{'name'} = $match->{'name'};
    $cd->getN->{'category'} = $match->{'category'};
    my $artist = Info::Artist->find(owner => $cd->owner, name => $match->{'artist'}, create => 1);
    $cd->getN->{'artist'} = $artist;
    $cd->getN->{'artist_ioid'} = $artist->ioid;
    foreach my $mtrack (values(%{ $match->{'tracks'} })) {
	my $track = (grep { $_->index == $mtrack->{'index'} } @tracks)[0];
	$track->setN;
	$track->getN->{'name'} = $mtrack->{'name'};
	$track->save;
	$track->sync;
	
    }
    $cd->getN->{'cddb match'} = $match;

    $cd->save;
    return 1;
}

sub remoteResult {
    my $cd = shift;
    return $cd->get->{'remoteResult'} if (defined($cd->get->{'remoteResult'}));
    return { status => Never };
}

sub remoteFinished {
    my $cd = shift;
    my $status = $cd->remoteResult->{'status'};
    return (($status eq Success) || ($status eq Failure) || ($status eq Multiple));
}

# remoteSave() - Submit the Cd info 

sub remoteSave {
    my $cd = shift;
    my %parms = @_;

    if (defined($cd->get->{'remoteResult'}) && ($cd->get->{'remoteResult'}->{'status'} eq Underway)) {
	Trace::error('Remote operation is already in progress');
	return;
    }
    
    my $matches = [];
    my $priorResult = $cd->get->{'remoteResult'};
    my $result = { status => Underway, message => 'Doing remote save' };
    $cd->setN;
    $cd->getN->{'remoteResult'} = $result;
    $cd->sync;

    if (!Cddb::configured) {
	$cd->getN->{'remoteResult'}->{'status'} = Failure;
	$cd->getN->{'remoteResult'}->{'message'} = 'Remote database is not configured';
	$cd->sync;
	return;
    }

    Info::Persist::disconnect;
    my $fork = Thing->context->fork;
    return 1 if ($fork != 0);

    $result->{'status'} = Failure;

    if (!_cddbInit()) {
	$result->{'message'} = 'Problems with CDDB configuration:'."\n".Trace::lastError;
	Trace::debug($result->{'message'}) if Trace::check(Trace::DEBUG4);
	$cd->sync;
	_exit(1);
    }

    if (!$_remoteConnectArgs{'user'} || !$_remoteConnectArgs{'host'}) {
	$result->{'message'} = 'Missing CDDBP user or host as required for submission';
	Trace::debug($result->{'message'}) if Trace::check(Trace::DEBUG4);
	$cd->sync;
	_exit(1);
    }

    if (!defined($cd->get->{'cddb match'})) {
	# Redo the lookup for cds whose remote info we don't know,
	# this way we won't lose any EXT info.
	$result->{'message'} = 'Doing remote lookup';
	$result->{'status'} = Underway;
	$cd->sync;
	if (!$cd->_remoteLookup(forking => 0, chooseExact => 0)) {
	    # If we can't execute a look up we can't save
	    $cd->sync;
	    _exit(1);
	}
	$result->{'message'} = 'Doing remote save';
	$result->{'status'} = Underway;
	$cd->getN->{'remoteResult'} = $result;
	$cd->sync;
	# A single perfect match can be used for modification, or
	# multiple exact matches where the category matches
	# ours. Otherwise the CD will be submitted as a new entry,
	# what the folks at freedb want.
	if (defined($result->{'matches'}) && ($#{ $result->{'matches'} } != -1)) {
	    foreach my $match (@{ $result->{'matches'} }) {
		if (($match->{'cddb_id'} eq $cd->cddb_id) && ($match->{'category'} eq $cd->category)) {
		    $cd->get->{'cddb match'} = $match;
		    $cd->getN->{'cddb match'} = $match;
		    last;
		}
	    }
	}
	$result->{'status'} = Failure;
    }

    my @tracks = $cd->tracks;
    my $tracksH = $cd->tracksH;
    my $submission = '';
    $submission .= '# xmcd'."\n";
    $submission .= '#'."\n";
    $submission .= '# Track frame offsets:'."\n";
    foreach my $track (@tracks) {
	next if ($track->index == Info::Track::LEADOUTTN);
	next if ($track->index == Info::Track::MSTN);
	$submission .= '#        '.$track->lba."\n";
    }
    $submission .= '#'."\n";
    $submission .= '# Disc length: '.$cd->running_time.' seconds'."\n";
    $submission .= '#'."\n";

    if (defined($cd->get->{'cddb match'}) && ($cd->get->{'cddb match'}->{'category'} eq $cd->category)) { 

	Trace::debug('Before:') if Trace::check(Trace::DEBUG6);
	Trace::dump(Trace::DEBUG6, $cd->get->{'cddb match'}->{'cddb read data'});

	# Modification of remote entry, do substitution to avoid
	# mangling EXT and DISCID links and to increase
	# Revision. Rewrite the track offsets and length though as the
	# data returned often does not match the CD.

	my $cddb_id = $cd->cddb_id;
	my @lines = split(/\n/, $cd->get->{'cddb match'}->{'cddb read data'});
	my $line;
	while ($line = shift @lines) {
	    last if ($line =~ /^\#\s+Revision/);
	}
	unshift @lines, $line;
	while ($line = shift @lines) {
	    if ($line =~ /^\#\s+Revision\:\s+(\d+)/) {
		$submission .= '# Revision: '.($1 + 1);
	    }
	    elsif ($line =~ /^\#\s+Submitted\ via\:/) {
		$submission .= '# Submitted via: emptytree-seedy '.Thing::VERSION;
	    }
	    elsif ($line =~ /^\#\s+Processed/) {
		next;
	    }
	    elsif (($line =~ /^DISCID/) && ($line !~ /$cddb_id/)) {
		$submission .= $cddb_id;
	    }
	    elsif ($line =~ /^DTITLE/) {
		$submission .= 'DTITLE='.$cd->artist->name.' / '.$cd->name;
	    }
	    elsif ($line =~ /^TTITLE(\d+)/) {
		$submission .= 'TTITLE'.$1.'='.$tracksH->{$1 + 1}->name;
	    }
	    else {
		$submission .= $line;
	    }
	    $submission .= "\n";
	}
	Trace::debug('After:') if Trace::check(Trace::DEBUG6);
	Trace::dump(Trace::DEBUG6, $submission);
    }
    else {

	$submission .= '# Revision: 0'."\n";
	$submission .= '# Submitted via: emptytree-seedy '.Thing::VERSION."\n";
	$submission .= '#'."\n";
	$submission .= 'DISCID='.$cd->cddb_id."\n";
	$submission .= 'DTITLE='.$cd->artist->name.' / '.$cd->name."\n";
	for (my $index = $cd->firstTrackIndex; $index <= $cd->lastTrackIndex; $index++) {
	    $submission .= 'TTITLE'.($index - 1).'='.$tracksH->{ $index }->name."\n";
	}
	$submission .= 'EXTD='."\n";
	for (my $index = $cd->firstTrackIndex; $index <= $cd->lastTrackIndex; $index++) {
	    $submission .= 'EXTT'.($index - 1).'='."\n";
	}
	$submission .= 'PLAYORDER=';

	Trace::debug('Submission:') if Trace::check(Trace::DEBUG6);
	Trace::dump(Trace::DEBUG6, $submission);
    }

    if (!defined(Cddb::submit(Data => $submission, Category => $cd->category, CddbId => $cd->cddb_id))) {
	$result->{'message'} = Trace::lastError;
    }
    else {
	$result->{'message'} = 'Remote save was successful';
	$result->{'status'} = Success;
    }
    $cd->sync;
    _exit(1);
}

sub firstTrackIndex {
    my $cd = shift;
    $cd->{'firstTrackIndex'};
}

sub lastTrackIndex {
    my $cd = shift;
    $cd->{'lastTrackIndex'};
}

sub firstAudioTrackIndex {
    my $cd = shift;
    $cd->{'firstAudioTrackIndex'};
}

sub lastAudioTrackIndex {
    my $cd = shift;
    $cd->{'lastAudioTrackIndex'};
}

sub nextAudioTrackIndex {
    my $cd = shift;
    my $from = shift;

    foreach my $index (@{ $cd->{'audioIndices'} }) {
	if ($from < $index) {
	    return $index;
	}
    }
    return undef;
}

sub prevAudioTrackIndex {
    my $cd = shift;
    my $from = shift;

    foreach my $index (reverse @{ $cd->{'audioIndices'} }) {
	if ($from > $index) {
	    return $index;
	}
    }
    return undef;
}

sub tracksH {
    my $cd = shift;
    return $cd->{'tracksH'} if defined($cd->{'tracksH'});
    if (!Info::Persist::configured) {
	Trace::error('Tracks must be specified at creation for non persistence');
	return;
    }
    my @tracks = Info::Track->find(owner => $cd->owner, album_ioid => $cd->ioid);
    $cd->{'tracksH'} = {};
    foreach my $track (@tracks) {
	$cd->{'tracksH'}->{ $track->index } = $track;
    }
    return $cd->{'tracksH'};
}

sub tracks {
    my $cd = shift;
    return sort { $a->index <=> $b->index } values %{ $cd->tracksH };
}

sub track {
    my $cd = shift;
    my $index = shift;
    my $tracksH = $cd->tracksH;
    return $tracksH->{ $index } if defined($tracksH->{ $index });
    return;
}

sub known {
    my $cd = shift;
    my $artist = $cd->artist;

    return 0 if (($artist->name eq 'Unknown') && ($cd->name =~ /^Unknown/));
    return 1;
}

sub equalInfo {
    my $cdA = shift;
    my $cdB = shift;
    
    return 0 if (($cdA->name ne $cdB->name) || ($cdA->artist ne $cdB->artist));

    my @tracksA = values(%{ $cdA->{'tracks'} });
    my @tracksB = values(%{ $cdB->{'tracks'} });

    return 0 if ($#tracksA != $#tracksB);

    foreach my $index (keys(%{ $cdA->{'tracks'} })) {
	if (!defined($cdB->{'tracks'}->{ $index })) {
	    return 0;
	}
	if (defined($cdA->{'tracks'}->{ $index }->{'name'}) xor
	    defined($cdB->{'tracks'}->{ $index }->{'name'})) {
	    return 0;
	}
	if (defined($cdA->{'tracks'}->{ $index }->{'name'}) &&
	    defined($cdB->{'tracks'}->{ $index }->{'name'}) &&
	    ($cdA->{'tracks'}->{ $index }->{'name'} ne $cdB->{'tracks'}->{ $index }->{'name'})) {
	    return 0;
	}
    }    
    return 1;
}

# _removeWhitespace() - Remove proceeding and trailing whitespaces

sub _removeWhitespace {
    my $str = shift;

    while ($str =~ /^\s(.*)$/) {
	$str = $1;
    }
    while ($str =~ /^(.*)\s$/) {
	$str = $1;
    }
    return($str);
}

# _uppercaseWords() - Make all words uppercase

sub _uppercaseWords {
    my $str = shift;
    my @words = ();

    # Words are typically seperated by a ' ' but they can also be
    # seperated with a '(' and perhaps others?

    my @newwords = ();
    @words = split(/\s/, $str);
    foreach my $word (@words) {
	push (@newwords, ucfirst $word);
    }
    $str = join(' ', @newwords);
    @newwords = ();
    @words = split(/\(/, $str);
    foreach my $word (@words) {
	push (@newwords, ucfirst $word);
    }
    $str = join('(', @newwords);
    return $str;
}

# _removeDoubleSpace - Replace all doube whitespace occurences with a
# single ' '

sub _removeDoubleSpace {
    my $str = shift;
    while ($str =~ m/^(.*)\s\s(.*)$/) {
	$str = $1.' '.$2;
    }
    return $str;
}

# _swapArticle - Put them at the end after a ,

sub _swapArticle {
    my $str = shift;
    if (($str =~ m/^(The) (.*)$/) ||
	($str =~ m/^(An) (.*)$/) ||
	($str =~ m/^(A) (.*)$/)) {
	return $2.', '.$1;
    }
    return $str;
}

# _unSwapArticle - Put them back at the front.

sub _unSwapArticle {
    my $str = shift;
    if (($str =~ m/^(.*)\, (The)$/) ||
	($str =~ m/^(.*)\, (An)$/) ||
	($str =~ m/^(.*)\, (A)$/)) {
	return $2.' '.$1;
    }
    return $str;
}

sub _cddbInit {
    my $user = $_remoteConnectArgs{'user'} || 'unknown';
    my $host = $_remoteConnectArgs{'host'} || 'unknown';

    return 1 if (!defined($_remoteConnectArgs{'URL'}) || ($_remoteConnectArgs{'URL'} eq 'None'));

    return Cddb::configure(URL => $_remoteConnectArgs{'URL'},
			   User => $user,
			   Host => $host,
			   Program => ($_remoteConnectArgs{'URL'} =~ /cddb\.com/) ? 'xmcd' : 'emptytree-seedy', # cddb.com doesn't like
			   Version => Thing::VERSION);
}

1;
