# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.4 $

package Info::List;

sub table { 'lists'; };

use Thing::Info;
@ISA = qw( Thing::Info);
use Info::Persist;
use Info::Track;
use RepoH;
use Trace;
use strict;

# sync `msg's
sub DeleteTracks    { 'Info__List__DeleteTracks'; } 
sub AddTracks       { 'Info__List__AddTracks'; } 
sub MoveTrack       { 'Info__List__MoveTrack'; } 

sub find {
    my $proto = shift;
    my $type = ref($proto) || $proto;
    my %parms = @_;

    if (!defined($parms{'owner'}) || ($parms{'owner'}->name ne Repo::NAME)) {
	Trace::error('Object must be owned by Repo');
	return undef;
    }

    if (Info::Persist::configured) {

	# Local database in use for persistence

	my @selects;
	if (defined($parms{'ioid'})) {
	    push @selects, 'ioid='.$parms{'ioid'};
	}
	if (defined($parms{'name'})) {
	    push @selects, 'name='.Info::Persist::quote($parms{'name'});
	}
	my $select = Info::Persist::buildSelect(table => Info::List::table, wheres => \@selects);

	my $dbh = Info::Persist::connect;
	return if (!defined($dbh));
    
	my $sth;
	if (!defined($sth = $dbh->prepare($select)) || !defined($sth->execute)) {
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return;
	}
	
	my $data = $sth->fetchall_arrayref({});
	$sth->finish;
	if (defined($data) && ($#{ $data } >= 0)) {

	    Trace::debug('Found %d lists', $#{ $data } + 1) if Trace::check(Trace::DEBUG6);

	    # track_ioids needs conversion, from string to array
	    map { 
		if (!defined($_->{'track_ioids'}) || !UNIVERSAL::isa($_->{'track_ioids'}, 'ARRAY')) {
		    if (defined($_->{'track_ioids'})) {
			$_->{'track_ioids'} =~ /\{(.*)\}/; 
			my @tioids = split(',', $1); 
			$_->{'track_ioids'} = \@tioids;
		    }
		    else {
			$_->{'track_ioids'} = [];
		    }
		}
	    } @{ $data };

	    if (!wantarray) {
		my $list = $type->new(owner => $parms{'owner'}, ioid => $data->[0]->{'ioid'}, data => $data->[0]);
		Trace::debug('Found list: %s: %s', $list->name, $list->ioid) if Trace::check(Trace::DEBUG4);
		return $list;
	    }
	    else {
		my @lists = map { $type->new(owner => $parms{'owner'}, ioid => $_->{'ioid'}, data => $_) } @{ $data };
		return @lists;
	    }
	}
    }

    Trace::debug('No matching locally persistent list') if Trace::check(Trace::DEBUG4);

    if ($parms{'create'}) {
	$parms{'_bypassfind'} = 1;
	delete $parms{'create'};
	return $type->create(%parms);
    }
    return;
}

sub create {
    my $proto = shift;
    my $type = ref($proto) || $proto;
    my %parms = @_;

    if (!defined($parms{'owner'}) || ($parms{'owner'}->name ne Repo::NAME)) {
	Trace::error('Object must be owned by Repo');
	return undef;
    }

    my $list;

    if (!defined($parms{'name'})) {
	Trace::debug('Missing list name for create') if Trace::check(Trace::DEBUG4);
	return undef;
    }

    if (Info::Persist::configured) {

	# Local database in use for persistence

	if (!$parms{'_bypassfind'}) {
	    my $list = $type->find(owner => $parms{'owner'}, name => $parms{'name'});
	    return $list if (defined($list));
	}

	my $dbh = Info::Persist::connect;
	return undef if (!defined($dbh));
    
	my $cmd = 'INSERT INTO lists ';
	my @fields;
	my @values;
	
	push @fields, 'name';
	push @values, Info::Persist::quote($parms{'name'});
	if (defined($parms{'track_ioids'})) {
	    push @fields, 'track_ioids';
	    push @values, '\'{'.join(',', @{ $parms{'track_ioids'} }).'}\'';
	}
	
	$cmd .= '('.join(', ', @fields).')';
	$cmd .= ' VALUES ('.join(', ', @values).');';
	
	if (!$dbh->do($cmd)) {
	    Trace::error($cmd);
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return undef;
	}

	# Find the assigned ioid

	my @lists = $type->find(owner => $parms{'owner'}, name => $parms{'name'});
	$list = shift @lists; # Take the lowest ioid
	if ($#lists >= 0) {
	    # Race, something else created this list
	    map { $_->dbDelete } @lists;
	}
    }
    else {
	$list = $type->new(owner => $parms{'owner'}, 
			    data =>
			    {
				name => $parms{'name'},
				track_ioids => $parms{'track_oids'},
			    },
			    );
    }
    Trace::debug('Created list: %s: %s', $list->name, $list->ioid) if Trace::check(Trace::DEBUG4);
    return $list;
}

# save() - Save modified (getN) values.

sub save {
    my $list = shift;

    return 1 if (!Info::Persist::configured);

    Trace::debug('Update list: %s: %s', $list->name, $list->ioid) if Trace::check(Trace::DEBUG3);
	
    my $dbh = Info::Persist::connect;
    return if (!defined($dbh));

    my $cmd = 'UPDATE lists SET ';
    my @sets;
    if ($list->saveCheck('name')) {
	push @sets, 'name='.Info::Persist::quote($list->getN->{'name'});
    }
    push @sets, 'track_ioids=\'{'.join(',', @{ $list->getN->{'track_ioids'} }).'}\'';
    return 1 if (!@sets);

    $cmd .= join(', ', @sets);
    $cmd .= ' WHERE ioid='.$list->ioid.';';

    if (!$dbh->do($cmd)) {
	Trace::error($DBI::errstr);
	Info::Persist::disconnect;
	return;
    }
    return 1;
}

sub delete {
    my $list = shift;

    Trace::debug('Delete list: %s: %s', $list->name, $list->ioid) if Trace::check(Trace::DEBUG3);

    return if (!Info::Persist::configured);

    Trace::debug('Delete list (database part): %s: %s', $list->name, $list->ioid) if Trace::check(Trace::DEBUG3);

    my $dbh = Info::Persist::connect;
    return if (!defined($dbh));

    if (!$dbh->do('DELETE FROM lists WHERE ioid='.$list->ioid.';')) {
	Trace::error($DBI::errstr);
	Info::Persist::disconnect;
	return 0;
    }

    return 1;
}

sub name { $_[0]->get->{'name'}; }

sub track_ioids { $_[0]->get->{'track_ioids'}; }

sub change_name {
    my $list = shift;
    my $name = shift;

    return 1 if ($name eq $list->name);

    if (Info::Persist::configured) {

	# Cannot use the name of an existing list

	my $exists = Info::List->find(owner => $list->owner, name => $name, create => 0);
	if ($exists) {
	    Trace::error('There is already a list named %s', $name);
	    return 0;
	}
    }

    $list->setN;
    $list->getN->{'name'} = $name;
    $list->save;
    $list->sync;
    return 1;
}

sub addTracks {
    my $list = shift;
    my %parms = @_;

    if (!defined($parms{'trackIoids'}) || ($#{ $parms{'trackIoids'} } == -1)) {
	Trace::mandatory('trackIoids');
	return;
    }
    my @add = @{ $parms{'trackIoids'} };

    Trace::debug('Add tracks for: %s, ioids: %s', $list->name, join(',', @add)) if Trace::check(Trace::DEBUG3);

    my @ioids = @{ $list->track_ioids };
    if (!@ioids) {
	@ioids = ();
    }

    if (!defined($parms{'at'})) {
	push @ioids, @add;
    }
    elsif ($#ioids < $parms{'at'} - 1) {
	Trace::error('Position (%d) is outside this lists tracks', $parms{'at'});
	return;
    }
    else {
	splice @ioids, $parms{'at'}, 0, @add;
    }

    $list->setN;
    $list->getN->{'track_ioids'} = \@ioids;
    $list->save;
    $list->sync(msg => AddTracks, %parms);
    return 1;
}

sub deleteTracks {
    my $list = shift;
    my %parms = @_;

    if (!defined($parms{'trackPositions'}) || ($#{ $parms{'trackPositions'} } == -1)) {
	Trace::mandatory('trackPositions');
	return;
    }

    Trace::debug('Delete tracks for: %s, at: %s', $list->name, join(',', @{ $parms{'trackPositions'} })) if Trace::check(Trace::DEBUG3);

    my @del = sort { $a <=> $b } @{ $parms{'trackPositions'} };
    my @deleted = ();

    if (!defined($list->track_ioids)) {
	Trace::error('No existing tracks to delete');
	return;
    }
    my @ioids = @{ $list->track_ioids };

    while (@del) {
	my $delete = pop @del;
	if ($delete > $#ioids) {
	    Trace::error('Requested track (%d) to delete does not exist', $delete);
	    return;
	}
	push @deleted, $ioids[ $delete ];
	splice @ioids, $delete, 1;
    }

    $list->setN;
    $list->getN->{'track_ioids'} = \@ioids;
    $list->save;
    $list->sync(msg => DeleteTracks, trackIoids => \@deleted, %parms);
    return 1;
}

sub moveTrack {
    my $list = shift;
    my %parms = @_;

    if (!defined($parms{'from'})) {
	Trace::mandatory('from');
	return;
    }
    my $to = $parms{'to'};
    my $from = $parms{'from'};

    return 1 if (defined($to) && ($to == $from));

    my @ioids = @{ $list->track_ioids };

    return 1 if (!defined($to) && ($from == $#ioids));

    my $hold = $ioids[ $from ];
    splice @ioids, $from, 1;
    if (!defined($to)) {
	push @ioids, $hold;
    }
    else {
	if ($to > $from) {
	    $to--;
	}
	splice @ioids, $to, 0, $hold;
    }
    $list->setN;
    $list->getN->{'track_ioids'} = \@ioids;
    $list->save;
    $list->sync(msg => MoveTrack, %parms);
    return 1;
}

1;
