# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.9 $

package Info::Persist;

# General persistant info objects

use Trace;
use RepoH;
use Utility;
use strict;

BEGIN {
    eval {
	require DBI;
    };
    if ($@) {
	require StubDBI;
    }
}

use vars qw( @_dbConnectArgs $_dbConnectRef );

# configure() - Configure the database for persistence.

sub configure {
    my %parms = @_;

    if (!defined($parms{'db'})) {
	Trace::mandatory('db');
	return 0;
    }

    @_dbConnectArgs = ();
    $_dbConnectRef = $parms{'db'};
    my $db = $parms{'db'};
    if ($db->{'type'} eq 'None') {	
    }
    elsif ($db->{'type'} eq 'Postgresql') {
	@_dbConnectArgs = ('dbi:Pg:dbname=', undef, undef);
	if (defined($db->{'name'}) && ($db->{'name'} ne '')) {
	    $_dbConnectArgs[0] .= $db->{'name'};
	}
	else {
	    Trace::error('A database name must be specified for the database');
	    @_dbConnectArgs = ();
	    return 0;
	}
	if (defined($db->{'host'}) && ($db->{'host'} ne '')) {
	    $_dbConnectArgs[0] .= ';host='.$db->{'host'};
	}
	if (defined($db->{'port'}) && ($db->{'port'} ne '')) {
	    $_dbConnectArgs[0] .= ';port='.$db->{'port'};
	}
	if (!defined($db->{'user'})) {
	    Trace::error('A user name must be specified for the database');
	    @_dbConnectArgs = ();
	    return 0;
	}
	$_dbConnectArgs[1] = $db->{'user'};
	if (!defined($db->{'password'})) {
	    Trace::error('A password must be specified for the database user');
	    @_dbConnectArgs = ();
	    return 0;
	}
	$_dbConnectArgs[2] = $db->{'password'};
	$_dbConnectArgs[3] = { AutoCommit => 1, PrintError => 1 };
	$_dbConnectRef = $parms{'db'};
	if (!Info::Persist::_databaseCreate()) {
	    return 0;
	}
	if (!Info::Persist::_databaseUpgrade()) {
	    return 0;
	}
    }
    else {
	Trace::error('Unsupported database type: %s', $db->{'type'});
	return 0;
    }
    return 1;
}

sub dbConnectArgs { @_dbConnectArgs; }
sub dbConnectRef { $_dbConnectRef; }

use vars qw( $_dbh );

sub connect {
    $ENV{'PGDATESTYLE'} = 'ISO'; # Allows for stringwise date comparisons

    return $_dbh if (defined($_dbh));

    if ($#_dbConnectArgs == -1) {
	Trace::debug('The database is not configured') if Trace::check(Trace::DEBUG4);
	return undef;
    }

    if (!defined($_dbh = DBI->connect(@_dbConnectArgs))) {
	Trace::error($DBI::errstr);
	return undef;
    }
    #$_dbh->trace(1) if (Trace::level >= Trace::DEBUG4);
    
    return $_dbh;
}

sub disconnect {
    if (defined($_dbh)) {
	$_dbh->disconnect;
    }
    $_dbh = undef;
    return 1;
}

sub dbh { $_dbh; }

# configured()

sub configured {
    if ($#_dbConnectArgs == -1) {
	Trace::debug('The database is not configured') if Trace::check(Trace::DEBUG4);
	return 0;
    }
    return 1;
}

# quote - postgresql escaping

sub quote {
    my $value = shift;
    $value =~ s/\'/\'\'/g;
    return '\''.$value.'\'';
}

# buildSelect

sub buildSelect {
    my %parms = @_;

    my $select = 'SELECT * FROM '.$parms{'table'};

    if (defined($parms{'wheres'}) && ($#{ $parms{'wheres'} } >= 0)) {
	my $where = ' WHERE (('.join(') AND (', @{ $parms{'wheres'} }).'))';
	$select .= $where;
    }
    $select .= ';';
    return $select;
}

# find - General database query method, return a `type' of object(s)
# based on the select statement.

sub find {
    my %parms = @_;

    if (!defined($parms{'owner'}) || ($parms{'owner'}->name ne Repo::NAME)) {
	Trace::error('Objects must be owned by Repo');
	return undef;
    }

    if (!Info::Persist::configured) {
	Trace::error('No persistence');
	return;
    }

    if (!defined($parms{'select'})) {
	Trace::mandatory('select');
	return;
    }

    if (!defined($parms{'type'})) {
	Trace::mandatory('type');
	return undef;
    }

    my $select = $parms{'select'};
    my $type = $parms{'type'};

    my $dbh = Info::Persist::connect;
    return if (!defined($dbh));
    
    my $sth;
    if (!defined($sth = $dbh->prepare($select)) || !defined($sth->execute)) {
	Trace::error($DBI::errstr);
	return;
    }

    my $data = $sth->fetchall_arrayref({});
    $sth->finish;
    if (defined($data) && ($#{ $data } >= 0)) {
	Trace::debug('Found %d objects', $#{ $data } + 1) if Trace::check(Trace::DEBUG6);
	if (!wantarray) {
	    my $object = $type->new(owner => $parms{'owner'}, ioid => $data->[0]->{'ioid'}, data => $data->[0]);
	    return $object;
	}
	else {
	    return map { $type->new(owner => $parms{'owner'}, ioid => $_->{'ioid'}, data => $_) } @{ $data };
	}
    }
    return;
}

# _databaseCreate - Attempts to create the local database if it doesn't exist.

sub _databaseCreate {

    return 1 if ($_dbh);

    if (Info::Persist::connect) {
	Info::Persist::disconnect;
	return 1;
    }

    # One reason the connect might fail is if the database is not
    # created. Attempt to create it.

    my @connectArgs = ('dbi:Pg:dbname=template1', undef, undef);
    if (defined($_dbConnectRef->{'host'})) {
	$connectArgs[0] .= ';host='.$_dbConnectRef->{'host'};
    }
    if (defined($_dbConnectRef->{'port'})) {
	$connectArgs[0] .= ';port='.$_dbConnectRef->{'port'};
    }
    $connectArgs[1] = $_dbConnectRef->{'user'};
    $connectArgs[2] = $_dbConnectRef->{'password'};

    my $dbh;
    if (!defined($dbh = DBI->connect(@connectArgs))) {
	Trace::error($DBI::errstr);
	return 0;
    }
    #$dbh->trace(1) if (Trace::level >= Trace::DEBUG4);

    if (!$dbh->do('CREATE DATABASE "'.$_dbConnectRef->{'name'}.'";')) {
	Trace::error($DBI::errstr);
	$dbh->disconnect if defined($dbh);
	return 0;
    }

    $dbh->disconnect;

    # Connect to the new database and create the tables.

    if (!defined($dbh = DBI->connect(@_dbConnectArgs))) {
	Trace::error($DBI::errstr);
	return 0;
    }
    #$dbh->trace(1) if (Trace::level >= Trace::DEBUG4);

    if (!$dbh->do('CREATE SEQUENCE ioid_seq;')) {
	Trace::error($DBI::errstr);
	$dbh->disconnect if defined($dbh);
	return 0;
    }

    if (!$dbh->do('CREATE TABLE artists (
        ioid            INT4 DEFAULT nextval(\'ioid_seq\'),
        name            TEXT,
        play_date       DATETIME
        );')) {
	Trace::error($DBI::errstr);
	$dbh->disconnect if defined($dbh);
	return 0;
    }

    if (!$dbh->do('CREATE TABLE albums (
        ioid            INT4 DEFAULT nextval(\'ioid_seq\'),
        name            TEXT,
        artist_ioid     INT4,
        cddb_id         TEXT,
        category        TEXT,
        track_count     INT4,
        running_time    INT4,
        play_count      INT4 DEFAULT 0,
        play_date       DATETIME
        );')) {
	Trace::error($DBI::errstr);
	$dbh->disconnect if defined($dbh);
	return 0;
    }

    if (!$dbh->do('CREATE TABLE tracks (
        ioid            INT4 DEFAULT nextval(\'ioid_seq\'),
        name            TEXT,
        index           INT4,
        album_ioid      INT4,
        lba             INT4,
        data            INT4,
        duration        INT4,
        play_count      INT4 DEFAULT 0,
        play_date       DATETIME
        );')) {
	Trace::error($DBI::errstr);
	$dbh->disconnect if defined($dbh);
	return 0;
    }

    if (!$dbh->do('CREATE TABLE lists (
        ioid            INT4 DEFAULT nextval(\'ioid_seq\'),
        name            TEXT,
        track_ioids     INT4[]
        );')) {
	Trace::error($DBI::errstr);
	$dbh->disconnect if defined($dbh);
	return 0;
    }

    if (!$dbh->do('CREATE TABLE version (name TEXT);
                   CREATE UNIQUE INDEX track_ioid ON tracks (ioid);
                   CREATE UNIQUE INDEX track_album_ioid_index ON tracks (album_ioid, index);
                   CREATE UNIQUE INDEX artist_ioid ON artists (ioid);
                   CREATE UNIQUE INDEX artist_name ON artists (name);
                   CREATE UNIQUE INDEX album_ioid ON albums (ioid);
                   CREATE INDEX album_artist_ioid_name ON albums (ioid);')) {
	Trace::error($DBI::errstr);
	$dbh->disconnect if defined($dbh);
	return 0;
    }

    foreach my $table ('artists', 'albums', 'tracks', 'lists', 'ioid_seq', 'version',
		       'track_ioid', 'track_album_ioid_index', 'artist_ioid', 'artist_name', 'album_ioid', 'album_artist_ioid_name') {

	if (!$dbh->do('GRANT ALL ON '.$table.' TO '.$_dbConnectRef->{'user'}.';')) {
	    Trace::error($DBI::errstr);
	    $dbh->disconnect if defined($dbh);
	    return 0;
	}
    }

    if (!$dbh->do('INSERT INTO version (name) VALUES (\'1.5.1.0\');')) {
	Trace::error($DBI::errstr);
	$dbh->disconnect if defined($dbh);
	return 0;
    }
    $dbh->disconnect if defined($dbh);
    return 1;
}

# _databaseUpgrade - Attempts to upgrade the database format between releases.

sub _databaseUpgrade {

    my $dbh = Info::Persist::connect;
    if (!$dbh) {
	return;
    }

    my $version;
    my $sth;
    if (!defined($sth = $dbh->prepare('SELECT * FROM version;')) || !defined($sth->execute)) {
	$sth->finish if defined($sth); 
	# Perhaps the version table does not exist, prior release < 1.3.1
	if (!$dbh->do('CREATE TABLE version (name TEXT);')) {
	    # More like the database is misconfigured
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return;
	}
	if (!$dbh->do('GRANT ALL ON version TO '.$_dbConnectRef->{'user'}.';')) {
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return;
	}
	if (!$dbh->do('INSERT INTO version (name) VALUES (\'1.3.0\');')) {
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return;
	}
	$version = '1.3.0';
    }
    else {
	my $data = $sth->fetchall_arrayref({});
	$sth->finish;
	if (defined($data) && ($#{ $data } == 0)) {
	    $version = $data->[0]->{'name'};
	}
	else {
	    Trace::error('Cannot determin database version');
	    return;
	}
    }

    return 1 if (Utility::cmpVersions($version, '1.5.1.0') >= 0);

    if (Utility::cmpVersions($version, '1.3.0') <= 0) {
	# Add play_count and play_date to tracks
	if (!$dbh->do('ALTER TABLE tracks ADD COLUMN play_count INT4;')) {
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return;
	}
	#if (!$dbh->do('ALTER TABLE tracks ALTER play_count SET DEFAULT 0;')) {
	#    Trace::error($DBI::errstr);
	#    $dbh->disconnect if defined($dbh);
	#    return;
	#}
	if (!$dbh->do('UPDATE tracks SET play_count=0;')) {
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return;
	}
	if (!$dbh->do('ALTER TABLE tracks ADD COLUMN play_date DATETIME;')) {
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return;
	}
    }	

    if (Utility::cmpVersions($version, '1.3.1') <= 0) {
	# Add play_count and play_date to albums
	if (!$dbh->do('ALTER TABLE albums ADD COLUMN play_count INT4;')) {
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return;
	}
	if (!$dbh->do('UPDATE albums SET play_count=0;')) {
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return;
	}
	if (!$dbh->do('ALTER TABLE albums ADD COLUMN play_date DATETIME;')) {
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return;
	}
	# Add play_date to artists
	if (!$dbh->do('ALTER TABLE artists ADD COLUMN play_date DATETIME;')) {
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return;
	}
    }	

    if (Utility::cmpVersions($version, '1.4.0') <= 0) {

	if (!$dbh->do('CREATE TABLE lists (
                       ioid            INT4 DEFAULT nextval(\'ioid_seq\'),
                       name            TEXT,
                       track_ioids     INT4[]
                       );')) {
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return;
	}
	if (!$dbh->do('GRANT ALL ON lists TO '.$_dbConnectRef->{'user'}.';')) {
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return;
	}
    }

    if (Utility::cmpVersions($version, '1.5.0') <= 0) {
	if (!$dbh->do('CREATE UNIQUE INDEX track_ioid ON tracks (ioid);
                   CREATE UNIQUE INDEX track_album_ioid_index ON tracks (album_ioid, index);
                   CREATE UNIQUE INDEX artist_ioid ON artists (ioid);
                   CREATE UNIQUE INDEX artist_name ON artists (name);
                   CREATE UNIQUE INDEX album_ioid ON albums (ioid);
                   CREATE INDEX album_artist_ioid_name ON albums (ioid);')) {
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return;
	}

	foreach my $table ('track_ioid', 'track_album_ioid_index', 'artist_ioid', 'artist_name', 'album_ioid', 'album_artist_ioid_name') {
	    if (!$dbh->do('GRANT ALL ON '.$table.' TO '.$_dbConnectRef->{'user'}.';')) {
		Trace::error($DBI::errstr);
		Info::Persist::disconnect;
		return;
	    }
	}
    }

    if (!$dbh->do('UPDATE version SET name=\'1.5.1.0\';')) {
	Trace::error($DBI::errstr);
	Info::Persist::disconnect;
	return;
    }
    return 1;
}

1;
