# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.8 $

package Info::Track;

sub table { 'tracks'; };

# Special track indices

sub MSTN       { -1; }
sub LEADOUTTN  { 170; }

use Thing::Info;
@ISA = qw( Thing::Info );
use Info::Persist;
use Info::Album;
use Info::Cd;
use RepoH;
use Trace;
use strict;

sub find {
    my $proto = shift;
    my $type = ref($proto) || $proto;
    my %parms = @_;

    if (!defined($parms{'owner'}) || ($parms{'owner'}->name ne Repo::NAME)) {
	Trace::error('Object must be owned by Repo');
	return undef;
    }

    if (Info::Persist::configured) {

	# Local database in use for persistence

	my @selects;
	if (defined($parms{'ioid'})) {
	    push @selects, 'ioid='.$parms{'ioid'};
	}
	if (defined($parms{'album_ioid'})) {
	    push @selects, 'album_ioid='.$parms{'album_ioid'};
	}
	elsif (defined($parms{'album'})) {
	    push @selects, 'album_ioid='.$parms{'album'}->ioid;
	}
	if (defined($parms{'index'})) {
	    push @selects, 'index='.$parms{'index'};
	}
	if (defined($parms{'name'})) {
	    push @selects, 'name='.Info::Persist::quote($parms{'name'});
	}
	my $select = Info::Persist::buildSelect(table => 'tracks', wheres => \@selects);
	
	my $dbh = Info::Persist::connect;
	return if (!defined($dbh));
    
	my $sth;
	if (!defined($sth = $dbh->prepare($select)) || !defined($sth->execute)) {
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return;
	}

	my $data = $sth->fetchall_arrayref({});
	$sth->finish;
	if (defined($data) && ($#{ $data } >= 0)) {
	    Trace::debug('Found %d tracks', $#{ $data } + 1) if Trace::check(Trace::DEBUG6);
	    if (!wantarray) {
		my $track = $type->new(owner => $parms{'owner'}, ioid => $data->[0]->{'ioid'}, data => $data->[0]);
		Trace::debug('Found track: %s: %s', $track->name, $track->ioid) if Trace::check(Trace::DEBUG4);
		if (defined($parms{'album'})) {
		    $track->{'album'} = $parms{'album'};
		}
		return $track;
	    }
	    else {
		my @tracks = map { $type->new(owner => $parms{'owner'}, ioid => $_->{'ioid'}, data => $_) } @{ $data };
		if (defined($parms{'album'})) {
		    map { $_->{'album'} = $parms{'album'} } @tracks;
		}
		return @tracks;
	    }
	}
    }
	
    Trace::debug('No matching locally persistent track') if Trace::check(Trace::DEBUG4);

    if ($parms{'create'}) {
	$parms{'_bypassfind'} = 1;
	delete $parms{'create'};
	return $type->create(%parms);
    }
    return;
}

sub create {
    my $proto = shift;
    my $type = ref($proto) || $proto;
    my %parms = @_;

    if (!defined($parms{'owner'}) || ($parms{'owner'}->name ne Repo::NAME)) {
	Trace::error('Object must be owned by Repo');
	return undef;
    }

    my $album_ioid = defined($parms{'album_ioid'}) ? $parms{'album_ioid'} : defined($parms{'album'}) ? $parms{'album'}->ioid : undef;
    if (!defined($album_ioid)) {
	Trace::debug('Missing album IOID for create') if Trace::check(Trace::DEBUG4);
	return undef;
    }

    if (!defined($parms{'index'})) {
	Trace::debug('Missing track index for create') if Trace::check(Trace::DEBUG4);
	return undef;
    }

    if (Info::Persist::configured) {

	# Local database in use for persistence

	if (!$parms{'_bypassfind'}) {
	    my $track = $type->find(owner => $parms{'owner'}, album_ioid => $album_ioid, index => $parms{'index'});
	    return $track if (defined($track));
	}

	my $dbh = Info::Persist::connect;
	return if (!defined($dbh));

	my $cmd = 'INSERT INTO tracks ';
	my @fields;
	my @values;
	
	push @fields, 'album_ioid';
	push @values, $album_ioid;
	push @fields, 'index';
	push @values, $parms{'index'};
	push @fields, 'name';
	if (defined($parms{'name'})) {
	    push @values, Info::Persist::quote($parms{'name'});
	}
	else {
	    push @values, 'Unknown';
	}
	if (defined($parms{'lba'})) {
	    push @fields, 'lba';
	    push @values, $parms{'lba'};
	}
	if (defined($parms{'data'})) {
	    push @fields, 'data';
	    push @values, $parms{'data'};
	}
	if (defined($parms{'duration'})) {
	    push @fields, 'duration';
	    push @values, $parms{'duration'};
	}
	if (defined($parms{'play_count'})) {
	    push @fields, 'play_count';
	    push @values, $parms{'play_count'};
	}
	else {
	    push @fields, 'play_count';
	    push @values, 0;
	}
	if (defined($parms{'play_date'})) {
	    push @fields, 'play_date';
	    push @values, Info::Persist::quote($parms{'play_date'});
	}
	
	$cmd .= '('.join(', ', @fields).')';
	$cmd .= ' VALUES ('.join(', ', @values).');';
	
	if (!$dbh->do($cmd)) {
	    Trace::error($cmd);
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return undef;
	}

	# Find the assigned ioid
	
	my $track;
	my @tracks = sort { $a->ioid <=> $b->ioid } $type->find(owner => $parms{'owner'}, album_ioid => $album_ioid, index => $parms{'index'});
	$track = shift @tracks; # Take the lowest ioid
	if ($#tracks >= 0) {
	    # Race
	    map { $_->dbDelete } @tracks;
	}
	if (defined($parms{'album'})) {
	    $track->{'album'} = $parms{'album'};
	}
	Trace::debug('Created track: %s: %s', $track->name, $track->ioid) if Trace::check(Trace::DEBUG4);
	return $track;
    }
    else {
	if (!defined($parms{'album'})) {
	    Trace::debug('Missing album for create') if Trace::check(Trace::DEBUG4);
	    return undef;
	}
	my $track = $type->new(owner => $parms{'owner'}, 
			       data =>
			       {
				   album_ioid => $parms{'album'}->ioid,
				   index => $parms{'index'},
				   name => $parms{'name'},
				   lba => $parms{'lba'},
				   data => $parms{'data'},
				   duration => $parms{'duration'},
				   play_count => $parms{'play_count'},
				   play_date => $parms{'play_date'},
			       }
			       );
	# Tracks won't change albums, so this does not need to be placed in the sync space.
	$track->{'album'} = $parms{'album'};
	Trace::debug('Created track: %s: %s', $track->name, $track->ioid) if Trace::check(Trace::DEBUG4);
	return $track;
    }
}

# save() - Save modified (getN) values.

sub save {
    my $track = shift;

    return 1 if (!Info::Persist::configured);

    Trace::debug('Update track: %s: %s', $track->name, $track->ioid) if Trace::check(Trace::DEBUG3);
	
    my $dbh = Info::Persist::connect;
    return if (!defined($dbh));

    my $cmd = 'UPDATE tracks SET ';
    my @sets;
    if ($track->saveCheck('name')) {
	push @sets, 'name='.Info::Persist::quote($track->getN->{'name'});
    }
    if ($track->saveCheck('index')) {
	push @sets, 'index='.$track->getN->{'index'};
    }
    if ($track->saveCheck('album_ioid')) {
	push @sets, 'album_ioid='.$track->getN->{'album_ioid'};
    }
    if ($track->saveCheck('lba')) {
	push @sets, 'lba='.$track->getN->{'lba'};
    }
    if ($track->saveCheck('duration')) {
	push @sets, 'duration='.$track->getN->{'duration'};
    }
    if ($track->saveCheck('data')) {
	push @sets, 'data='.$track->getN->{'data'};
    }
    if ($track->saveCheck('play_count')) {
	push @sets, 'play_count='.$track->getN->{'play_count'};
    }
    if ($track->saveCheck('play_date')) {
	push @sets, 'play_date='.Info::Persist::quote($track->getN->{'play_date'});
    }

    if ($#sets >= 0) {
	$cmd .= join(', ', @sets);
	$cmd .= ' WHERE ioid='.$track->ioid.';';

	Trace::debug('%s', $cmd) if (Trace::check(Trace::DEBUG3));
	
	if (!$dbh->do($cmd)) {
	    Trace::error($DBI::errstr);
	    Info::Persist::disconnect;
	    return;
	}
    }
    return 1;
}

sub delete {
    my $track = shift;

    Trace::debug('Delete track: %s: %s', $track->name, $track->ioid) if Trace::check(Trace::DEBUG3);

    # The repo does the work
    Thing::context->message(name => Repo::Delete, to => $track->owner, parms => { track => $track });
    return 1;
}

sub dbDelete {
    my $track = shift;

    my $dbh = Info::Persist::connect;
    return if (!defined($dbh));

    Trace::debug('Delete track (database part): %s: %s', $track->name, $track->ioid) if Trace::check(Trace::DEBUG3);

    if (!$dbh->do('DELETE FROM tracks WHERE ioid='.$track->ioid.';')) {
	Trace::error($DBI::errstr);
	Info::Persist::disconnect;
	return 0;
    }

    return 1;
}

sub album_ioid { $_[0]->get->{'album_ioid'}; }

sub index { $_[0]->get->{'index'}; }

sub name { $_[0]->get->{'name'}; }

sub lba { $_[0]->get->{'lba'}; }

sub msf {
    my $track = shift;
    my $lba = $track->get->{'lba'};
    return (int($lba / (60 * 75)), int(($lba % (60 * 75)) / 75), $lba % 75);
}

sub data { $_[0]->get->{'data'}; }

sub duration { $_[0]->get->{'duration'}; }

sub play_count { $_[0]->get->{'play_count'}; }

sub play_date { $_[0]->get->{'play_date'}; }

sub album {
    my $track = shift;
    return $track->{'album'} if defined($track->{'album'});
    if (!Info::Persist::configured) {
	Trace::error('Album must be specified at creation for non persistence');
	return;
    }
    # Tracks won't change albums, so this does not need to be placed in the sync space.
    $track->{'album'} = Info::Album->find(ioid => $track->album_ioid, owner => $track->owner);
    if ($track->ismonitor) {
	$track->{'album'} = $track->{'album'}->monitor;
    }
    return $track->{'album'};
}

sub change_name {
    my $track = shift;
    my $name = shift;

    return 1 if ($name eq $track->name);

    # The repo must do the work

    Thing::context->message(name => Repo::Change, to => $track->owner, parms => { object => $track, name => $name });
    return 1;
}

# updateStatsAfterPlay() - Update the play statistics after a full play of a track.

sub updateStatsAfterPlay {
    my $track = shift;

    # Count is incremented

    $track->setN;
    $track->getN->{'play_count'} = defined($track->play_count) ? $track->play_count + 1 : 1;
    $track->getN->{'play_date'} = POSIX::strftime('%G-%m-%d %H:%M:%S', localtime(time));
    $track->save;
    $track->sync;

    # Album play date, last time album was completely played i.e. earliest track play date

    my $albumMod = 0;
    if (my $date = $track->album->discover_play_date) {
	if (!defined($track->album->play_date) || ($date ne $track->album->play_date)) {
	    $albumMod = 1;
	    $track->album->setN;
	    $track->album->getN->{'play_date'} = $date;
	}
    }

    # Album play count

    if (my $count = $track->album->discover_play_count) {
	if (!defined($track->album->play_count) || ($count != $track->album->play_count)) {
	    $track->album->setN if (!$albumMod);
	    $albumMod = 1;
	    $track->album->getN->{'play_count'} = $count;
	}
    }
    if ($albumMod) {
	$track->album->save;
	$track->album->sync;
    }

    # Artist play date, most recently played song

    $track->album->artist->setN;
    $track->album->artist->getN->{'play_date'} = $track->getN->{'play_date'};
    $track->album->artist->save;
    $track->album->artist->sync;

}

sub monitor {
    my $track = shift;
    $track = Thing::Info::monitor($track, @_);
    if (defined($track->{'album'})) {
	$track->{'album'} = $track->{'album'}->monitor;
    }
    return $track;
}

1;
