# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.21.2.6 $

package Player;

use Thing;
@ISA = qw( Thing );
use PlayerH;
use Format::Mp3;
use Format::Wav;
use Info::Track;
use Trace;
use IO::File qw();
use POSIX qw(floor);
use POSIX qw(:fcntl_h);
use POSIX qw(:stdio_h);
use strict;

# OSS ioctls

BEGIN {
  use vars qw( $sndctl_dsp_getfmts $sndctl_dsp_getblksize $sndctl_dsp_setfmt $sndctl_dsp_speed $sndctl_dsp_channels 
	       $sndctl_dsp_post $afmt_s16_le $oss_ioctls );

  my $sys = `uname -smp`;
  chomp $sys;

  $oss_ioctls = 1;

  if ($sys =~ /Linux/) {

    # Only the upper nibble differs and I don't think it matters but
    # I'll leave it here for completeness as I don't have these
    # machines to test with.

    if ($sys =~ /Linux alpha/) {
      $sndctl_dsp_getfmts = 0x4004500b;
      $sndctl_dsp_getblksize = 0xc0045004;
      $sndctl_dsp_setfmt = 0xc0045005;
      $sndctl_dsp_speed = 0xc0045002;
      $sndctl_dsp_channels = 0xc0045006;
      $sndctl_dsp_post = 0x20005008;
      $afmt_s16_le = 0x10;
    }
    elsif ($sys =~ /Linux ppc/) {
      $sndctl_dsp_getfmts = 0x4004500b;
      $sndctl_dsp_getblksize = 0xc0045004;
      $sndctl_dsp_setfmt = 0xc0045005;
      $sndctl_dsp_speed = 0xc0045002;
      $sndctl_dsp_channels = 0xc0045006;
      $sndctl_dsp_post = 0x20005008;
      $afmt_s16_le = 0x10;
    }
    elsif ($sys =~ /Linux sparc64/) {
      $sndctl_dsp_getfmts = 0x2004500b;
      $sndctl_dsp_getblksize = 0x60045004;
      $sndctl_dsp_setfmt = 0x60045005;
      $sndctl_dsp_speed = 0x60045002;
      $sndctl_dsp_channels = 0x60045006;
      $sndctl_dsp_post = 0x5008;
      $afmt_s16_le = 0x10;
    }
    else {
      # From i686
      $sndctl_dsp_getfmts = 0x8004500b;
      $sndctl_dsp_getblksize = 0xc0045004;
      $sndctl_dsp_setfmt = 0xc0045005;
      $sndctl_dsp_speed = 0xc0045002;
      $sndctl_dsp_channels = 0xc0045006;
      $sndctl_dsp_post = 0x5008;
      $afmt_s16_le = 0x10;
    }
  }
  elsif ($sys =~ /FreeBSD/) {
    $sndctl_dsp_getfmts = 0x4004500b;
    $sndctl_dsp_getblksize = 0x40045004;
    $sndctl_dsp_setfmt = 0xc0045005;
    $sndctl_dsp_speed = 0xc0045002;
    $sndctl_dsp_channels = 0xc0045006;
    $sndctl_dsp_post = 0x20005008;
    $afmt_s16_le = 0x10;
  }
  else {
    $oss_ioctls = 0;
  }
}

sub SNDCTL_DSP_GETFMTS ()     { $sndctl_dsp_getfmts; }
sub SNDCTL_DSP_GETBLKSIZE ()  { $sndctl_dsp_getblksize; }
sub SNDCTL_DSP_SETFMT ()      { $sndctl_dsp_setfmt; }
sub SNDCTL_DSP_SPEED ()       { $sndctl_dsp_speed; }
sub SNDCTL_DSP_CHANNELS ()    { $sndctl_dsp_channels; }
sub SNDCTL_DSP_POST ()        { $sndctl_dsp_post; }
sub AFMT_S16_LE ()            { $afmt_s16_le; }


# DATA
#
# play - attributes of playing, flags etc
#
#   endOfSpool - All files are read and copied to the decoder buffer
#
#   endOfPlay - Last piece of data has been read from the decoder and
#     written to the dsp.
#
#   searching - Play is in search mode
#
#   direction - Direction of search
#
#   granularity - Limits the time spent in the play machine
#
#   status - hold status check when playing or paused
#
# decoder - attributes of players decoder process, that which takes in 
#           mp3 and outputs decoded audio
#
#   read - read handle if applicable
#   write - write handle
#   error - error handle
#
#   buffer - data to write to the decoder
#
#     frames - raw MP3 frames read in order from spool
#     chunks - to write to decoder in order
#     data - partial data that is not a full frame
#     length - of above
#     fh - Current file
#     fh_length - of above
#     fh_offset - read pointer in above
#
#   maxChunks - Maximum number of chunks to buffer
#
# dsp - attributes of DSP device, how we write to it etc
#
#   name - filename of the device
#   handle - to the device
#   blocksize - size of data block to write to device
#   blocksecs - audio seconds of above
#   buffer - array of blocks to write
#   maxBlocks - upper bound of above
#   partial - partial data read from decoder that is not a full block
#   length - of above
#   bufferStatus - hash of stats about the audio buffer
#
# spool - spool of tracks to play in order
# position - position of track to play or being played
# seek - seek to here on next file read
# builtin - Linux OSS driver is used
#
# mpg123BufferK - buffer values with mpg123 command line -b switch
# mpg123BufferSecs - seconds above equates to

# Implementation Note: Flow Control
#
# Two processes are forked. A decoder (mpg123) and optionally a
# soundcard driver. The original process reads frames from a spool of
# MP3 files and writes the data to the decoder. It blocks when the
# pipe to the decoder is full. The soundcard driver process reads from
# the decoder and writes raw PCM audio to the DSP device. It blocks on
# a read from the decoder if it has no audio buffered. Typically flow
# control is maintained by blocking on writes to the DSP device.

# NEW

sub new {
  my $proto = shift;
  my $type = ref($proto) || $proto;
  my $player = $type->SUPER::new(@_);
  if ($player->error) {
    return $player;
  }

  $player->{'spool'} = [];
  $player->{'position'} = 0;
  $player->{'state'} = Stopped;
  $player->{'attributes'} = {};
  $player->{'cause'} = None;
  $player->{'configured'} = 0;
  $player->{'monitors'} = {};
  $player->{'tag'} = 0;
  $player->{'builtin'} = 0;

  $player->register(name => Spool, handler => 'spool');
  $player->register(name => Clear, handler => 'clear');
  $player->register(name => State, handler => 'state');
  $player->register(name => Play, handler => 'play');
  $player->register(name => Stop, handler => 'stop');
  $player->register(name => Pause, handler => 'pause');
  $player->register(name => Seek, handler => 'seek');
  $player->register(name => Search, handler => 'search');
  $player->register(name => '_playMachine', handler => '_playMachine');

  # Driver
  $player->register(name => Thing::Started, handler => '_driverStarted');
  $player->register(name => '_driver_error', handler => '_driverError');
  $player->register(name => '_driver_buffered', handler => '_driverBuffered');
  $player->register(name => '_driver_EOP', handler => '_driverEOP');

  $player->register(name => Thing::ConfigureRequest, handler => '_configure');
  $player->register(name => Thing::ShutdownRequest, handler => '_shutdown');

  return $player;
}

# _configure

# playUsing => 'oss' use the builtin OSS driver to play decoded audio,
#   defaults to using mpg123 or the decoder command line if specified
#   for play.

# decoderCommandLine => allow full control, overriding default 'mpg123 -q -'.
#   Provides the ability to use another command line decoder other
#   than mpg123 or pass other options to mpg123. However the decoder
#   must accept input from the STDIN and output to the STDOUT when using
#   the OSS driver.

# device => specify an audio device other than the default /dev/dsp.

# decodeBufferSecs => seconds of encoded audio to buffer from files,
#   may improve play when files are accessed over a congested network.

# playBufferSecs => seconds of decoded audio to buffer. Higher values
#   may improve performance on machines doing other tasks.

# searchSkipSecs => length of audio to skip when searching in seconds.

# searchPlaySecs => length of audio to play when searching in seconds.

# db => database parameters.

sub _configure {
  my $player = shift;
  my %tParms = @_;
  my $parms = $tParms{'parms'};

  return 1 if !defined($parms);

  my $errorS = sub { 
    if (defined($parms->{'error'})) {
      $parms->{'error'} .= "\n".$_[0];
    }
    else {
      $parms->{'error'} = "Configuration error:\n".$_[0];
    }
  };
  my $stateC = 0;
  my $priorState;

  if (defined($parms->{'db'}) && !Info::Persist::configure(db => $parms->{'db'}->{'local'})) {
    &$errorS(Trace::lastError);
  }

  my $builtin = 0;
  if ($parms->{'playUsing'}) {
    if ($parms->{'playUsing'} eq 'oss') {
      $builtin = 1;
    }
    else {
      &$errorS('Play method: '.$parms->{'playUsing'}.', not supported.');
    }
  }

  if ($builtin && !$oss_ioctls) {
    &$errorS('No OSS ioctls are defined for this type of system. OSS cannot be used to play the decoded audio.')
  }

  if ($builtin != $player->{'builtin'}) {
    $priorState = $player->{'state'};
    $player->_stop();
    $stateC = 1 if ($priorState ne $player->{'state'});
  }
  $player->{'builtin'} = $builtin;

  if ($parms->{'decoderCommandLine'}) {
    $player->{'decoderCommandLine'} = $parms->{'decoderCommandLine'};
  }
  else {
    delete $player->{'decoderCommandLine'};
  }

  if ($parms->{'device'}) {
    $player->{'device'} = $parms->{'device'};
  }
  else {
    $player->{'device'} = '/dev/dsp';
  }
  
  if ($parms->{'decodeBufferSecs'}) {
    $player->{'decodeBufferSecs'} = $parms->{'decodeBufferSecs'};
  }
  else {
    $player->{'decodeBufferSecs'} = 5.0;
  }

  if ($parms->{'playBufferSecs'}) {
    $player->{'playBufferSecs'} = $parms->{'playBufferSecs'};
  }
  else {
    $player->{'playBufferSecs'} = 5.0;
  }

  if ($player->{'playBufferSecs'} && !$builtin) {
    $player->{'mpg123BufferK'} = POSIX::floor(($parms->{'playBufferSecs'} * 44100 * 4) / 1024);
    $player->{'mpg123BufferK'} = 1 if (!$player->{'mpg123BufferK'});
    $player->{'mpg123BufferSecs'} = POSIX::floor(($player->{'mpg123BufferK'} * 1024) / (44100 * 4));
  }
  else {
    $player->{'mpg123BufferK'} = 0;
    $player->{'mpg123BufferSecs'} = 0;
  }

  if ($parms->{'searchSkipSecs'}) {
    $player->{'searchSkipSecs'} = $parms->{'searchSkipSecs'};
  }
  else {
    $player->{'searchSkipSecs'} = 2.0;
  }

  if ($parms->{'searchPlaySecs'}) {
    $player->{'searchPlaySecs'} = $parms->{'searchPlaySecs'};
  }
  else {
    $player->{'searchPlaySecs'} = 0.5;
  }

  if ($parms->{'error'}) {
    $player->{'configured'} = 0;
    $player->{'configureError'} = $parms->{'error'};
  }
  else {
    $player->{'configured'} = 1;
    delete $player->{'configureError'};
  }

  if ($stateC || defined($parms->{'error'})) {
    $player->_state();
  }

  return 1;
}

# _shutdown - work is done

sub _shutdown {
  my $player = shift;
  $player->_stop;
}

# spool() - Create a spool from `tracks'. An optional argument `at'
# can specify to `append' the tracks to the existing spool or `insert'
# them at `position'.

sub spool {
  my $player = shift;
  my %tParms = @_;
  my $parms = $tParms{'parms'};

  $player->_addToMonitors($tParms{'from'});
  return $player->_state(to => $tParms{'from'}) if !$player->{'configured'};

  if (defined($parms->{'tracks'})) {
    foreach my $track (@{ $parms->{'tracks'} }) {
      my $format = Format::Mp3::bestQuality(values(%{ $track->{'formatsH'} }));
      if (!$format) {
	Trace::error('No format to play: %s', $track->name);
	$player->{'cause'} = Error;
	$player->{'attributes'}->{'diagnostic'} = Trace::lastError;
	$player->_state(to => $tParms{'from'});
	return 0;
      }
      $track->{'format'} = $format;
    }
  }

  if (defined($parms->{'at'}) && defined($parms->{'tracks'})) {
    if ($parms->{'at'} eq 'append') {
      push @{ $player->{'spool'} }, map { $_->monitor } @{ $parms->{'tracks'} };
    }
    elsif ($parms->{'at'} eq 'insert') {

      my $pos = (!defined($parms->{'position'}) || ($parms->{'position'} > $#{ $player->{'spool'} })) ? 
	undef : $parms->{'position'};

      if (defined($pos)) {
	splice @{ $player->{'spool'} }, $parms->{'position'}, 0, map { $_->monitor } @{ $parms->{'tracks'} };
      }
      else {
	push @{ $player->{'spool'} }, map { $_->monitor } @{ $parms->{'tracks'} };
      }
    }
  }
  elsif (defined($parms->{'tracks'})) {
    $player->infoDemonitorAll;
    if ($#{ $parms->{'tracks'} } >= 0) {
      $player->{'spool'} = [ map { $_->monitor } @{ $parms->{'tracks'} } ];
    }
    else {
      $player->{'spool'} = [];
    }
  }

  foreach my $dest (values(%{ $player->{'monitors'} })) {
    if (!$player->message(name => Spooled, to => $dest, parms => { tracks => $player->{'spool'}, }, )) {
      Trace::debug('Removed unresponsive state monitor: %s', $dest->unique) if (Trace::check(Trace::DEBUG3));
      delete $player->{'monitors'}->{ $dest->unique };
    }
  }
}

# clear() - Clear the spool. Stopping play if necessary and inform any
# changes in state.

sub clear {
  my $player = shift;
  my %tParms = @_;
  my $parms = $tParms{'parms'};

  $player->_addToMonitors($tParms{'from'});
  return $player->_state(to => $tParms{'from'}) if !$player->{'configured'};

  if (!defined($parms->{'rows'})) {
    $player->infoDemonitorAll;
    $player->_stop();
    $player->{'spool'} = [];
    $player->_state();
  }
  else {
    my @rows = sort @{ $parms->{'rows'} };
    while (@rows) {
      my $row = pop @rows;
      splice @{ $player->{'spool'} }, $row, 1;
    }
  }

  foreach my $dest (values(%{ $player->{'monitors'} })) {
    if (!$player->message(name => Spooled, to => $dest, parms => { tracks => $player->{'spool'}, }, )) {
      Trace::debug('Removed unresponsive state monitor: %s', $dest->unique) if (Trace::check(Trace::DEBUG3));
      delete $player->{'monitors'}->{ $dest->unique };
    }
  }
}

# state() - Inform sender of state info.

sub state {
  my $player = shift;
  my %tParms = @_;

  $player->_addToMonitors($tParms{'from'});
  $player->_state(to => $tParms{'from'});
}

sub _state {
  my $player = shift;
  my %parms = @_;

  my $status = {};
  $status->{'state'} = $player->{'state'};

  if (!$player->{'configured'}) {
    $status->{'cause'} = Error;
    $status->{'attributes'} = { diagnostic => $player->{'configureError'}, time => time };
  }
  else {
    $status->{'cause'} = $player->{'cause'};
    my $attr = $status->{'attributes'} = $player->{'attributes'};

    if ($player->{'state'} eq Stopped) {
      if (defined($player->{'spool'}) && @{ $player->{'spool'} }) {
	if (defined($player->{'position'})) {
	  $attr->{'position'} = $player->{'position'};
	}
	else {
	  $attr->{'position'} = 0;
	}
	if (defined($player->{'seek'})) {
	  $attr->{'seek'} = $player->{'seek'};
	}
	else {
	  $attr->{'seek'} = 0;
	}
      }
      $attr->{'time'} = time;
    }
    elsif (($player->{'state'} eq Playing) || ($player->{'state'} eq Paused)) {
      my $real;
      if (!defined($player->{'seek'}) && defined($player->{'play'}->{'status'})) {
	$real = $player->{'play'}->{'status'};
      }
      else {
	$real = $player->_playingAttributes;
      }
      $attr->{'position'} = $real->{'position'};
      $attr->{'seek'} = $real->{'seek'};
      $attr->{'endTime'} = $real->{'endTime'};
      $attr->{'time'} = $real->{'time'};
      $attr->{'searching'} = $player->{'play'}->{'searching'};
      $attr->{'direction'} = $player->{'play'}->{'direction'};

      Trace::debug('State: point: %d, seek: %f', $attr->{'position'}, $attr->{'seek'}) 
	if Trace::check(Trace::DEBUG3);
    }

    Trace::debug('State is %s due to %s', $status->{'state'}, $status->{'cause'}) if Trace::check(Trace::DEBUG3);
  }

  # Send on down
  my @dest = ();
  if (defined($parms{'to'})) {
    push @dest, $parms{'to'};
  }
  else {
    @dest = values(%{ $player->{'monitors'} });
  }
  foreach my $dest (@dest) {
    if (!$player->message(name => Status, to => $dest, parms => $status)) {
      Trace::debug('Removed unresponsive state monitor: %s', $dest->unique) if (Trace::check(Trace::DEBUG3));
      delete $player->{'monitors'}->{ $dest->unique };
    }
  }
  $player->{'cause'} = None;
  $player->{'attributes'} = {};
}

# PLAY

# play() - Play the spool. An optional `position' indicates a starting
# track. If current state is paused resume play. Inform any changes in
# state.

sub play {
  my $player = shift;
  my %tParms = @_;
  my $parms = $tParms{'parms'};

  $player->_addToMonitors($tParms{'from'});
  return $player->_state(to => $tParms{'from'}) if !$player->{'configured'};

  if ($#{ $player->{'spool'} } < 0) {
    $player->{'cause'} = Invalid;
    $player->{'attributes'}->{'diagnostic'} = 'Nothing to play';
    $player->_state(to => $tParms{'from'});
    return;
  }

  if (defined($parms->{'position'}) && (($parms->{'position'} > $#{ $player->{'spool'} }) || ($parms->{'position'} < 0))) {
    $player->{'cause'} = Invalid;
    $player->{'attributes'}->{'diagnostic'} = 'Play position is outside the spooled tracks';
    $player->_state(to => $tParms{'from'});
    return;
  }

  if (defined($parms->{'seek'}) && (($parms->{'seek'} > 100) || ($parms->{'seek'} < 0))) {
    $player->{'cause'} = Invalid;
    $player->{'attributes'}->{'diagnostic'} = 'Seek position should between 0 and 100';
    $player->_state(to => $tParms{'from'});
    return;
  }

  $player->{'cause'} = Message;

  if ($player->{'state'} eq Paused) {
    Trace::debug('Continuing play') if Trace::check(Trace::DEBUG3);
    $player->_continue();
  }
  else {
    $player->_stop() if ($player->{'state'} eq Playing);

    if (defined($parms->{'position'})) {
      $player->{'position'} = $parms->{'position'};
    }
    if (defined($parms->{'seek'})) {
      $player->{'seek'} = $parms->{'seek'};
    }

    Trace::debug('Play request at position %d', $player->{'position'}) if Trace::check(Trace::DEBUG3);
    $player->_play();
  }
  $player->_state();
}

sub _play {
  my $player = shift;

  my $errorSub = sub { 
    Trace::error(@_) if ($#_ != -1);

    $player->{'cause'} = Error;
    $player->{'attributes'}->{'diagnostic'} = Trace::lastError;

    if (defined($player->{'decoder'})) {
      $player->{'decoder'}->{'write'}->close if defined($player->{'decoder'}->{'write'});
      $player->{'decoder'}->{'read'}->close if defined($player->{'decoder'}->{'read'});
      delete $player->{'decoder'};
    }

    if (defined($player->{'dsp'})) {
      $player->{'dsp'}->{'handle'}->close;
      delete $player->{'dsp'};
    }
    return 0;
  };

  if ($player->{'state'} ne Stopped) {
    Trace::error(@_) if ($#_ != -1);
    $player->{'cause'} = Error;
    $player->{'attributes'}->{'diagnostic'} = Trace::lastError;
    return &$errorSub('Player must be stopped to play');
  }

  $player->{'decoder'} = my $decoder = {};
  $player->{'play'} = my $play = { endOfSpool => 0, endOfPlay => 0, searching => 0, direction => Forward };

  # Some tunables:
  # Limits the time spent in the play machine, affects response to user requests.
  $play->{'granularity'} = 0.1;

  # Max number of encoded chunks to buffer, a chunk will be searchPlay
  # seconds long. At least one chunk will be buffered.
  $decoder->{'maxChunks'} = POSIX::ceil($player->{'decodeBufferSecs'} / $player->{'searchPlaySecs'});

  if ($player->{'builtin'}) {

    # Linux OSS driver

    # TEST: Write to file
    if (defined($ENV{'DECODER_READ_TEST'})) {
      my $th = $player->{'testhandle'} = IO::File->new;
      if (!$th->open($ENV{'DECODER_READ_TEST'}, O_WRONLY|O_CREAT)) {
    	return &$errorSub('Cannot open the test file: %s', $!);
      }
    }

    # Ready the dsp

    my $handle = IO::File->new;
    my $name = $player->{'device'};

    if (!$handle->open($name, O_WRONLY)) {
      return &$errorSub('Cannot open the sound device: %s, %s', $name, $!);
    }

    $player->{'dsp'} = my $dsp = { name => $name, handle => $handle, buffer => [] };

    my $formats = pack('L', 0);
    if (!defined(ioctl($handle, SNDCTL_DSP_GETFMTS(), $formats))) {
      return &$errorSub('Cannot get sound devices supported formats: %s', $!);
    }
    $formats = unpack('L', $formats);

    if (!$formats & AFMT_S16_LE()) {
      return &$errorSub('16 bit sound is not supported by this sound device');
    }

    my $format = pack('L', AFMT_S16_LE());
    if (!defined(ioctl($handle, SNDCTL_DSP_SETFMT(), $format))) {
      return &$errorSub('Cannot set the play format: %s', $!);
    }
    $format = unpack('L', $format);

    if ($format != AFMT_S16_LE()) {
      return &$errorSub('16 bit sound could not be set');
    }

    my $speed = pack('L', 44100);
    if (!defined(ioctl($handle, SNDCTL_DSP_SPEED(), $speed))) {
      return &$errorSub('Cannot set speed: %s', $!);
    }
    $speed = unpack('L', $speed);

    if ($speed != 44100) {
      return &$errorSub('Speed could not be set');
    }

    my $channels = pack('L', 2);
    if (!defined(ioctl($handle, SNDCTL_DSP_CHANNELS(), $channels))) {
      return &$errorSub('Cannot set to stereo mode: %s', $!);
    }
    $channels = unpack('L', $channels);

    if ($channels != 2) {
      return &$errorSub('Stereo mode could not be set');
    }

    my $blksize = pack('L', 0);
    if (!defined(ioctl($handle, SNDCTL_DSP_GETBLKSIZE(), $blksize))) {
      return &$errorSub('Cannot get sound devices block size: %s', $!);
    }
    $blksize = unpack('L', $blksize);
    $dsp->{'blocksize'} = $blksize;

    # Some constants:
    # How many seconds is a dsp block
    $dsp->{'blocksecs'} = $player->{'dsp'}->{'blocksize'} / (44100 * 4.0);
    # How many blocks in the buffer. At least one.
    $dsp->{'maxBlocks'} = POSIX::ceil($player->{'playBufferSecs'} / $player->{'dsp'}->{'blocksecs'});
    # How many blocks on each write to the dsp
    $dsp->{'writeBlocks'} = POSIX::ceil($play->{'granularity'} / $player->{'dsp'}->{'blocksecs'});
  }

  if (!pipe(CR, PW) || !pipe(PR, CW) || !pipe(ER, EW)) {
    return &$errorSub('Cannot create decoder pipes: %s', $!); 
  }

  my $pid = CORE::fork;
  if (!defined($pid)) {
    return &$errorSub('Cannot fork: %s', $!);
  }

  if ($pid) {

    # Parent

    $decoder->{'pid'} = $pid;
    $decoder->{'write'} = *PW;
    $decoder->{'read'} = *PR;
    $decoder->{'error'} = *ER;

    close CR;
    close CW;
    close EW;
  }
  else {

    # Decoder child

    close PW;
    close PR;
    close ER;

    POSIX::dup2(*CR->fileno, *STDIN->fileno) || die $!;
    POSIX::dup2(*CW->fileno, *STDOUT->fileno) || die $!;
    POSIX::dup2(*EW->fileno, *STDERR->fileno) || die $!;

    my $cmd;
    if (defined($player->{'decoderCommandLine'})) {
      $cmd = $player->{'decoderCommandLine'};
    }
    else {
      $cmd = 'mpg123 ';
      if ($player->{'builtin'}) {
	$cmd .= '-qs -';
      }
      else {
	if ($player->{'mpg123BufferK'}) {
	  $cmd .= '-b '.$player->{'mpg123BufferK'}.' ';
	}
	$cmd .= '-a '.$player->{'device'}.' ';
	$cmd .= '-q -';
      }
    }

    if ($player->{'builtin'}) {
      $player->{'dsp'}->{'handle'}->close;
    }

    exec($cmd);
    die "Cannot execute $cmd: $!";
  }

  if ($player->{'builtin'}) {
    if (!Thing::daemon(type => 'Player::Driver', name => $player->name.'_driver', loaded => 1, 
		       player => $player, close => $player,
		       dsp => $player->{'dsp'}, decoder => $player->{'decoder'}, play => $player->{'play'})) {
      return &$errorSub();
    }
    $decoder->{'read'}->close;
    delete $decoder->{'read'};
  }
  else {
    Trace::debug('Starting play machine') if Trace::check(Trace::DEBUG4);
    $player->message(name => '_playMachine', to => $player, parms => { tag => ++$player->{'tag'} });
  }

  # Writing to the decoder is blocking for flow control
  $decoder->{'write'}->fcntl(F_SETFL, 0);
  $decoder->{'oflags'} = 0;
  $decoder->{'error'}->fcntl(F_SETFL, 0);

  # Now in play mode
  $player->{'state'} = Playing;

  return 1;
}

sub _playMachine {
  my $player = shift;
  my %tParms = @_;
  my $parms = $tParms{'parms'};

  # Check for a change in play.
  return 0 if ($player->{'tag'} ne $parms->{'tag'});
  return 0 if ($player->{'state'} ne Playing);

  my $errorSub = sub {
    Trace::error(@_) if ($#_ != -1);
    $player->{'cause'} = Error;
    $player->{'attributes'}->{'diagnostic'} = Trace::lastError;
    $player->_stop();
    $player->_state();
    return 0;
  };

  my $play = $player->{'play'};
  my $decoder = $player->{'decoder'};

  if (!defined($decoder->{'buffer'})) {
    $decoder->{'buffer'} = { frames => [], chunks => [], data => '', length => 0 };
  }

  if (!$player->_framesReadFromFile || !defined($player->_framesWriteToDecoder($play->{'granularity'}))) {
    return &$errorSub();
  }

  if (!$player->{'builtin'}) {
    if ($play->{'endOfSpool'} && !@{ $decoder->{'buffer'}->{'chunks'} }) {
      $player->{'cause'} = End;
      $player->_stop;
      $player->_state;
      return;
    }
  }

  if (!defined($play->{'status'})) {
    $play->{'status'} = $player->_playingAttributes;
  }
  if ($play->{'status'}->{'time'} < time) {
    # Update the stats every second and check for a track switchover
    my $oldstatus = $play->{'status'};
    $play->{'status'} = $player->_playingAttributes;
    if ($oldstatus->{'position'} != $play->{'status'}->{'position'}) {
      $player->_state;
    }
  }
  $player->message(name => '_playMachine', to => $player, parms => { tag => $player->{'tag'} });

  return 1;
}

# _continue() - Resume play.

sub _continue {
  my $player = shift;

  # Resume the _playMachine
  $player->message(name => '_playMachine', to => $player, parms => { tag => ++$player->{'tag'} });
  $player->{'state'} = Playing;
  if ($player->{'builtin'}) {
    $player->message(name => '_driver_play', to => $player->{'dsp'}->{'driver'}) if defined($player->{'dsp'}->{'driver'});
  }

  return 1;
}

# STOP

# stop() - Stop play. Inform any changes in state. Silently ignore
# when already stopped.

sub stop {
  my $player = shift;
  my %tParms = @_;
  $player->_addToMonitors($tParms{'from'});
  return $player->_state(to => $tParms{'from'}) if !$player->{'configured'};

  Trace::debug('Stop request') if Trace::check(Trace::DEBUG3);

  if ($player->{'state'} ne Stopped) {
    $player->{'cause'} = Message;
    $player->_stop();
    $player->_state();
  }
  return 1;
}

# _stop() - Stops play and reset state.

sub _stop {
  my $player = shift;

  $player->{'position'} = 0;

  delete $player->{'seek'};
  delete $player->{'search'};
  delete $player->{'play'};

  if ($player->{'builtin'}) {
    if (defined($player->{'dsp'}->{'driver'}) &&
	$player->message(name => Thing::ShutdownRequest, to => $player->{'dsp'}->{'driver'})) {
      sleep 1; # Give it a chance to close the DSP
    }
    $player->{'dsp'}->{'handle'}->close if defined($player->{'dsp'}->{'handle'});
    delete $player->{'dsp'};

    if (defined($ENV{'DECODER_READ_TEST'})) {
      $player->{'testhandle'}->close;
    }
  }

  if (defined($player->{'decoder'})) {

    my $decoder = $player->{'decoder'};

    if (($player->{'cause'} eq Error) && defined($decoder->{'error'})) {

      # Read decoder output for diagnostics

      my $diagnostic;
      if (!defined($player->{'attributes'})) {
	$player->{'attributes'} = { diagnostic => '' };
      }
      elsif (!defined($player->{'attributes'}->{'diagnostic'})) {
	$player->{'attributes'}->{'diagnostic'} = '';
      }
      else {
	$player->{'attributes'}->{'diagnostic'} .= "\n";
      }

      $diagnostic = $player->{'attributes'}->{'diagnostic'};
      $diagnostic .= 'Output from the decoder:'."\n";

      my $data;
      while ((length $diagnostic < 8192) && sysread($decoder->{'error'}, $data, 1024)) {
	$diagnostic .= $data;
      }

      $player->{'attributes'}->{'diagnostic'} = $diagnostic;
    }

    $decoder->{'error'}->close if defined($decoder->{'error'});
    $decoder->{'write'}->close if defined($decoder->{'write'});
    $decoder->{'read'}->close if defined($decoder->{'read'});
    $decoder->{'buffer'}->{'fh'}->close if (defined($decoder->{'buffer'}) && defined($decoder->{'buffer'}->{'fh'}));

    if (!$player->{'builtin'} && defined($decoder->{'pid'})) {
      # Wait for the decoder to die to open up dsp
      waitpid($decoder->{'pid'}, 0);
      delete $decoder->{'pid'};
    }

    delete $player->{'decoder'};
  }

  $player->{'state'} = Stopped;

  Trace::debug('Stopped play') if Trace::check(Trace::DEBUG4);
  return 1;
}

# PAUSE

# pause() - Pause play. Inform any changes in state. Silently ignore if not playing.

sub pause {
  my $player = shift;
  my %tParms = @_;

  $player->_addToMonitors($tParms{'from'});
  return $player->_state(to => $tParms{'from'}) if !$player->{'configured'};

  Trace::debug('Pause request') if Trace::check(Trace::DEBUG3);

  if ($player->{'state'} eq Playing) {
    $player->{'cause'} = Message;
    $player->{'state'} = Paused; # Stops the _playMachine
    delete $player->{'play'}->{'status'}; # Remove prior status to ensure generation of new
    
    if ($player->{'builtin'}) {
      $player->message(name => '_driver_stop', to => $player->{'dsp'}->{'driver'}) if defined($player->{'dsp'}->{'driver'});
    }
    
    Trace::debug('Paused') if Trace::check(Trace::DEBUG4);
    
    $player->_state();
  }
  return 1;
}

# SEEK

# seek() - Change point of play, `position' in spool and percentage 'seek' into the track.

sub seek {
  my $player = shift;
  my %tParms = @_;
  my $parms = $tParms{'parms'};

  $player->_addToMonitors($tParms{'from'});
  return $player->_state(to => $tParms{'from'}) if !$player->{'configured'};

  if (defined($parms->{'position'}) && (($parms->{'position'} > $#{ $player->{'spool'} }) || ($parms->{'position'} < 0))) {
    $player->{'cause'} = Invalid;
    $player->{'attributes'}->{'diagnostic'} = 'Play position is outside the spooled tracks';
    $player->_state(to => $tParms{'from'});
    return;
  }

  if (defined($parms->{'seek'}) && (($parms->{'seek'} > 100) || ($parms->{'seek'} < 0))) {
    $player->{'cause'} = Invalid;
    $player->{'attributes'}->{'diagnostic'} = 'Seek position should be between 0 and 100';
    $player->_state(to => $tParms{'from'});
    return;
  }

  if (!defined($parms->{'position'}) && !defined($parms->{'seek'})) {
    $player->{'cause'} = Invalid;
    $player->{'attributes'}->{'diagnostic'} = 'Seeking requires somewhere to seek to';
    $player->_state(to => $tParms{'from'});
    return;
  }

  Trace::debug('Seek request') if Trace::check(Trace::DEBUG3);

  my $oldPosition;
  if (defined($parms->{'position'})) {
    $oldPosition = $player->{'position'};
    $player->{'position'} = $parms->{'position'};
  }
  if (defined($parms->{'seek'})) {
    $player->{'seek'} = $parms->{'seek'};
  }
  else {
    $player->{'seek'} = 0.0;
  }

  if (defined($player->{'decoder'}) && defined($player->{'decoder'}->{'buffer'})) {

    my $buffer = $player->{'decoder'}->{'buffer'};

    # Discard all but current chunk

    if ($player->{'play'}->{'direction'} eq Forward) {
      $buffer->{'chunks'} = [ $buffer->{'chunks'}->[0] ];
    }
    else {
      $buffer->{'chunks'} = [ $buffer->{'chunks'}->[ $#{ $buffer->{'chunks'} } ] ];
    }

    $buffer->{'frames'} = [];
    $buffer->{'data'} = '';
    $buffer->{'length'} = 0;

    # Close file

    if ((defined($oldPosition) && ($oldPosition != $player->{'position'})) && defined($buffer->{'fh'})) {
      $buffer->{'fh'}->close;
      delete $buffer->{'fh'};
      delete $buffer->{'fh_length'};
      delete $buffer->{'fh_offset'};
    }

    Trace::debug('Dropped decoder buffer') if Trace::check(Trace::DEBUG4);
  }

  if ($player->{'builtin'} && defined($player->{'dsp'})) {

    # Discard builtin player buffer

    $player->message(name => '_driver_discard', to => $player->{'dsp'}->{'driver'}) if defined($player->{'dsp'}->{'driver'});
    $player->{'dsp'}->{'buffered'} = 0.0;

    Trace::debug('Dropping dsp buffer') if Trace::check(Trace::DEBUG4);
  }

  # Clear end conditions before seek

  my $play = $player->{'play'};
  $play->{'endOfSpool'} = 0;

  delete $play->{'status'}; # Remove prior status to ensure generation of new
  $player->{'cause'} = Message;
  $player->_state();

  return 1;
}

# SEARCH

# search() - Turn on or off `search' mode in `direction'.

sub search {
  my $player = shift;
  my %tParms = @_;
  my $parms = $tParms{'parms'};

  $player->_addToMonitors($tParms{'from'});
  return $player->_state(to => $tParms{'from'}) if !$player->{'configured'};

  if (($player->{'state'} ne Playing) || !defined($player->{'play'})) {
    $player->{'cause'} = Ignored;
    $player->{'attributes'}->{'diagnostic'} = 'Cannot search, not playing';
    $player->_state(to => $tParms{'from'});
    return;
  }

  if (defined($parms->{'direction'}) && ($parms->{'direction'} ne Forward) && ($parms->{'direction'} ne Reverse)) {
    $player->{'cause'} = Invalid;
    $player->{'attributes'}->{'diagnostic'} = 'Search direction should be forward or reverse';
    $player->_state(to => $tParms{'from'});
    return;
  }

  Trace::debug('Search request') if Trace::check(Trace::DEBUG3);

  my $status = $player->_playingAttributes;
  my $play = $player->{'play'};
  my $oldDirection = $play->{'direction'};

  if ($parms->{'search'}) {
    $play->{'searching'} = 1;
  }
  else {
    $play->{'searching'} = 0;
    $play->{'direction'} = Forward;
  }

  if (defined($parms->{'direction'})) {
    $play->{'direction'} = $parms->{'direction'};
  }

  # Search from the current estimated play position not the file read
  # point.

  $player->seek(from => $tParms{'from'}, parms => { position => $status->{'position'}, seek => $status->{'seek'} });

  return 1;
}

# FRAMES

# _framesReadFromFile - Read a decode chunk worth of
# frames. Return is 0 on failure, 1 on success.

sub _framesReadFromFile {
  my $player = shift;

  my $play = $player->{'play'};
  my $decoder = $player->{'decoder'};
  my $buffer = $decoder->{'buffer'};

  return 1 if ($play->{'endOfSpool'} && !@{ $buffer->{'frames'} }); # All frames read
  return 1 if ($#{ $buffer->{'chunks'} } + 1 >= $decoder->{'maxChunks'}); # Full

  my $have = 0.0;
  map { $have += $_->{'secs'} } @{ $buffer->{'frames'} };

  # In search mode need to read more than will be played

  my $need;
  if ($play->{'searching'}) {
    $need = $player->{'searchPlaySecs'} + $player->{'searchSkipSecs'};
  }
  else {
    $need = $player->{'searchPlaySecs'};
  }

  Trace::debug('Request for %f seconds, have %f', $need, $have) if Trace::check(Trace::DEBUG4);

  my $blocksize = 8192;
  my $partial = undef;
  my $initialRead = 0;
  my $file;
  my $fh;
  my $track;
  my $seek;
  my $whence;
  my $data;
  my $bytesRead;

  while (($need > $have) && !$play->{'endOfSpool'}) {

    # Open a file if need be

    if (!defined($buffer->{'fh'})) {

      Trace::timestamp('file open') if (Trace::performance && Trace::check(Trace::DEBUG1));

      $file = $player->{'spool'}->[ $player->{'position'} ]->{'format'}->file;
      $fh = $buffer->{'fh'} = IO::File->new;

      Trace::debug('Handle creation: %f', Trace::timediff('file open')) 
	if (Trace::performance && Trace::check(Trace::DEBUG1));

      if (!$fh->open($file)) {
	Trace::error('Error opening file: %s, %s', $file, $!);
	delete $buffer->{'fh'};
	return 0;
      }

      Trace::debug('Open part: %f', Trace::timediff('file open')) 
	if (Trace::performance && Trace::check(Trace::DEBUG1));

      if (!defined($buffer->{'fh_length'} = sysseek $fh, 0, 2)) {
	Trace::error('Error seeking in file: %s, %s', $file, $!);
	return 0;
      }

      Trace::debug('Seek part: %f', Trace::timediff('file open')) 
	if (Trace::performance && Trace::check(Trace::DEBUG1));

      Trace::debug('Now using file: %s, %d bytes', $file, $buffer->{'fh_length'}) if Trace::check(Trace::DEBUG3);

      if ($play->{'direction'} eq Reverse) {
	if ($buffer->{'fh_length'} < $blocksize) {
	  $partial = $buffer->{'fh_length'};
	  $whence = 0;
	  $seek = 0;
	}
	else {
	  $partial = undef;
	  $whence = 1;
	  $seek = -$blocksize;
	}
	$initialRead = 1;
      }
      else {
	$whence = 0;
	$seek = 0;
      }

      if (!defined($buffer->{'fh_offset'} = sysseek $buffer->{'fh'}, $seek, $whence)) {
	Trace::error('Error seeking in file: %s, %s', $file, $!);
	return 0;
      }
      Trace::debug('Seek to: %d', $buffer->{'fh_offset'}) if Trace::check(Trace::DEBUG4);
    }

    # External Seek Request

    if (defined($player->{'seek'})) {

      $seek = int (($buffer->{'fh_length'} * $player->{'seek'}) / 100);
      if ($play->{'direction'} eq Reverse) {
	$seek -= $blocksize;
	if ($seek < 0) {
	  $seek = 0;
	}
 	elsif ($buffer->{'fh_length'} < $blocksize) {
	  $partial = $buffer->{'fh_length'};
	  $seek = 0;
	}
      }
      if (!defined($buffer->{'fh_offset'} = sysseek $buffer->{'fh'}, $seek, 0)) {
	Trace::error('Error seeking in file: %s, %s', $file, $!);
	return 0;
      }
      Trace::debug('Seek to: %d', $buffer->{'fh_offset'}) if Trace::check(Trace::DEBUG4);
      delete $player->{'seek'};
    }

    # Reverse seek after read

    elsif (($play->{'direction'} eq Reverse) && !$initialRead) {

      if ($buffer->{'fh_offset'} < (2 * $blocksize)) {
	$partial = $buffer->{'fh_offset'} - $blocksize;
	$whence = 0;
	$seek = 0;
      }
      else {
	$whence = 1;
	$seek = -(2 * $blocksize);
      }

      if (!defined($buffer->{'fh_offset'} = sysseek $buffer->{'fh'}, $seek, $whence)) {
	Trace::error('Error seeking in file: %s, %s', $file, $!);
	return 0;
      }
      Trace::debug('Seek to: %d', $buffer->{'fh_offset'}) if Trace::check(Trace::DEBUG4);
    }
    elsif ($initialRead) {
      $initialRead = 0;
    }

    # Read

    $bytesRead = sysread($buffer->{'fh'}, $data, defined($partial) ? $partial : $blocksize);
    if (!defined($bytesRead) || ($bytesRead < 0)) {
      Trace::error('Error reading from file: %s, %s', $file, $!);
      return 0;
    }
    $buffer->{'length'} += $bytesRead;
    $buffer->{'fh_offset'} += $bytesRead;
    if ($play->{'direction'} eq Forward) {
      $buffer->{'data'} .= $data;
    }
    else {
      $buffer->{'data'} = $data.$buffer->{'data'};
    }

    Trace::debug('Read block of mp3, %d/%d bytes', $bytesRead, $buffer->{'length'}) if Trace::check(Trace::DEBUG4);

    if ((($play->{'direction'} eq Forward) && (($bytesRead < $blocksize) || ($buffer->{'fh_offset'} == $buffer->{'fh_length'}))) ||
	(($play->{'direction'} eq Reverse) && (($buffer->{'fh_offset'} - $bytesRead) == 0))) {

      $track = $player->{'spool'}->[ $player->{'position'} ];
      if ($play->{'direction'} eq Forward) {
	if ($player->{'position'} < $#{ $player->{'spool'} }) {
	  # Move to next track
	  $player->{'position'}++;
	}
	else {
	  Trace::debug('End of spool') if Trace::check(Trace::DEBUG2);
	  $play->{'endOfSpool'} = 1;
	}
      }
      else {
	if ($player->{'position'} && ($player->{'position'} <= $#{ $player->{'spool'} })) {
	  # Move to previous track
	  $player->{'position'}--;
	}
	else {
	  Trace::debug('Beginning of spool') if Trace::check(Trace::DEBUG2);
	  $play->{'endOfSpool'} = 1;
	}
      }

      $buffer->{'fh'}->close;
      $track->updateStatsAfterPlay if (!$play->{'searching'});
      delete $buffer->{'fh'};
      delete $buffer->{'fh_length'};
      delete $buffer->{'fh_offset'};
      $partial = undef;

      if (!$play->{'endOfSpool'}) {
	Trace::debug('Moved spool position to %d', $player->{'position'}) if Trace::check(Trace::DEBUG3);
      }
      else {
	if ($player->{'builtin'}) {
	  $player->message(name => '_driver_EOS', to => $player->{'dsp'}->{'driver'}) if defined($player->{'dsp'}->{'driver'});
	}
      }
    }

    # Find the frames within this data and move them to the raw frames array

    $player->_framesFind;

    $have = 0.0;
    map { $have += $_->{'secs'} } @{ $buffer->{'frames'} };

    Trace::debug('Now have %f seconds', $have) if Trace::check(Trace::DEBUG4);
  }

  my $chunks = $buffer->{'chunks'};

  # Have enough frames to create a chunk or it is the end of spool.
  # A chunk is frames with,
  # wp - Write pointer to frame to write to decoder
  # woffset - offset of last write from the above frame

  my $chunk = { 'frames' => [], wp => 0, woffset => 0 };
  my $frames = $chunk->{'frames'};
  my $frame;

  if ($play->{'searching'}) {
    $have = 0.0;
    while (@{ $buffer->{'frames'} } && ($have < $player->{'searchSkipSecs'})) {
      if ($play->{'direction'} eq Forward) {
	# Forward searching drop from front
	$frame = shift @{ $buffer->{'frames'} };
      }
      else {
	# Reverse searching drop from end
	$frame = pop @{ $buffer->{'frames'} };
      }
      Trace::debug('Dropped frame [0x%08x]', $frame) if Trace::check(Trace::DEBUG5);
      $have += $frame->{'secs'};
    }
  }

  $have = 0.0;
  while (@{ $buffer->{'frames'} } && ($have < $player->{'searchPlaySecs'})) {
    if ($play->{'direction'} eq Forward) {
      push @{ $frames }, shift @{ $buffer->{'frames'} };
      $have += $frames->[0]->{'secs'};
    }
    else {
      unshift @{ $frames }, pop @{ $buffer->{'frames'} };
      $have += $frames->[ $#{ $frames } ]->{'secs'};
    }
  }

  if ($play->{'direction'} eq Forward) {
    push @{ $chunks }, $chunk;
  }
  else {
    unshift @{ $chunks }, $chunk;
  }

  Trace::debug('Taken %d frames (%f seconds) for decode chunk', $#{ $frames } + 1, $have) if Trace::check(Trace::DEBUG4);
  map { Trace::debug('Taken frame [0x%08x]', $_); } @{ $frames } if Trace::check(Trace::DEBUG5);

  return 1;
}

# _framesFind - Find valid MP3 frames in the data and place them in
# frames, works out the seconds of audio in the frames and the
# expected size of generated audio.

sub _framesFind {
  my $player = shift;

  my $buffer = $player->{'decoder'}->{'buffer'};
  my $play = $player->{'play'};

  my $frames = _dataToFrames($buffer->{'data'}, $buffer->{'length'});
  return 0 if (!$frames || ($#{ $frames } == -1));

  # Clean up possibly missing sizes
  my $pos = 0;
  foreach my $frame (@{ $frames }) {
    if (!defined($frame->{'size'})) {
      if ($pos < $#{ $frames }) {
	$frame->{'size'} = $frames->[ $pos + 1 ]->{'offset'} - $frame->{'offset'};
      }
    }
    $pos++;
  }

  my $last = $frames->[ $#{ $frames } ];
  if (!defined($last->{'size'})) {
    # No size on last frame is not much help
    pop @{ $frames };
  }

  $last = $frames->[ $#{ $frames } ];
  if ($last->{'offset'} + $last->{'size'} > $buffer->{'length'}) {
    # last frame outside of data is not much help either
    pop @{ $frames };
  }

  return 0 if (!$frames || ($#{ $frames } == -1));

  # Take the data

  foreach my $frame (@{ $frames }) {
    $frame->{'data'} = substr($buffer->{'data'}, $frame->{'offset'}, $frame->{'size'});
    $frame->{'secs'} = ($frame->{'size'} * 8.0) / $frame->{'bitrate'};
  }

  if ($play->{'direction'} eq Forward) {
    # Keep unused data at end of buffer
    $last = $frames->[ $#{ $frames } ];
    if ($last->{'offset'} + $last->{'size'} < $buffer->{'length'}) {
      $buffer->{'length'} = $buffer->{'length'} - ($last->{'offset'} + $last->{'size'});
      $buffer->{'data'} = substr($buffer->{'data'}, $last->{'offset'} + $last->{'size'}, $buffer->{'length'});
    }
    else {
      $buffer->{'length'} = 0;
      $buffer->{'data'} = '';
    }

    # Frames get tacked onto the end of any existing ones.
    if (defined($buffer->{'frames'})) {
      push @{ $buffer->{'frames'} }, @{ $frames };
    }
    else {
      $buffer->{'frames'} = $frames;
    }
  }
  else {
    # Keep starting dribble
    my $first = $frames->[ 0 ];
    if ($first->{'offset'}) {
      $buffer->{'length'} = $first->{'offset'};
      $buffer->{'data'} = substr($buffer->{'data'}, 0, $buffer->{'length'});
    }
    else {
      $buffer->{'length'} = 0;
      $buffer->{'data'} = '';
    }

    # Frames get tacked at the front of any existing ones.
    if (defined($buffer->{'frames'})) {
      unshift @{ $buffer->{'frames'} }, @{ $frames };
    }
    else {
      $buffer->{'frames'} = \@{ $frames };
    }
  }

  return 1;
}

my @t_bitrate =
  ([
    [0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448],
    [0, 32, 48, 56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 384],
    [0, 32, 40, 48,  56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320],
   ],
   [
    [0, 32, 48, 56,  64,  80,  96, 112, 128, 144, 160, 176, 192, 224, 256],
    [0,  8, 16, 24,  32,  40,  48,  56,  64,  80,  96, 112, 128, 144, 160],
    [0,  8, 16, 24,  32,  40,  48,  56,  64,  80,  96, 112, 128, 144, 160],
   ]);

my @t_sampling_freq =
  (
   [11025, 12000,  8000],
   [undef, undef, undef], # reserved
   [22050, 24000, 16000],
   [44100, 48000, 32000]
  );

# _dataToFrames - Looks for MP3 frames in the data. Frames must be
# properly stacked i.e. the size of each frame correctly indicates the
# start of another MP3 frame. Reduces the possibility of reporting
# invalid frames. Data must be large enough to contain at least two
# frames.

sub _dataToFrames {
  #my $data = $_[0];
  my $length = $_[1];

  my @frames = ();
  my $offset = 0;
  my $roffset = 0;
  my $f;
  my $lookAhead;
  my $word;

  while (($length - $offset) >= 4) {

    $word = substr($_[0], $offset, 4);
    $word = unpack('N', $word);

    # Find sync bits
    if ((($word & 0xFFE00000) == 0xFFE00000) && ($f = _is_mp3($word))) {
      # Might be an MP3 frame
      $f->{'offset'} = $offset;
      push @frames, $f;
      Trace::debug('Possible frame [0x%08x] at %d, size %d', $f, $offset, $f->{'size'}) if Trace::check(Trace::DEBUG5);
      # Skip ahead to the next frame
      $offset += $f->{'size'};
    }
    else {
      # Drop any found frames and try from the next byte.
      $roffset++;
      $offset = $roffset;
      @frames = ();
    }
  }

  return \@frames if ($#frames > 0);
  return 0;
}

sub _is_mp3 {
  my $word = shift;

  my $id = ($word >> 19) &0x3;
  return 0 if ($id == 1);
  my $layer = ($word >> 17) &0x3;
  return 0 if ($layer == 0);
  $layer = (0, 3, 2, 1)[ $layer ];

  my $protected = !($word >> 16) &0x1;

  my $bitrateIndex = ($word >> 12) &0xf;
  return 0 if ($bitrateIndex == 0xf);

  my $sampleRateIndex = ($word >> 10) &0x3;
  return 0 if ($sampleRateIndex == 3);

  my $padded = ($word >> 9) &0x1;

  my $emphasis = ($word) &0x3;
  return 0 if ($emphasis == 2);

  my $bitrate = 1000 * $t_bitrate[($id == 3) ? 0 : 1][$layer - 1][$bitrateIndex];
  return 0 if !$bitrate;

  my $sampleRate = $t_sampling_freq[$id][$sampleRateIndex];

  my $frameSize;
  $frameSize = POSIX::floor((($id == 3) ? 144 : 72) * $bitrate / $sampleRate);
  $frameSize += ($layer == 1) ? 4 : 1 if ($padded);
  $frameSize += 4 if ($protected);

  return { size => $frameSize, bitrate => $bitrate, };
}

# _framesWriteToDecoder - Write frames to the decoder. 
#
# Can operate in a number of modes based on the manner in which the
# decoder pipe is opened;
#
#   blocking: will write at least `seconds' of audio.
#   non-blocking: will write as much audio as possible or up to `seconds' 
#     before a partial write occurs.
#
# Returns the seconds written on success, undefined on error.

sub _framesWriteToDecoder {
  my $player = shift;
  my $seconds = shift;

  my $play = $player->{'play'};
  my $decoder = $player->{'decoder'};
  my $buffer = $decoder->{'buffer'};
  my $chunks = $buffer->{'chunks'};

  if ($play->{'endOfSpool'} && !@{ $chunks }) {
    Trace::debug('All chunks written to decoder') if Trace::check(Trace::DEBUG4);
    return 1;
  }

  my @order;
  my $chunk;
  my $offset;
  my $frame;
  my $bytesWritten;
  my $written = 0.0;

  while (@{ $chunks }) {

    if ($play->{'direction'} eq Forward) {
      $chunk = $chunks->[0];
    }
    else {
      $chunk = $chunks->[ $#{ $chunks } ];
    }
    $offset = $chunk->{'woffset'};

    while ($chunk->{'wp'} <= $#{ $chunk->{'frames'} }) {

      $frame = $chunk->{'frames'}->[ $chunk->{'wp'} ];

      do {
	$bytesWritten = syswrite($decoder->{'write'}, $frame->{'data'}, $frame->{'size'} - $offset, $offset);

	if ($decoder->{'oflags'} & O_NONBLOCK) {
	  if (defined($bytesWritten) && ($bytesWritten < 0)) {
	    Trace::error('Error writing to the decoder');
	    return undef;
	  }
	}
	else {
	  if (!defined($bytesWritten) || ($bytesWritten < 0)) {
	    Trace::error('Error writing to the decoder');
	    return undef;
	  }
	}

	if ($bytesWritten) {
	  $offset += $bytesWritten;
	}
      } while ($bytesWritten && ($offset != $frame->{'size'}));

      # TEST: write to seperate files.
      if (defined($ENV{'DECODER_WRITE_TEST'})) {
	my $dwt = $ENV{'DECODER_WRITE_TEST'};
	my $file = 'dwt/'.sprintf('%08d', $dwt).'.mp3';
	my $fh = IO::File->new;
	$fh->open('>'.$file) || die $!;
	syswrite($fh, $frame->{'data'}, $frame->{'size'});
	$fh->close;
	$ENV{'DECODER_WRITE_TEST'}++;
	Trace::debug('Saved frame to %s', $file) if Trace::check(Trace::DEBUG5);
      }

      Trace::debug('Wrote frame [0x%08x] to decoder, %d/%d bytes', $frame, $offset, $frame->{'size'}) 
	if Trace::check(Trace::DEBUG5);

      if ($offset == $frame->{'size'}) {
	$offset = 0;
	$written += $frame->{'secs'};
	if ($seconds && ($written >= $seconds)) {
	  Trace::debug('Wrote requested amount to decoder') if Trace::check(Trace::DEBUG5);
	  $chunk->{'wp'}++;
	  $chunk->{'woffset'} = $offset;
	  return $written;
	}
      }
      else {
	# Blocking on write, get out
	$chunk->{'woffset'} = $offset;
	return 1;
      }

      $chunk->{'wp'}++;
    }

    if ($chunk->{'wp'} > $#{ $chunk->{'frames'} }) {
      Trace::debug('Wrote last frame from chunk to decoder') if Trace::check(Trace::DEBUG4);
      if ($play->{'direction'} eq Forward) {
	shift @{ $chunks };
      }
      else {
	pop @{ $chunks };
      }
    }
  }

  if ($play->{'endOfSpool'} && !@{ $chunks }) {
    # If every frame has been written to the decoder and this is
    # the end of the spool close the decoder.
    Trace::debug('All files in spool written to decoder') if Trace::check(Trace::DEBUG2);
    $decoder->{'write'}->close;
    delete $decoder->{'write'};
    if (!$player->{'builtin'}) {
      # Wait for the decoder to die to open up dsp
      if (defined($decoder->{'pid'})) {
	waitpid($decoder->{'pid'}, 0);
	delete $decoder->{'pid'};
      }
    }
  }

  return 1;
}

# _playingAttributes - Estimate the true play position when playing or
# paused. With various buffering and direction taken into
# consideration.

sub _playingAttributes {
  my $player = shift;
  
  Trace::debug('Refreshing play attributes') if Trace::check(Trace::DEBUG5);

  return undef if (($player->{'state'} ne Playing) && ($player->{'state'} ne Paused));

  my $play = $player->{'play'};
  my $duration = $player->{'spool'}->[ $player->{'position'} ]->{'format'}->duration;
  my $rate = $play->{'searching'} ? ($player->{'searchSkipSecs'} + $player->{'searchPlaySecs'}) / $player->{'searchPlaySecs'} : 1.0;

  my $buffer;
  if (defined($player->{'decoder'}) && defined($player->{'decoder'}->{'buffer'})) {
    $buffer = $player->{'decoder'}->{'buffer'};
  }

  if (!defined($buffer) || defined($player->{'seek'})) {
    Trace::debug('Not started to play or a seek value has been set but not executed') if Trace::check(Trace::DEBUG3);
    my $seek = defined($player->{'seek'}) ? $player->{'seek'} : 0.0;
    my $endTime = ($player->{'state'} eq Playing) ? 
      time + ($duration * (( ($play->{'direction'} eq Forward) ? 100.0 - $seek : $seek ) / 100.0 )) / $rate :
      undef;
    return { position => $player->{'position'}, seek => $seek, endTime => $endTime, time => time };
  }

  # Buffered from file:

  my $bufferedTime = 0.0;
  if (defined($buffer->{'chunks'})) {
    map { map { $bufferedTime += $_->{'secs'} } @{ $_->{'frames'} } } @{ $buffer->{'chunks'} };
    if ($play->{'direction'} eq Forward) {
      my $chunk = $buffer->{'chunks'}->[0]; # Current chunk
      if ($chunk->{'wp'}) {
	for (my $i = 0; $i < $chunk->{'wp'}; $i++) {
	  $bufferedTime -= $chunk->{'frames'}->[ $i ]->{'secs'}; # Already written
	}
      }
    }
    else {
      my $chunk = $buffer->{'chunks'}->[ $#{ $buffer->{'chunks'} } ]; # Current chunk
      if (defined($chunk->{'wp'}) && ($chunk->{'wp'} < $#{ $chunk->{'frames'} })) {
	for (my $i = $#{ $chunk->{'frames'} }; $i > $chunk->{'wp'}; $i--) {
	  $bufferedTime -= $chunk->{'frames'}->[ $i ]->{'secs'}; # Already written
	}
      }
    }
  }

  if (defined($buffer->{'frames'})) {
    map { $bufferedTime += $_->{'secs'} } @{ $buffer->{'frames'} };
  }

  # Buffered in pipes: TODO: Could estimate the duration buffered in the various PIPES

  # Buffered at player:

  if ($player->{'builtin'}) {
    $bufferedTime += defined($player->{'dsp'}->{'buffered'}) ? $player->{'dsp'}->{'buffered'} : 0.0;
  }
  else {
    # Assume it is fully buffered
    if (defined($player->{'mpg123BufferSecs'})) {
      $bufferedTime += $player->{'mpg123BufferSecs'};
    }
  }

  Trace::debug('Buffered %.2f', $bufferedTime) if Trace::check(Trace::DEBUG4);

  # When searching the buffers represent much more.
  $bufferedTime *= $rate;

  # Estimated
  my $trueTime = 0.0; # Portion of track played
  my $truePosition = $player->{'position'};
  my $endTime; # Relative to time of day
  my $fileTime; # Read point in file

  if (defined($buffer->{'fh'})) {
    $fileTime = $duration * ($buffer->{'fh_offset'} / $buffer->{'fh_length'}); 
  }
  elsif ($play->{'endOfSpool'}) {
    Trace::debug('End of spool') if Trace::check(Trace::DEBUG5); # position points to next track
    if ($play->{'direction'} eq Forward) {
      $fileTime = $duration;
      $truePosition = $#{ $player->{'spool'} };
    }
    else {
      $fileTime = 0.0;
      $truePosition = 0;
    }
  }
  else {
    Trace::debug('No open file') if Trace::check(Trace::DEBUG5); # position points to next track
    if ($play->{'direction'} eq Forward) {
      $fileTime = 0.0;
    }
    else {
      $fileTime = $duration;
    }
  }

  Trace::debug('Read point: %d, %.2f / %.2f', $player->{'position'}, $fileTime, $duration) if Trace::check(Trace::DEBUG5);

  if ($play->{'direction'} eq Forward) {

    # e.g forward play, position == n+1, truePosition = n
      
    #           track n                       track n+1
    #                    ...............
    #  mmmmmmmmmmmmmmmmmmmmmmmmmmmm|mmmmmmmmmmmmmmmmmmmmmmmm
    #                    ^trueTime     ^fileTime

    $trueTime = $fileTime - $bufferedTime;
    while ($trueTime < 0.0) {
      # Negative offset, on a previous track
      $truePosition--;
      if ($truePosition < 0) {
	Trace::debug('Position moved past 0!') if Trace::check(Trace::DEBUG5);
	$trueTime = 0.0;
	$truePosition = 0;
      }
      else {
	$duration = $player->{'spool'}->[ $truePosition ]->{'format'}->duration;
	$trueTime += $duration;
      }
    }
    $endTime = time + ($duration - $trueTime) / $rate;
  }
  else {

    # e.g reverse play, position == n, truePosition = n+1

    #           track n                       track n+1
    #  mmmmmmmmmmmmmmmmmmmmmmmmmmmm|mmmmmmmmmmmmmmmmmmmmmmmm
    #                    ^fileTime        ^trueTime

    $trueTime = $fileTime + $bufferedTime;
    while ($trueTime > $duration) {
      $truePosition++;
      if ($truePosition > $#{ $player->{'spool'} }) {
	Trace::debug('Position moved past end!') if Trace::check(Trace::DEBUG5);
	$trueTime = $duration;
	$truePosition = $#{ $player->{'spool'} };
      }
      else {
	$trueTime -= $duration;
	$duration = $player->{'spool'}->[ $truePosition ]->{'format'}->duration;
      }
    }
    $endTime = time + $trueTime / $rate;
  }

  Trace::debug('True point: %d, %.2f / %.2f', $truePosition, $trueTime, $duration) if Trace::check(Trace::DEBUG5);

  my $seek = ($trueTime / $duration) * 100;
  if ($seek < 0) { $seek = 0.0; } elsif ($seek > 100) { $seek = 100; }

  Trace::debug('Play attributes; p: %d, s%%: %.2f', $truePosition, $seek) if Trace::check(Trace::DEBUG5);

  return { position => $truePosition, seek => $seek, endTime => ($player->{'state'} eq Playing) ? $endTime : undef, time => time };
}

# _addToMonitors - Anything that sends us a command should be informed
# of state or spool changes.

sub _addToMonitors {
  my $player = shift;
  my $monitor = shift;

  my $monitors = $player->{'monitors'};
  if (!defined($monitors->{ $monitor->unique })) {
    Trace::debug('New state monitor: %s', $monitor->unique) if (Trace::check(Trace::DEBUG3));
    $monitors->{ $monitor->unique } = $monitor;
  }
  return 1;
}

# OSS Driver

# _driverStarted

sub _driverStarted {
  my $player = shift;
  my %tParms = @_;
  my $parms = $tParms{'parms'};

  if (defined($parms->{'error'})) {
    $player->{'cause'} = Error;
    $player->{'attributes'} = { diagnostic => $parms->{'error'} };
    $player->_stop();
    $player->_state();
    return 0;
  }
  if (!defined($parms->{'handle'})) {
    $player->{'cause'} = Error;
    $player->{'attributes'} = { diagnostic => 'No handle to communicate with driver' };
    $player->_stop();
    $player->_state();
    return 0;
  }
  $player->{'dsp'}->{'driver'} = $parms->{'handle'};

  if ($player->{'state'} eq Playing) {

    Trace::debug('Starting play machine') if Trace::check(Trace::DEBUG4);
    $player->message(name => '_playMachine', to => $player, parms => { tag => ++$player->{'tag'} });
    $player->message(name => '_driver_play', to => $player->{'dsp'}->{'driver'});
  }

  return 1;
}

# _driverEOP

sub _driverEOP {
  my $player = shift;

  $player->{'cause'} = End;
  $player->_stop;
  $player->_state;

  return 1;
}

# _driverError

sub _driverError {
  my $player = shift;
  my %tParms = @_;

  if (defined($tParms{'error'})) {
    $player->{'cause'} = Error;
    $player->{'attributes'} = { diagnostic => $tParms{'error'} };
    $player->_stop();
    $player->_state();
  }

  return 1;
}

# _driverBuffered

sub _driverBuffered {
  my $player = shift;
  my %tParms = @_;

  if (defined($tParms{'buffered'})) {
    $player->{'dsp'}->{'buffered'} = $tParms{'buffered'};
  }

  return 1;
}

package Player::Driver;

# OSS Driver process. Simple dedicated process to handle reading from
# the decoder and writing to the OSS dsp.

# Messages:

# play
# stop
# discard - discard buffered audio and clear any end of flags
# EOS - All tracks written to decoder, end is near.

# Response:

# EOP - Play finished normally
# buffered - audio buffered by the driver
# error 

use Thing;
@Player::Driver::ISA = qw( Thing );
use Trace;
use IO::File qw();
use POSIX qw(:fcntl_h);
use POSIX qw(:stdio_h);
use strict;

sub new {
  my $proto = shift;
  my %parms = @_;

  my $type = ref($proto) || $proto;
  my $driver = $type->SUPER::new(@_);
  if ($driver->error) {
    $driver->message(name => Thing::Started, to => $parms{'player'}, parms => { error => $driver->error });
    return $driver;
  }

  $driver->{'dsp'} = $parms{'dsp'};
  $driver->{'decoder'} = $parms{'decoder'};
  $driver->{'play'} = $parms{'play'};
  $driver->{'player'} = $parms{'player'}->handle;
  $driver->{'tag'} = 1;
  $driver->{'dsp'}->{'bufferStatus'} = { time => time, buffered => 0.0 };
  $driver->{'decoder'}->{'write'}->close;

  $driver->{'decoder'}->{'read'}->fcntl(F_SETFL, O_NONBLOCK);

  $driver->register(name => '_driver_play', handler => '_play');
  $driver->register(name => '_driver_stop', handler => '_stop');
  $driver->register(name => '_driver_discard', handler => '_discard');
  $driver->register(name => '_driver_EOS', handler => '_EOS');
  $driver->register(name => '_playMachine', handler => '_playMachine');
  $driver->register(name => Thing::ShutdownRequest, handler => '_shutdown');

  $driver->message(name => Thing::Started, to => $driver->{'player'}, parms => { handle => $driver->handle });

  return $driver;
}

sub _play {
  my $driver = shift;

  $driver->message(name => '_playMachine', to => $driver, parms => { tag => ++$driver->{'tag'} });

  Trace::debug('Play requested') if Trace::check(Trace::DEBUG3);

  return 1;
}

sub _playMachine {
  my $driver = shift;
  my %tParms = @_;
  my $parms = $tParms{'parms'};

  # Check for a change in play.
  return 0 if ($driver->{'tag'} ne $parms->{'tag'});

  my $play = $driver->{'play'};
  my $dsp = $driver->{'dsp'};
  my $decoded = 0;

  # Try and read at least one block more than may be played to ramp up the buffer
  if (!defined($decoded = $driver->_pcmReadFromDecoder($dsp->{'writeBlocks'} + 1))) {
    return $driver->_error;
  }

  my $p = -1;
  if ($#{ $dsp->{'buffer'} } + 1 >= $dsp->{'maxBlocks'}) {
    # Fully buffered
    $p = $dsp->{'writeBlocks'};
  }
  elsif (!$decoded) {
    if (@{ $dsp->{'buffer'} }) {
      # Nothing from decoder but some buffering, use it
      $p = 1;
    }
  }
  else {
    # Play less than decoded
    $p = $decoded - 1;
    if (($p <= 0) && @{ $dsp->{'buffer'} }) {
      # One block decoded, play only if there is buffering
      $p = 1;
    }
  }

  if ($p > 0) {
    if (!defined($driver->_pcmPlay($p))) {
      return $driver->_error;
    }
  }
  else {
    Trace::debug('Nothing decoded or playable') if Trace::check(Trace::DEBUG3);
  }

  if ($play->{'endOfPlay'} && !@{ $dsp->{'buffer'} }) {
    Trace::debug('End of play reached') if Trace::check(Trace::DEBUG2);
    $driver->{'dsp'}->{'handle'}->close;
    delete $driver->{'dsp'}->{'handle'};
    $driver->message(name => '_driver_EOP', to => $driver->{'player'});
  }

  if ($dsp->{'bufferStatus'}->{'time'} < time) {

    my $bufferStatus = $dsp->{'bufferStatus'};
    $bufferStatus->{'time'} = time;
    my $buffered = 0.0;

    if (@{ $dsp->{'buffer'} }) {
      $buffered = ($#{ $dsp->{'buffer'} } + 1) * $dsp->{'blocksecs'};
      Trace::debug('Buffered: %.2f', $buffered) if Trace::check(Trace::DEBUG4);
    }

    if (abs($bufferStatus->{'buffered'} - $buffered) >= 0.5) {
      Trace::debug('Buffered update: %.2f', $buffered) if Trace::check(Trace::DEBUG3);
      $driver->message(name => '_driver_buffered', to => $driver->{'player'}, buffered => $buffered);
      $bufferStatus->{'buffered'} = $buffered;
    }
  }

  $driver->message(name => '_playMachine', to => $driver, parms => { tag => $driver->{'tag'} });

  return 1;
}

sub _stop {
  my $driver = shift;

  # Supposedly make a pause smoother
  ioctl($driver->{'dsp'}->{'handle'}, Player::SNDCTL_DSP_POST, 0) if defined($driver->{'dsp'}->{'handle'});

  $driver->{'tag'} = 1;

  Trace::debug('Stop requested') if Trace::check(Trace::DEBUG2);

  return 1;
}  

sub _EOS {
  my $driver = shift;

  $driver->{'play'}->{'endOfSpool'} = 1;
  # Block to flush decoder output
  $driver->{'decoder'}->{'read'}->fcntl(F_SETFL, 0);
  
  Trace::debug('End of spool notification') if Trace::check(Trace::DEBUG2);

  return 1;
}

sub _discard {
  my $driver = shift;

  $driver->{'play'}->{'endOfSpool'} = 0;
  $driver->{'play'}->{'endOfPlay'} = 0;
  $driver->{'dsp'}->{'buffer'} = [];

  Trace::debug('Discard requested') if Trace::check(Trace::DEBUG2);

  return 1;
}

sub _shutdown {
  my $driver = shift;

  $driver->{'dsp'}->{'handle'}->close if defined($driver->{'dsp'}->{'handle'});
  
  return 1;
}

sub _error {
  my $driver = shift;
  my $error;

  if ($#_ == -1) {
    $error = Trace::lastError;
  }
  else {
    $error = shift;
  }

  $driver->message(name => '_driver_error', to => $driver->{'player'}, error => $error);
  $driver->_stop;

  return 1;
}

# _pcmReadFromDecoder - Read the requested amount of audio from the
# decoder. Returns the amount of aduio read or undef on error.

sub _pcmReadFromDecoder {
  my $driver = shift;
  my $request = shift;

  my $decoder = $driver->{'decoder'};
  my $dsp = $driver->{'dsp'};
  my $play = $driver->{'play'};

  my $blocksize = $dsp->{'blocksize'};
  my $blocksecs = $dsp->{'blocksecs'};
  my $decoded = 0;
  my $blocking = 0;

  # Read from the decoder.

  my $offset;
  my $bytesRead;
  my $data;

  if ($dsp->{'length'}) {
    # Prior partial
    $data = $dsp->{'partial'};
    $offset = $dsp->{'length'};
    $dsp->{'length'} = 0;
  }
  else {
    $offset = 0;
  }

  # Block when empty
  if (!@{ $dsp->{'buffer'} }) {
    $driver->{'decoder'}->{'read'}->fcntl(F_SETFL, 0);
    $blocking = 1;
  }
  
  BLOCK: while (!$play->{'endOfPlay'} && ($decoded < $request) && (($#{ $dsp->{'buffer'} } + 1) < $dsp->{'maxBlocks'})) {

    my $data if !$offset; # New data for dataR

    # Read decoded

    do {
      $bytesRead = sysread($decoder->{'read'}, $data, $blocksize - $offset, $offset);
      if (defined($bytesRead) && ($bytesRead < 0)) {
	Trace::error('Error reading from the decoder');
	return undef;
      }
      if ($bytesRead) {
	$offset += $bytesRead;
      }
    } while ($bytesRead && ($offset != $blocksize));

    if ($offset != $blocksize) {
      if ($play->{'endOfSpool'} && !$decoded) {
	# Empty read on ready signal, end of play
	Trace::debug('Last decoder output') if Trace::check(Trace::DEBUG5);
	$play->{'endOfPlay'} = 1;
	push @{ $dsp->{'buffer'} }, { dataR => \$data, length => $offset };
      }
      else {
	# Partial read from decoder
	$dsp->{'partial'} = $data;
	$dsp->{'length'} = $offset;
	Trace::debug('Read partial decoded block: %d', $offset) if Trace::check(Trace::DEBUG5);
      }
      last BLOCK;
    }
    else {
      Trace::debug('Read full decoded block') if Trace::check(Trace::DEBUG5);
      push @{ $dsp->{'buffer'} }, { dataR => \$data, length => $blocksize };
    }

    $offset = 0;
    $decoded++;
  }

  Trace::debug('Read %d blocks from decoder', $decoded) if Trace::check(Trace::DEBUG4);

  $driver->{'decoder'}->{'read'}->fcntl(F_SETFL, O_NONBLOCK) if $blocking;

  return $decoded;
}

# _pcmPlay - Play the requested amount of audio if
# possible. Returns the amount of audio played or undef on error.

sub _pcmPlay {
  my $driver = shift;
  my $request = shift;

  my $dsp = $driver->{'dsp'};

  my $blocksecs = $dsp->{'blocksecs'};
  my $played = 0;
  my $bytesWritten;
  my $write;
  my $dataR;

  # Write to dsp

  while (@{ $dsp->{'buffer'} } && ($played < $request)) {
    $write = shift @{ $dsp->{'buffer'} };
    $dataR = $write->{'dataR'};
    $bytesWritten = syswrite($dsp->{'handle'}, $$dataR, $write->{'length'});
    if (!defined($bytesWritten) || ($bytesWritten != $write->{'length'})) {
      Trace::error('Error writing to sound device: %s', $!);
      return undef;
    }

    # TEST: Write to file
    if (defined($ENV{'DECODER_READ_TEST'})) {
      syswrite($driver->{'testhandle'}, $$dataR, $write->{'length'});
    }

    Trace::debug('Wrote pcm, %d bytes', $bytesWritten) if Trace::check(Trace::DEBUG5);
    $played++;
  }

  Trace::debug('Played %d block', $played) if Trace::check(Trace::DEBUG4);

  return $played;
}

1;
