# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.20.6.4 $

package Repo;

use Thing;
@ISA = qw( Thing );
use RepoH;
use Command;
use Format::Wav;
use Format::Mp3;
use Info::Persist;
use Info::Artist;
use Info::Album;
use Info::Cd;
use Info::Track;
use Utility;
use Trace;
use strict;

sub new {
    my $proto = shift;
    my $type = ref($proto) || $proto;
    my $repo = $type->SUPER::new(@_);
    if ($repo->error) {
	return $repo;
    }
    $repo->{'dirs'} = [];
    $repo->{'store_forking'} = 1;	# DEBUG
    $repo->{'query_forking'} = 0;	# DEBUG

    $repo->register(name => 'store', handler => 'store');
    $repo->register(name => 'query', handler => 'query');
    $repo->register(name => Repo::Change, handler => 'change');
    $repo->register(name => Repo::Delete, handler => 'delete');
    $repo->register(name => Repo::Reparent, handler => 'reparent');

    $repo->register(name => Thing::ConfigureRequest, handler => '_configure');
    $repo->register(name => Thing::ShutdownRequest, handler => '_shutdown');

    return $repo;
}

# _configure

sub _configure {
    my $repo = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};

    my %parms = @_;

    my $errorS = sub { if (defined($parms->{'error'})) { $parms->{'error'} .= "\n".$_[0]; } else { $parms->{'error'} = $_[0]; } };

    if (!defined($parms->{'dirs'}) || (ref($parms->{'dirs'}) ne 'ARRAY') || ($#{ $parms->{'dirs'} } == -1)) {
	Trace::error('Music directories must be specified.');
	&$errorS(Trace::lastError);
    }
    else {

	my $newdirs = [];

	DIR: foreach my $dir (@{ $parms->{'dirs'} }) {

	    # Don't initialise existing directories
	    
	    foreach my $edir (@{ $repo->{'dirs'} }) {
		if ($dir eq $edir) {
		    push @{ $newdirs }, $edir;
		    next DIR;
		}
	    }

	    # Must be absolute and non root

	    if ($dir !~ m/^\/.+\/*$/) {
		Trace::error("Music directory ($dir) must be absolute (begin with a /) but cannot be /.");
		&$errorS(Trace::lastError);
		next DIR;
	    }
	
	    # Strip trailing ////
	    
	    while ($dir =~ m/^(.*)\/$/) {
		$dir = $1;
	    }
	    
	    # Check accessibility
	    
	    if (!(-r $dir && -w $dir && -x $dir && -d $dir)) {
		Trace::error('Cannot access music directory (%s).', $dir);
		&$errorS(Trace::lastError);
		next DIR;
	    }

	    # Create a .tmp area, see musicDirsToTmpDirs in RepoH.pm

	    if (! -e $dir.'/.tmp') {
		if (!mkdir($dir.'/.tmp', 0755)) {
		    Trace::error('Cannot create temp area (%s).', $dir.'/.tmp');
		    &$errorS(Trace::lastError);
		    next DIR;
		}
	    }
	
	    push @{ $newdirs }, $dir;

	    next DIR if (!opendir(DIR, $dir));
	    ARTIST: foreach my $artist (readdir(DIR)) {
		next ARTIST if (($artist =~ m/^\.$/) || ($artist =~ m/^\.\.$/) || ($artist =~ m/^\.tmp$/) || ($artist =~ m/^lost\+found$/));
		next if (!opendir(ARTIST, $dir.'/'.$artist));

		ALBUM: foreach my $album (readdir(ARTIST)) {
		    next ALBUM if (($album =~ m/^\.$/) || ($album =~ m/^\.\.$/));
		    next if (!opendir(ALBUM, $dir.'/'.$artist.'/'.$album));
		    
		    TRACK: foreach my $file (readdir(ALBUM)) {
			next TRACK if (($file =~ m/^\.$/) || ($file =~ m/^\.\.$/) || ($file !~ m/^__(.*)$/));
			Trace::debug('Removing tmpfile: %s', $dir.'/'.$artist.'/'.$album.'/'.$file) if (Trace::check(Trace::DEBUG3));
			unlink $dir.'/'.$artist.'/'.$album.'/'.$file;
		    }
		    closedir(ALBUM);
		}
		closedir(ARTIST);
	    }
	    closedir(DIR);
	}

	$repo->{'dirs'} = $newdirs;
    }

    if (defined($parms->{'db'}) && !Info::Persist::configure(db => $parms->{'db'}->{'local'})) {
	&$errorS(Trace::lastError);
    }

    # Rebuild the cache
    $repo->_artistCache;	

    return 1;
}

sub _shutdown {
    my $repo = shift;
}

# store(track, format, file, remove)

sub store {
    my $repo = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};
    my $from = $tParms{'from'};

    my $track = $parms->{'track'};
    my $format = $parms->{'format'};
    my $file = $parms->{'file'};
    if (!defined($track) || !defined($format) || !defined($file)) {
	$parms->{'error'} = 'Problems with parameters';
	$repo->message(name => 'failure', to => $from, parms => $parms);
	return;
    }
    my $artist = _encodeFileName($track->album->artist->name);
    my $album = _encodeFileName($track->album->name);
    my $name = _encodeFileName($track->name);
    my $index = $track->index;

    # Add to artist cache
    if (!defined($repo->{'artistCache'}->{ $artist })) {
	$repo->{'artistCache'}->{ $artist } = $track->album->artist;
    }

    if ($repo->{'store_forking'}) {
	Info::Persist::disconnect;
	my $pid = $repo->fork;
	if ($pid == -1) {
	    $parms->{'error'} = Trace::lastError;
	    $repo->message(name => 'failure', to => $from, parms => $parms);
	    return;
	}
	return if ($pid);
    }

    # Music is stored in an artist / album sub directory hierarchy. 

    my $storePath = $artist.'/'.$album;
    my $storeFile = sprintf('%02d_', $index).$name.$format->suffix(unique => 1);
    my $stored = 0;

    # We first try a rename to each repo dir to save copies across filesystems. 

    DIR: foreach my $rdir (@{ $repo->{'dirs'} }) {
	my $createdArtistDir = 0;
	my $path = $rdir.'/'.$artist;
	if (! -e $path) {
	    if (!mkdir($path, 0755)) {
		Trace::error('Cannot create %s', $path);
		next DIR;
	    }
	    $createdArtistDir = 1;
	}	
	my $createdAlbumDir = 0;
	$path = $rdir.'/'.$storePath;
	if (! -e $path) {
	    if (!mkdir($path, 0755)) {
		Trace::error('Cannot create %s', $path);
		next DIR;
	    }
	    $createdAlbumDir = 1;
	}

	if (rename($file, $path.'/'.$storeFile)) {
	    # Renamed directly
	    Trace::debug('Stored %s as %s', $file, $path.'/'.$storeFile) if Trace::check(Trace::DEBUG4);
	    $stored = $path.'/'.$storeFile;
	    last DIR;
	}

	# Don't want empty dirs left arround
	if ($createdAlbumDir) {
	    unlink $path;
	}
	if ($createdArtistDir) {
	    unlink $rdir.'/'.$artist;
	}
    }

    if (!$stored) {

	# Store it wherever there is enough space

	my $size = -s $file;
	DIR: foreach my $rdir (@{ $repo->{'dirs'} }) {
	    my $space = Utility::available($rdir);
	    next if ($space < $size);

	    my $path = $rdir.'/'.$artist;
	    if (! -e $path) {
		if (!mkdir($path, 0755)) {
		    Trace::error('Cannot create %s', $path);
		    next DIR;
		}
	    }	
	    $path = $rdir.'/'.$storePath;
	    if (! -e $path) {
		if (!mkdir($path, 0755)) {
		    Trace::error('Cannot create %s', $path);
		    next DIR;
		}
	    }

	    # Copy to a tempfile so it doesn't get picked up in a query

	    my $tmpName = $path.'/__'.$storeFile;
	    my $str = 'cp '.Command::quotemeta($file).' '.Command::quotemeta($tmpName);
	    my $cmd = Command->new(Shell => $str, Timeout => 60 * 20);    
	    if (!$cmd || !$cmd->run || ($cmd->state != Command::FINISHED) || (defined($cmd->return) && ($cmd->return != 0))) {
		Trace::debug('Cannot copy %s to %s', $file, $tmpName) if Trace::check(Trace::DEBUG4);
		Trace::dump($cmd->output);
		unlink($tmpName);
		$parms->{'error'} = "Cannot copy $file to $tmpName.\n".$cmd->output;
		$repo->message(name => 'failure', to => $from, parms => $parms);
		Thing::exit(1) if ($repo->{'store_forking'});
		return if (!$repo->{'store_forking'});
	    }
	    else {
		# Rename temp file to real file
		if (!rename($tmpName, $path.'/'.$storeFile)) {
		    Trace::error("Cannot rename: $!");
		    unlink($tmpName);
		    $parms->{'error'} = "Cannot rename $tmpName to $path/$storeFile: $!.";
		    $repo->message(name => 'failure', to => $from, parms => $parms);
		    Thing::exit(1) if ($repo->{'store_forking'});
		    return if (!$repo->{'store_forking'});
		}
		$stored = $path.'/'.$storeFile;
		Trace::debug('Stored %s as %s', $file, $stored) if Trace::check(Trace::DEBUG4);
		# Copy was successful so we can remove the source if asked.
		if ($parms->{'remove'}) {
		    Trace::debug('Removed source: %s', $file) if Trace::check(Trace::DEBUG4);
		    unlink($file);
		}
	    }
	}
    }

    if ($stored) {
	$format->file($stored);
	$repo->message(name => 'stored', to => $from, parms => $parms);
    }
    else {
	$parms->{'error'} = 'Not enough space to store file, '.(-s $file).' bytes needed';
	$repo->message(name => 'failure', to => $from, parms => $parms);
    }
    Thing::exit(1) if ($repo->{'store_forking'});
    return if (!$repo->{'store_forking'});
}

# delete - Delete an artist, album, track or file

#   artist - deletes all files associated with the artist.

#   album - deletes all files associated with the album.

#   track - deletes all files associated with the track.

#   track & format - deletes all files of this format for this track, should be one.

sub delete {
    my $repo = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};

    Trace::debug('Delete request') if Trace::check(Trace::DEBUG3);

    if (defined($parms->{'artist'})) {
	my $artist = $parms->{'artist'};

	# Safety checks

	if (!defined($artist->name) || ($artist->name eq '')) {
	    Trace::error('Badly formed artist');
	    return;
	}

	Trace::debug('Deleting artist: %s', $artist->name) if Trace::check(Trace::DEBUG4);

	# Get rid of all files

	my $artistname = _encodeFileName($artist->name);
	DIR: foreach my $dir (@{ $repo->{ 'dirs' } }) {
	    next DIR if (!opendir(ARTIST, $dir.'/'.$artistname));
	    ALBUM: foreach my $albumname (readdir(ARTIST)) {
		next ALBUM if (($albumname =~ m/^\.$/) || ($albumname =~ m/^\.\.$/));
		next ALBUM if (!opendir(ALBUM, $dir.'/'.$artistname.'/'.$albumname));
		FILE: foreach my $file (readdir(ALBUM)) {
		    next FILE if (($file =~ m/^\.$/) || ($file =~ m/^\.\.$/));
		    Trace::debug('Deleting file: %s', $file) if Trace::check(Trace::DEBUG6);
		    unlink $dir.'/'.$artistname.'/'.$albumname.'/'.$file;
		}
		closedir(ALBUM);
		Trace::debug('Deleting album: %s', $albumname) if Trace::check(Trace::DEBUG6);
		rmdir $dir.'/'.$artistname.'/'.$albumname;
	    }
	    closedir(ARTIST);
	    Trace::debug('Deleting artist: %s', $artistname) if Trace::check(Trace::DEBUG6);
	    rmdir $dir.'/'.$artistname;
	}
	
	# Remove from cache
	delete $repo->{'artistCache'}->{ $artistname };

	# Remove from database
	$artist->dbDelete;
    }
    elsif (defined($parms->{'album'})) {
	my $album = $parms->{'album'};
	my $artist = $album->artist;

	# Safety checks

	if (!defined($album->name) || ($album->name eq '')) {
	    Trace::error('Badly formed album');
	    return;
	}
	if (!defined($artist->name) || ($artist->name eq '')) {
	    Trace::error('Badly formed artist');
	    return;
	}

	Trace::debug('Deleting album: %s, for: %s', $album->name, $artist->name) if Trace::check(Trace::DEBUG4);

	# Get rid of all files

	my $artistname = _encodeFileName($artist->name);
	my $albumname = _encodeFileName($album->name);
	DIR: foreach my $dir (@{ $repo->{ 'dirs' } }) {
	    next DIR if (!opendir(ALBUM, $dir.'/'.$artistname.'/'.$albumname));
	    FILE: foreach my $file (readdir(ALBUM)) {
		next FILE if (($file =~ m/^\.$/) || ($file =~ m/^\.\.$/));
		Trace::debug('Deleting file: %s', $file) if Trace::check(Trace::DEBUG6);
		unlink $dir.'/'.$artistname.'/'.$albumname.'/'.$file;
	    }
	    closedir(ALBUM);
	    Trace::debug('Deleting album: %s', $albumname) if Trace::check(Trace::DEBUG6);
	    rmdir $dir.'/'.$artistname.'/'.$albumname;
	}

	# Remove from database
	$album->dbDelete;

	$artist->sync(force => 1);
    }
    elsif (defined($parms->{'track'}) && !defined($parms->{'format'})) {
	my $track = $parms->{'track'};
	my $album = $track->album;
	my $artist = $album->artist;

	# Safety checks

	if (!defined($track->index)) {
	    Trace::error('Badly formed track');
	    return;
	}
	if (!defined($album->name) || ($album->name eq '')) {
	    Trace::error('Badly formed album');
	    return;
	}
	if (!defined($artist->name) || ($artist->name eq '')) {
	    Trace::error('Badly formed artist');
	    return;
	}

	Trace::debug('Deleting track: %s, from: %s, for: %s', $track->name, $album->name, $artist->name) if Trace::check(Trace::DEBUG4);

	# Get rid of all files

	my $artistname = _encodeFileName($artist->name);
	my $albumname = _encodeFileName($album->name);
	DIR: foreach my $dir (@{ $repo->{ 'dirs' } }) {
	    next DIR if (!opendir(ALBUM, $dir.'/'.$artistname.'/'.$albumname));
	    FILE: foreach my $file (readdir(ALBUM)) {
		next FILE if (($file =~ m/^\.$/) || ($file =~ m/^\.\.$/) || ($file !~ m/^\d\d_.*\..*$/));
		$file =~ /^(\d\d)_/;
		my $thisIndex = int($1);
		next FILE if ($thisIndex != $track->index);
		Trace::debug('Deleting file: %s', $file) if Trace::check(Trace::DEBUG6);
		unlink $dir.'/'.$artistname.'/'.$albumname.'/'.$file;
	    }
	    closedir(ALBUM);
	}

	# Remove from database
	$track->dbDelete;

	$album->sync(force => 1);
    }
    elsif (defined($parms->{'track'}) && defined($parms->{'format'})) {
	my $track = $parms->{'track'};
	my $format = $parms->{'format'};
	my $album = $track->album;
	my $artist = $album->artist;

	# Safety checks

	if (!defined($track->index)) {
	    Trace::error('Badly formed track');
	    return;
	}
	if (!defined($album->name) || ($album->name eq '')) {
	    Trace::error('Badly formed album');
	    return;
	}
	if (!defined($artist->name) || ($artist->name eq '')) {
	    Trace::error('Badly formed artist');
	    return;
	}

	Trace::debug('Deleting track: %s, of format: %s, from: %s, for: %s', 
		     $track->name, $format->type, $album->name, $artist->name) if Trace::check(Trace::DEBUG4);

	my $file = sprintf('%02d_', $track->index);
	$file .= _encodeFileName($track->name);
	$file .= $format->suffix(unique => 1);

	my $artistname = _encodeFileName($artist->name);
	my $albumname = _encodeFileName($album->name);
	DIR: foreach my $dir (@{ $repo->{ 'dirs' } }) {
	    my $path = $dir.'/'.$artistname.'/'.$albumname.'/'.$file;
	    if (-e $path) {
		Trace::debug('Deleted file: %s', $path) if Trace::check(Trace::DEBUG6);
		unlink $path;
	    }
	}

	# Check for persistence delete
	my $otherFormat = 0;
	DIR: foreach my $dir (@{ $repo->{ 'dirs' } }) {
	    next DIR if (!opendir(ALBUM, $dir.'/'.$artistname.'/'.$albumname));
	    FILE: foreach my $file (readdir(ALBUM)) {
		next FILE if (($file =~ m/^\.$/) || ($file =~ m/^\.\.$/) || ($file !~ m/^\d\d_.*\..*$/));
		$file =~ /^(\d\d)_/;
		my $thisIndex = int($1);
		next FILE if ($thisIndex != $track->index);
		Trace::debug('Other format exists for: %s', $track->name) if Trace::check(Trace::DEBUG6);
		$otherFormat = 1;
		last DIR;
	    }
	    closedir(ALBUM);
	}

	$track->dbDelete if (!$otherFormat);

	$album->sync(force => 1) if (!$otherFormat);
    }
    else {
	Trace::error('Unhandled delete request from: %s', $tParms{'from'}->key);
    }
}

# change - process a change to `object'

sub change {
    my $repo = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};

    if (!defined($parms->{'object'})) {
	Trace::mandatory('object');
	return undef;
    }
    my $object = $parms->{'object'};

    Trace::debug('Change request') if Trace::check(Trace::DEBUG3);

    if ($object->isa('Info::Track')) {
	my $track = $object;
	Trace::debug('Track change from %s to %s', $track->name, $parms->{'name'}) if Trace::check(Trace::DEBUG4);
	return if (!defined($parms->{'name'}));

	$repo->trackQuery($track);
	my $name = $parms->{'name'};
	foreach my $format (values %{ $track->{'formatsH'} }) {
	    my $file = $format->file;
	    my $namepart = _encodeFileName($track->name);
	    if ($file =~ /(.*)$namepart(.*)/) {
		my $newfile = $1._encodeFileName($name).$2;
		Trace::debug('Renaming %s to %s', $file, $newfile) if Trace::check(Trace::DEBUG3);
		if (!rename($file, $newfile)) {
		    Trace::error('Cannot rename %s to %s: %s', $file, $newfile, $!);
		}
		else {
		    Trace::debug('File renamed from %s to %s', $file, $newfile) if Trace::check(Trace::DEBUG6);
		}
	    }
	    else {
		Trace::error('Format does not match old name: %s, %s', $file, _encodeFileName($track->name));
	    }
	}
	$track->setN;
	$track->getN->{'name'} = $name;
	$track->save;
	$track->sync;
    }
    elsif ($object->isa('Info::Album')) {
	my $album = $object;
	Trace::debug('Album change from %s to %s', $album->name, $parms->{'name'}) if Trace::check(Trace::DEBUG4);
	return if (!defined($parms->{'name'}));

	my $newname = _encodeFileName($parms->{'name'});
	my $artistname = _encodeFileName($album->artist->name);
	my $currname = _encodeFileName($album->name);

	my $exists = 0;
	foreach my $dir (@{ $repo->{'dirs'} }) {
	    next if (! -e $dir.'/'.$artistname.'/'.$newname);
	    $exists = 1;
	}
	if ($exists) {
	    Trace::debug('An album with this name already exists') if Trace::check(Trace::DEBUG4);
	}
	else {
	    # Rename directories where album resides 
	    foreach my $dir (@{ $repo->{ 'dirs' } }) {
		next if (! -e $dir.'/'.$artistname.'/'.$currname);
		if (!rename($dir.'/'.$artistname.'/'.$currname, $dir.'/'.$artistname.'/'.$newname)) {
		    Trace::error('Cannot rename %s to %s: %s', $currname, $newname, $!);
		}
	    }
	    $album->setN;
	    $album->getN->{'name'} = $parms->{'name'};
	    $album->save;
	    $album->sync;
	}
    }
    elsif ($object->isa('Info::Artist')) {
	my $artist = $object;
	Trace::debug('Artist change from %s to %s', $artist->name, $parms->{'name'}) if Trace::check(Trace::DEBUG4);
	return if (!defined($parms->{'name'}));

	my $newname = _encodeFileName($parms->{'name'});
	my $currname = _encodeFileName($artist->name);
	if (defined($repo->{'artistCache'}->{ $newname })) {
	    Trace::debug('An artist with this name already exists') if Trace::check(Trace::DEBUG4);
	}
	else {
	    foreach my $dir (@{ $repo->{ 'dirs' } }) {
		next if (! -e $dir.'/'.$currname);
		if (!rename($dir.'/'.$currname, $dir.'/'.$newname)) {
		    Trace::error('Cannot rename %s to %s: %s', $currname, $newname, $!);
		}
	    }
	    delete $repo->{'artistCache'}->{ $currname };
	    $repo->{'artistCache'}->{ $newname } = $artist;
	    $artist->setN;
	    $artist->getN->{'name'} = $parms->{'name'};
	    $artist->save;
	    $artist->sync;
	}
    }
    else {
	Trace::error('Unhandled change request from: %s', $tParms{'from'}->key);
    }
}

# reparent - process a reparent of `object'

sub reparent {
    my $repo = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};

    if (!defined($parms->{'object'})) {
	Trace::mandatory('object');
	return undef;
    }
    if (!defined($parms->{'parent'})) {
	Trace::mandatory('parent');
	return undef;
    }
    my $object = $parms->{'object'};

    if ($object->isa('Info::Album')) {
	my $album = $object;
	my $newartist = $parms->{'parent'};
	my $newname = _encodeFileName($newartist->name);
	my $albumname = _encodeFileName($album->name);

	Trace::debug('Assigning album (%s) from artist (%s) to (%s)', $album->name, $album->artist->name, $newartist->name) if Trace::check(Trace::DEBUG3);

	my $currartist = $album->artist;
	my $currname = _encodeFileName($currartist->name);

	if ($currname eq $newname) {
	    Trace::error('Album (%s) already assigned to artist (%s)', $album->name, $album->artist->name);
	    return;
	}

	# Create target directories on the same filesystem
	foreach my $dir (@{ $repo->{ 'dirs' } }) {
	    next if (! -e $dir.'/'.$currname.'/'.$albumname);
	    if (! -e $dir.'/'.$newname) {
		if (!mkdir($dir.'/'.$newname, 0755)) {
		    Trace::error('Cannot create %s: %s', $dir.'/'.$newname, $!);
		    return;
		}
	    }
	}
	    
	# Reparent the album
	foreach my $dir (@{ $repo->{ 'dirs' } }) {
	    next if (! -e $dir.'/'.$currname.'/'.$albumname);
	    if (!rename($dir.'/'.$currname.'/'.$albumname, $dir.'/'.$newname.'/'.$albumname)) {
		Trace::error('Cannot rename %s to %s: %s', $currname.'/'.$albumname, $newname.'/'.$albumname, $!);
	    }
	}

	# New cache entry
	$repo->{'artistCache'}->{ $newname } = $newartist;	

	$album->setN;
	$album->getN->{'artist_ioid'} = $newartist->ioid;
	$album->getN->{'artist'} = $newartist;
	$album->save;
	$album->sync;

	# Check for play_date change

	$currartist->setN;
	my $date;
	if (($date = $currartist->discover_play_date) && ($date ne $currartist->play_date)) {
	    $currartist->getN->{'play_date'} = $date;
	    $currartist->save;
	}
	$newartist->setN;
	if (($date = $newartist->discover_play_date) && ($date ne $newartist->play_date)) {
	    $newartist->getN->{'play_date'} = $date;
	    $newartist->save;
	}

	# Possibly no change but let monitors know they have lost or gained an album

	$newartist->sync(force => 1);
	$currartist->sync(force => 1);
    }
    else {
	Trace::error('Unhandled reparent request from: %s', $tParms{'from'}->key);
    }
}

# query - process a query

sub query {
    my $repo = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};
    my $from = $tParms{'from'};

    my $reply = $parms;

    if (!defined($repo->{'dirs'}) || ($#{ $repo->{'dirs'} } == -1)) {
	Trace::error('No music directories defined, cannot build directory');
	$reply->{'error'} = Trace::lastError;
	$repo->message(name => 'info', to => $from, parms => $reply);    
	return;
    }

    Trace::timestamp('queryprocess') if (Trace::performance && Trace::check(Trace::DEBUG1));

    if ($repo->{'query_forking'}) {
	Info::Persist::disconnect;
	my $pid = $repo->fork;
	if ($pid == -1) {
	    $reply->{'error'} = Trace::lastError;
	    $repo->message(name => 'info', to => $from, parms => $reply);
	    return;
	}
	return if ($pid);
    }

    Trace::debug('%s query', $parms->{'type'}) if Trace::check(Trace::DEBUG2);
    Trace::debug('Query forking took: %f', Trace::timediff('queryprocess')) if (Trace::performance && Trace::check(Trace::DEBUG1));
    Trace::timestamp('queryprocess') if (Trace::performance && Trace::check(Trace::DEBUG1));

    # All tracks of a particular format
    if ($parms->{'type'} eq 'all') {
	if (!defined($parms->{'format'})) {
	    Trace::error('Message missing format parameter');
	    $reply->{'error'} = Trace::lastError;
	}
	else {
	    $reply->{'tracks'} = $repo->allQuery($parms->{'format'});
	}
    }
    # All artists
    elsif ($parms->{'type'} eq 'artists') {
	$reply->{'artists'} = $repo->artistsQuery();
    }
    # All albums for a particular artist
    elsif ($parms->{'type'} eq 'albums') {
	if (!defined($parms->{'artist'})) {
	    Trace::error('Message missing artist parameter to albums query');
	    $reply->{'error'} = Trace::lastError;
	}
	else {
	    $reply->{'albums'} = $repo->albumsQuery($parms->{'artist'});
	}
    }
    # All tracks for a particular album
    elsif ($parms->{'type'} eq 'tracks') {
	if (!defined($parms->{'album'})) {
	    Trace::error('Message missing album parameter to tracks query');
	    $reply->{'error'} = Trace::lastError;
	}
	else {
	    $reply->{'tracks'} = $repo->tracksQuery($parms->{'album'});
	}
    }
    # Everything about a particular track, or group of tracks
    elsif ($parms->{'type'} eq 'track') {
	if (!defined($parms->{'track'}) && !defined($parms->{'tracks'})) {
	    Trace::error('Message missing track or tracks parameter');
	    $reply->{'error'} = Trace::lastError;
	}
	else {
	    my @tracks = ();
	    if (defined($parms->{'track'})) {
		push @tracks, $parms->{'track'};
	    }
	    else {
		@tracks = @{ $parms->{'tracks'} };
	    }
	    foreach my $track (@tracks) {
		$repo->trackQuery($track);
	    }
	}
    }
    else {
	Trace::error('Unhandled query type: %s', $parms->{'type'});
	$reply->{'error'} = Trace::lastError;
    }

    Trace::debug('Querying took: %f', Trace::timediff('queryprocess')) if (Trace::performance && Trace::check(Trace::DEBUG1));
    Trace::timestamp('queryprocess') if (Trace::performance && Trace::check(Trace::DEBUG1));

    $repo->message(name => 'info', to => $from, parms => $reply);    

    Trace::debug('Query response took: %f', Trace::timediff('queryprocess')) 
      if (Trace::performance && Trace::check(Trace::DEBUG1));

    Thing::exit(1) if ($repo->{'query_forking'});
    return if (!$repo->{'query_forking'});
}

# allQuery - find all tracks of a particular format.

sub allQuery {
    my $repo = shift;
    my $format = shift;
    return if (!defined($format));

    my $suffix = $format->suffix(unique => 1);
    my $sufflen = length $suffix;
    my @tracks = ();
    my @files = (); # NFS duplicate check

    DIR: foreach my $dir (@{ $repo->{ 'dirs' } }) {
	next if (!opendir(DIR, $dir));

        ARTIST: foreach my $artistname (readdir(DIR)) {
	    next ARTIST if (($artistname =~ m/^\.$/) || ($artistname =~ m/^\.\.$/) || 
			    ($artistname =~ m/^\.tmp$/) || ($artistname =~ m/^lost\+found$/));
	    next ARTIST if (!opendir(ARTIST, $dir.'/'.$artistname));

	    ALBUM: foreach my $albumname (readdir(ARTIST)) {
		next ALBUM if (($albumname =~ m/^\.$/) || ($albumname =~ m/^\.\.$/));
		next ALBUM if (!opendir(ALBUM, $dir.'/'.$artistname.'/'.$albumname));

		TRACK: foreach my $file (readdir(ALBUM)) {
		    next TRACK if (($file =~ m/^\.$/) || ($file =~ m/^\.\.$/) || ($file !~ m/^\d\d_.*\..*$/) || ($file !~ m/$suffix$/));
		    my $ffile = $dir.'/'.$artistname.'/'.$albumname.'/'.$file;
		    next if (grep { $_ eq $ffile } @files);
		    push @files, $ffile;
		    my $prefix = substr $file, 0, -$sufflen;
		    $prefix =~ /(\d\d)_(.*)/;
		    my $index = int($1);
		    my $name = $2;
		    my $fformat = $format->new(file => $ffile);
		    if (!defined($fformat)) {
			Trace::debug('Problems with %s', $ffile) if Trace::check(Trace::DEBUG6);
			next TRACK;
		    }

		    my $artist = Info::Artist->find(owner => $repo->handle, name => _decodeFileName($artistname), create => 1);
		    next ARTIST if (!defined($artist));

		    my $album = Info::Album->find(owner => $repo->handle, artist => $artist, name => _decodeFileName($albumname), create => 1);
		    next ALBUM if (!defined($album));

		    my $track = Info::Track->find(owner => $repo->handle, album => $album, index => $index, name => _decodeFileName($name),
						  duration => $fformat->duration, data => 0, create => 1);
		    next TRACK if (!defined($track));

		    $track->{'formatsH'} = {};
		    $track->{'formatsH'}->{ $fformat->descriptor } = $fformat;		    
		    push @tracks, $track;
		    Trace::debug('Found match at %s', $ffile) if Trace::check(Trace::DEBUG6);
		}
		closedir(ALBUM);
	    }
	    closedir(ARTIST);
	}
	closedir(DIR);
    }
    my @sorted = sort {
	my $rc;
	if (($rc = ($a->album->artist->name cmp $b->album->artist->name)) == 0) {
	    if (($rc = ($a->album->name cmp $b->album->name)) == 0) {
		$rc = $a->index <=> $b->index;
	    }
	}
	return $rc;
    } @tracks;

    Trace::debug('Found %d matches', $#tracks + 1) if Trace::check(Trace::DEBUG4);

    return (\@sorted);
}

# trackQuery(track) - Find what we know about this track

sub trackQuery {
    my $repo = shift;
    my $track = shift;
    my @files = (); # NFS duplicate check

    my $artistname = _encodeFileName($track->album->artist->name);
    my $albumname = _encodeFileName($track->album->name);

    $track->{'formatsH'} = {};

    DIR: foreach my $dir (@{ $repo->{ 'dirs' } }) {
	next DIR if (! -d $dir.'/'.$artistname);

	Trace::debug('Artist match found at: %s', $dir.'/'.$artistname) if Trace::check(Trace::DEBUG6);
	Trace::debug('Looking for album: %s', $albumname) if Trace::check(Trace::DEBUG6);

	next DIR if (! -d $dir.'/'.$artistname.'/'.$albumname);
		
	Trace::debug('Album match found at: %s', $dir.'/'.$artistname.'/'.$albumname) if Trace::check(Trace::DEBUG6);
	Trace::debug('Looking for track: %d', $track->index) if Trace::check(Trace::DEBUG6);

	next DIR if (!opendir(ALBUM, $dir.'/'.$artistname.'/'.$albumname));
	TRACK: foreach my $file (readdir(ALBUM)) {
	    next TRACK if (($file =~ m/^\.$/) || ($file =~ m/^\.\.$/) || ($file !~ m/^\d\d_.*\..*$/));
	    my $tfile = $dir.'/'.$artistname.'/'.$albumname.'/'.$file;
	    next TRACK if (grep { $_ eq $tfile } @files);
	    push @files, $tfile;
	    $file =~ /^(\d\d)_/;
	    my $thisIndex = int($1);
	    next TRACK if ($thisIndex != $track->index);
	    my $format = Format->new(file => $tfile);
	    if (!defined($format)) {
		Trace::debug('Problems with %s', $tfile) if Trace::check(Trace::DEBUG6);
		next TRACK;
	    }
	    
	    $track->{'formatsH'}->{ $format->descriptor } = $format;		    
	    Trace::debug('Found match at %s', $tfile) if Trace::check(Trace::DEBUG6);
	}
	closedir(ALBUM);
    }

    return 1;
}

# artistsQuery() - All artists

sub artistsQuery {
    my $repo = shift;
    my @artists = map { $repo->{'artistCache'}->{ $_ } } sort { $a cmp $b } keys(%{ $repo->{'artistCache'} });
    Trace::debug('Found %d matches', $#artists + 1) if Trace::check(Trace::DEBUG4);
    return \@artists;
}

# albumsQuery() - All albums by a particular artist

sub albumsQuery {
    my $repo = shift;
    my $artist = shift;
    my $artistname = _encodeFileName($artist->name);
    
    my %albumsH;
    return undef if (!defined($repo->{'artistCache'}->{ $artistname }));

    DIR: foreach my $dir (@{ $repo->{ 'dirs' } }) {
	next DIR if (!opendir(ARTIST, $dir.'/'.$artistname));
	ALBUM: foreach my $albumname (readdir(ARTIST)) {
	    next ALBUM if (($albumname =~ m/^\.$/) || ($albumname =~ m/^\.\.$/));
	    next ALBUM if (defined($albumsH{ $albumname }));
	    my $album = Info::Album->find(owner => $repo->handle, artist => $artist, name => _decodeFileName($albumname), create => 1);
	    next ALBUM if (!defined($album));
	    $albumsH{ $albumname } = $album;
	}
	closedir(ARTIST);
    }
    my @albums = sort { $a->name cmp $b->name } values %albumsH;
    Trace::debug('Found %d matches', $#albums + 1) if Trace::check(Trace::DEBUG4);
    return \@albums;
}

# tracksQuery() - All tracks on a particular album

sub tracksQuery {
    my $repo = shift;
    my $album = shift;
    my $albumname = _encodeFileName($album->name);
    my $artistname = _encodeFileName($album->artist->name);
    
    my %tracks = ();
    my @files = (); # NFS duplicate check

    return undef if (!defined($repo->{'artistCache'}->{ $artistname }));

    foreach my $dir (@{ $repo->{ 'dirs' } }) {
	next if (!opendir(ALBUM, $dir.'/'.$artistname.'/'.$albumname));
	TRACK: foreach my $file (readdir(ALBUM)) {
	    next TRACK if (($file =~ m/^\.$/) || ($file =~ m/^\.\.$/) || ($file !~ m/^\d\d_.*\..*$/));
	    my $tfile = $dir.'/'.$artistname.'/'.$albumname.'/'.$file;
	    next if (grep { $_ eq $tfile } @files); # NFS duplicate check
	    push @files, $tfile;
	    my $format = Format->new(file => $tfile);
	    if (!defined($format)) {
		Trace::debug('Problems with %s', $tfile) if Trace::check(Trace::DEBUG6);
		next TRACK;
	    }
	    my $suffix = $format->suffix(unique => 1);
	    my $sufflen = length $suffix;
	    my $prefix = substr $file, 0, -$sufflen;
	    $prefix =~ /(\d\d)_(.*)/;
	    my $index = int($1);
	    my $name = $2;

	    my $track;
	    if (defined($tracks{ $index })) {
		$track = $tracks{ $index };
	    }
	    else {
		$track = Info::Track->find(owner => $repo->handle, album => $album, index => $index, name => _decodeFileName($name), 
					   duration => $format->duration, data => 0, create => 1);
		next TRACK if (!defined($track));
		$tracks{ $index } = $track;
		$track->{'formatsH'} = {};
	    }
	    $track->{'formatsH'}->{ $format->descriptor } = $format;		    
	    Trace::debug('Found match at %s', $tfile) if Trace::check(Trace::DEBUG6);
	}
	closedir(ALBUM);
    }
    my @sorted = map { $tracks{ $_ } } sort { $a <=> $b } keys(%tracks);
    return \@sorted;
}

# verifyAgainstDb() - Compare the directory contents against the Cd database. Debug purposes.

sub verifyAgainstDb {
    my $repo = shift;

    die "Needs work due to changes in cache";

    $repo->_artistCache;

    CD: for (my $cd = Cd->first; defined($cd); $cd = $cd->next) {
	TRACK: for (my $ti = $cd->firstAudioTrackIndex; defined($ti); $ti = $cd->nextAudioTrackIndex($ti)) {

	    my $track = $cd->track($ti);
	    my $artist = _encodeFileName( $track->{'artist'} );
	    my $album = _encodeFileName( $track->{'album'} );

	    if (!defined($repo->{'artistCache'}->{ $artist })) {
		Trace::debug('No such artist: %s', $artist) if Trace::check(Trace::DEBUG6);
		next CD;
	    }
	    my $cache = $repo->{'artistCache'}->{ $artist };

	    my $prefix = sprintf('%02d_', $track->{'index'});
	    $prefix .= _encodeFileName( $track->{'name'} );
	    $prefix = Command::quotemeta($prefix);

	    my $match = 0;
	    foreach my $dir (@{ $cache->{ 'dirs' } }) {
		next if (!opendir(ALBUM, $dir.'/'.$artist.'/'.$album));
		FILE: foreach my $file (readdir(ALBUM)) {
		    next FILE if (($file =~ m/^\.$/) || ($file =~ m/^\.\.$/) || ($file !~ m/^\d\d_.*\..*$/));
		    next FILE if ($file !~ m/^$prefix/);

		    $match = 1;
		    my $tfile = $dir.'/'.$artist.'/'.$album.'/'.$file;
		    my $format = Format->new(file => $tfile);
		    if (!defined($format)) {
			Trace::debug('Problems with %s', $tfile) if Trace::check(Trace::DEBUG6);
			next TRACK;
		    }
		    if (int($format->duration) != int($track->{'duration'})) {
			Trace::debug('Duration of file (%d seconds: %s) does not match track (%d)', 
				     int($format->duration), $tfile, int($track->{'duration'})) if Trace::check(Trace::DEBUG6);
			next TRACK;
		    }
		}
	    }
	    if (!$match) {
		Trace::debug('No file found for %s/%s/%s', $track->{'artist'}, $track->{'album'}, $track->{'name'}) if Trace::check(Trace::DEBUG6);
	    }
	}
    }
}

# _artistCache() - Caches artists, increases perfomance for some queries.

sub _artistCache {
    my $repo = shift;

    my $cache = $repo->{'artistCache'} = {};

    DIR: foreach my $dir (@{ $repo->{ 'dirs' } }) {
	next DIR if (!opendir(DIR, $dir));
        ARTIST: foreach my $artistname (readdir(DIR)) {
	    next ARTIST if (($artistname =~ m/^\.$/) || ($artistname =~ m/^\.\.$/) || ($artistname =~ m/^\.tmp$/));
	    if (!defined($cache->{ $artistname })) {
		my $artist = Info::Artist->find(owner => $repo->handle, name => _decodeFileName($artistname), create => 1);
		if ($artist) {
		    $cache->{ $artistname } = $artist;
		}
		else {
		    Trace::error('Could not find artist matching: %s', $artistname);
		}
	    }
	}
	closedir(DIR);
    }
}

sub _encodeFileName {
    my $filename = shift;
    while ($filename =~ m/^(.*)\/(.*)$/) {
	$filename = $1.'%2f%'.$2;
    }
    return $filename;
}

sub _decodeFileName {
    my $filename = shift;
    while ($filename =~ m/^(.*)\%2f\%(.*)$/) {
	$filename = $1.'/'.$2;
    }
    return $filename;
}

1;
