# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.37.2.4 $

package Thing;

# Things, useful event driven objects with a number of capabilities.

#   Message trading over TCP networks with other `things'.

#   Delayed messages sending.

#   Monitoring of other things.

#   Shared information objects. see Thing/Info.pm.

# Things can be instantiated and discovered by communicating with
# Thing::Service.

# When inherited, things can use the HASH ref object directly to store
# local state. The key `_thing' is reserved. Methods may be augmented
# as listed below. The prefix _tlm_ is used for local methods in this
# class to avoid possible unintended conflicts with your inherited
# class. The prefix _tm_ is used for methods shared with Thing/*
# classes.

# Object Method API: new(), name(), port(), host(), handle(), error(),
# register(), deregister(), message(), at(), work(), loop(), fork(),
# monitor(), monitorChange(), infoFind().

# Object Class API: daemon().

# External Message API, handlers are registered last.

# ConfigureRequest, parms are thing dependent, sub classes should
# register a handler to do the configuration and set parms->{'error'}
# if appropriate. _tlm_configure() is registered last and generates a
# ConfigureResponse which returns the same parms.

sub ConfigureRequest ()  { 'Thing__ConfigureRequest'; } # handler: _tlm_configure
sub ConfigureResponse () { 'Thing__ConfigureResponse'; }

sub ShutdownRequest ()   { 'Thing__ShutdownRequest'; } # handler: _tlm_shutdown
sub ShutdownResponse ()  { 'Thing__ShutdownResponse'; }

sub Started ()           { 'Thing__Started'; }

# Asynchronous response to a message() call when the outcome is not
# immediately apparent, parms indicate the original `message' and its
# send `outcome'.

sub SentMessage ()       { 'Thing__SentMessage'; }

# Response to a received `bad message', parms will indicate a `reason'.

sub BadMessage ()        { 'Thing__BadMessage'; }

# Messages bodies are serialized created hashes with the following keys:
#
#   from, a handle to the sending thing
#   name, the message name
#   parms, message dependant parameters

# Messages are in the form:

#   length, 4 bytes giving the body length in bytes.
#   mac, 8 byte authentication code, the MD5 digest of the body
#        encrypted with a shared secret.
#   body, serialized message.

use Thing::Handle;
@ISA = qw( Thing::Handle );
use Thing::ServiceH;
use Utility;
use Thing::Info;
use IO::Socket qw();
use IO::Select qw();
use IO::File qw();
use File::stat qw( stat );
use Socket qw();
use Trace;
use strict;
use POSIX qw(:errno_h :fcntl_h :sys_wait_h);
use POSIX qw(_exit);
BEGIN {
    use vars qw($_digest);
    eval {
	require Digest::MD5;
	$_digest = 'Digest::MD5';
    };
    if ($@) {
	require Digest::Perl::MD5;
	$_digest = 'Digest::Perl::MD5';
    }		
}
BEGIN {
    use vars qw($_crypt);
    eval {
	require Crypt::Blowfish;
	$_crypt = 'Crypt::Blowfish';
    };
    if ($@) {
	require Crypt::Blowfish_PP;
	$_crypt = 'Crypt::Blowfish_PP';
    }		
}

sub VERSION { $ENV{'THING_VERSION'} || '0.0.0'  }

# Special handlers

sub Ignore { 'Thing__Ignore'; }

# How to `serialize' a message:

# Use Storable module for serializing messages. Storable doesn't give
# an appreciable performance increase to warrant all the version
# troubles it creates when operating in a multi host environment (but
# I like it so the option exists to use it with THING_USE_STORABLE).

# Use FreezeThaw module for serializing messages. Freeze thaw isn't
# eval security dangerous like the local _unravel.

use vars qw( $_useStorable $_useFreezeThaw );

BEGIN {
    $_useFreezeThaw = 1;
    $_useStorable = 0;
    if (defined($ENV{'THING_USE_STORABLE'}) && $ENV{'THING_USE_STORABLE'}) {
	eval {
	    require Storable;
	};
	if (!$@) {
	    $_useFreezeThaw = 0;
	    $_useStorable = 1;
	}
    }
    if ($_useFreezeThaw) {
	require FreezeThaw;
    }
}

# Store the last blessed Thing, useful when daemonizing Things where
# only one Thing is instantiated, gives dependent packages a context
# in which to send messages.
my $_context;

# new() - Thing instantiation. 
#
#   `name' - The name to use for this thing, should be host unique for
#   lookups and configuration to work, defaults to the lowercase of
#   the type.
#
#   `port' - The TCP port to use, defaults to random selection.
#
#   `loner' - This thing will not register with Thing::Service.
#
# Sets $self->error() on failure.

sub new {
    my $proto;
    my $type;
    if ($#_ == -1) {
	$type = 'Thing';
    }
    else {
	$proto = shift;
	$type = ref($proto) || $proto;
    }
    my %parms = @_;
    
    $SIG{PIPE} = 'IGNORE';
    # Another attempt to workarround the seg faults generated by
    # reentrant mallocs. May not be very portable.
    $SIG{CHLD} = 'IGNORE';
    #$SIG{CHLD} = \&Thing::REAPER;
    $SIG{'__WARN__'} = sub 
    { 
	my $msg = shift;
	chop $msg;
	if ($msg =~ /Gtk|Socket|DBI/) {
	    Trace::debug($msg) if Trace::check(Trace::DEBUG6); 
	}
	else {
	    $SIG{'__WARN__'} = 'IGNORE';
	    Trace::error($msg); 
	}
    };
    $SIG{ABRT} = sub 
    { 
	$SIG{'__WARN__'} = 'IGNORE';
	Trace::error('Abort signal'); 
    };

    my $thing = {};

    my $name = $parms{'name'};
    if (!defined($name)) {
	$name = $type;
    }

    my $host = Thing::Service::HOST;
    my $self = $type->SUPER::new(name => $name, host => $host || 'unknown');
    $self->{'_thing'} = $thing;
    if (!defined($host)) {
	$thing->{'error'} = Trace::lastError;
	return $self;
    }

    my $pwd = Thing::readPwd();
    if (!$pwd) {
	$thing->{'error'} = Trace::lastError;
	return $self;
    }
    if (length $pwd < 8) {
	$pwd .= '        ';
    }
    $thing->{'crypt'} = $_crypt->new($pwd);

    $thing->{'read_select'} = IO::Select->new();
    $thing->{'write_select'} = IO::Select->new();

    if (!defined($thing->{'socket'} = IO::Socket::INET->new(Listen    => 5,
							    LocalPort => $parms{'port'} || undef,
							    Proto     => 'tcp',
							    Reuse => 1, ))) {
	Trace::error('Cannot open listening socket: %s', $!);
	if (defined($parms{'port'})) {
	    if ($parms{'port'} == Thing::Service::PORT) {
		$thing->{'error'} = 'Port '.$parms{'port'}.' is probably already in use, perhaps Thing::Service is already running.'."\n".
		                    'Use the environment variable THING_SERVICEPORT to specify a different one.'."\n".
				    'e.g. env THING_SERVICEPORT=5678 ....';
	    }
	    else {
		$thing->{'error'} = 'Port '.$parms{'port'}.' is probably already in use';
	    }
	}
	else {
	    $thing->{'error'} = Trace::lastError;
	}
	return $self;
    }
    else {
	$thing->{'socket'}->fcntl(F_SETFL, O_NONBLOCK);
	$self->port($thing->{'socket'}->sockport);
	Trace::debug('Opened socket on port %d', $self->port) if Trace::check(Trace::DEBUG5);
    }

    $thing->{'read_select'}->add($thing->{'socket'});
    $thing->{'registry'} = {};
    $thing->{'regIndex'} = 0;
    $thing->{'connections'} = [];
    $thing->{'atq'} = [];
    $thing->{'localq'} = [];
    $thing->{'id'} = 1;
    $thing->{'canReceive'} = 1;
    $thing->{'monitor'} = { queue => [], active => 0 };
    $thing->{'stdio'} = undef;
    $thing->{'trace'} = undef;
    $thing->{'info'} = { ioid => 1, owner => {}, monitor => {} };

    $self->register(name => ShutdownRequest, handler => '_tlm_shutdown', last => 1);
    $self->register(name => ConfigureRequest, handler => '_tlm_configure', last => 1);

    $self->register(name => Thing::Info::Monitor, handler => '_tlm_infoMonitor');
    $self->register(name => Thing::Info::Demonitor, handler => '_tlm_infoDemonitor');
    $self->register(name => Thing::Info::Sync, handler => '_tlm_infoSync');
    $self->register(name => Thing::Info::Update, handler => '_tlm_infoUpdate', first => 1);

    $self->register(name => Thing::Service::MonitorResponse, handler => '_tlm_monitorResponse', last => 1);

    if (!$parms{'loner'} && ($self->name ne Thing::Service::NAME)) {

	# Register with Thing::Service
	
	my $service = Thing::Service::Handle;
	$self->{'_serviceUp'} = 0;
	my $regid = $self->register(name => Thing::Service::RegisterResponse, handler => sub { $self->{'_serviceUp'} = 1; });    
	my $regid2 = $self->register(name => Thing::BadMessage, handler => sub { $self->{'_serviceUp'} = 2; 
										 shift; my %tP = @_; 
										 $self->{'_badMessage'} = $tP{'parms'}->{'reason'}; });
	if ($self->message(to => $service, name => Thing::Service::RegisterRequest)) {
	    my $clock = 0;
	    $self->work();
	    while (!$self->{'_serviceUp'}) {
		$self->work();
		select(undef, undef, undef, 0.1);
		$clock++;
		last if ($clock > 300);
	    }
	}
	
	if ($self->{'_serviceUp'} == 0) {
	    Trace::error('Cannot register with Thing::Service');
	    $thing->{'error'} = Trace::lastError;
	    return $self;
	}
	elsif ($self->{'_serviceUp'} == 2) {
	    Trace::error('Error from service: %s', $self->{'_badMessage'});
	    $thing->{'error'} = Trace::lastError;
	    return $self;
	}

	$self->deregister(name => Thing::Service::RegisterResponse, index => $regid);
	$self->deregister(name => Thing::BadMessage, index => $regid2);
	    
	delete $self->{'_serviceUp'} if (defined($self->{'_serviceUp'}));
	delete $self->{'_startError'} if (defined($self->{'_startError'}));
    }

    $_context = $self;

    return ($self);
}

# destroy()

sub destroy {
    my $self = shift;
    $self->message(name => Thing::Service::DeregisterRequest, to => Thing::Service::Handle);
    $self->_tlm_closeall;
    return 1;
}

sub context { return $_context; }

# error() - Set or retrieve an error message that occurred when the
# Thing was instantiated.

sub error { 
    if ($#_ == 1) {
	$_[0]->{'_thing'}->{'error'} = $_[1];
    }
    return $_[0]->{'_thing'}->{'error'};
}

# trace(), stdio() - Return the trace and stdio parameters used and
# set when the thing was daemonized.

sub trace { $_[0]->{'_thing'}->{'trace'}; }
sub stdio { $_[0]->{'_thing'}->{'stdio'}; }

# daemon() - Fork a thing process. All parms are passed to new()
# except where noted.

#   `type' - Use this to try and find the Class file and to
#   instantiate the thing. Defaults to a vanilla Thing.

#   `file' - Use this filename instead. Optional.

#   `loaded' - Don't try and load the Class file.

#   `starter' - Inform this thing of the start status via a Started
#   message. The parms are returned in the message and are polluted
#   with a `handle' to the object and `error' if something went wrong.

#   `trace' - Turn trace on with these parms otherwise the daemon
#   inherits the trace settigns of its creator.

#   `stdio' - Redirect the stdio to this file.

#   `close' - Close the open file handles of this thing after
#   forking. Not passed. Used by forked daemons to forget about their
#   creator.

sub daemon {
    my %parms = @_;

    my $pid = CORE::fork;
    if (!defined($pid)) {
	die "Cannot fork: $!";
    }

    if ($pid == 0) {
	# Fork again for proper disconnect and seemingly safer SIG_CHLDs.
	$pid = CORE::fork;
	if (!defined($pid)) {
	    die "Cannot fork: $!";
	}
	Thing::exit(1) if ($pid);
    }
    else {
	return 1;
    }

    # Close parents open sockets
    if (defined($parms{'close'})) {
	$parms{'close'}->_tlm_closeall;
	delete $parms{'close'};
    }

    # Redirect to log file
    if (defined($parms{'stdio'})) {
	my $fh = IO::File->new('>> '.$parms{'stdio'}) || die $!;
	POSIX::dup2($fh->fileno, 0);
	POSIX::dup2($fh->fileno, 1);
	POSIX::dup2($fh->fileno, 2);
	$| = 1;
    }

    # Turn on trace as desired
    if (defined($parms{'trace'})) {
	Trace::on(@{ $parms{'trace'} });
    }

    # Load library
    if (defined($parms{'type'})) {
	if (!$parms{'loaded'}) {
	    # Check it was not done already
	    my $file;
	    if (!defined($parms{'file'})) {
		$file = $parms{'type'}.'.pm';
		$file =~ s/\:\:/\//g;
	    }
	    else {
		$file = $parms{'file'};
	    }
	    if (!$INC{ $file }) {
		eval {
		    require $file;
		};
		if ($@) {
		    Trace::error('Cannot read thing library: %s: %s', $file, $@.Trace::lastError);
		    die Trace::lastError;
		}
	    }
	}
    }
    else {
	$parms{'type'} = 'Thing';
    }

    # Disconnect
    POSIX::setsid();

    if ($parms{'debug'}) {
	$DB::single = 1;
	delete $parms{'debug'};
    }

    # Start
    my $thing = $parms{'type'}->new(%parms);
    if (!defined($thing)) {
	die Trace::lastError;
    }
    $parms{'handle'} = $thing->handle;

    if (!$thing->error) {
	if (defined($parms{'starter'})) {
	    $parms{'starter'} = $parms{'starter'}->handle;
	    $thing->message(name => Started, to => $parms{'starter'}, parms => \%parms);
	}
	$0 = $thing->name;
	$thing->{'_thing'}->{'stdio'} = $parms{'stdio'};
	$thing->{'_thing'}->{'trace'} = $parms{'trace'};
	$thing->loop();
    }
    else {
	if (defined($parms{'starter'})) {
	    $parms{'error'} = $thing->error;
	    $parms{'starter'} = $parms{'starter'}->handle;
	    $thing->message(name => Started, to => $parms{'starter'}, parms => \%parms);
	}
	Trace::debug('Cannot create thing') if Trace::check(Trace::DEBUG2);
    }
    Thing::exit(1);
}

# register() - Registers a `handler' method for received messages of
# `name'. `handler' can be a subroutine name or an anonymous code
# reference. It will be invoked when messages are recieved from
# another thing as:
#
#   $self->`handler'(from => <thing>, parms => {}),
# 
# where `parms' are those passed by the other thing in its call to
# message(). Multiple handlers may registered for the same named
# message and will be invoked in a last registered first called
# manner. However setting `last' makes this handler the last to be
# invoked. An index is returned which is needed to deregister this
# handle.

sub register {
    my $self = shift;
    my $thing = $self->{'_thing'};
    my %parms = @_;

    my $name = $parms{'name'};
    if (!defined($name)) {
	Trace::mandatory('name');
	return undef;
    }
    my $handler = $parms{'handler'};
    if (!defined($handler)) {
	Trace::mandatory('handler');
	return undef;
    }

    $thing->{'regIndex'}++;
    $thing->{'registry'}->{ $name } = [] if (!defined($thing->{'registry'}->{ $name }));
    my $reg = { name => $name, handler => $handler, index => $thing->{'regIndex'}};
    
    if ($parms{'last'}) {
	push @{ $thing->{'registry'}->{ $name } }, $reg;
    }
    else {
	unshift @{ $thing->{'registry'}->{ $name } }, $reg;
    }

    Trace::debug('Registered message handler for: %s', $name) if Trace::check(Trace::DEBUG4);

    return $thing->{'regIndex'};
}

# deregister() - remove previously registerd handler at 'index' or all
# handlers of `name ' if no `index' is specified.

sub deregister {
    my $self = shift;
    my $thing = $self->{'_thing'};
    my %parms = @_;

    my $name = $parms{'name'};
    if (!defined($name)) {
	Trace::mandatory('name');
	return undef;
    }

    my $index = $parms{'index'};
    if (!defined($index)) {
	delete $thing->{'registry'}->{ $name };
	Trace::debug('Deregistered all message handlers for: %s', $name) if Trace::check(Trace::DEBUG4);
	return 1;
    }

    my $pos = 0;
    foreach my $reg (@{ $thing->{'registry'}->{ $name } }) {
	if ($reg->{'index'} == $index) {
	    splice @{ $thing->{'registry'}->{ $name } }, $pos, 1;
	    Trace::debug('Deregistered message handlers for: %s, at index: %d', $name, $index) if Trace::check(Trace::DEBUG4);
	    return 1;
	}
	$pos++;
    }
    Trace::debug('No registered message handlers for: %s, at index: %d', $name, $index) if Trace::check(Trace::DEBUG4);
    return 0;
}

# message() - Send a message `name' to recipient `to', including
# `parms'. Returns 1 on success, 0 on error, undef if the outcome is
# not yet known. For an unknown outcome set parms `outcome' and a
# SentMessage will be sent indicating `message' and `outcome' status
# in its parms.

sub message {
    my $self = shift;
    my $thing = $self->{'_thing'};
    my %parms = @_;

    if (!defined($parms{'to'}) || !UNIVERSAL::isa($parms{'to'}, 'Thing::Handle')) {
	Trace::mandatory('to');
	return undef;
    }
    my $to = $parms{'to'}->handle;
    if (!defined($parms{'name'})) {
	Trace::mandatory('name');
	return undef;
    }

    if (!defined($to->port) || ($to->port == -1)) {
	Trace::error('Cannot send to %s, needs looking up', $to->name);
	return undef;
    }

    Trace::debug('Sending message %s to %s', $parms{'name'}, _tlm_describeHandle($to)) if Trace::check(Trace::DEBUG4);

    # Check for local message

    if (($self == $to) || (($self->host eq $to->host) && ($self->port == $to->port))) {
	Trace::debug('Sent message locally') if Trace::check(Trace::DEBUG4);
	push @{ $thing->{'localq'} }, { from => $self->handle(), name => $parms{'name'}, parms => $parms{'parms'} };
	return 1;
    }

    # Check for existing connection

    my $connection;
    foreach my $check (@{ $thing->{'connections'} }) {
	next if (!defined($check->{'thing'}));
	if (($to->host eq $check->{'thing'}->host) && ($to->port == $check->{'thing'}->port)) {
	    $connection = $check;
	    last;
	}
    }

    # No existing connection

    if (!defined($connection)) {
	# Create new connection
	$connection = {};
	if (!($connection->{'socket'} = IO::Socket::INET->new(PeerAddr => $to->host, PeerPort => $to->port, Proto => 'tcp'))) {
	    Trace::debug('Message %s never sent to %s: %s', $parms{'name'}, _tlm_describeHandle($to), $!)
	      if Trace::check(Trace::DEBUG3);
	    return 0;
	}
	$connection->{'thing'} = $to;
	$connection->{'socket'}->fcntl(F_SETFL, O_NONBLOCK);
	$connection->{'id'} = $thing->{'id'}++;
	$connection->{'read'} = {};
	$connection->{'write'} = {};
	$connection->{'direction'} = 'out';

	$thing->{'read_select'}->add($connection->{'socket'});
	push @{ $thing->{'connections'} }, $connection;

	Trace::debug('New connection to: %s', _tlm_describeConnection($connection)) if Trace::check(Trace::DEBUG4);
    }

    # Put it on the write queue

    if (!defined($connection->{'write'}->{'queue'})) {
	$connection->{'write'}->{'queue'} = [];
    }

    push @{ $connection->{'write'}->{'queue'} }, \%parms;

    #require 'dumpvar.pl';
    #main::dumpValue(\%parms);

    # Process the write queue

    my $rc = $self->_tlm_processWriteQueue($connection, 0);

    if (!defined($rc)) {
	# Error
	return 0;
    }
    elsif (($#{ $connection->{'write'}->{'queue'} } != -1) || defined($connection->{'write'}->{'partial'}))  {
	# Some still pending
        #return undef;

        # The concept of pending messages is not supported by using
        # code. Cleaning up for 1.7.0 for now assume success.
        return 1;
    }

    # Sent
    return 1;
}

# at() - send a message at `time' seconds.

sub at {
    my $self = shift;
    my $thing = $self->{'_thing'};
    my %parms = @_;

    my $time = $parms{'time'};
    if (!defined($time)) {
	Trace::mandatory('time');
	return 0;
    }

    push @{ $thing->{'atq'} }, \%parms;

    return 1;
}

# work([readers, writers]) - Check for received messages and execute
# any registered handlers. Send any queued write messages or write
# ready sockets. Send any ready queued via at() messages.

# Process is as follows: Attempt to read one message from any ready
# socket, Check for any more ready sockets. Repeat until no messages
# are read. Gives a fair look in to each connection, while still
# processing all messages.

sub work {
    my $self = shift;
    my $readers = shift;
    my $writers = shift;
    my $thing = $self->{'_thing'};

    # Forked process can't read from the listening socket.
    return 1 if (!$thing->{'canReceive'});

    my @list;
    my @priorReaders = ();
    my @possiblyDead = ();
    my $firstPass = 1;
    my $pending;

    do {
	$pending = 0;

	# Check for and send pending at() messages.

	my $now = time;
	if (@{ $thing->{'atq'} }) {
	  my @undone = grep 
	    {
	      if ($_->{'time'} <= $now) {
		$self->message(%{ $_ });
		0;
	      }
	      else {
		1;
	      }
	    } @{ $thing->{'atq'} };
	  $thing->{'atq'} = \@undone;
	}

	# A single pass is made through each queue ready for
	# reading. Subsequent passes will take all available data for
	# queues that reported ready. Following that queues will not
	# necessarily report ready if only part of their available
	# data is read. The single pass gives write a chance.

	# Get the reader ready list

	if (!$firstPass || !defined($readers)) {
	    my @ready = $self->{'_thing'}->{'read_select'}->can_read(0.0001);
	    $readers = \@ready;
	}

	# Derive ready connections

	@list = ();
	READER: foreach my $ready (@{ $readers }) {
	    if ($ready != $thing->{'socket'}) {
		foreach my $c (@{ $thing->{'connections'} }) {
		    if ($c->{'socket'} == $ready) {
			$c->{'readySignal'} = 1;
			push @list, $c;
			next READER;
		    }
		}
		Trace::debug('Not one of my sockets: %s',  _tlm_describeSocket($ready)) if Trace::check(Trace::DEBUG6);
	    }
	    else {
		# Check for new connections
		while (my $new = $thing->{'socket'}->accept) {
		    last if ($new == EAGAIN);
		    die if (ref($new) ne 'IO::Socket::INET');
		    $new->fcntl(F_SETFL, O_NONBLOCK);
		    my $connection = {
			socket => $new,
			id => $thing->{'id'}++,
			read => {},
			write => {},
			direction => 'in',
		    };
		    $thing->{'read_select'}->add($new);
		    push @{ $thing->{'connections'} }, $connection;
		    push @list, $connection; # Should be ready for a read
		    Trace::debug('New connection from: %s', _tlm_describeConnection($connection)) if Trace::check(Trace::DEBUG4);
		}
	    }
	}

	# Use prior read connections not now in the ready list

	PRIOR: foreach my $prior (@priorReaders) {
	    CONN: foreach my $connection (@list) {
		if ($prior == $connection) {
		    Trace::debug('Prior on ready list: %s', _tlm_describeConnection($prior)) if Trace::check(Trace::DEBUG6);
		    next PRIOR;
		}
	    }
	    Trace::debug('Prior not on ready list: %s', _tlm_describeConnection($prior)) if Trace::check(Trace::DEBUG6);
	    push @list, $prior;
	}
	@priorReaders = ();

	# DEBUG: Check for duplicates

	if (Trace::check(Trace::DEBUG6)) {
	    CONN: foreach my $connection (@list) {
		my $count = 0;
		foreach my $dupcheck (@list) {
		    if ($dupcheck == $connection) {
			$count++;
			if ($count >= 2) {
			    Trace::debug('Duplicate on read list: %s', _tlm_describeConnection($dupcheck));
			    next CONN;
			}
		    }
		}
	    }
	}

	push @list, $thing->{'localq'}; # Check for messages on the local queue

	CONN: foreach my $connection (@list) {

	    my $message;
	    if ($connection != $thing->{'localq'}) {
		my $rc = $self->_tlm_readMessage($connection, $connection->{'readySignal'}, \@possiblyDead, \$message);
		if (!$rc) {
		    next CONN; # Error or no message
		}
		if (!defined($message)) {
		    $pending++; 
		    push @priorReaders, $connection;
		    $connection->{'readySignal'} = 0;
		    next CONN; # Partial
		}
	    }
	    elsif ($#{ $thing->{'localq'} } != -1) {
		$message = shift @{ $thing->{'localq'} };
	    }
	    next CONN if (!$message);

	    my $handlers = $thing->{'registry'}->{ $message->{'name'} };
	    if (defined($handlers)) {
		foreach my $reg (@{ $handlers }) {
		    if ($reg->{'handler'} eq Thing::Ignore) {
			Trace::debug('Message ignored: %s', $message->{'name'}) if Trace::check(Trace::DEBUG6);
		    }
		    else {
			my $handler = $reg->{'handler'};
			if (UNIVERSAL::isa($handler, 'CODE')) {
			    &{ $handler }($self, %{ $message });
			}
			else {
			    Trace::debug('Message handler: %s', $handler) if Trace::check(Trace::DEBUG6);
			    $self->$handler(%{ $message });
			}
		    }
		}
	    }
	    else {
		Trace::debug('Handler for %s undefined', $message->{'name'}) if Trace::check(Trace::DEBUG2);
	    }
	
	    if ($connection != $thing->{'localq'}) {
		push @priorReaders, $connection;
		$connection->{'readySignal'} = 0;
	    }
	    $pending++; # Possibly
	}

	# Build the writers list

	if (!$firstPass || !defined($writers)) {
	    my @ready = $self->{'_thing'}->{'write_select'}->can_write(0.0001);
	    $writers = \@ready;
	}

	@list = ();
	WRITER: foreach my $ready (@{ $writers }) {
	    foreach my $c (@{ $thing->{'connections'} }) {
		if ($c->{'socket'} == $ready) {
		    push @list, $c;
		    next WRITER;
		}
	    }
	    Trace::debug('Not one of my sockets: %s',  _tlm_describeSocket($ready)) if Trace::check(Trace::DEBUG6);
	}

	# A single pass is made through each queue ready for writing
	# attempting to write all outstanding messages. The single
	# pass gives read another chance.

	foreach my $connection (@list) {

	    if ($connection == $thing->{'socket'}) {
		Trace::debug('Write ready reported on my listener') if Trace::check(Trace::DEBUG6);
		next;
	    }
	    $self->_tlm_processWriteQueue($connection, 1);
	}

	if ($thing->{'write_select'}->count) {
	    $pending++;
	}

	$firstPass = 0;
    } while ($pending || ($#{ $thing->{'localq'} } != -1));

    # Possibly dead things are used for monitoring by Thing::Service

    DEAD: foreach my $dead (@possiblyDead) {
	next if !defined($dead->{'thing'});
	# Make sure it did not find its way back on the connection list
	foreach my $c (@{ $thing->{'connections'} }) {
	    if (defined($c->{'thing'}) && Thing::Handle::cmpNameHost($c->{'thing'}->handle, $dead->{'thing'}->handle)) {
		Trace::debug('Possibly dead connection is back on the connection list: %s',  _tlm_describeConnection($dead)) if Trace::check(Trace::DEBUG6);
		next DEAD;
	    }
	}
	$self->_tm_possiblyDead($dead->{'thing'}->handle);
    }

    return 1;
}
    
# loop() - Work continually.
    
sub loop {
    my $self = shift;
    my $thing = $self->{'_thing'};
    my $readers; my $writers;

    while (1) {

	# Work out a correct timeout or block until something comes in.

	my $nap = undef; # Block
	if ($#{ $thing->{'localq'} } == -1) {
	    my $now = time;
	    if (defined($thing->{'wait'}) && defined($thing->{'wait'}->{'time'})) {
		my $diff = $thing->{'wait'}->{'time'} - $now;
		if ($diff == 0) {
		    $diff = 0.0001;
		}
		$nap = $diff;
	    }
	    foreach my $pending (@{ $thing->{'atq'} }) {
		my $diff = $pending->{'time'} - $now;
		if ($diff == 0) {
		    $diff = 0.0001;
		}
		if (!defined($nap) || ($nap > $diff)) {
		    $nap = $diff;
		}
	    }
	}
	else {
	    # Stuff in the local queue
	    $nap = 0.0001;
	}
	if (defined($nap) && ($nap < 0)) {
	    $nap = 0.0001;
	}

	Trace::debug('Waiting to read or write messages: %s', defined($nap) ? $nap.' seconds' : 'blocking') if Trace::check(Trace::DEBUG6);
	($readers, $writers) = IO::Select::select($thing->{'read_select'}, $thing->{'write_select'}, undef, $nap);
	if (defined($nap) || ($#{ $readers } >= 0) || ($#{ $writers } >= 0)) {
	    $self->work($readers, $writers);
	}
	if (defined($thing->{'wait'})) {
	    if (&{ $thing->{'wait'}->{'event'} }) {
		return 1;
	    }
	    if (defined($thing->{'wait'}->{'time'}) && ($thing->{'wait'}->{'time'} <= time)) {
		return 0;
	    }
	}
    }
}

# wait - Work until `event' becomes true or until `timeout'
# seconds. Returns 1 if the event happened, 0 if it timed out first
# and undefined on error. Nested waits are supported but the last wait
# executed takes precedence.

sub wait {
    my $self = shift;
    my $thing = $self->{'_thing'};
    my %parms = @_;

    if (!defined($parms{'event'})) {
	Trace::mandatory('event');
	return;
    }
    
    if (&{ $parms{'event'} }) {
	# Already occurred
	return 1;
    }

    if (!defined($thing->{'waitStack'})) {
	$thing->{'waitStack'} = [];
    }

    my $wait = { event => $parms{'event'} };
    if (defined($parms{'timeout'})) {
	$wait->{'time'} = time + $parms{'timeout'};
    }
    push @{ $thing->{'waitStack'} }, $wait;
    $thing->{'wait'} = $wait;

    my $rc = $self->loop;

    pop @{ $thing->{'waitStack'} };
    if ($#{ $thing->{'waitStack'} } >= 0) {
	$thing->{'wait'} = $thing->{'waitStack'}->[ $#{ $thing->{'waitStack'} } ];
    }
    else {
	delete $thing->{'wait'};
    }

    return $rc;
}

# fork() - Fork a child process cleanly - Close all connections, used
# by things that fork to ensure that messages the child process might
# send go via a clean separate pipe.

# Incidentally a child process cannot receive messages.

sub fork {
    my $self = shift;

    my $pid = CORE::fork;
    if ($pid == -1) {
	Trace::error('Cannot fork child thing: %s', $!);
	return -1;
    }
    return 1 if ($pid);

    # Fork again for proper disconnect and seemingly safer SIG_CHLDs.
    $pid = CORE::fork;
    if (!defined($pid)) {
	die "Cannot fork: $!";
    }
    Thing::exit(1) if ($pid);
    # Disconnect
    POSIX::setsid();
    # Close parents connections
    $self->{'_thing'}->{'socket'}->close;
    foreach my $connection (@{ $self->{'_thing'}->{'connections'} }) {
	$connection->{'socket'}->close;
    }
    delete $self->{'_thing'}->{'socket'};
    $self->{'_thing'}->{'connections'} = [];
    $self->{'_thing'}->{'canReceive'} = 0;

    return 0;
}

# exit() - exit call for forked or daemonised Things

sub exit {
    if (defined($DB::profile) && $DB::profile) {
      CORE::exit($_[0]);
    }
    else {
      # Avoid destructor END calls which have caused problems
      _exit($_[0]);
    }
}

# PASSWORD

# readPwd() - Class function to read the password from the pwd file or
# ENV var THING_PWD. Returns the passowrd on success, 0 on failure and
# sets Trace::lastError. 

sub readPwd {

    if (defined($ENV{'THING_PWD'})) {
	return $ENV{'THING_PWD'};
    }
    if (!defined(Thing::Service::configPath())) {
	Trace::error(Thing::Service::configPathError());
	return 0;
    }
    my $pwdfile = Thing::Service::configPath().'/pwd';
    my $sb = stat($pwdfile) if (-e $pwdfile);
    if (!defined($sb)) {
	Trace::error('Cannot stat password file: %s: %s', $pwdfile, $!);
	return 0;
    }
    my $fh = IO::File->new;
    if (!$fh->open($pwdfile)) {
	Trace::error('Cannot open password file: %s: %s', $pwdfile, $!);
	return 0;
    }
    my $pwd = <$fh>;
    $fh->close;
    if (!defined($pwd) || (length $pwd < 8)) {
	Trace::error('Bad password found in password file: %s', $pwdfile);
	return 0;
    }
    return $pwd;
}

# MONITORING

# monitor(handle) - Used to monitor the state of a host uniquely named
# thing as identified by `handle'.

# States is initially Unknown and varies between Dead and Alive. State
# change is notified by a call to $self->monitorChange(handle) which
# should be overwritten.

sub monitor {
    my $self = shift;
    my $thing = $self->{'_thing'};
    my $monitor = $thing->{'monitor'};
    my $monitorThing = shift;

    Trace::debug('Monitor request: %s', $monitorThing->key) if Trace::check(Trace::DEBUG4);

    foreach my $check ( @{ $monitor->{'queue'} } ) {
	if (Thing::Handle::cmpNameHost($check, $monitorThing)) {
	    Trace::debug('Already monitoring: %s', $monitorThing->key) if Trace::check(Trace::DEBUG4);
	    return $check;
	}
    }
    $monitorThing->currentState(Thing::Handle::Unknown);
    $monitorThing->previousState(Thing::Handle::Unknown);
    push @{ $monitor->{'queue'} }, $monitorThing;
    
    $self->message(name => Thing::Service::MonitorRequest, to => Thing::Service::Handle, 
		   parms => { for => $self->handle, handle => $monitorThing });

    return $monitorThing;
}

# ismonitor(handle)

sub ismonitor {
    my $self = shift;
    my $thing = $self->{'_thing'};
    my $monitor = $thing->{'monitor'};
    my $monitorThing = shift;

    foreach my $check ( @{ $monitor->{'queue'} } ) {
	return $check if (Thing::Handle::cmpNameHost($check, $monitorThing));
    }
    return 0;
}

# demonitor(handle)

sub demonitor {
    my $self = shift;
    my $thing = $self->{'_thing'};
    my $monitor = $thing->{'monitor'};
    my $monitorThing = shift;

    Trace::debug('Demonitor request: %s ', $monitorThing->key) if Trace::check(Trace::DEBUG4);

    my $pos = 0;
    foreach my $check ( @{ $monitor->{'queue'} } ) {
	if (Thing::Handle::cmpNameHost($check, $monitorThing)) {
	    splice @{ $monitor->{'queue'} }, $pos, 1;
	    $self->message(name => Thing::Service::DemonitorRequest, to => Thing::Service::Handle, 
			   parms => { for => $self->handle, handle => $monitorThing });
	    return 1;
	}
	$pos++;
    }
    return 0;
}

# monitorChange(handle) - Called when the status of one of the monitored
# things has changed. Typically overwritten by subclasses.

sub monitorChange {
    my $self = shift;
    my $thing = $self->{'_thing'};
    my $monitorThing = shift;
}

# _tlm_monitorResponse - Monitor response from service

sub _tlm_monitorResponse {
    my $self = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};
    my $from = $tParms{'from'};
    my $thing = $self->{'_thing'};
    my $monitor = $thing->{'monitor'};

    foreach my $check ( @{ $monitor->{'queue'} } ) {
	if (Thing::Handle::cmpNameHost($check, $parms->{'handle'})) {
	    $check->port($parms->{'handle'}->port);
	    Trace::debug('Monitor response: %s: %s -> %s', $check->unique, $check->previousState, $parms->{'handle'}->currentState) if Trace::check(Trace::DEBUG4);
	    if ($check->stateChange($parms->{'handle'}->currentState)) {
		$self->monitorChange($check);
	    }
	    return;
	}
    }
}

# _tm_possiblyDead - Used by Thing::Service to monitor things

sub _tm_possiblyDead {
    my $self = shift;
    my $check = shift;
}

# INFO

# infoFind() - Find an information object we are monitoring by its ioid.

sub infoFind {
    my $self = shift;
    my $thing = $self->{'_thing'};
    my %parms = @_;

    my $ioid = $parms{'ioid'};
    if (!defined($ioid)) {
	Trace::mandatory('ioid');
	return undef;
    }
    return $thing->{'info'}->{'monitor'}->{ $ioid };
}

# _tm_infoAllocIoid() - Allocate a unique object id.

sub _tm_infoAllocIoid {
    my $self = shift;
    my $thing = $self->{'_thing'};
    return $self->handle->unique.'_'.$$.'_'.$thing->{'info'}->{'ioid'}++;
}

# _tm_infoRegisterMonitor() - Called by Thing::Info->monitor, Remember
# that we are monitoring this info object and tell its owner the
# same. If we are already monitoring this object return it.

sub _tm_infoRegisterMonitor {
    my $self = shift;
    my $thing = $self->{'_thing'};
    my $info = shift;

    if (defined($thing->{'info'}->{'monitor'}->{ $info->ioid })) {
	return $thing->{'info'}->{'monitor'}->{ $info->ioid };
    }
    $thing->{'info'}->{'monitor'}->{ $info->ioid } = $info;
    $self->message(to => $info->owner, name => Thing::Info::Monitor, parms => { ioid => $info->ioid });
    return $info;
}

# _tm_infoDeregisterMonitor() - Called by Thing::Info->demonitor.

sub _tm_infoDeregisterMonitor {
    my $self = shift;
    my $thing = $self->{'_thing'};
    my $info = shift;

    delete $thing->{'info'}->{'monitor'}->{ $info->ioid } if (defined($thing->{'info'}->{'monitor'}->{ $info->ioid }));
    $self->message(to => $info->owner, name => Thing::Info::Demonitor, parms => { ioid => $info->ioid });
    return 1;
}

# infoDemonitorAll() - Stop monitoring all monitored objects, quick and easy garbage cleaner.

sub infoDemonitorAll {
    my $self = shift;
    my $thing = $self->{'_thing'};

    foreach my $info (values(%{ $thing->{'info'}->{'monitor'} })) {
	delete $thing->{'info'}->{'monitor'}->{ $info->ioid };
	$self->message(to => $info->owner, name => Thing::Info::Demonitor, parms => { ioid => $info->ioid });
    }
    return 1;
}

# _tlm_infoMonitor - As the owner of this object add `from' as a monitor.

sub _tlm_infoMonitor {
    my $self = shift;
    my $thing = $self->{'_thing'};
    my %tParms = @_;
    my $from = $tParms{'from'};
    my $parms = $tParms{'parms'};

    my $ioid = $parms->{'ioid'};
    if (!defined($thing->{'info'}->{'owner'}->{ $ioid })) {
	$thing->{'info'}->{'owner'}->{ $ioid } = { ioid => $ioid, monitors => {} };
    }
    $thing->{'info'}->{'owner'}->{ $ioid }->{'monitors'}->{ $from->unique } = $from;

    # Send the latest data

    if (defined($thing->{'info'}->{'owner'}->{ $ioid }->{'data'})) {
	$self->message(to => $from, name => Thing::Info::Update, parms => { ioid => $ioid,
									    new => $thing->{'info'}->{'owner'}->{ $ioid }->{'data'} });
    }
}

# _tlm_infoDemonitor - As the owner of this object remove `from' as a monitor.

sub _tlm_infoDemonitor {
    my $self = shift;
    my $thing = $self->{'_thing'};
    my %tParms = @_;
    my $from = $tParms{'from'};
    my $parms = $tParms{'parms'};

    my $ioid = $parms->{'ioid'};
    if (defined($thing->{'info'}->{'owner'}->{ $ioid }) && defined($thing->{'info'}->{'owner'}->{ $ioid }->{'monitors'}->{ $from->unique })) {
	delete $thing->{'info'}->{'owner'}->{ $ioid }->{'monitors'}->{ $from->unique };
	Trace::debug('%s stopped monitoring: %s', $from->key, $ioid) if Trace::check(Trace::DEBUG4);
	if (! scalar keys %{ $thing->{'info'}->{'owner'}->{ $ioid }->{'monitors'} }) {
	    # Nothing left monitoring this object
	    Trace::debug('Nothing left monitoring: %s', $ioid) if Trace::check(Trace::DEBUG4);
	    delete $thing->{'info'}->{'owner'}->{ $ioid };
	}
    }
}

# _tlm_infoSync() - Receipt of Sync message for an object we own. A
# holder of this object wants to modify its data. We track changes to
# the object to ensure that all monitors always hold the most up to
# date data.

sub _tlm_infoSync {
    my $self = shift;
    my $thing = $self->{'_thing'};
    my %tParms = @_;
    my $from = $tParms{'from'};
    my $parms = $tParms{'parms'};

    my $ioid = $parms->{'ioid'};
    if (!defined($thing->{'info'}->{'owner'}->{ $ioid })) {
	$thing->{'info'}->{'owner'}->{ $ioid } = { ioid => $ioid, monitors => {} };
    }
    my $mo = $thing->{'info'}->{'owner'}->{ $ioid };

    Trace::debug('Info sync from %s for: %s', $from->key, $ioid) if Trace::check(Trace::DEBUG5);

    # Only update changes. Allows for single field changes and sub
    # objects with more fields than supers. WHY: We can get a sync
    # when instantiating a super class of an object which has already
    # been instantiated via a sub class. We don't want to lose the sub
    # class data. ASSUMES: the sub class augments the super data wise.

    my $diff = Utility::Hash::augment(existing => $mo->{'data'}, new => $parms->{'new'}, 
				      special => [ { compareFn => sub { Thing::Info::compare(@_); } } ]);

    if (!defined($diff)) {
	if (defined($parms->{'new'})) {
	    # Something other than a hash or previously undefined
	    $mo->{'data'} = $parms->{'new'};
	}
	elsif (!defined($parms->{'new'}) && (!defined($parms->{'force'}) || !$parms->{'force'})) {
	    # No change
	    return 1;
	}
    }
    else {
	if (!$diff && (!defined($parms->{'force'}) || !$parms->{'force'})) {
	    # No change
	    return 1;
	}
    }

    foreach my $monitor (values %{ $mo->{'monitors'} }) {
	if (!$self->message(to => $monitor, name => Thing::Info::Update, parms => $parms)) {
	    delete $mo->{'monitors'}->{ $monitor->unique };
	    Trace::debug('%s stopped monitoring: %s', $from->key, $ioid) if Trace::check(Trace::DEBUG4);
	}
    }

    if (! scalar keys %{ $thing->{'info'}->{'owner'}->{ $ioid }->{'monitors'} }) {
	# Nothing left monitoring this object
	Trace::debug('Nothing left monitoring: %s', $ioid) if Trace::check(Trace::DEBUG4);
	delete $thing->{'info'}->{'owner'}->{ $ioid };
    }
    return 1;
}

# _tlm_infoUpdate() - Receipt of Update message.

sub _tlm_infoUpdate {
    my $self = shift;
    my $thing = $self->{'_thing'};
    my %tParms = @_;
    my $from = $tParms{'from'};
    my $parms = $tParms{'parms'};

    my $ioid = $parms->{'ioid'};
    my $info = $thing->{'info'}->{'monitor'}->{ $ioid };

    if (!defined($info)) {
	Trace::error('Object is not monitored by me: %s', $ioid);
	return 0;
    }

    return $info->_im_update($parms);
}

# LOCAL

#
# Message Writing
#

# _tlm_processWriteQueue() - Process the write queue for a ready
# connection. Returns 1 if something was written, 0 if nothing was
# written and undef on error.

sub _tlm_processWriteQueue {
    my $self = shift;
    my $thing = $self->{'_thing'};
    my $connection = shift;
    my $workContext = shift;

    my $write = $connection->{'write'};

    my $errorFree = 0; # Count for this socket
    my $newSocket = 0; # Allow one attempt to reconnect

    my $errorSub = sub {
	Trace::debug(@_) if (($#_ != -1) && Trace::check(Trace::DEBUG6));
	if (!$errorFree && !$newSocket) {
	    # Socket has not had an error free write for this call try with a new one
	    $thing->{'read_select'}->remove($connection->{'socket'});
	    $thing->{'write_select'}->remove($connection->{'socket'});
	    $connection->{'socket'}->close;
	    if (!($connection->{'socket'} = IO::Socket::INET->new(PeerAddr => $connection->{'thing'}->host, 
								  PeerPort => $connection->{'thing'}->port, Proto => 'tcp'))) {
		$self->_tlm_socketError($connection);
		return undef;
	    }
	    $connection->{'direction'} = 'out';
	    $newSocket = 1;
	    return 1;
	}
	else {
	    # Had recent writes to this socket, the connection is bad.
	    $self->_tlm_socketError($connection);
	    return undef;
	}
    };

    my $dataWritten = 0;
    TRYAGAIN: while (1) {

	# Partial prior write

	if (defined($write->{'partial'})) {
	    my $partial = $write->{'partial'};
	    my $written;
	    if (!defined($written = _tlm_writeBytes($connection->{'socket'}, $partial->{'dataR'}, $partial->{'length'}, $partial->{'offset'}))) {
		return undef if !&$errorSub('Problems finishing write');
		next TRYAGAIN;
	    }
	    $partial->{'offset'} += $written;
	    $dataWritten += $written;
	    if ($partial->{'offset'} == $partial->{'length'}) {
		# Written
		my $message = $partial->{'message'};
		$self->message(to => $self, name => SentMessage, parms => { message => $message, outcome => 1 }) if ($message->{'outcome'});
		Trace::debug('Sent message (%d bytes) to %s', $partial->{'length'}, _tlm_describeConnection($connection)) if Trace::check(Trace::DEBUG4);
		$thing->{'write_select'}->remove($connection->{'socket'});
		$connection->{'lastSend'} = time;
		delete $write->{'partial'};
	    }
	    else {
		Trace::debug('Partial write of message (%d/%d bytes)', $partial->{'offset'}, $partial->{'length'}) if Trace::check(Trace::DEBUG5);
	    }
	    $errorFree++;
	}

	# No partial prior write or the last one was sent

	if (!defined($write->{'partial'})) {
	    MESSAGE: while (my $message = shift (@{ $write->{'queue'} })) {

		my $data; my $total; my $body; my $mac; my $length;
		if ($message->{'name'} ne BadMessage) {
		    my $mref = { from => $self->handle(), name => $message->{'name'}, parms => $message->{'parms'}, };
		    $self->_tlm_buildMessage($mref, \$body, \$mac, \$length);
		}
		else {
		    $body = pack('Z*', $message->{'reason'});
		    $mac = '__0000__';
		    $length = length $body;
		}

		my $plength = pack('L', $length);
		$data = $plength.$mac.$body;
		
		$total = (length $plength) + (length $mac) + $length;

		my $written;
		if (!defined($written = _tlm_writeBytes($connection->{'socket'}, \$data, $total, 0))) {
		    unshift @{ $write->{'queue'} }, $message;
		    return undef if !&$errorSub('Problems with message write');
		    next TRYAGAIN;
		}
		$errorFree++;
		$dataWritten += $written;
		if ($written != $total) {
		    # Partial write
		    $write->{'partial'} = { dataR => \$data, length => $total, offset => $written, message => $message };
		    # This connection now needs write checking
		    $thing->{'write_select'}->add($connection->{'socket'});
		    last MESSAGE;
		}
		else {
		    # Written - send the outcome, last messages in the
		    # queue may still be able to get a successful
		    # return value if called from a message() context.
		    if (($#{ $write->{'queue'} } != -1) || $workContext) {
			$self->message(to => $self, name => SentMessage, parms => { message => $message, outcome => 1 }) if ($message->{'outcome'});
		    }
		    Trace::debug('Sent message (%d bytes) to %s', $total, _tlm_describeConnection($connection)) if Trace::check(Trace::DEBUG4);
		    $connection->{'lastSend'} = time;
		}
	    }
	}
	last TRYAGAIN;
	
    }
    return $dataWritten ? 1 : 0;
}

# _tlm_buildMessage() - Take a hash ref and return the message components.

sub _tlm_buildMessage {
    my $self = shift;
    my $message = shift;
    my $frozen = shift;
    my $mac = shift;
    my $length = shift;

    if ($_useStorable) {
	$$frozen = Storable::nfreeze($message);
    }
    elsif ($_useFreezeThaw) {
	$$frozen = FreezeThaw::freeze($message);
    }
    else {
	$$frozen = _unravel($message);
    }

    # Create message authentication code

    # Hash digest

    my $md5 = $_digest->new;
    $md5->add($$frozen);
    my $digest = $md5->digest;

    # Encrypt with shared secret key

    $$mac = $self->{'_thing'}->{'crypt'}->encrypt(substr $digest, 0, 8);
    $$length = length $$frozen;
}

# _tlm_sendBadMessage() - Unauthenticated error messages.

sub _tlm_sendBadMessage {
    my $self = shift;
    my $connection = shift;
    my $reason = shift;

    if (!defined($connection->{'write'}->{'queue'})) {
	$connection->{'write'}->{'queue'} = [];
    }
    
    push @{ $connection->{'write'}->{'queue'} }, { name => BadMessage, reason => $reason };
    return 1;
}

# _tlm_writeBytes() - Non blocking io, returns undef on error or the
# number of bytes written.

sub _tlm_writeBytes {
    my $socket = shift;
    my $dataR = shift;
    my $length = shift;
    my $offset = shift;

    my $thisWrite = syswrite($socket, $$dataR, $length - $offset, $offset);
    if (defined($thisWrite) && ($thisWrite < 0)) {
	Trace::debug('Negative from syswrite on: %s', _tlm_describeSocket($socket)) if Trace::check(Trace::DEBUG6);
	return undef;
    }
    elsif (!defined($thisWrite)) {
	Trace::debug('Undefined syswrite on: %s', _tlm_describeSocket($socket)) if Trace::check(Trace::DEBUG6); # OK in non blocking
	return 0;
    }

    Trace::debug('Wrote %d bytes on: %s', $thisWrite, _tlm_describeSocket($socket)) if Trace::check(Trace::DEBUG6);
    return $thisWrite;
}

#
# Message Reading
#

# _tlm_readMessage(connection, readySignal, possiblyDead, messageRef)
#

# Tries to read a message from the connection. Returns 1 and sets the
# message reference on a successful read. Returns 1 and does not set a
# message ref on a partial read. Returns undef on error and closes the
# connection. Returns 0 if there is nothing there. Will send a bad
# message indication if the MAC does not match.

sub _tlm_readMessage {
    my $self = shift;
    my $thing = $self->{'_thing'};
    my $connection = shift;
    my $readySignal = shift;
    my $possiblyDead = shift;
    my $msgR = shift;

    my $socket = $connection->{'socket'};

    my $dataRead = 0;
    my $partial;
    my $header;
    if (defined($connection->{'read'}->{'partial'})) {
	# Try and finish a partial header read
	$partial = $connection->{'read'}->{'partial'};
	$header = $partial->{'header'};
    }
    else {
	my $data = '';
	$header = { dataR => \$data, length => 12, offset => 0 };
    }

    # Try and read 4 byte length and 8 byte authentication code

    if (!defined($header->{'mac'})) {

	my $dataR = $header->{'dataR'};
	my $bytesRead = _tlm_readBytes($socket, $dataR, $header->{'length'}, $header->{'offset'});
	if (!defined($bytesRead)) {
	    $self->_tlm_socketError($connection);
	    return undef;
	}
	if ($bytesRead == 0) {
	    if ($readySignal) {
		# Reported ready but really not, good indication that this thing has died
		Trace::debug('Ready fault for %s', _tlm_describeConnection($connection)) if Trace::check(Trace::DEBUG6);
		if (defined($connection->{'thing'}) && !(grep { $_ == $connection } @{ $possiblyDead })) {
		    push @{ $possiblyDead }, $connection;
		}
		$self->_tlm_socketError($connection);
	    }
	    # Otherwise its a subsequent pass for this socket and there is simply no message.
	    return 0;
	}
	    
	$header->{'offset'} += $bytesRead;
	if ($bytesRead != 12) {
	    # Partial read
	    Trace::debug('Partial read of message header (%d/12 bytes)', $header->{'offset'}); # if Trace::check(Trace::DEBUG5);
	    if (!defined($connection->{'read'}->{'partial'})) {
		$connection->{'read'}->{'partial'} = {};
	    }
	    $connection->{'read'}->{'partial'}->{'header'} = $header;
	    return 1;
	}
	$readySignal = 0;
	$dataRead = 1;
    }

    # Take MAC and length from header

    my $length;
    my $mac;
    if (!defined($header->{'mac'})) {
	my $dataR = $header->{'dataR'};
	$length = substr $$dataR, 0, 4;
	$length = unpack('L', $length);
	$mac = substr $$dataR, 4;
	$header->{'mac'} = $mac;
	$header->{'bodylength'} = $length;
	Trace::debug('Read message header, message is %d bytes', $length) if Trace::check(Trace::DEBUG5);
    }
    else {
	$mac = $header->{'mac'};
	$length = $header->{'bodylength'};
    }

    my $body;
    if (defined($partial->{'body'})) {
	# Finish a partial body read
	$body = $partial->{'body'};
    }
    else {
	my $data = '';
	$body = { dataR => \$data, length => $header->{'bodylength'}, offset => 0 };
    }

    # Read the message body
	    
    my $dataR = $body->{'dataR'};
    my $bytesRead = _tlm_readBytes($socket, $dataR, $body->{'length'}, $body->{'offset'});
    if (!defined($bytesRead)) {
	$self->_tlm_socketError($connection);
	return undef;
    }
    if ($bytesRead == 0) {
	if ($readySignal) {
	    # Reported ready but really not, good indication that this thing has died
	    Trace::debug('Ready fault for %s', _tlm_describeConnection($connection)) if Trace::check(Trace::DEBUG6);
	    if (defined($connection->{'thing'}) && !(grep { $_ == $connection } @{ $possiblyDead })) {
		push @{ $possiblyDead }, $connection;
	    }
	    $self->_tlm_socketError($connection);
	}
	# Otherwise its a subsequent pass for this socket and there is
	# simply no message. Return depending on wether something was
	# read for the header.
	return $dataRead;
    }

    if ($bytesRead + $body->{'offset'} != $body->{'length'}) {
	# Partial read
	$body->{'offset'} += $bytesRead;
	Trace::debug('Partial read of message body (%d/%d bytes)', $body->{'offset'}, $body->{'length'}) if Trace::check(Trace::DEBUG5);
	if (!defined($connection->{'read'}->{'partial'})) {
	    $connection->{'read'}->{'partial'} = {};
	}
	$connection->{'read'}->{'partial'}->{'header'} = $header;
	$connection->{'read'}->{'partial'}->{'body'} = $body;
	return 1;
    }
    delete $connection->{'read'}->{'partial'};

    my $msgblockR = $body->{'dataR'};
    my $msgblock = $$msgblockR;
	    
    #require 'dumpvar.pl';
    #main::dumpValue($msgblock);

    my $message;
    if ($mac ne '__0000__') {

	Trace::timestamp('digest received message') if (Trace::performance && Trace::check(Trace::DEBUG1));

	my $md5 = $_digest->new;
	$md5->add($msgblock);
	my $digest = $md5->digest;
	my $ed = $thing->{'crypt'}->encrypt(substr $digest, 0, 8);

	Trace::debug('Digesting read message took: %f', Trace::timediff('digest received message')) 
	  if (Trace::performance && Trace::check(Trace::DEBUG1));

	if ($mac ne $ed) {
	    Trace::error('Authentication code is incorrect. [%d:%s] ==> [%d:%s]', length $mac, unpack('h32', $mac), length $ed, unpack('h32', $ed));
	    $self->_tlm_sendBadMessage($connection, 'Authentication failure');
	    $self->_tlm_socketError($connection);
	    return undef;
	}

	Trace::timestamp('thaw received message') if (Trace::performance && Trace::check(Trace::DEBUG1));

	if ($_useStorable) {
	    if (!defined(eval { $message = Storable::thaw($msgblock); })) {
		Trace::error($@);
	    }
	}
	elsif ($_useFreezeThaw) {
	    #Trace::dump(Trace::DEBUG1, $msgblock);
	    my @thawed;
	    if (!defined(eval { @thawed = FreezeThaw::thaw($msgblock); })) {
		Trace::error($@);
	    }
	    elsif ($#thawed != 0) {
		Trace::error('Incorrect message structure');
	    }
	    else {
		$message = $thawed[0];
	    }
	    #require 'dumpvar.pl';
	    #main::dumpValue($message);
	}
	else {
	    #Trace::dump($msgblock);
	    if (!defined(eval $msgblock)) {
		Trace::error($@);
		Trace::dump($msgblock);
	    }
	    #require 'dumpvar.pl';
	    #main::dumpValue($message);
	}
	
	Trace::debug('Thawing read message took: %f', Trace::timediff('thaw received message')) 
	  if (Trace::performance && Trace::check(Trace::DEBUG1));

	if (!defined($message)) {
	    $self->_tlm_sendBadMessage($connection, 'Bad message: '.Trace::lastError);
	    $self->_tlm_socketError($connection);
	    return undef;
	}
	
	if (!defined($connection->{'thing'})) {
	    $connection->{'thing'} = $message->{'from'};
	}
		
	Trace::debug('Received message %s from %s', Trace::defStr($message->{'name'}), _tlm_describeConnection($connection)) 
	    if Trace::check(Trace::DEBUG4);
    }
    else {
	# Bad message indication received
	$message = { name => Thing::BadMessage, parms => { reason => unpack('Z*', $msgblock) } };
	Trace::debug('Received bad message: %s', $message->{'parms'}->{'reason'}) if Trace::check(Trace::DEBUG4);
    }
	    
    $connection->{'lastReceive'} = time;
    $$msgR = $message;
    return 1;
}

# _tlm_readBytes() - Non blocking io, returns undef on error or the number of bytes read.

sub _tlm_readBytes {
    my $socket = shift;
    my $dataR = shift;
    my $length = shift;
    my $offset = shift;

    my $thisRead = sysread $socket, $$dataR, $length - $offset, $offset;
    if (defined($thisRead) && ($thisRead < 0)) {
	Trace::debug('Negative from sysread on: %s', _tlm_describeSocket($socket)) if Trace::check(Trace::DEBUG6);
	return undef;
    }
    elsif (!defined($thisRead)) {
	Trace::debug('Undefined sysread on: %s', _tlm_describeSocket($socket)) if Trace::check(Trace::DEBUG6); # OK in non blocking
	return 0;
    }
    Trace::debug('Read %d bytes on: %s', $thisRead, _tlm_describeSocket($socket)) if Trace::check(Trace::DEBUG6);

    return $thisRead;
}

sub _tlm_socketError {
    my $self = shift;
    my $thing = $self->{'_thing'};
    my $errorConnection = shift;

    my $position = 0;
    foreach my $connection (@{ $thing->{'connections'} }) {
	if ($errorConnection->{'id'} == $connection->{'id'}) {

	    Trace::debug('Dropped connection with %s', _tlm_describeConnection($connection)) if Trace::check(Trace::DEBUG5);

	    # Remove from can_read and can_write

	    if ($connection->{'socket'}) {
		$thing->{'read_select'}->remove($connection->{'socket'});
		$thing->{'write_select'}->remove($connection->{'socket'});
		$connection->{'socket'}->close;
	    }

	    # Send failure outcomes for pending writes

	    if (defined($connection->{'write'}->{'queue'})) {
		foreach my $message (@{ $connection->{'write'}->{'queue'} }) {
		    $self->message(to => $self, name => SentMessage, parms => { message => $message, outcome => 0 }) if ($message->{'outcome'});
		    Trace::debug('Message %s never sent', $message->{'name'}) if Trace::check(Trace::DEBUG3);
		}
	    }
	    if (defined($connection->{'write'}->{'partial'})) {
		my $message = $connection->{'write'}->{'partial'}->{'message'};
		$self->message(to => $self, name => SentMessage, parms => { message => $message, outcome => 0 }) if ($message->{'outcome'});
		Trace::debug('Message %s never sent', $message->{'name'}) if Trace::check(Trace::DEBUG3);
	    }

	    splice @{ $thing->{'connections'} }, $position, 1;
	    return;
	}
	$position++;
    }
}

sub _tlm_describeConnection {
    my $connection = shift;
    if (defined($connection)) {
	if (defined($connection->{'thing'})) {
	    return sprintf('%s over %s', _tlm_describeHandle($connection->{'thing'}), 
			   _tlm_describeSocket($connection->{'socket'}, $connection->{'direction'}));
	}
	return _tlm_describeSocket($connection->{'socket'}, $connection->{'direction'});
    }
    return 'undefined';
}

sub _tlm_describeSocket {
    my $socket = shift;
    my $direction = shift;
    if (defined($socket)) {
	return sprintf('[%s:%s %s %s:%s]', 
		       defined($socket->sockhost) ? $socket->sockhost : 'u', defined($socket->sockport) ? $socket->sockport : 'u',
		       !defined($direction) ? '<->' : (($direction eq 'out') ? '-->' : '<--'),
		       defined($socket->peerhost) ? $socket->peerhost : 'u', defined($socket->peerport) ? $socket->peerport : 'u');
    }
    return 'undefined';
}

sub _tlm_describeHandle {
    my $handle = shift;
    if (defined($handle)) {
	return sprintf('%s [%s:%s]', $handle->name, $handle->host, $handle->port);
    }
    return 'undefined';
}

# _tlm_configure() - Handle a configuration message, returns a
# ConfigureResponse to sender. Registered last so that a sub class can
# register a handler to do the configuration and set $parms->{'error'}
# if there is a problem.

sub _tlm_configure {
    my $self = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};
    my $from = $tParms{'from'};
    $self->message(name => ConfigureResponse, to => $from, parms => $parms);
}

# _tlm_shutdown()

sub _tlm_shutdown {
    my $self = shift;
    my %tParms = @_;
    my $from = $tParms{'from'};
    $self->message(name => ShutdownResponse, to => $from, parms => $tParms{'parms'});
    $self->destroy;
    Thing::exit(0);
}

# _tlm_closeall() - Close all open sockets.

sub _tlm_closeall {
    my $self = shift;
    $self->{'_thing'}->{'socket'}->close;
    foreach my $connection (@{ $self->{'_thing'}->{'connections'} }) {
	$connection->{'socket'}->close;
    }
}

sub REAPER {
    #$trace->debug('Dead child');
    while ((my $waitedpid = waitpid(-1, WNOHANG)) > 0) {
	#$trace->debug("Dead child $waitedpid");
    }
    $SIG{CHLD} = \&Thing::REAPER;
}

# _unravel(data) - Make a serialised version of data.

sub _unravel {
    my $data = shift;

    # First pass - find the shared references

    my @refstack = ();
    my @ref = ( $data );
    while (@ref) {
	my $check = pop @ref;
	my $found = 0;
	foreach my $se (@refstack) {
	    if ($se->{ ref } == $check) {
		$found = $se;
		last;
	    }
	}
	if ($found) {
	    $found->{ count }++;
	    next;
	}
	else {
	    push @refstack, { ref => $check, count => 0, label => '$r'.($#refstack + 1) };
	}

	if (UNIVERSAL::isa($check, 'HASH')) { 
	    push @ref, grep { ref($_) } values(%{ $check });
	} 
	elsif (UNIVERSAL::isa($check, 'ARRAY')) { 
	    push @ref, grep { ref($_) } @{ $check };
	}
	else {
	    Trace::error('Unhandled element: %s', ref($data));
	    die;
	}
    }
    my @used = grep { $_->{ count } } @refstack;

    # Pass two - unravel

    return join('', map { 'my '.$_->{label}.';' } @used).' $message='.__unravel($data, 0, \@used).';';
}

# __unravel(data, depth, refstack)

sub __unravel {
    my $data = shift;
    my $depth = shift;
    my $refstack = shift;

    my $segment = '';
    if (ref($data)) {

	my $found = 0;
	foreach my $se (@{ $refstack }) {
	    if ($se->{ ref } == $data) {
		$found = $se;
		last;
	    }
	}
	if (!$found || !defined($found->{ declared })) {

	    if ($found) {
		$found->{ declared } = 1;
	    }

	    # Unravel its value

	    if (UNIVERSAL::isa($data, 'HASH')) { 
		$segment .= "{\n";
		foreach my $key (keys(%{ $data })) {
		    $segment .= _depth($depth).'\''.$key.'\''.' => '.__unravel($data->{ $key }, $depth + 1, $refstack).",\n";
		}
		$segment .= _depth($depth - 1)."}";
		if (UNIVERSAL::isa($data, 'Thing::Handle')) {
		    $segment = 'eval { bless '.$segment.', \'Thing::Handle\'}';
		}
		elsif (UNIVERSAL::isa($data, 'Info::Cd')) {
		    $segment = 'eval { bless '.$segment.', \'Info::Cd\'}';
		}
		elsif (UNIVERSAL::isa($data, 'Info::Artist')) {
		    $segment = 'eval { bless '.$segment.', \'Info::Artist\'}';
		}
		elsif (UNIVERSAL::isa($data, 'Info::Album')) {
		    $segment = 'eval { bless '.$segment.', \'Info::Album\'}';
		}
		elsif (UNIVERSAL::isa($data, 'Info::Track')) {
		    $segment = 'eval { bless '.$segment.', \'Info::Track\'}';
		}
		elsif (UNIVERSAL::isa($data, 'Info::List')) {
		    $segment = 'eval { bless '.$segment.', \'Info::List\'}';
		}
		elsif (UNIVERSAL::isa($data, 'Format::Mp3')) {
		    $segment = 'eval { bless '.$segment.', \'Format::Mp3\'}';
		}
		elsif (UNIVERSAL::isa($data, 'Format::Wav')) {
		    $segment = 'eval { bless '.$segment.', \'Format::Wav\'}';
		}
	    } 
	    elsif (UNIVERSAL::isa($data, 'ARRAY')) { 
		$segment .= "[\n";
		foreach my $value (@{ $data }) {
		    $segment .= _depth($depth).__unravel($value, $depth + 1, $refstack).",\n";
		}
		$segment .= _depth($depth - 1)."]";
	    } 
	    else {
		Trace::error('Unhandled element: %s', ref($data));
		die;
		$segment = _depth($depth).'undef'; 
	    }

	    if ($found) {
		$segment = $found->{ label }.'='.$segment;
	    }

	}
	else {

	    # Use an existing ref
	    $segment .= $found->{ label };
	}
    }	
    else {
	if (defined($data)) {
	    if ($data eq '') {
		$segment .= '\'\'';
	    }
	    elsif ($data =~ /\D/) {
		if ($data =~ /\'/) {
		    my $s = $data;
		    $s =~ s#'#\\'#g; #'
		    $segment .= '\''.$s.'\'';
		}
		else {
		    $segment .= '\''.$data.'\'';
		}
	    }
	    else {
		$segment .= $data;
	    }
	}
	else {
	    $segment .= 'undef';
	}
    }
    return $segment;
}

sub _depth { 
    my $depth = shift;
    my $depthStr = '';
    for (my $i = 0; $i < $depth; $i++) {
	$depthStr .= '    ';
    }
    return $depthStr;
}

1;
