# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.4 $

package Thing::Handle;

use Trace;
use strict;

# Monitored States

sub Unknown    { 'Thing__Handle__Unknown'; }
sub Alive      { 'Thing__Handle__Alive'; }
sub Dead       { 'Thing__Handle__Dead'; }

# new() - A handle to another thing, possibly on a remote system. Used
# to cleanly identify things in messages. Explicitly pollutable HASH
# ref namespace, except by using keys with an initial _.

#   `name' - The name to use for this thing. Required.

#   `host' - The name of the host upon which this remote thing
#   resides. Required.

#   `port' - The TCP port the remote thing is listening too.

sub new {
    my $proto = shift;
    my $type = ref($proto) || $proto;
    my %parms = @_;

    my $self = {};
    $self->{'_handle'} = {};
    bless $self, $type;

    if (!defined($parms{'name'})) {
	Trace::mandatory('name');
	return undef;
    }
    $self->{'_handle'}->{'name'} = $parms{'name'};
    if (!defined($parms{'host'})) {
	Trace::mandatory('host');
	return undef;
    }
    $self->{'_handle'}->{'host'} = $parms{'host'};
    if (defined($parms{'port'})) {
	$self->{'_handle'}->{'port'} = $parms{'port'};
    }
    else {
	$self->{'_handle'}->{'port'} = -1; # Dirty
    }
    $self->{'_handle'}->{'previousState'} = Unknown;
    $self->{'_handle'}->{'currentState'} = Unknown;

    return $self;
}

sub name { 
    if ($#_ == 1) {
	$_[0]->{'_handle'}->{'name'} = $_[1];
    }
    return $_[0]->{'_handle'}->{'name'};
}
    
sub host { 
    if ($#_ == 1) {
	$_[0]->{'_handle'}->{'host'} = $_[1];
    }
    return $_[0]->{'_handle'}->{'host'};
}
    
sub port { 
    if ($#_ == 1) {
	$_[0]->{'_handle'}->{'port'} = $_[1];
    }
    return $_[0]->{'_handle'}->{'port'};
}    

# handle - Return a vanilla reference to the same handle and without any pollutants.

sub handle {
    my $handle = shift;
    my $pfree = {};
    $pfree->{'_handle'} = $handle->{'_handle'};
    bless $pfree, 'Thing::Handle';
    return $pfree;
}

# key - Handy name@host creator

sub key {
    return $_[0]->{'_handle'}->{'name'}.'@'.$_[0]->{'_handle'}->{'host'};
}

# unique - Unique string to identify alive things

sub unique {
    return $_[0]->{'_handle'}->{'name'}.'@'.$_[0]->{'_handle'}->{'host'}.':'.$_[0]->{'_handle'}->{'port'};
}

# currentState(), previousState() - used by Thing::monitor to store
# the state of what is being monitored.

sub currentState { 
    if ($#_ == 1) {
	$_[0]->{'_handle'}->{'currentState'} = $_[1];
    }
    return $_[0]->{'_handle'}->{'currentState'};
}    

sub previousState { 
    if ($#_ == 1) {
	$_[0]->{'_handle'}->{'previousState'} = $_[1];
    }
    return $_[0]->{'_handle'}->{'previousState'};
}    

sub stateChange {
    if ($#_ == 1) {
	$_[0]->{'_handle'}->{'previousState'} = $_[0]->{'_handle'}->{'currentState'};
	$_[0]->{'_handle'}->{'currentState'} = $_[1];
    }
    return ($_[0]->{'_handle'}->{'previousState'} ne $_[0]->{'_handle'}->{'currentState'});
}

# cmpNameHost() 

sub cmpNameHost {
    if ($#_ != 1) {
	Trace::error('Too many (%d) or too few paramaeters', $#_);
	return undef;
    }

    my $a = shift;
    my $b = shift;

    return 1 if (!defined($a) && !defined($b));
    return 0 if !defined($a);
    return 0 if !defined($b);

    return 0 if (defined($a->host) xor defined($b->host));
    return 0 if (defined($a->name) xor defined($b->name));

    return (($a->host eq $b->host) && ($a->name eq $b->name));
}

sub equal {
    if ($#_ != 1) {
	Trace::error('Too many (%d) or too few paramaeters', $#_);
	return undef;
    }

    my $a = shift;
    my $b = shift;

    return 1 if (!defined($a) && !defined($b));
    return 0 if !defined($a);
    return 0 if !defined($b);

    return 0 if (defined($a->host) xor defined($b->host));
    return 0 if (defined($a->name) xor defined($b->name));

    return (($a->host eq $b->host) && ($a->name eq $b->name) && ($a->port == $b->port));
}

1;
