# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.9 $

package Thing::Info;

# Information objects, used to synchronize data among things.

use Trace;
use Utility;
use strict;

# External messages that can be registered for

# 'ioid' has recently been modified, other parameters are
# implementation dependant and may be used to qualify the
# change. Register 'last'.
sub Update      { 'Thing__Info__Update'; } 

# Internal messages for use with Thing.pm.

sub Sync        { 'Thing__Info__Sync'; }
sub Monitor     { 'Thing__Info__Monitor'; }
sub Demonitor   { 'Thing__Info__Demonitor'; }

# new() - Create a new info object.
#
#   owner - The thing that owns this object.
#
#   ioid - Optionally specify an id for this object, allows for systems
#   to define their own id scheme.
#
#   data - The data for this object. Optional. See get() and set()
#   below.

sub new {
    my $proto;
    my $type;
    if ($#_ == -1) {
	$type = 'Thing::Info';
    }
    else {
	$proto = shift;
	$type = ref($proto) || $proto;
    }
    my %parms = @_;

    my $owner = $parms{'owner'};
    if (!defined($owner)) {
	$owner = Thing->context->handle;
    }

    my $self;
    my $ioid = $parms{'ioid'};
    if (!defined($ioid)) {
	$ioid = Thing->context->_tm_infoAllocIoid;
    }

    $self = {};
    bless $self, $type;

    Trace::debug('Instantiated info object of type %s', $type) if Trace::check(Trace::DEBUG6);

    my $info = $self->{'_info'} = {};
    $info->{'owner'} = Thing->context->monitor($owner->handle);
    $info->{'ioid'} = $ioid;
    $info->{'data'} = { current => undef, new => undef, };

    if (defined($parms{'data'})) {
	$info->{'data'}->{'current'} = $parms{'data'};
    }
    return $self;
}

sub ioid { $_[0]->{'_info'}->{'ioid'}; }
sub owner { $_[0]->{'_info'}->{'owner'}; }

# get() - Get the current data. If this is a reference what it refers
# to can be modified if it is known to be only in local use. Otherwise
# setN, getN and sync should be used.

sub get { $_[0]->{'_info'}->{'data'}->{'current'}; }

# getN() - Get the new data. If the new data is a reference this can
# be used to modify the data prior to sync.

sub getN { $_[0]->{'_info'}->{'data'}->{'new'}; }

# setN() - Set the new data. With no parameters and a hash reference
# current data, this will prep the new data to an empty hash
# reference. This can then be used by getN to modify the new
# data. Otherwise the new data will be assigned from the first
# parameter.

sub setN {
    my $self = shift;
    if (($#_ == -1) && UNIVERSAL::isa($self->{'_info'}->{'data'}->{'current'}, 'HASH')) {
	$self->{'_info'}->{'data'}->{'new'} = {};
    }
    else {
	$self->{'_info'}->{'data'}->{'new'} = shift;
    }
}

# sync() - Synchronize any changes to new data to all monitors of this
# object. Use 'force' to send an update when no changes have been
# made. Other parameters are passed through. This can be augmented by
# sub classes to check the validity of the changes perhaps. In
# augmentation return 0 or undef to signify a problem.

sub sync {
    my $self = shift;
    my $info = $self->{'_info'};

    Trace::debug('Generating sync for: %s', $info->{'ioid'}) if Trace::check(Trace::DEBUG5);

    return Thing->context->message(to => $info->{'owner'}, name => Thing::Info::Sync, 
				   parms => { ioid => $info->{'ioid'}, new => $info->{'data'}->{'new'}, @_ });
}

# _im_update() - Called by the monitoring thing upon receipt of an
# update request. Parms are 'new' indicating the change. This is
# private and should not be augmented. Changes to monitored objects
# can be explicitly noticed by registering a handler for Update
# messages. (not however 'first' or the change won't have been made).

sub _im_update {
    my $self = shift;
    my $info = $self->{'_info'};
    my $parms = shift;

    Trace::debug('Object update: %s at 0x%08x', $self->ioid, $self) if Trace::check(Trace::DEBUG6);

    my $new = $parms->{'new'};
    if (!defined($new)) {
	return 1;
    }
    my $current = $self->{'_info'}->{'data'}->{'current'};

    my $diff = Utility::Hash::augment(existing => $current, new => $new,
				      special => [ { compareFn => sub { Thing::Info::compare(@_); } } ]);

    if (!defined($diff) && defined($new)) {
	# Something other than a hash or previously undefined
	$self->{'_info'}->{'data'}->{'current'} = $new;
    }
    return 1;
}

# monitor() - Monitor changes to this info object. If this object is
# already monitored, a reference to that object is returned. The data
# of the passed object takes precedence and the pollutants of an
# existing object take precendence.

# usage is $object = $object->monitor to maintain a single shared
# reference.

sub monitor {
    my $self = shift;
    my $object = Thing->context->_tm_infoRegisterMonitor($self);
    if ($object != $self) {
	# Already monitored
	if (!UNIVERSAL::isa($object, ref($self))) {
	    # Monitored under a different class.
	    # ASSUME its monitored as a super class and that this sub
	    # class only extends the super. There can only be one path
	    # of extension. 
	    Trace::debug('Object of type %s is reblessed to its subclass %s: %s at 0x%08x', ref($object), ref($self), $object->ioid, $object) if (Trace::check(Trace::DEBUG3));
	    bless $object, ref($self);
	}
	# Passed objects data takes precedence
	# $object->{'_info'}->{'data'} = $self->{'_info'}->{'data'};

	# Save pollution from passed object but maintain pollution
	# where there are no clashes.
	foreach my $key (keys(%{ $self })) {
	    next if ($key eq '_info');
	    $object->{ $key } = $self->{ $key };
	}
    }
    Trace::debug('Monitoring object: %s at 0x%08x', $object->ioid, $object) if Trace::check(Trace::DEBUG6);
    return $object;
}

# demonitor() - Stop monitoring changes to this info object.

sub demonitor {
    my $self = shift;
    return Thing->context->_tm_infoDeregisterMonitor($self);
}

# ismonitor() - Are we monitoring this object?

sub ismonitor {
    my $self = shift;
    return Thing->context->infoFind(ioid => $self->ioid);
}

# saveCheck - Determin if an objects field needs saving. For use in a
# persistent environment where the saved part of the object is a
# single level hash.

sub saveCheck {
    my $object = shift;
    my $field = shift;

    return 0 if (!defined($object->getN));
    return 0 if (!defined($object->getN->{ $field }));
    return 1 if (!defined($object->get->{ $field }));
    return 1 if ($object->get->{ $field } ne $object->getN->{ $field });
    return 0;
}

# compare - Used by augmentation.

sub compare {
    my $a = shift; my $b = shift;
    return if (!UNIVERSAL::isa($a, 'Thing::Info'));
    return if (!UNIVERSAL::isa($b, 'Thing::Info'));
    return ($a->ioid eq $b->ioid) ? 0 : 1; # Nop if same object, if different assign
}

1;
