# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.13 $

package Thing::Service;

# Service for things. Including thing start, lookup and configuration
# save and restore. Works with Thing::Service on other hosts to
# provide a dependable response to requests despite possible host
# outages. 

# Security: 
# Service can operate in two modes: trusted and not. Trusted allows
# thing invocation which is fundamentally insecure. Trusted also
# allows writing of configuration files, again fundamentally
# insecure.

# Service should be used via messages passed to the
# Thing::Service::Handle to allow a single local service destination
# for requests (dependable response), rather than a remote one.

# Internal Messages

sub MonitorListRequest  { 'Thing__Service__MonitorListRequest'; }

use Thing;
@ISA = qw( Thing );
use Thing::ServiceH;
use Trace;
use strict;

# new()

sub new {
    my $proto = shift;
    my $type = ref($proto) || $proto;
    my $self = $type->SUPER::new(@_, port => PORT);
    if ($self->error) {
	return $self;
    }

    $self->{'registry'}->{ $self->name } = $self->handle;

    $self->{'monitor'} = {};
    $self->handle->stateChange(Thing::Handle::Alive);
    $self->{'monitor'}->{ $self->host } = { $self->name => { handle => $self->handle, monitors => {} } };

    $self->{'hosts'} = {};
    $self->{'config'} = {};
    $self->{'pendingStarts'} = [];
    $self->{'pendingRemote'} = [];
    $self->{'id'} = 0;
    $self->{'timeoutRunning'} = 0;
    $self->{'localTimeout'} = 10;
    $self->{'remoteTimeout'} = 20;

    # Alive Poll with peers
    $self->{'aliveTimeout'} = 90;
    $self->{'aliveInterval'} = 60;
    $self->register(name => '_aliveCheck', handler => '_aliveCheck');
    $self->register(name => '__Alive__', handler => '_alive');
    $self->at(name => '_aliveCheck', to => $self, time => time + $self->{'aliveInterval'});
    
    if (!defined(Thing::Service::configPath)) {
	Trace::error(Thing::Service::configPathError);
	$self->error(Trace::lastError);
	return $self;
    }
    my $cfg = Thing::Service::readConfig;
    if (!$cfg) {
	$self->error(Trace::lastError);
	return $self;
    }
    $self->{'config'} = $cfg->{'config'};
    $self->{'trusted'} = $cfg->{'trusted'};

    $self->register(name => RegisterRequest, handler => 'registerRequest');
    $self->register(name => DeregisterRequest, handler => 'deregisterRequest');
    $self->register(name => StartRequest, handler => 'startRequest');
    $self->register(name => Thing::Started, handler => '_forwardResponse');
    $self->register(name => StartResponse, handler => '_forwardResponse');
    $self->register(name => LookupRequest, handler => 'lookupRequest');
    $self->register(name => LookupResponse, handler => '_forwardResponse');
    $self->register(name => ConfigLoadRequest, handler => 'configLoadRequest');
    $self->register(name => ConfigLoadResponse, handler => '_forwardResponse');
    $self->register(name => ConfigSaveRequest, handler => 'configSaveRequest');
    $self->register(name => ConfigSaveResponse, handler => '_forwardResponse');
    $self->register(name => Thing::ShutdownRequest, handler => 'shutdownRequest');
    $self->register(name => Thing::ConfigureRequest, handler => '_configure');
    $self->register(name => MonitorRequest, handler => 'monitorRequest');
    $self->register(name => MonitorResponse, handler => 'monitorResponse');
    $self->register(name => DemonitorRequest, handler => 'demonitorRequest');
    $self->register(name => MonitorListRequest, handler => 'monitorListRequest');
    $self->register(name => '_timeout', handler => '_timeout');
    return $self;
}

# _configure() - Like all things the service has to be explicitly configured.

sub _configure {
    my $self = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};

    foreach my $host (@{ $parms->{'hosts'} }) {
	next if ($host eq $self->host);
	if (!defined($self->{'hosts'}->{ $host })) {
	    my $service = Thing::Service::Handle; 
	    $service->host($host);
	    $self->{'hosts'}->{ $host } = { handle => $service };
	    # New host, in case of reboot check for monitors
	    if ($self->message(name => MonitorListRequest, to => $service)) {
		$service->stateChange(Thing::Handle::Alive);
	    }
	    else {
		$service->stateChange(Thing::Handle::Dead);
	    }
	    $self->serviceChange($service);
	}
    }
}

sub serviceChange {
    my $self = shift;
    my $handle = shift;

    Trace::debug('Possible state change for service: %s: %s -> %s', $handle->unique, 
		 $handle->previousState, $handle->currentState) if Trace::check(Trace::DEBUG3);

    return if ($handle->name ne Thing::Service::NAME);

    # If a peer service has gone dead we mark all on it as dead and
    # inform its monitors.

    if (($handle->previousState ne Thing::Handle::Dead) && ($handle->currentState eq Thing::Handle::Dead)) {

	return if !defined($self->{'monitor'}->{ $handle->host }); # None configured for this host
	my $host = $self->{'monitor'}->{ $handle->host };

	foreach my $check (values(%{ $host })) {
	    my $handle = $check->{'handle'};	
	    $handle->stateChange(Thing::Handle::Dead);
	    $handle->port(-1);
	    foreach my $monitor (values(%{ $check->{'monitors'} })) {
		if (!$self->message(name => Thing::Service::MonitorResponse, to => $monitor, parms => { handle => $handle })) {
		    delete $check->{'monitors'}->{ $monitor->unique };
		}
	    }
	}
    }

    # If a peer service has gone alive we ask it to tell us the state
    # of things we are monitoring on it.

    elsif (($handle->previousState ne Thing::Handle::Alive) && ($handle->currentState eq Thing::Handle::Alive)) {

	return if !defined($self->{'monitor'}->{ $handle->host }); # Not monitoring anything on this host
	my $host = $self->{'monitor'}->{ $handle->host };
	foreach my $check (values(%{ $host })) {
	    $self->message(name => Thing::Service::MonitorRequest, to => $handle, parms => { for => $self->handle, handle => $check->{'handle'} });
	}
    }
}

sub registerRequest {
    my $self = shift;
    my %tParms = @_;
    my $from = $tParms{'from'};

    Trace::debug('Register from: %s', $from->unique) if Trace::check(Trace::DEBUG2);

    $self->{'registry'}->{ $from->name } = $from;
    $self->message(name => RegisterResponse, to => $from);

    # Check for existing monitors

    return if ($from->host ne $self->host);
    return if !defined($self->{'monitor'}->{ $from->host });
    my $host = $self->{'monitor'}->{ $from->host };
    return if !defined($host->{ $from->name });
    return if ($host->{ $from->name }->{'handle'}->currentState eq Thing::Handle::Alive);

    $host->{ $from->name }->{'handle'}->stateChange(Thing::Handle::Alive);
    $host->{ $from->name }->{'handle'}->port($from->port);

    foreach my $monitor (values(%{ $host->{ $from->name }->{'monitors'} })) {
	if (!$self->message(name => Thing::Service::MonitorResponse, to => $monitor, parms => { handle => $host->{ $from->name }->{'handle'} })) {
	    delete $host->{ $from->name }->{'monitors'}->{ $monitor->unique };
	}
    }
}

sub deregisterRequest {
    my $self = shift;
    my %tParms = @_;
    my $from = $tParms{'from'};

    Trace::debug('Deregister from: %s', $from->unique) if Trace::check(Trace::DEBUG2);

    delete $self->{'registry'}->{ $from->name } if (defined($self->{'registry'}->{ $from->name }));
}

# startRequest() - Start a thing. Parms are as for the Thing::daemon()
# call with the addition of:

#   `host' - Either the local or a remote host.

sub startRequest {
    my $self = shift;
    my %tParms = @_;
    my $from = $tParms{'from'};
    my $parms = $tParms{'parms'};

    Trace::debug('Start request') if Trace::check(Trace::DEBUG2);

    if (!defined($parms->{'host'}) || ($parms->{'host'} eq $self->host)) {

	# Local

	if (!$self->{'trusted'}) {
	    Trace::error('Will not start when in untrusted mode');
	    $parms->{'error'} = Trace::lastError;
	    $self->message(name => StartResponse, to => $from, parms => $parms);
	}
	else {

	    $self->{'id'}++;
	    if (!$self->{'timeoutRunning'}) {
		$self->{'timeoutRunning'} = 1;
		$self->at(name => '_timeout', to => $self, time => time + ($self->{'localTimeout'} / 2));
	    }
	    
	    $parms->{'starter'} = $self;
	    $parms->{'close'} = $self;
	    $parms->{'_service'} = {} if (!defined($parms->{'_service'}));
	    $parms->{'_service'}->{'localId'} = $self->{'id'};
	    Thing::daemon(%{ $parms });
	    delete $parms->{'close'};
	    delete $parms->{'starter'};
	    push @{ $self->{'pendingStarts'} }, { id => $self->{'id'}, parms => $parms, from => $from, time => time, };
	}
    }
    else {

	# Remote

	$self->_forwardRequest(type => StartRequest, response => StartResponse, from => $from, host => $parms->{'host'}, parms => $parms);
    }
}

# lookupRequest() - Take a thing `name' and optional `host' and
# respond via a LookupResponse message with a thing `handle'.

sub lookupRequest {
    my $self = shift;
    my %tParms = @_;
    my $from = $tParms{'from'};
    my $parms = $tParms{'parms'};

    if (!defined($parms->{'host'}) || ($parms->{'host'} eq $self->host)) {

	# Local lookup

	my $entry = $self->{'registry'}->{ $parms->{'name'} };
	if (!defined($entry)) {
	    Trace::debug('%s is not registered', $parms->{'name'}) if Trace::check(Trace::DEBUG3);
	    $parms->{'error'} = ThingNotRegistered;
	}
	else {
	    $parms->{'handle'} = $entry;
	}
	$self->message(name => LookupResponse, to => $from, parms => $parms);
    }
    else {

	# Remote lookup

	$self->_forwardRequest(type => LookupRequest,
			       response => LookupResponse,
			       from => $from,
			       host => $parms->{'host'},
			       parms => $parms);
    }
}

# configLoadRequest() - Take a thing `handle' and respond via a
# ConfigLoadResponse message with the saved `config' for this
# thing. Take a `host' and respond with all the `configs' for this
# host. Take a `type' and respond with all the configs for things of
# that type.

sub configLoadRequest {
    my $self = shift;
    my %tParms = @_;
    my $from = $tParms{'from'};
    my $parms = $tParms{'parms'};

    my $host;
    if (!defined($parms->{'handle'})) {
	if (!defined($parms->{'host'})) {
	    $host = $self->host;
	}
	else {
	    $host = $parms->{'host'};
	}
    }
    else {
	$host = $parms->{'handle'}->host;
    }

    if ($host eq $self->host) {

	# Local
	
	my $name;
	if (!defined($parms->{'handle'})) {
	    if (defined($parms->{'name'})) {
		$name = $parms->{'name'};
	    }
	}
	else {
	    $name = $parms->{'handle'}->name;
	}

	if (defined($name)) {
	    $parms->{'config'} = $self->{'config'}->{ $name };
	}
	else {
	    $parms->{'configs'} = {};
	    if (defined($parms->{'type'})) {
		foreach my $config (values(%{ $self->{'config'} })) {
		    if ($parms->{'type'} eq $config->{'type'}) {
			$parms->{'configs'}->{ $config->{'name'} } = $config;
		    }
		}
	    }
	    else {
		$parms->{'configs'} = $self->{'config'};
	    }
	}
	$self->message(name => ConfigLoadResponse, to => $from, parms => $parms);
    }
    else {

	# Remote

	$self->_forwardRequest(type => ConfigLoadRequest,
			       response => ConfigLoadResponse,
			       from => $from,
			       host => $host,
			       parms => $parms);
    }
}

# configSaveRequest() - Take a thing `handle' and its `config', save
# it to a non volatile spot and respond via a ConfigSaveResponse
# message.

sub configSaveRequest {
    my $self = shift;
    my %tParms = @_;
    my $from = $tParms{'from'};
    my $parms = $tParms{'parms'};

    if (!defined($parms->{'handle'})) {
	Trace::mandatory('handle');
	$parms->{'error'} = Trace::lastError;
	$self->message(name => ConfigSaveResponse, to => $from, parms => $parms);
	return;
    }

    if ($parms->{'handle'}->host eq $self->host) {

	# Local

	Trace::debug('Save request for %s', $parms->{'handle'}->unique) if Trace::check(Trace::DEBUG3);

	if (!$self->{'trusted'}) {
	    Trace::error('Will not save configuration when in untrusted mode');
	    $parms->{'error'} = Trace::lastError;
	}
	else {
	    $self->{'config'}->{ $parms->{'handle'}->name } = $parms->{'config'};
	    if (!Thing::Service::writeConfig({ config => $self->{'config'}, trusted => $self->{'trusted'}, version => Thing::VERSION() })) {
		$parms->{'error'} = WriteFailed;
	    }
	}
	$self->message(name => ConfigSaveResponse, to => $from, parms => $parms);
    }
    else {

	# Remote

	$self->_forwardRequest(type => ConfigSaveRequest, response => ConfigSaveResponse, from => $from, host => $parms->{'handle'}->host, parms => $parms);
    }
}

# shutdownRequest 

sub shutdownRequest {
    my $self = shift;
    my %tParms = @_;
    my $parms = $tParms{'parms'};

    if (defined($parms->{'qualifier'}) && ($parms->{'qualifier'} eq 'all')) {
	foreach my $thing (values(%{ $self->{'registry'} })) {
	    $self->message(name => Thing::ShutdownRequest, to => $thing);	
	}
    }
}

# monitorRequest() 

sub monitorRequest {
    my $self = shift;
    my %tParms = @_;
    my $from = $tParms{'from'};
    my $parms = $tParms{'parms'};

    my $for = $parms->{'for'}; # Who wants to monitor
    my $handle = $parms->{'handle'}; # What it wants to monitor

    Trace::debug('Monitor request from: %s, for: %s', $for->unique, $handle->unique) if Trace::check(Trace::DEBUG2);

    my $host = $self->{'monitor'}->{ $handle->host };
    if (!defined($self->{'monitor'}->{ $handle->host })) {
	$host = $self->{'monitor'}->{ $handle->host } = {};
    }
    if (!defined($host->{ $handle->name })) {
	$host->{ $handle->name } = { handle => $handle, monitors => {} };
    }
    $host->{ $handle->name }->{'monitors'}->{ $for->unique } = $for;

    $handle = $host->{ $handle->name }->{'handle'};

    if ($handle->host eq $self->host) {

	# Local

	if ($handle->currentState eq Thing::Handle::Unknown) {
	    if (defined($self->{'registry'}->{ $handle->name })) {
		$handle->stateChange(Thing::Handle::Alive);
		$handle->port($self->{'registry'}->{ $handle->name }->port);
	    }
	    else {
		$handle->stateChange(Thing::Handle::Dead);
		$handle->port(-1);
	    }
	}
	$self->message(name => Thing::Service::MonitorResponse, to => $from, parms => { handle => $handle });
    }
    else {
	
	# Remote

	if ($handle->currentState eq Thing::Handle::Unknown) {
	    my $service = Thing::Service::Handle; 
	    $service->host($handle->host);
	    if (!$self->message(name => Thing::Service::MonitorRequest, to => $service, parms => { for => $self->handle, handle => $handle })) {
		# Remote service is dead
		$handle->stateChange(Thing::Handle::Dead);
		$handle->port(-1);
		$self->message(name => Thing::Service::MonitorResponse, to => $from, parms => { handle => $handle });
	    }
	}
	else {
	    $self->message(name => Thing::Service::MonitorResponse, to => $from, parms => { handle => $handle });
	}
    }	
}

# monitor

sub monitor { die 'service cannot monitor'; }

# monitorResponse() - From other service, send to registered monitors.

sub monitorResponse {
    my $self = shift;
    my %tParms = @_;
    my $from = $tParms{'from'};
    my $parms = $tParms{'parms'};

    my $handle = $parms->{'handle'}; # Whats being monitored
    return if !defined($self->{'monitor'}->{ $handle->host }); # None configured for this host
    my $host = $self->{'monitor'}->{ $handle->host };
    return if !defined($host->{ $handle->name }); # None configured for this thing
    $host->{ $handle->name }->{'handle'} = $handle;

    foreach my $monitor (values(%{ $host->{ $handle->name }->{'monitors'} })) {
	if (!$self->message(name => Thing::Service::MonitorResponse, to => $monitor, parms => { handle => $handle })) {
	    delete $host->{ $handle->name }->{'monitors'}->{ $monitor->unique };
	}
    }
}

# demonitorRequest()

sub demonitorRequest {
    my $self = shift;
    my %tParms = @_;
    my $from = $tParms{'from'};
    my $parms = $tParms{'parms'};

    my $for = $parms->{'for'}; # Who wants to demonitor
    my $handle = $parms->{'handle'}; # What it wants to demonitor

    Trace::debug('Demonitor request from: %s, for: %s', $for->unique, $handle->unique) if Trace::check(Trace::DEBUG2);

    delete $self->{'monitor'}->{ $handle->host }->{ $handle->name }->{'monitors'}->{ $for->unique };
}

# monitorListRequest() - Request from other service for a refresh of
# what we are monitoring on it.

sub monitorListRequest {
    my $self = shift;
    my %tParms = @_;
    my $from = $tParms{'from'};

    if (!defined($self->{'hosts'}->{$from->host})) {
	Trace::debug('List request from non peered service: %s', $from->unique) if Trace::check(Trace::DEBUG3);
	return;
    }
    $self->{'hosts'}->{ $from->host }->{'lastHeard'} = time;
    my $service = $self->{'hosts'}->{ $from->host }->{'handle'};
    $service->stateChange(Thing::Handle::Alive);
    $self->serviceChange($service);

    return if !defined($self->{'monitor'}->{ $from->host }); # Not monitoring anything on this host

    my $host = $self->{'monitor'}->{ $from->host };
    foreach my $check (values(%{ $host })) {
	$self->message(name => Thing::Service::MonitorRequest, to => $from, parms => { for => $self->handle, handle => $check->{'handle'} });
    }
}

# _tm_possiblyDead()

sub _tm_possiblyDead {
    my $self = shift;
    my $check = shift;

    if (($check->host ne $self->host) && ($check->name ne $self->name)) {
	# Not on this host or not a remote service
	Trace::debug('Dead handle not our resposibility: %s', $check->unique) if Trace::check(Trace::DEBUG3);
	return;
    }

    # Is if definately dead?
    
    return if ($self->message(name => '__AliveTrip__', to => $check));

    if ($check->name ne $self->name) {

	# Local thing non service

	return if !defined($self->{'monitor'}->{ $check->host }); # None configured for this host

	my $host = $self->{'monitor'}->{ $check->host };
	return if !defined($host->{ $check->name }); # No one monitoring this thing

	$check = $host->{ $check->name }->{'handle'};
    
	$check->stateChange(Thing::Handle::Dead);
	$check->port(-1);
    
	foreach my $monitor (values(%{ $host->{ $check->name }->{'monitors'} })) {
	    if (!$self->message(name => Thing::Service::MonitorResponse, to => $monitor, parms => { handle => $check })) {
		delete $host->{ $check->name }->{'monitors'}->{ $monitor->unique };
	    }
	}
    }
    else {
	# Remote service

	if (!defined($self->{'hosts'}->{ $check->host })) {
	    Trace::debug('Death from non peered service: %s', $check->unique) if Trace::check(Trace::DEBUG3);
	    return;
	}
	my $service = $self->{'hosts'}->{ $check->host }->{'handle'};
	$service->stateChange(Thing::Handle::Dead);
	$self->serviceChange($service);
    }
}

# _alive() - An alive poll has come from a peer service

sub _alive {
    my $self = shift;
    my %tParms = @_;
    my $from = $tParms{'from'};

    if (!defined($self->{'hosts'}->{ $from->host })) {
	Trace::debug('Alive from non peered service: %s', $from->unique) if Trace::check(Trace::DEBUG3);
	return;
    }
    my $otherService = $self->{'hosts'}->{ $from->host };
    $otherService->{'lastHeard'} = time;
    my $service = $otherService->{'handle'};
    $service->stateChange(Thing::Handle::Alive);
    $self->serviceChange($service);
}

# _aliveCheck() - Ensure our peered services are alive

sub _aliveCheck {
    my $self = shift;

    foreach my $host (values(%{ $self->{'hosts'} })) {
	my $dead = 0;
	if (!$self->message(name => '__Alive__', to => $host->{'handle'})) {
	    Trace::debug('Could not send alive to peer service on: %s', $host->{'handle'}->host) if Trace::check(Trace::DEBUG3);
	    $dead = 1;
	}
	if (defined($host->{ 'lastHeard' }) && (($host->{ 'lastHeard' } + $self->{'aliveTimeout'}) < time)) {
	    Trace::debug('Service has timed out on: %s', $host->{'handle'}->host) if Trace::check(Trace::DEBUG3);
	    $dead = 1;
	}

	if ($dead) {
	    my $service = $host->{'handle'};
	    $service->stateChange(Thing::Handle::Dead);
	    $self->serviceChange($service);
	}
    }

    $self->at(name => '_aliveCheck', to => $self, time => time + $self->{'aliveInterval'});
}

# _forwardRequest() - Forward a request to another service entity on a
# remote host. Set a timer or send a response as appropriate.

sub _forwardRequest {
    my $self = shift;
    my %parms = @_;

    $self->{'id'}++;
    Trace::debug('Remote request %d', $self->{'id'}) if Trace::check(Trace::DEBUG2);
    my $remote = Thing::Handle->new(name => NAME, host => $parms{'host'}, port => PORT);

    $parms{'parms'}->{'_service'} = {} if (!defined($parms{'parms'}->{'_service'}));
    $parms{'parms'}->{'_service'}->{'remoteId'} = $self->{'id'};

    if (!$self->{'timeoutRunning'}) {
	$self->{'timeoutRunning'} = 1;
	$self->at(name => '_timeout', to => $self, time => time + ($self->{'localTimeout'} / 2));
    }

    # Forward request
    
    if ($self->message(name => $parms{'type'}, to => $remote, parms => $parms{'parms'})) {
	push @{ $self->{'pendingRemote'} }, { id => $self->{'id'},
					      tries => 1,
					      type => $parms{'type'},
					      response => $parms{'response'},
					      host => $parms{'host'},
					      parms => $parms{'parms'},
					      remote => $remote,
					      from => $parms{'from'},
					      time => time, };
    }
    else {

	# Service is probably dead at the other end.

	$parms{'parms'}->{'error'} = HostUnreachable;
	$self->message(name => $parms{'response'}, to => $parms{'from'}, parms => $parms{'parms'});
    }
}

# _forwardResponse() - Response received, forward to correct point.

sub _forwardResponse {
    my $self = shift;
    my %tParms = @_;
    my $from = $tParms{'from'};
    my $parms = $tParms{'parms'};

    if (!defined($parms->{'_service'})) {
	Trace::error('Not a request that originated from me');
	return;
    }

    # Check local startups

    my @list = @{ $self->{'pendingStarts'} };
    my $pos = 0;
    foreach my $startup (@list) {

	if ($parms->{'_service'}->{'localId'} == $startup->{'id'}) {
	    # Response for this 
	    $self->message(name => StartResponse, to => $startup->{'from'}, parms => $parms);
	    splice @{ $self->{'pendingStarts'} }, $pos, 1;
	    return;
	}
	$pos++;
    }

    # Check remote requests

    @list = @{ $self->{'pendingRemote'} };
    $pos = 0;
    foreach my $pending (@list) {

	if ($parms->{'_service'}->{'remoteId'} == $pending->{'id'}) {
	    # Response for this 
	    $self->message(name => $pending->{'response'}, to => $pending->{'from'}, parms => $parms);
	    splice @{ $self->{'pendingRemote'} }, $pos, 1;
	    return;
	}
	$pos++;
    }

    Trace::error('Not a request that originated from me');
}

# _timeout() - Check for pending starts and remote requests that have
# timed out.

sub _timeout {
    my $self = shift;

    my @list = @{ $self->{'pendingStarts'} };
    my $pos = 0;
    my $time = time;
    foreach my $startup (@list) {
	if ($startup->{'time'} + $self->{'localTimeout'} < time) {
	    # No response after timeout
	    Trace::error('Start request %d timed out', $startup->{'id'});
	    $startup->{'parms'}->{'error'} = TimedOut;
	    $self->message(name => StartResponse, to => $startup->{'from'}, parms => $startup->{'parms'});
	    splice @{ $self->{'pendingStarts'} }, $pos, 1;
	}
	$pos++;
    }

    @list = @{ $self->{'pendingRemote'} };
    $pos = 0;
    $time = time;
    foreach my $pending (@list) {
	if ($pending->{'time'} + $self->{'remoteTimeout'} < time) {
	    # No response after timeout
	    Trace::error('Request %d timed out', $pending->{'id'});
	    $pending->{'parms'}->{'error'} = TimedOut;
	    $self->message(name => $pending->{'response'}, to => $pending->{'from'}, parms => $pending->{'parms'});
	    splice @{ $self->{'pendingRemote'} }, $pos, 1;
	}
	$pos++;
    }

    if (($#{ $self->{'pendingStarts'} } >= 0) || ($#{ $self->{'pendingRemote'} } >= 0)) {
	$self->at(name => '_timeout', to => $self, time => time + ($self->{'localTimeout'} / 2));
    }
    else {
	$self->{'timeoutRunning'} = 0;
    }
}

1;
