# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.8 $

package Thing::Service;

use Thing::Handle;

# NAME, PORT, Handle - Provides an access point to service. PORT is
# designed to be easily overridden by the user and precedence is as
# follows (from high to low):
#
#   Environment variable THING_SERVICEPORT
#   Configuration file variable $servicePort (if Thing::Service::readConfig was performed)
#   2222

sub NAME { 'Thing::Service'; }
sub PORT {
    if (!defined($_servicePort)) {
	if (defined($ENV{THING_SERVICEPORT})) {
	    $_servicePort = $ENV{THING_SERVICEPORT};
	}
	else {
	    $_servicePort = 2222;
	}
    }
    return $_servicePort;
}
sub HOST {
  if (!defined($_hostname)) {
    if (defined($ENV{THING_HOSTNAME})) {
      $_hostname = $ENV{THING_HOSTNAME};
    }
    else {
      my $hostname = `uname -n`;
      if (!defined($hostname)) {
	Trace::error('Cannot determin host name with \'uname -n\': ', $!);
	return undef;
      }
      chop $hostname;
      $_hostname = $hostname;
    }
  }
  return $_hostname;
}

sub Handle {
    return Thing::Handle->new(name => Thing::Service::NAME, host => Thing::Service::HOST, port => Thing::Service::PORT);
}

sub RegisterRequest    { 'Thing__Service__RegisterRequest'; }
sub RegisterResponse   { 'Thing__Service__RegisterResponse'; }
sub DeregisterRequest  { 'Thing__Service__DeregisterRequest'; }

sub StartRequest       { 'Thing__Service__StartRequest'; }
sub StartResponse      { 'Thing__Service__StartResponse'; }

sub LookupRequest      { 'Thing__Service__LookupRequest'; }
sub LookupResponse     { 'Thing__Service__LookupResponse'; }

sub ConfigLoadRequest      { 'Thing__Service__ConfigLoadRequest'; }
sub ConfigLoadResponse     { 'Thing__Service__ConfigLoadResponse'; }

sub ConfigSaveRequest      { 'Thing__Service__ConfigSaveRequest'; }
sub ConfigSaveResponse     { 'Thing__Service__ConfigSaveResponse'; }

sub ShutdownRequest        { 'Thing__Service__ShutdownRequest'; }

# Monitor messages

sub MonitorRequest      { 'Thing__Service__MonitorRequest'; }
sub DemonitorRequest    { 'Thing__Service__DemonitorRequest'; }
sub MonitorResponse     { 'Thing__Service__MonitorResponse'; }

# Lookup errors

sub ThingNotRegistered { 'ThingNotRegistered'; }

# Config save errors

sub WriteFailed { 'WriteFailed'; }

# Generic errors

sub HostUnreachable    { 'HostUnreachable'; }
sub TimedOut           { 'TimedOut'; }

# readConfig, writeConfig - Read and write the config file, exported
# for boot, upgrade and access to common parms.

sub readConfig {
    my $configfile;
    if ($#_ == 0) {
	$configfile = shift;
    }
    else {
	if (!defined(configPath())) {
	    Trace::error(configPathError());
	    return;
	}
	$configfile = configFile();
    }
    use vars qw( $config $trusted $version $servicePort );
    $config = undef; $trusted = 0; $version = '';
    if (-e $configfile) {
	if (! -r $configfile) {
	    Trace::error('Cannot read config file %s', $configfile);
	    return;
	}
	eval {
	    do $configfile;
	};
	if ($@) {
	    Trace::error('Error in config file: %s %s', $configfile, $@);
	    return;
	}
	else {
	    Trace::debug('Loaded config file: %s', $configfile) if Trace::check(Trace::DEBUG3);
	}
    }
    else {
	Trace::debug('Config file %s does not exist', $configfile) if Trace::check(Trace::DEBUG3);
	return { config => {}, trusted => $trusted, version => Thing::VERSION() };
    }
    if (!defined($ENV{THING_SERVICEPORT}) && defined($servicePort) && $servicePort) {
	$ENV{THING_SERVICEPORT} = $servicePort;
    }
    return { config => $config, trusted => $trusted, version => $version };
}

sub writeConfig {
    my $cfg = shift;

    if ($#_ == 0) {
	$configfile = shift;
    }
    else {
	if (!defined(configPath())) {
	    Trace::error(configPathError());
	    return;
	}
	$configfile = configFile();
    }

    my $fh = IO::File->new;    
    if (!$fh->open('>'.$configfile)) {
	Trace::error('Cannot open config file %s: %s', $configfile, $!);
	return;
    }

    my $text = '$version=\''.$cfg->{'version'}."\';\n".
	       '$trusted='.$cfg->{'trusted'}.";\n".
		(defined($cfg->{'servicePort'}) ? '$servicePort='.$cfg->{'servicePort'}.";\n" : '').
	       '$config = '.unravel($cfg->{'config'}, 0).';1;';
    if (!$fh->print($text)) {
	Trace::error('Cannot write to config file %s: %s', $configfile, $!);
	$fh->close;
	unlink $configfile;
	return;
    }
    $fh->close;
    return 1;
}

sub configFile { 
    if (!defined(configPath())) {
	Trace::error(configPathError());
	return;
    }
    return configPath().'/config.pl';
}

sub configPath { 
    if (!defined($ENV{'THING_CONFPATH'})) {
	if (!defined($ENV{'HOME'})) {
	    Trace::debug('Cannot determin a configuration directory') if Trace::check(Trace::DEBUG6);
	    return undef;
	}
	my $hostname = `uname -n`;
	chop $hostname;
	$ENV{'THING_CONFPATH'} = $ENV{'HOME'}.'/.emptytree-seedy/'.$hostname;
    }
    return $ENV{'THING_CONFPATH'};
}

sub configPathError { 'Must declare a config path with either HOME or THING_CONFPATH'; }

# unravel(data, depth)

sub unravel {
    my $segment = '';
    if (ref($_[0])) {
	if (ref($_[0]) eq 'Thing::Handle') {
	    $segment = 'Thing::Handle->new(name => \''.
		$_[0]->name.'\', host => \''.$_[0]->host.'\')';
	}
	elsif (UNIVERSAL::isa($_[0], 'HASH')) { 
	    $segment .= "{\n";
	    foreach my $key (keys(%{ $_[0] })) {
		$segment .= _depth($_[1]).'\''.$key.'\''.' => '.unravel($_[0]->{ $key }, $_[1] + 1).",\n";
	    }
	    $segment .= _depth($_[1] - 1)."}";
	} 
	elsif (UNIVERSAL::isa($_[0], 'ARRAY')) { 
	    $segment .= "[\n";
	    foreach my $value (@{ $_[0] }) {
		$segment .= _depth($_[1]).unravel($value, $_[1] + 1).",\n";
	    }
	    $segment .= _depth($_[1] - 1)."]";
	} 
	else {
	    Trace::error('Unhandled element in config: %s', ref($_[0]));
	    $segment = _depth($_[1]).'undef';
	}
    }	
    else {
	if (defined($_[0])) {
	    $segment = "'".$_[0]."'";
	}
	else {
	    $segment = 'undef';
	}
    }
    return $segment;
}

sub _depth { 
    my $depth = '';
    for (my $i = 0; $i < $_[0]; $i++) {
	$depth .= '    ';
    }
    return $depth;
}

# INTERNAL

my $_hostname;
my $_servicePort;

1;
