# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.13.2.1 $

package Trace;

use strict;
use IO::File qw();
use POSIX qw();

sub DEBUG6 () {7;}
sub DEBUG5 () {6;}
sub DEBUG4 () {5;}
sub DEBUG3 () {4;}
sub DEBUG2 () {3;}
sub DEBUG1 () {2;}
sub ERROR  () {1;}
sub OFF ()    {0;}

my $_lastError;
my $_lastDebug;
my $_state = ERROR;
my $_pid = 0;
my $_packages = undef;

sub on {
    $_state = shift;
    $_pid = shift;
    $_packages = shift;
}

sub off {
    $_state = 0;
    $_pid = 0;
    $_packages = undef;
}

sub level { $_state; }

sub performance { 0; }
#use Time::HiRes;
#my $_stamps = {};
#sub timestamp { $_stamps->{ $_[0] } = [Time::HiRes::gettimeofday] }
#sub timediff { Time::HiRes::tv_interval($_stamps->{ $_[0] }); }

sub check { my $level = shift || die; return 0 if ($_state < $level); return 0 if (!_package()); return 1; }

sub error {
    my $str = shift;
    $str = $_lastError = sprintf($str, @_);
    return undef if ($_state < ERROR);
    print STDOUT _caller(), 'ERROR', ': ', $str, "\n";
    print STDOUT _stack(), "\n";
    return undef;
}

sub mandatory {
    my $str = shift;
    $_lastError = "Missing parameters $str";
    return undef if ($_state < ERROR);
    return undef if (!_package());
    printf(STDOUT _caller()."MANDATORY: $_lastError\n");
    print STDOUT _stack(), "\n";
    return undef;
}

sub warn {
    return undef if ($_state < DEBUG6);
    return undef if (!_package());
    my $str = shift;
    printf(STDOUT _caller()."WARN: $str\n", @_);
    return undef;
}

sub debug {
    my $str = shift;
    $str = $_lastDebug = sprintf($str, @_);
    print STDOUT _caller(), 'DEBUG: ', $str, "\n";
    return undef;
}

sub point {
    return undef if ($_state < DEBUG4);
    return undef if (!_package());
    printf(STDOUT _caller()."POINT\n");
    return undef;
}

sub dump {
    my $level;
    if ($#_ == 1) {
	$level = shift;
    }
    else {
	$level = DEBUG6;
    }
    return undef if ($_state < $level);
    my @lines = split("\n", $_[0]);
    my $c = 0;
    foreach my $line (@lines) {
	print STDOUT 'DUMP: ', ++$c, ': ', $line, "\n";
    }
    return undef;
}

sub lastError { $_lastError; }
sub lastDebug { $_lastDebug; }

sub defStr { if (!defined($_[0])) { 'unknown'; } else { $_[0]; } }

sub defNum { if (!defined($_[0])) { 0; } else { $_[0]; } }

sub file {
    my $file = shift;
    my $fh = IO::File->new('>'.$file) || die $!;
    POSIX::dup2($fh->fileno, 0);
    POSIX::dup2($fh->fileno, 1);
    POSIX::dup2($fh->fileno, 2);
    $| = 1;
}

sub _caller {
    my $pkg; 
    my $filename; 
    my $line; 
    my $subroutine;
    my $dummy;
    my $lineInfo;

    ($pkg, $dummy, $line, $subroutine) = caller(2);
    ($dummy, $filename, $line, $dummy) = caller(1);
    if (defined($subroutine)) {
	$lineInfo = $subroutine.':'.$line;
    } 
    elsif (defined($filename)) {
	$lineInfo = $filename.':'.$line;
    }	
    if (length($lineInfo) > 30) {
	# Shorten the front
	$lineInfo = substr $lineInfo, -30;
    }
    my $pid;
    if ($_pid) {
	$pid = $$.' ';
    }
    else {
	$pid = '';
    }
    return sprintf('%s%s: %-30.30s ', $pid, POSIX::strftime('%d/%m %H:%M:%S', localtime(time)), $lineInfo);
}

sub _stack {
    my $line; 
    my $subroutine;
    my $dummy;

    my @stack = ();
    my $level = 2;
    do {
	($dummy, $dummy, $dummy, $subroutine) = caller($level);
	($dummy, $dummy, $line, $dummy) = caller($level - 1);
	if (defined($subroutine)) {
	    push @stack, $subroutine.':'.$line;
	}
	$level++;
    } while (defined($subroutine) && ($level < 7));
    if (@stack) {
	return join(', ', @stack);
    }    
    return 'EMPTY STACK';
}

sub _package {
    my $pkg; 
    my $line; 
    my $subroutine;
    my $dummy;

    ($pkg, $dummy, $line, $subroutine) = caller(2);

    if (defined($_packages) && defined($subroutine)) {
	return grep { $subroutine =~ /^$_\:\:/ } @{ $_packages };
    }
    return 1;
}

1;
