# Copyright (C) 2000 Gavin Jefferies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Revision: 1.8 $

package Utility;

use Command;
use Trace;
use strict;

# available() - Calculate how many bytes are available in the
# filesystem that holds `path'.

sub available {
    my $path = shift;

    my $cmd = Command->new('Shell' => 'df '.$path, 'Timeout' => 20);    
    if (!$cmd || 
	!$cmd->run || 
	($cmd->state != Command::FINISHED) ||
	(defined($cmd->return) && ($cmd->return != 0))) {
	Trace::error('Cannot determin available disk space');
	Trace::error($cmd->output);
	return undef;
    }

    my @lines = split(/\n/, $cmd->output);
    if ($lines[1] =~ m/^\S+\s+(\d+)\s+(\d+)\s+(\d+)\s+/) {

	# Available space is the free space minus whats left to be
	# used up in each reserved tmpfile.  `df' reports (space -
	# used) as different from available. This is to do with the
	# space reserved for root in the ext2 case.

	# my $freeBytes = 1024 * ($1 - $2);
	# my $freeBytes = 1024 * $3;
	# Blocks are supposedly 1024 bytes, but with overhead(?) the
	# real amount works out arround 1019	

	my $availableBytes = 1019 * $3;
	Trace::debug('%d bytes free in %s', $availableBytes, $path) if (Trace::check(Trace::DEBUG3));
	return $availableBytes;
    }
    
    return undef;
}

# executes() - See if a binary can execute in the path, 
# 127 - command not found
# 126 - permission denied
# 139 - segmentation violation
# Assume 255 (-1), 0 or 1 are good

sub executes {
    my $cmd = Command->new('Shell' => shift, 'Timeout' => 20);    
    if (!$cmd || 
	!$cmd->run || 
	($cmd->state != Command::FINISHED) ||
	(defined($cmd->return) && (($cmd->return != 255) &&
				   ($cmd->return != 0) &&
				   ($cmd->return != 1)))) {
	return $cmd->output || 1;
    }
    return 0;
}

# cmpVersions(a, b) - Compare two version strings, returns -1 a is
# less than b, 0 a is equal to b, 1 a is greater than b. Versions are
# a sequence of dot seperated digits. undefined values are considered
# 0. Thus 1.3.1 is equal to 1.3.1.0.

sub cmpVersions {
    my $astr = shift; my $bstr = shift;

    my @a = split(/\./, $astr);
    my @b = split(/\./, $bstr);

    for (my $i = 0; $i <= (($#a > $#b) ? $#a : $#b); $i++) {
	my $aelem = defined($a[$i]) ? $a[$i] : 0;
	my $belem = defined($b[$i]) ? $b[$i] : 0;

	if ($aelem > $belem) {
	    return 1;
	}
	elsif ($aelem < $belem) {
	    return -1;
	}
    }
    return 0;
}

package Utility::Array;

# subset - is b a subset of a?

sub subset {
    my $a = shift;
    my $b = shift;

    my %ah;

    foreach my $elem (@{ $a }) {
	$ah{ $elem } = 1;
    }
    foreach my $elem (@{ $b }) {
	return 0 if (!$ah{ $elem });
    }
    return 1;
}

# same - unordered same array

sub same {
    my $a = shift;
    my $b = shift;

    return 0 if (!subset($a, $b));
    return ($#{ $a } == $#{ $b });
}

# sameOrder - ordered same array

sub sameOrder {
    my $a = shift;
    my $b = shift;

    return 0 if ($#{ $a } != $#{ $b });
    for (my $i = 0; $i <= $#{ $a }; $i++) {
	return 0 if ($a->[$i] ne $b->[$i]);
    }
    return 1;
}

# find - return element where sub is true

sub find {
    my $a = shift;
    my $s = shift;

    foreach my $element (@{ $a }) {
	return $element if (&$s($element));
    }
    return undef;
}

package Utility::Hash;

# augment(existing, new, special) - Augment an existing hash with data
# from new, values in new replace values in existing unless the
# existing value is a hash ref in which case that hash too is
# augmented from its corresponding value in new. 1 is returned if
# changes were made, 0 if not and undef if either existing or new are
# not hash refs.

# special is an array of compare functions. The compare functions
# check the refs for special objects that you don't want descended and
# compared. They return undef if the ref is not special, 1 if it is
# and they are different objects, 0 if they are the same.

sub augment {
    my %parms = @_;
    my $existing = $parms{'existing'};
    my $new = $parms{'new'};
    my $special = $parms{'special'} || []; # [ { compareFn -> \& } ]

    return if (!defined($existing));
    return if (!defined($new));
    return if (!UNIVERSAL::isa($existing, 'HASH'));
    return if (!UNIVERSAL::isa($new, 'HASH'));

    my $diff = 0;
    my @stack = ();
    my $printStack =  sub { @stack ? join(',', map { $_->{'key'} } @stack) : 'empty' };
    my @keys = keys(%{ $new });
    return 0 if (!@keys);
    TOP: while (1) {
	my $key = shift @keys;
	Trace::debug('[%s] Trying %s', &$printStack, $key) if Trace::check(Trace::DEBUG6);
	if (defined($existing->{ $key })) {
	    # Handlers for `special' objects
	    my $descend = 1;
	    SPECIAL: foreach my $sp (@{ $special }) {
		my $rc = &{ $sp->{'compareFn'} }($existing->{ $key }, $new->{ $key });
		next SPECIAL if (!defined($rc));
		$descend = 0;
		if ($rc) {
		    Trace::debug('[%s] Special difference %s', &$printStack, $key) if Trace::check(Trace::DEBUG6);
		    # Objects that we don't want to descend
		    $existing->{ $key } = $new->{ $key };
		    $diff = 1;
		}
		else {
		    # No diff
		}
		last SPECIAL;
 	    }
	    if ($descend) {
		if (UNIVERSAL::isa($new->{ $key }, 'HASH') && UNIVERSAL::isa($existing->{ $key }, 'HASH')) {
		    # Both hashes, step in
		    my @dup = @keys;
		    push @stack, { key => $key, new => $new, existing => $existing, keys => \@dup };
		    $new = $new->{ $key };
		    $existing = $existing->{ $key };
		    @keys = keys(%{ $new });
		    Trace::debug('[%s] Descending %s', &$printStack, $key) if Trace::check(Trace::DEBUG6);
		}
		elsif ($new->{ $key } ne $existing->{ $key }) {
		    # Only a stringwise compare all other refs will be always considered different
		    Trace::debug('[%s] Difference %s ne %s', &$printStack, $existing->{ $key }, $new->{ $key }) if Trace::check(Trace::DEBUG6);
		    $existing->{ $key } = $new->{ $key };
		    $diff = 1;
		}
	    }
	}
	else {
	    Trace::debug('[%s] Difference !defined %s', &$printStack, $key) if Trace::check(Trace::DEBUG6);
	    $existing->{ $key } = $new->{ $key };
	    $diff = 1;
	}
	LEVEL: while ($#keys == -1) {
	    # Back up the stack
	    my $se = pop @stack;
	    last TOP if (!$se);
	    $new = $se->{'new'};
	    $existing = $se->{'existing'};
	    @keys = @{ $se->{'keys'} };
	    Trace::debug('[%s] Back to %s', &$printStack, $se->{'key'}) if Trace::check(Trace::DEBUG6);
	}
    }
    return $diff;
}

1;
